/*
 * The SOSI OSS Library - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of the SOSI OSS Library
 * is copyright to National Board of e-Health.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 */
package dk.sosi.demo.client.pacf;

import dk.sosi.demo.client.vault.CredentialVaultUtil;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.security.PrivateKey;
import java.security.Signature;
import java.security.cert.X509Certificate;
import java.util.Properties;

/**
 * The default controller for the SOSI Client PACF agent. The public methods on
 * this class are the <code>actions</code> that can be pressed upon the agent.
 * Actions usually results in <code>reactions</code> being "fired" and
 * "catched" by reactor-realizations such as GUI formatters or unit-test
 * asserters.
 *
 * @author Jan Riis
 * @see SOSIClientActions
 * @see SOSIClientReactions
 * @since 1.0
 */
public class SOSIClientController implements SOSIClientActions {

	protected SOSIClientReactions reactor = null;

	protected ServiceConnector connector = null;

	// ==========================================
	// Action methods
	// ==========================================
	public void initialize(ServiceConnector connectorImpl, SOSIClientReactions reactorImpl) {
		this.reactor = reactorImpl;
		this.connector = connectorImpl;
		reactorImpl.initialize();
	}

	public IDCard requestIDCard(int level, String cpr, String givenName, String surName, String occupation, String email, String role,
			String systemName, String orgName, String orgCVR, boolean print) throws Exception {

		// Build request ...
        Properties properties = SignatureUtil.setupCryptoProviderForJVM();
		SOSIFactory factory = CredentialVaultUtil.createSOSIFactory(new SOSITestFederation(properties));
		SecurityTokenRequest securityTokenRequest = createSecurityTokenRequest(
				level,
					occupation,
					cpr,
					givenName,
					surName,
					email,
					role,
					systemName,
					orgName,
					orgCVR, factory);

		Document doc = securityTokenRequest.serialize2DOMDocument();
		if (level == AuthenticationLevel.MOCES_TRUSTED_USER.getLevel()) {
			// Sign the document - should be done externally with user Moces signature
			byte[] siBytes = securityTokenRequest.getIDCard().getBytesForSigning(doc);
			Signature jceSign = Signature.getInstance("SHA1withRSA", SignatureUtil.getCryptoProvider(factory.getProperties(), SOSIFactory.PROPERTYNAME_SOSI_CRYPTOPROVIDER_SHA1WITHRSA));
			PrivateKey key = factory.getCredentialVault().getSystemCredentialPair().getPrivateKey();
			jceSign.initSign(key);
			jceSign.update(siBytes);
			String signature = XmlUtil.toBase64(jceSign.sign());
			X509Certificate certificate = factory.getCredentialVault().getSystemCredentialPair().getCertificate();
			securityTokenRequest.getIDCard().injectSignature(signature, certificate);
		}
		// Call Service
		System.out.println("Calling Tomcat connector with IDCard " + securityTokenRequest.getIDCard().toString());
		SecurityTokenResponse reply = connector.callIdpService(factory, doc);

		if (!reply.isFault()) {
			if (print) {
				reactor.newIDCard(connector.getRequest(), connector.getResponse(), reply.getIDCard());
			}
			return reply.getIDCard();
		} else {
			reactor.fault(connector.getRequest(), connector.getResponse());
			return null;
		}
	}

	public void requestService(String arg) throws Exception {
        Properties properties = SignatureUtil.setupCryptoProviderForJVM();
		SOSIFactory factory = CredentialVaultUtil.createSOSIFactory(new SOSITestFederation(properties));
		// Fetch an IDCard, call service using replyDocument
		IDCard idCard = requestIDCard(
				4,
					"1111111118",
					"Per",
					"Petersen",
					"Doktor",
					"a@a.com",
					"doctor",
					"SOSITEST",
					"SOSI project",
					"19343634",
					false);
		Request request = factory.createNewRequest(false, XmlUtil.createGUID());
		request.setIDCard(idCard);

		Document doc = XmlUtil.createEmptyDocument();
		Element body = doc.createElement("serverTime");
		request.setBody(body);

		request.serialize2DOMDocument(doc);
		connector.callProviderService(factory, doc);

		reactor.serviceResponse(connector.getRequest(), connector.getResponse(), "Read returned document below");
	}

	private SecurityTokenRequest createSecurityTokenRequest(int level, String occupation, String cpr, String givenName, String surName, String email,
			String role, String systemName, String orgName, String orgCVR, SOSIFactory factory) {

		IDCard idCard = null;
		if (level == AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel()) {
			idCard = factory.createNewSystemIDCard(systemName, new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, orgCVR, orgName), AuthenticationLevel.VOCES_TRUSTED_SYSTEM,
					null, null, factory.getCredentialVault().getSystemCredentialPair().getCertificate(), null);
		} else if (level == AuthenticationLevel.MOCES_TRUSTED_USER.getLevel()) {
			idCard = factory.createNewUserIDCard(
					systemName,
					new UserInfo(cpr, givenName, surName, email, occupation, role,"2101"),
					new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, orgCVR, orgName),
					AuthenticationLevel.MOCES_TRUSTED_USER,
					null,
					null,
					null,
					null);
		} else if (level == AuthenticationLevel.NO_AUTHENTICATION.getLevel()) {
			idCard = factory.createNewSystemIDCard(systemName, new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, orgCVR, orgName),
					AuthenticationLevel.NO_AUTHENTICATION, null, null, null, null);

		} else {
			System.out.println("NOT IMPLEMENTED YET");
		}

		SecurityTokenRequest securityTokenRequest = factory.createNewSecurityTokenRequest();

		securityTokenRequest.setIDCard(idCard);
		return securityTokenRequest;
	}
}