package dk.sosi.demo.sbo;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.model.dombuilders.IDCardToOIOSAMLAssertionRequestDOMBuilder;
import dk.sosi.seal.model.dombuilders.UnsolicitedResponseDOMBuilder;
import dk.sosi.seal.modelbuilders.IDCardToOIOSAMLAssertionResponseModelBuilder;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.testtools.testrequest.SoapClient;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.net.MalformedURLException;
import java.net.URL;
import java.security.*;
import java.util.Properties;

public class SBOTestClient {

    private static final String HOST = "test1-cnsp.ekstern-test.nspop.dk:8080";

    private static final String STS_ENDPOINT = "http://" + HOST + "/sts/services/NewSecurityTokenService";
    private static final String SOSI_2_OIOSAML_ENDPOINT = "http://" + HOST + "/sts/services/Sosi2OIOSaml";

    public static void main(String[] args) throws Exception {
        getUnsolicitedResponse();
    }

    public static String getUnsolicitedResponseBase64() throws Exception {
        return XmlUtil.toBase64(getUnsolicitedResponse().getBytes());
    }

    private static String getUnsolicitedResponse() throws Exception {
        Properties properties = System.getProperties();
        SOSIFactory sosiFactory = new SOSIFactory(new SOSITestFederation(properties), new EmptyCredentialVault(), properties);

        OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
        CredentialVault vocesVault = CredentialVaultTestUtil.getVocesCredentialVault();

        UserIDCard idCard = getSTSSignedUserIDCard(sosiFactory);
        Element encryptedAssertion = getEncryptedAssertion(oiosamlFactory, vocesVault, idCard);

        UnsolicitedResponseDOMBuilder builder = oiosamlFactory.createUnsolicitedResponseDOMBuilder();
        builder.setIssuer("Seal Java SBO Test Client");
        builder.setEncryptedAssertion(encryptedAssertion);
        builder.requireRandomInResponseTo();
        Document unsolicitedResponseDoc = builder.build();
        //System.out.println(XmlUtil.node2String(unsolicitedResponseDoc, true, true));
        return  XmlUtil.node2String(unsolicitedResponseDoc);
    }

    private static Element getEncryptedAssertion(OIOSAMLFactory oiosamlFactory, CredentialVault signingVault, UserIDCard idCard) throws MalformedURLException {
        IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = oiosamlFactory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();
        domBuilder.setSigningVault(signingVault);
        domBuilder.setUserIDCard(idCard);
        domBuilder.setAudience("https://saml.sositest.sundhed.dk");
        Document requestDoc = domBuilder.build();

        SoapClient client = new SoapClient(new URL(SOSI_2_OIOSAML_ENDPOINT));
        Document responseDoc = client.execute(XmlUtil.node2String(requestDoc));
        IDCardToOIOSAMLAssertionResponseModelBuilder modelBuilder = oiosamlFactory.createIDCardToOIOSAMLAssertionResponseModelBuilder();
        IDCardToOIOSAMLAssertionResponse response = modelBuilder.build(responseDoc);

        //System.out.println(XmlUtil.node2String(responseDoc, true, true));

        return response.getEncryptedOIOSAMLAssertionElement();
    }

    private static UserIDCard getSTSSignedUserIDCard(SOSIFactory factory) throws Exception {
        CredentialVault mocesVault = CredentialVaultTestUtil.getCredentialVault();
        String cvrNumber = "30808460";

        UserIDCard idCard = factory.createNewUserIDCard(
                "SOSITEST",
                new UserInfo(null, "Fritz", "Muller", "fritz@muller.com", "fool", "doctor", null),
                new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, cvrNumber, "Hansens Praksis"),
                AuthenticationLevel.MOCES_TRUSTED_USER,
                null,
                null,
                mocesVault.getSystemCredentialPair().getCertificate(),
                null);

        SecurityTokenRequest request = factory.createNewSecurityTokenRequest();
        request.setIDCard(idCard);

        Document requestDoc = request.serialize2DOMDocument();
        String signature = createSignature(mocesVault, request, requestDoc, factory.getProperties());
        request.getIDCard().injectSignature(signature, mocesVault.getSystemCredentialPair().getCertificate());

        SoapClient client = new SoapClient(new URL(STS_ENDPOINT));
        Document responseDoc = client.execute(XmlUtil.node2String(requestDoc));
        SecurityTokenResponse response = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(responseDoc));
        //System.out.println(XmlUtil.node2String(responseDoc, true, true));
        return (UserIDCard) response.getIDCard();
    }

    private static String createSignature(CredentialVault vault, Message message, Document doc, Properties properties) throws NoSuchAlgorithmException, InvalidKeyException,
            SignatureException {

        byte[] siBytes = message.getIDCard().getBytesForSigning(doc);
//		Sign the bytes with a private key (in this case the key in credential
//		vault, but that makes no difference)
        Signature jceSign;
        try {
            jceSign = Signature.getInstance("SHA1withRSA", SignatureUtil.getCryptoProvider(properties,
                    SOSIFactory.PROPERTYNAME_SOSI_CRYPTOPROVIDER_SHA1WITHRSA));
        } catch (NoSuchProviderException e) {
            throw new ModelException("No Such Provider", e);
        }
        PrivateKey key = vault.getSystemCredentialPair().getPrivateKey();
        jceSign.initSign(key);
        jceSign.update(siBytes);
        return XmlUtil.toBase64(jceSign.sign());
    }

}

