/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import dk.sosi.seal.model.CitizenIdentityTokenBuilder;
import dk.sosi.seal.model.LibertyRequest;
import dk.sosi.seal.model.LibertyRequestDOMEnhancer;
import dk.sosi.seal.model.LibertyResponse;
import dk.sosi.seal.model.LibertyResponseDOMEnhancer;
import dk.sosi.seal.modelbuilders.LibertyRequestModelBuilder;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.vault.CredentialVault;

public class OIOIDWSFactory {

	protected final CredentialVault credentialVault;
	protected final Federation federation;

	/**
	 * Constructor for the <code>OIOIDWSFactory</code> class.
	 */
	protected OIOIDWSFactory() {
		this(null, null);
	}

	/**
	 * Constructor for the <code>OIOIDWSFactory</code> class.
	 * 
	 * @param federation The <code>Federation</code> to use. See {@link Federation}
	 *                   for more information.
	 */
	protected OIOIDWSFactory(Federation federation) {
		this(federation, null);
	}

	/**
	 * Constructor for the <code>OIOIDWSFactory</code> class.
	 * 
	 * @param credentialVault The <code>CredentialVault</code> to use. See
	 *                        {@link CredentialVault} for more information.
	 */
	public OIOIDWSFactory(CredentialVault credentialVault) {
		this(null, credentialVault);
	}

	/**
	 * Constructor for the <code>OIOIDWSFactory</code> class.
	 * 
	 * @param federation      The <code>Federation</code> to use. See
	 *                        {@link Federation} for more information.
	 * @param credentialVault The <code>CredentialVault</code> to use. See
	 *                        {@link CredentialVault} for more information.
	 */
	public OIOIDWSFactory(Federation federation, CredentialVault credentialVault) {
		this.federation = federation;
		this.credentialVault = credentialVault;
	}

	/**
	 * Creates a new <code>CitizenIdentityTokenBuilder</code>.
	 * 
	 * @return An <code>CitizenIdentityTokenBuilder</code> instance.
	 */
	public CitizenIdentityTokenBuilder createIdentityTokenBuilder() {
		return new CitizenIdentityTokenBuilder();
	}

	/**
	 * Creates a new <code>LibertyRequestModelBuilder</code>.
	 * 
	 * @return An <code>LibertyRequestModelBuilder</code> instance.
	 */
	public LibertyRequestModelBuilder createRequestModelBuilder() {
		if (federation == null) {
			NullPointerException npe = new NullPointerException("Federation is null");
			throw new ModelBuildException("This method cannot be called without Federation.", npe);
		}

		return new LibertyRequestModelBuilder(federation);
	}

	/**
	 * Creates a new <code>LibertyRequestModelBuilder</code>.
	 *
	 * @param federation The <code>Federation</code> to use. See {@link Federation}
	 *                   for more information.
	 * 
	 * @return An <code>LibertyRequestModelBuilder</code> instance.
	 */
	public static LibertyRequestModelBuilder createRequestModelBuilder(Federation federation) {
		return new OIOIDWSFactory(federation).createRequestModelBuilder();
	}

	/**
	 * Creates a new <code>LibertyRequest</code>.
	 * 
	 * @param node The <code>Document</code> or <code>Element</code> to transform
	 *             into a LibertyRequest object.
	 * 
	 * @return An <code>LibertyRequest</code> instance.
	 */
	public LibertyRequest createRequest(Node node) {
		if (node == null) {
			throw new ModelBuildException("This method cannot be called without: Document or Document element",
					new NullPointerException("Document/Element is null"));
		}

		return createRequestModelBuilder().build(node);
	}
	
	/**
	 * Creates a new <code>LibertyRequest</code>.
	 * 
	 * @param federation The <code>Federation</code> to use. See {@link Federation}
	 *                   for more information.
	 * @param keyInfoType The type of KeyInfo used in this OIO IDWS node.
	 *  
	 * @param document   The <code>Document</code> to transform into a
	 *                   LibertyRequest object.
	 * @return An <code>LibertyRequest</code> instance.
	 */
	public static LibertyRequest createRequest(Federation federation, Node node) {
		if (node == null) {
			throw new ModelBuildException("This method cannot be called without: Document or Document element",
					new NullPointerException("Document/Element is null"));
		}

		return createRequestModelBuilder(federation).build(node);
	}

	/**
	 * Creates a new <code>LibertyRequestDOMEnhancer</code>.
	 * 
	 * @return An <code>LibertyRequestDOMEnhancer</code> instance.
	 */
	public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
		/*Check af document og credentialVault sker i LibertyRequestDOMEnhancer*/
		return new LibertyRequestDOMEnhancer(credentialVault, document);
	}

	/**
	 * Creates a new <code>LibertyMessageDOMEnhancer</code>.
	 * 
	 * @return An <code>LibertyMessageDOMEnhancer</code> instance.
	 */
	public LibertyResponseDOMEnhancer createResponseDomEnhancer(Document document, boolean ensureLibertyFrameworkHeader) {
		if (document == null) {
			throw new ModelBuildException("This method cannot be called without: Document.",
					new NullPointerException("Document is null"));
		} else if (credentialVault == null) {
			throw new ModelBuildException("This method cannot be called without: CredentialVault.",
					new NullPointerException("CredentialVault is null"));
		}

		return new LibertyResponseDOMEnhancer(credentialVault, document, ensureLibertyFrameworkHeader);
	}

	/**
	 * Creates a new <code>LibertyMessageDOMEnhancer</code>.
	 * 
	 * @return An <code>LibertyMessageDOMEnhancer</code> instance.
	 */
	public static LibertyResponseDOMEnhancer createResponseDomEnhancer(CredentialVault vault, Document document, boolean ensureLibertyFrameworkHeader) {
		return new OIOIDWSFactory(vault).createResponseDomEnhancer(document, ensureLibertyFrameworkHeader);
	}

	/**
	 * 
	 * @param document
	 * @return
	 */
	public static LibertyResponse createResponse(Node node) {
		return new LibertyResponse(node);
	}
}
