/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import java.util.*;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import dk.sosi.seal.model.constants.Tag;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.xml.XmlUtil;

import static java.util.Optional.ofNullable;

/**
 * Abstract class used by classes directly modifying an underlying DOM model object.
 * 
 * @author ads
 */
public abstract class AbstractDOMInfoExtractor {

	/** The underlying DOM representation */
	protected final Element dom;
	private Map<String, String> namespaces;

	public AbstractDOMInfoExtractor(Node node) {
		if (node == null) {
			throw new IllegalArgumentException("Node[Document or Element] cannot be null");
		}

		Element elm = null;
		if (node.getNodeType() == Node.DOCUMENT_NODE) {
			Document doc = (Document) node;
			elm = doc.getDocumentElement();
		} else if (node.getNodeType() == Node.ELEMENT_NODE) {
			elm = (Element) node;
		} else {
			throw new ModelBuildException("Cannot create model with from NodeType:" + node.getNodeType());
		}

		if (elm == null) {
			throw new IllegalArgumentException("Document element cannot be null");
		}

		this.dom = elm;
	}

	/**
	 * Converts the attribute of the supplied <code>Element</code> to a
	 * <code>Date</code> using <code>XMLUtil.getDateFormat(true)...</code>.
	 * 
	 * @param element   The <code>Element</code> to retrieve the attribute from.
	 * @param attribute The name of the attribute to convert.
	 * @return The converted <code>Date</code> object.
	 */
	protected Date convertToDate(Element element, String attribute) {
		if (element == null) {
			throw new ModelException("Date element cannot be null");
		}
		String value;
		if (attribute != null) {
			value = element.getAttribute(attribute);
		} else {
			value = element.getTextContent();
		}

		try {
			return XmlUtil.parseZuluDateTime(value);
		} catch (Exception e) {
			throw new ModelException("Invalid date format of " + (attribute != null ? attribute : "text contents") + " (" + value + ")", e);
		}
	}

	/**
	 * Filter the supplied <code>NodeList</code> and retrieve the
	 * <code>Element</code> having the supplied attribute matching the supplied
	 * value.<br />
	 * 
	 * @param nl        The <code>NodeList</code> to filter.
	 * @param attribute The attribute to test.
	 * @param value     The value to filter for.
	 * @return The matching <code>Element</code> or <code>null</code> if no match
	 *         was found.
	 */
	protected Element getFilteredElement(List<Element> nl, String attribute, String value) {
		return getStaticFilteredElement(nl, attribute, value);
	}

	protected static Element getStaticFilteredElement(List<Element> nl, String attribute, String value) {
		for (Element tmp : nl) {
			String val = tmp.getAttribute(attribute);
			if (val != null) {
				if (val.equals(value)) {
					return tmp;
				}
			}
		}
		return null;
	}

	/**
	 * Retrieve the <code>Element</code> identified by the supplied tag path
	 * structure.<br />
	 * This method will traverse the DOM retrieve the first child for each specified
	 * tag.
	 * 
	 * @param tags List of tag names to traverse.
	 * @return The <code>Element</code> matching the final tag in the list.
	 */
	protected Element getTag(Tag... tags) {
		return getTag(dom, tags);
	}

	protected Element getTag(Element elm, Tag... tags) {
		if (tags.length == 1) {
			return elm;
		}
		return getFirstElement(getTags(elm, tags));
	}

	protected Optional<Element> safeGetTag(Element elm, Tag... tags) { return ofNullable(getTag(elm, tags)); }


	protected Element getDescendant(Element elm, Tag... tags) {
		Element current = elm;
		for (Tag tag : tags) {
			current = getFirstElement(getElements(current, tag, true));
		}
		return current;
	}

	/**
	 * Retrieve the <code>NodeList</code> identified by the supplied tag path
	 * structure.<br />
	 * This method will traverse the DOM retrieve the first child for each specified
	 * tag.
	 * 
	 * @param tags List of tag names to traverse.
	 * @return The <code>NodeList</code> matching the final tag in the list.
	 */
	protected List<Element> getTags(Tag... tags) {
		return getTags(dom, tags);
	}

	protected List<Element> getTags(Element elm, Tag... tags) {
		return getTags(getNamesSpaces(), elm, tags);
	}

	protected static List<Element> getTags(Map<String, String> ns, Element elm, Tag... tags) {
		for (int i = 1; i < tags.length - 1; i++) {
			Tag tag = tags[i];
			elm = getFirstElement(getElements(ns, elm, tag, true));
		}

		Tag tag = tags[tags.length - 1];
		return getElements(ns, elm, tag, false);
	}

	private List<Element> getElements(Element elm, Tag tag, boolean onlyFirst) {
		return getElements(getNamesSpaces(), elm, tag, onlyFirst);
	}

	private static List<Element> getElements(Map<String, String> ns, Element elm, Tag tag, boolean onlyFirst) {
		List<Element> res = new ArrayList<Element>();

		if (elm == null) {
			return res;
		}

		NodeList childNodes = elm.getChildNodes();
		for (int i = 0; i < childNodes.getLength(); i++) {
			Node childNode = childNodes.item(i);
			if (childNode instanceof Element) {
				Element childElm = (Element) childNode;
				if (matches(ns, childElm, tag)) {
					res.add(childElm);
					if (onlyFirst) {
						break;
					}
				}
			}
		}
		return res;
	}

	private static boolean matches(Map<String, String> ns, Element childElm, Tag tag) {
		if (childElm.getLocalName() != null) {
			if (childElm.getNamespaceURI() != null) {
				return childElm.getNamespaceURI().equals(tag.getNS()) && tag.getTagName().equals(childElm.getLocalName());
			}
			return tag.getTagName().equals(childElm.getLocalName());
		}

		String tn = childElm.getTagName();
		int colonIndex = tn.indexOf(':');

		if (colonIndex == -1) {
			return tag.getNS() == null && tag.getTagName().equals(tn);
		}

		String nsPrefix = ns.get(tag.getNS());
		return childElm.getTagName().equals(nsPrefix + ":" + tag.getTagName());
	}

	protected String safeGetTagTextContent(Tag... tags) {
		Element element = getTag(tags);
		return element != null ? element.getTextContent() : null;
	}

	protected String safeGetAttribute(String attributeName, Tag... tags) {
		Element element = getTag(tags);
		return element != null ? element.getAttribute(attributeName) : null;
	}

	protected Map<String, String> getNamesSpaces() {
		if (namespaces == null) {
			namespaces = getNamesSpaces(dom);
		}
		return namespaces;
	}

	protected static Map<String, String> getNamesSpaces(Element elm) {
		Map<String, String> ns = new HashMap<String, String>();

		NamedNodeMap attributes = elm.getAttributes();
		for (int i = 0; i < attributes.getLength(); i++) {
			Attr attribute = (Attr) attributes.item(i);

			String name = attribute.getName();
			String value = attribute.getValue();

			if (name.startsWith("xmlns:")) {
				ns.put(value, name.substring("xmlns:".length()));
			}
		}

		return ns;
	}

	private static Element getFirstElement(List<Element> nodeList) {
		if (nodeList.isEmpty()) {
			return null;
		}
		return nodeList.get(0);
	}
}