/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.xml.crypto.dsig.DigestMethod;

import dk.sosi.seal.model.constants.NameSpaces;
import org.apache.xml.security.signature.XMLSignature;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import dk.sosi.seal.model.OIOBSTSAMLAssertion.SAMLAttribute;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.model.constants.XSIAttributes;
import dk.sosi.seal.model.dombuilders.AbstractSAMLBuilder;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;

public abstract class AbstractOIOBSTSAMLAssertionBuilder<S extends AbstractOIOBSTSAMLAssertionBuilder<S, T>, T extends OIOBSTSAMLAssertion> extends AbstractSAMLBuilder<S, T> {
	private X509Certificate holderOfKeyCertificate;
	private CredentialVault signingVault;
	private Date notOnOrAfter;
	private String nameId, audience;
	private Map<SAMLAttribute, String> attributes = new LinkedHashMap<>();
	
	@SuppressWarnings("unchecked")
	public S setNameId(String nameId) {
		this.nameId = nameId;
		return (S)this;
	}
	
	@SuppressWarnings("unchecked")
	public S setAudience(String audience) {
		this.audience = audience;
		return (S) this;
	}
	
	@SuppressWarnings("unchecked")
	public S setNotOnOrAfter(Date notOnOrAfter) {
		this.notOnOrAfter = notOnOrAfter;
		return (S) this;
	}
	
    @SuppressWarnings("unchecked")
	public S setSigningVault(CredentialVault signingVault) {
        this.signingVault = signingVault;
        return (S) this;
    }
    
    @SuppressWarnings("unchecked")
	public S setHolderOfKeyCertificate(X509Certificate holderOfKeyCertificate) {
		this.holderOfKeyCertificate = holderOfKeyCertificate;
		return (S) this;
	}
        
    @Override
    public void validateBeforeBuild() throws ModelException {
    	validate("Assertion nameId", nameId);
    	validate("Assertion audience", audience);
		validate("Assertion notOnOrAfter", notOnOrAfter);
		validate("Assertion holderOfCertificate", holderOfKeyCertificate);
		validate("Assertion signingVault", signingVault);
		
        SAMLAttribute[] validAttributes = getAttributes();
        for (SAMLAttribute validAttribute : validAttributes) {
			if(validAttribute.isMandatory()) {
				validate("Assertion attribute "+validAttribute.getName(), getAttribute(validAttribute));
			}
		}
        super.validateBeforeBuild();
    }
    
    protected abstract SAMLAttribute[] getAttributes();
    
	@Override
	protected void appendToRoot(Document doc, Element root) {
		super.appendToRoot(doc, root);
		Node subjectNode = createSubject();
		root.appendChild(subjectNode);
        root.appendChild(createConditions());
		Node authnStatementNode = createAuthnStatement();
		if (authnStatementNode != null) {
			root.appendChild(authnStatementNode);
		}
        root.appendChild(createAttributeStatement(doc));
	}

    protected Node createSubject() {
        Element subjectElm = createElement(SAMLTags.subject);

        Element nameIdElm = createElement(SAMLTags.nameID);
        nameIdElm.setAttributeNS(null, SAMLAttributes.FORMAT, getNameIDFormat());
        nameIdElm.setTextContent(nameId);
        subjectElm.appendChild(nameIdElm);

        Element subjectConfirmationElm = createElement(SAMLTags.subjectConfirmation);
        subjectConfirmationElm.setAttributeNS(null, SAMLAttributes.METHOD, SAMLValues.CONFIRMATION_METHOD_HOLDER_OF_KEY);
        subjectElm.appendChild(subjectConfirmationElm);

        Element subjectConfirmationDataElm = createElement(SAMLTags.subjectConfirmationData);
        subjectConfirmationDataElm.setAttribute(XSIAttributes.TYPE_PREFIXED, NameSpaces.NS_SAML + ":KeyInfoConfirmationDataType");
        subjectConfirmationElm.appendChild(subjectConfirmationDataElm);
        
        Element keyInfoElm = createElement(DSTags.keyInfo);
        subjectConfirmationDataElm.appendChild(keyInfoElm);
        
        Element x509DataElm = createElement(DSTags.x509Data);
        keyInfoElm.appendChild(x509DataElm);
        
        Element x509CertificateElm = createElement(DSTags.x509Certificate);
        try {
            String encodedCert = XmlUtil.toBase64(holderOfKeyCertificate.getEncoded());
            x509CertificateElm.setTextContent(encodedCert);
        } catch (CertificateEncodingException e) {
            throw new ModelException("Unable to encode certificate", e);
        }
        x509DataElm.appendChild(x509CertificateElm);
        
        return subjectElm;
    }
    
    protected abstract String getNameIDFormat();
	
	protected Node createConditions() {
		Element conditionsElm = createElement(SAMLTags.conditions);
		conditionsElm.setAttributeNS(null, SAMLAttributes.NOT_ON_OR_AFTER, XmlUtil.getDateFormat(true).format(notOnOrAfter));

		Element audienceRestrictionElm = createElement(SAMLTags.audienceRestriction);
		conditionsElm.appendChild(audienceRestrictionElm);

		Element audienceElm = createElement(SAMLTags.audience);
		audienceElm.setTextContent(audience);
		audienceRestrictionElm.appendChild(audienceElm);

		return conditionsElm;
	}

	protected Node createAuthnStatement() {
		return null;
	}

    protected Element createAttributeStatement(Document doc) {
        Element attributeStatementElm = createElement(SAMLTags.attributeStatement);
        
        Set<SAMLAttribute> keys = attributes.keySet();
        for (SAMLAttribute key : keys) {
        	Element attributeElm = createAttributeElement(key, attributes.get(key));
        	attributeStatementElm.appendChild(attributeElm);		
		}
   
        return attributeStatementElm;
    }
    
    protected Element createAttributeElement(SAMLAttribute attribute, String value) {
    	return super.createAttributeElement(attribute.getName(), attribute.getNameFormat(), null, value);
    }

	protected void signAssertion(Document document) {
		SignatureConfiguration signatureConfiguration = getSignatureConfiguration();
		Node subjectNode = document.getDocumentElement().getElementsByTagName(SAMLTags.SUBJECT_PREFIXED).item(0);
		signatureConfiguration.setSignatureSiblingNode(subjectNode);
		SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(signingVault), document, signatureConfiguration);
	}

	protected SignatureConfiguration getSignatureConfiguration() {
		SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[] { assertionID }, assertionID, IDValues.Id);
		signatureConfiguration.setSignatureMethodAlgorithmUri(XMLSignature.ALGO_ID_SIGNATURE_RSA_SHA256);
		signatureConfiguration.setDigestMethodAlgorithmUri(DigestMethod.SHA256);
		return signatureConfiguration;
	}
	
    @Override
    public T build() throws ModelException {
        Document document = createDocument();

        // This makes it easier to pass the document around without messing up the signature.
		String documentString = XmlUtil.node2String(document.getDocumentElement());
		document = XmlUtil.readXml(new Properties(), documentString, false);

        signAssertion(document);
        return createAssertion((Element)document.getFirstChild());
    }
    
    protected abstract T createAssertion(Element elm);
    
    @SuppressWarnings("unchecked")
	protected S removeAttribute(SAMLAttribute attribute) {
    	this.attributes.remove(attribute);
    	return (S) this;
    }
    
    @SuppressWarnings("unchecked")
	protected S addAttribute(SAMLAttribute attribute, Object obj) {
    	this.attributes.put(attribute, obj != null ? String.valueOf(obj) : null);
    	return (S) this;
    }
    
    protected String getAttribute(SAMLAttribute key) {
    	return attributes.get(key);
    }
}
