/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.List;

import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import javax.xml.validation.Validator;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

public class AbstractOIOSAMLAssertion extends AbstractOIOSAMLToken {

	private static Schema schema;
	
	AbstractOIOSAMLAssertion(Element element, boolean validate) {
		super(element);
		if (validate) {
			validateElement(element);
			validateSchema(element);
		}
	}
	
	private void validateElement(Element element) {
        if (! (NameSpaces.SAML2ASSERTION_SCHEMA.equals(element.getNamespaceURI())
                && SAMLTags.ASSERTION.equals(element.getLocalName()))) {
            throw new IllegalArgumentException("Element is not a SAML assertion");
        }
    }

    private void validateSchema(Node node) {
        try {
            final Validator validator = getSchema().newValidator();
            validator.validate(new DOMSource(node));
        } catch (SAXException | IOException e) {
            throw new ModelBuildException("Error validating OIOSAMLAssertion. " + e.getMessage(), e);
        }
    }
    
    private static synchronized Schema getSchema() throws SAXException {
        if (schema == null) {
            schema = SchemaUtil.loadSchema("/oiosaml/standard-saml.xsd");
        }
        return schema;
    }
    
    
    /**
     * Checks the signature on the <code>OIOSAMLAssertion</code>.
     *
     * @param vault
     *            The CredentialVault containing trusted certificates used to check trust for the <code>OIOSAMLAssertion</code>.
     *
     * @throws dk.sosi.seal.model.ModelException
     *             Thrown if the signature on the <code>OIOSAMLAssertion</code> is invalid.
     */
    public void validateSignatureAndTrust(CredentialVault vault) {
        List<Element> signatureElements = XmlUtil.getChildElementsNS(dom, NameSpaces.DSIG_SCHEMA, DSTags.SIGNATURE);
        if (signatureElements.size() == 0) {
            throw new ModelException("OIOSAMLAssertion is not signed");
        }
        Element signatureElement = signatureElements.get(0);
        List<Element> referencedSignedElements = SignatureUtil.dereferenceSignedElements(signatureElement);
        if (! referencedSignedElements.contains(dom)) {
            throw new ModelException("OIOSAMLAssertion element is not referenced by contained signature");
        }
        if(! SignatureUtil.validate(signatureElement, null, vault, true)) {
            throw new ModelException("Signature on OIOSAMLAssertion is invalid");
        }
    }
    
    /**
     * Gets the signing certificate
     *
     * @return Returns the certificate contained in assertions XML signature or <code>null</code> if the assertion was not signed
     */
    public X509Certificate getSigningCertificate() {
        Element certificateElm = getTag(SAMLTags.assertion, DSTags.signature, DSTags.keyInfo, DSTags.x509Data, DSTags.x509Certificate);
        if (certificateElm == null) {
            return null;
        } else {
            return CertificateParser.asCertificate(XmlUtil.fromBase64(certificateElm.getTextContent()));
        }
    }
    
    public String getSubjectNameID() {
        Element nameID = getTag(SAMLTags.assertion, SAMLTags.subject, SAMLTags.nameID);
        if (nameID == null) {
            return null;
        }
        return nameID.getTextContent().trim();
    }

    public String getSubjectNameIDFormat() {
        Element nameID = getTag(SAMLTags.assertion, SAMLTags.subject, SAMLTags.nameID);
        if (nameID == null) {
            return null;
        }
        return nameID.getAttribute(SAMLAttributes.FORMAT);
    }
    
    public String getRecipient() {
        Element subjectConfirmationData = getTag(SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData);
        if (subjectConfirmationData == null) {
            return null;
        }
        return subjectConfirmationData.getAttribute(SAMLAttributes.RECIPIENT);
    }

    public Date getDeliveryNotOnOrAfter() {
        Element subjectConfirmationData = getTag(SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData);
        if (subjectConfirmationData == null) {
            return null;
        }
        return convertToDate(subjectConfirmationData, SAMLAttributes.NOT_ON_OR_AFTER);
    }
}
