/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.BasicPrivileges12Tags;
import dk.sosi.seal.model.constants.BasicPrivilegesTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

import static dk.sosi.seal.xml.XmlUtil.*;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.Collections.*;

public class BasicPrivileges {
	public enum Profile {
		V10, V12;
	}

	public static class Constraint {
		private final String name, value;

		public Constraint(final String name, final String value) {
			this.name = name;
			this.value = value;
		}

		public String getName() {
			return name;
		}

		public String getValue() {
			return value;
		}

		@Override
		public String toString() {
			return "{name=" + name + ", value=" + value + "}";
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || getClass() != o.getClass()) return false;
			Constraint other = (Constraint) o;
			return Objects.equals(name, other.name) && Objects.equals(value, other.value);
		}

		@Override
		public int hashCode() { return Objects.hash(name, value); }
	}

	public static BasicPrivileges decode(String base64Encoded) {
		return new BasicPrivileges(toElement(base64Encoded));
	}
	public static BasicPrivileges decode(String base64Encoded, Profile profile) {
		return new BasicPrivileges(toElement(base64Encoded), profile);
	}

	private static Element toElement(String base64Encoded) {
		String decodedString = new String(fromBase64(base64Encoded), UTF_8);
		return readXml(new Properties(), decodedString, false).getDocumentElement();
	}

	protected final Map<String, List<String>> privileges = new HashMap<>();
	protected final Map<String, List<Constraint>> constraints = new HashMap<>();

	public final Profile profile;
	
	public BasicPrivileges(Element root) {
		Profile p = Profile.V12;
		try { setup(root, p); }
		catch (ModelException e) { p = Profile.V10; setup(root, p); }
		profile = p;
	}

	public BasicPrivileges(Element root, Profile p) {
		this.profile = p;
		setup(root, p);
	}

	private void setup(Element root, Profile p) {
		if (p == Profile.V12) { validate(root); }
		parse(root, p);
	}

	private static void validate(Node privilegesXml) {
		try {
			Schema schema = SchemaUtil.loadSchema("/basic-privileges-1.2.xsd");
			schema.newValidator().validate(new DOMSource(privilegesXml));
		} catch (SAXException | IOException e) {
			throw new ModelException("Error validating basic privileges: " + e.getMessage(), e);
		}
	}

	protected void parse(Element root, Profile p) {
		/* root=PrivilegeList */
		/* PrivilegeGroup (0..*) */
		List<Element> privileGroups = getProfileGroups(root, p);

		for (Element group : privileGroups) {
			/* PrivilegeGroup Scope='?' */
			String scopeAtt =  group.getAttributeNS(null, BasicPrivilegesTags.SCOPE_ATT);
			/* Privilege (0..*) */
			createPrivileges(group, scopeAtt, p);

			if (p == Profile.V12) { createConstraints(group, scopeAtt); }
		}
	}

	private static List<Element> getProfileGroups(Element e, Profile p) {
		return p == Profile.V12
				? getChildElements(e, BasicPrivileges12Tags.PRIVILEGE_GROUP.getTagName())
				: getChildElementsNS(e, NameSpaces.OIO_BASIC_PRIVILEGES_PROFILE,
											 BasicPrivilegesTags.PRIVILEGE_GROUP.getTagName());
	}

	private void createPrivileges(Element group, String scopeAtt, Profile p) {
		List<String> privilegeValues = getPrivileges(group, p).stream()
				.map(XmlUtil::getTextNodeValue)
				.collect(Collectors.toList());
		privileges.put(scopeAtt, unmodifiableList(privilegeValues));
	}

	private static List<Element> getPrivileges(Element group, Profile profile) {
		return profile == Profile.V12
				? getChildElements(group, BasicPrivileges12Tags.PRIVILEGE.getTagName())
				: getChildElementsNS(group, NameSpaces.OIO_BASIC_PRIVILEGES_PROFILE,
									 BasicPrivilegesTags.PRIVILEGE.getTagName());
	}

	private void createConstraints(Element group, String scopeAtt) {
		List<Element> constraintElms = getChildElements(group, BasicPrivileges12Tags.CONSTRAINT.getTagName());
		List<Constraint> constraintValues = new ArrayList<>();
		for (Element constraintElm : constraintElms) {
			final String name = constraintElm.getAttributeNS(null, BasicPrivileges12Tags.NAME_ATT);
			final String value = getTextNodeValue(constraintElm);
			constraintValues.add(new Constraint(name, value));
		}
		constraints.put(scopeAtt, unmodifiableList(constraintValues));
	}

	public Set<String> getScopes() {
		return privileges.keySet();
	}

	public List<String> getPrivileges(String scope) {
		return privileges.get(scope);
	}

	public List<Constraint> getConstraints(String scope) {
		return constraints.get(scope);
	}
}