/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import dk.sosi.seal.model.BasicPrivileges.Constraint;
import dk.sosi.seal.model.BasicPrivileges.Profile;
import dk.sosi.seal.model.constants.BasicPrivileges12Tags;
import dk.sosi.seal.model.constants.BasicPrivilegesTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.dombuilders.AbstractDOMBuilder;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class BasicPrivilegesDOMBuilder extends AbstractDOMBuilder<Document> {

	private Map<String, List<String>> privileges = new HashMap<String, List<String>>();
	/* Constraints only used with Profile 1.2 */
	private Map<String, List<Constraint>> constraints = new HashMap<String, List<Constraint>>();

	private final Profile profile;

	public BasicPrivilegesDOMBuilder() {
		this(Profile.V10);
	}

	public BasicPrivilegesDOMBuilder(Profile profile) {
		this.profile = profile;
	}

	protected BasicPrivilegesDOMBuilder(BasicPrivileges basicPrivileges) {
		this(basicPrivileges.profile);

        this.privileges = basicPrivileges.privileges;
        if (profile == Profile.V12) { this.constraints = basicPrivileges.constraints; }
    }

	public BasicPrivilegesDOMBuilder addPrivilege(String scope, String privilege) {
		List<String> privilegesByScope = privileges.get(scope);
		if (privilegesByScope == null) {
			privilegesByScope = new ArrayList<String>();
			privileges.put(scope, privilegesByScope);
		}
		privilegesByScope.add(privilege);
		return this;
	}

	public BasicPrivilegesDOMBuilder addConstraint(String scope, Constraint constraint) {
		if (profile == Profile.V12) {
			List<Constraint> constraintsByScope = constraints.get(scope);
			if (constraintsByScope == null) {
				constraintsByScope = new ArrayList<Constraint>();
				constraints.put(scope, constraintsByScope);
			}
			constraintsByScope.add(constraint);
		}
		return this;
	}

	@Override
	public Document build() throws ModelException {
		return createDocument();
	}

	@Override
	protected void validateBeforeBuild() throws ModelException {
		if (privileges.isEmpty()) {
			throw new ModelException("No privileges defined");
		}
		for (String scope : privileges.keySet()) {
			if (scope == null) {
				throw new ModelException("PrivilegeGroup with missing scope");
			}
			// list cannot be empty due to constraint by API (#addPrivilege method)
			for (String privilege : privileges.get(scope)) {
				if (privilege == null) {
					throw new ModelException("Blank privilege for scope " + scope);
				}
			}
		}

		if (profile == Profile.V12) {
			for (String scope : constraints.keySet()) {
				if (scope == null) {
					throw new ModelException("Constraint missing scope");
				} else if (!privileges.containsKey(scope)) {
					throw new ModelException("Constraint scope without privileges. Scope=" + scope);
				}
				for (Constraint constraint : constraints.get(scope)) {
					if (constraint == null) {
						throw new ModelException("Blank constraint. Scope " + scope);
					} else if (constraint.getName() == null) {
						throw new ModelException("Constraint missing name. Scope " + scope);
					} else if (constraint.getValue() == null) {
						throw new ModelException("Constraint missing value. Scope " + scope);
					}
				}
			}
		}
	}

	@Override
	protected Element createRoot(Document doc) {
		return profile == Profile.V12 ? createElement(BasicPrivileges12Tags.PRIVILEGE_LIST) : createElement(BasicPrivilegesTags.PRIVILEGE_LIST);
	}

	@Override
	protected void addRootAttributes(Element root) {
		if(profile == Profile.V12) {
			addNS(root, NameSpaces.NS_BPP, NameSpaces.OIO_BASIC_PRIVILEGES_PROFILE_1_2);
		} else {
			addNS(root, NameSpaces.NS_BPP, NameSpaces.OIO_BASIC_PRIVILEGES_PROFILE);
		}
	}

	@Override
	protected void appendToRoot(Document localDoc, Element privilegeListElm) {
		for (String scope : privileges.keySet()) {
			/* PrivilegeGroup Scope="?" */
			Element groupElm = profile == Profile.V12 ? createElement(BasicPrivileges12Tags.PRIVILEGE_GROUP) : createElement(BasicPrivilegesTags.PRIVILEGE_GROUP);
			groupElm.setAttributeNS(null, BasicPrivilegesTags.SCOPE_ATT, scope);
			
			appendPrivileges(scope, groupElm);
			
			if(profile == Profile.V12) {
				appendConstraints(scope, groupElm);
			}
			privilegeListElm.appendChild(groupElm);
		}
	}

	private void appendPrivileges(String scope, Element groupElm) {
		for (String privilege : privileges.get(scope)) {
			Element privilegeElm = profile == Profile.V12 ? createElement(BasicPrivileges12Tags.PRIVILEGE) : createElement(BasicPrivilegesTags.PRIVILEGE);
			privilegeElm.setTextContent(privilege);
			groupElm.appendChild(privilegeElm);
		}

	}

	private void appendConstraints(String scope, Element groupElm) {
		for (Constraint constraint : nullCheck(constraints.get(scope))) {
			if (constraint == null || constraint.getName() == null) {
				continue;
			}
			Element constraintElm = createElement(BasicPrivileges12Tags.CONSTRAINT);
			constraintElm.setAttributeNS(null, BasicPrivileges12Tags.NAME_ATT, constraint.getName());
			constraintElm.setTextContent(constraint.getValue());
			groupElm.appendChild(constraintElm);
		}
	}

	private <E> Iterable<E> nullCheck(Iterable<E> iterable) {
		return (iterable == null) ? Collections.emptyList() : iterable;
	}
}
