/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import java.util.Date;

import org.w3c.dom.Element;

import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;

public class CitizenIdentityToken extends IdentityToken {
    /**
     * Constructor for the <code>IdentityToken</code> class.
     * 
     * @param dom
     *            DOM representation of the <code>IdentityToken</code> object.
     */
    public CitizenIdentityToken(Element dom) {
        super(dom);
    }

    /**
     * Constructor for the <code>IdentityToken</code> class.
     * 
     * @param dom
     *            DOM representation of the <code>IdentityToken</code> object.
     */
    protected CitizenIdentityToken(Element dom, boolean validateSchema) {
        super(dom, validateSchema);
    }

    @Override
    public String getCommonName() {
        return getAttribute(OIOSAMLAttributes.COMMON_NAME);
    }

    @Override
    public String getEmail() {
        return getAttribute(OIOSAMLAttributes.EMAIL);
    }

    @Override
    public String getSurName() {
        return getAttribute(OIOSAMLAttributes.SURNAME);
    }

    @Override
    public Date getUserAuthenticationInstant() {
        final Element auth = getTag(SAMLTags.assertion, SAMLTags.authnStatement);
        return auth != null ? convertToDate(auth, SAMLAttributes.AUTHN_INSTANT) : null;
    }

    protected Date getDeliveryNotOnOrAfter() {
        final Element subjectConfirmationData = getTag(SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData);
        if(subjectConfirmationData == null) {
            return null;
        }

        return convertToDate(subjectConfirmationData, "NotOnOrAfter");
    }

    protected String getRecipient() {
        return safeGetAttribute(SAMLAttributes.RECIPIENT, SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData);
    }

    protected String getSubjectNameID() {
        return safeGetTagTextContent(SAMLTags.assertion, SAMLTags.subject, SAMLTags.nameID);
    }

    protected String getSubjectNameIDFormat() {
        return safeGetAttribute(SAMLAttributes.FORMAT, SAMLTags.assertion, SAMLTags.subject, SAMLTags.nameID);
    }

    public byte[] getToken() {
        Element holderOfKeyCertificateElm = getHolderOfKeyCertificate();
        return XmlUtil.fromBase64(holderOfKeyCertificateElm.getTextContent());
    }

    private Element getHolderOfKeyCertificate() {
        final Element subjectConfirmation = getTag(SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation);

        final String method = subjectConfirmation != null ? subjectConfirmation.getAttribute("Method") : null;

        final boolean isHolderOfKeyMethod = SAMLValues.CONFIRMATION_METHOD_HOLDER_OF_KEY.equals(method);

        Element certificate = null;
        if(isHolderOfKeyMethod) {
            Element keyInfoElm = getTag(subjectConfirmation, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData, DSTags.keyInfo);

            if(TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.x509Data, DSTags.x509Certificate) != null) {
                certificate = TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.x509Data, DSTags.x509Certificate);
            } else if(TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.keyName) != null) {
                String keyNameVal = XmlUtil.getTextNodeValue(TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.keyName));
                Element elementByIdAndTagNameNS = XmlUtil.getElementByIdAndTagNameNS(DSTags.keyInfo.getTagName(), DSTags.keyInfo.getNS(), keyNameVal, dom.getOwnerDocument());
                certificate = TagUtil.getFirstChildElementNS(elementByIdAndTagNameNS, DSTags.x509Data, DSTags.x509Certificate);

            }
        }

        if (certificate == null) {
            throw new ModelBuildException("Could not find holder-of-key certificate element");
        }
        return certificate;
    }
}
