/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import java.text.ParseException;
import java.util.Date;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import dk.sosi.seal.model.SignatureConfiguration.KeyInfoType;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.constants.WSUTags;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;

/**
 */
public abstract class LibertyRequest extends AbstractDOMInfoExtractor {
	
	private final LibertySignatureValidator validator;

	public LibertyRequest(Federation federation, Node node) {
		super(node);
		
		final IdentityToken identityToken = getIdentityToken();
		identityToken.validateSignature(federation);
		identityToken.validateTimestamp();
		
		validator = validateSignature(new LibertySignatureValidator(federation, super.dom));	
	}

	protected LibertySignatureValidator validateSignature(LibertySignatureValidator validator) {
		setValidatorRequirements(validator);
		
		validator.validateSignature();
		validator.validateAndTrustHolderOfKeyRelation();
		
		return validator;
	}

	protected LibertySignatureValidator setValidatorRequirements(LibertySignatureValidator validator) {
		validator.requireWSAddressingAction();
		validator.requireWSSecurityUtilityTimestamp();
		validator.requireFrameworkHeader();
//		TODO: s.dk laver i øjeblikket requests uden reference til SecurityTokenReference, men i stedet direkte til Assertion elementet. Det skal undersøges nærmere.
//		validator.requireReferenceToIdentityToken();

		return validator;
	}

	public CitizenIdentityToken getIdentityToken() {
		final Element assertion = getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security, SAMLTags.assertion);
		if(assertion == null) {
			throw new ModelBuildException("Could not find SAML assertion element");
		}
		return new CitizenIdentityToken(assertion);
	}

	public Date getCreatedTimestamp() {
		final Element timestamp = validator.getTimestampElement();
		
		if(timestamp != null) {
			Element createdTimestamp = TagUtil.getFirstChildElementNS(timestamp, WSUTags.created);
			try {
				return XmlUtil.fromXMLTimeStamp(createdTimestamp.getTextContent());
			} catch (ParseException e) {
				throw new ModelBuildException("Could not parse wsu:Created element", e);
			}
		}
		
		return null;
	}

	public String getMessageID() {
		Element elm = validator.getWSAddressingMessageIDElement();
		return elm != null ? elm.getTextContent() : null;
	}
	
	public String getAction() {
		Element elm = validator.getWSAddressingActionElement();
		return elm != null ? elm.getTextContent() : null;
	}
	
	public String getTo() {
		Element elm = validator.getWSAddressingToElement();
		return elm != null ? elm.getTextContent() : null;
	}
	
	public String getRelatesTo() {
		Element elm = validator.getWSAddressingRelatesToElement();
		return elm != null ? elm.getTextContent() : null;
	}
	
	public KeyInfoType getKeyInfoType() {
		return validator.getKeyInfoType();
	}
}
