/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import java.security.cert.X509Certificate;
import java.text.ParseException;
import java.util.Date;

import org.apache.ws.security.WSDocInfo;
import org.apache.ws.security.WSDocInfoStore;
import org.apache.ws.security.WSSecurityException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import dk.sosi.seal.model.SignatureConfiguration.KeyInfoType;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.WSATags;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.constants.WSUTags;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.transform.internal.SecurityTokenReference;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

/**
 */
public class LibertyResponse extends AbstractDOMInfoExtractor {
    public LibertyResponse(Node node) {
        super(node);
    }

    public void validateSignatureAndTrust(CredentialVault vault) {
        LibertySignatureValidator validator = createLibertySignatureValidator(vault, null, super.dom);
        validateSignatureAndTrust(validator);
    }

    protected void validateSignatureAndTrust(Federation federation) {
        LibertySignatureValidator validator = createLibertySignatureValidator(null, federation, super.dom);
        validateSignatureAndTrust(validator);
    }

    protected void validateSignatureAndTrust(LibertySignatureValidator validator) {
        setValidatorRequirements(validator).validateSignatureAndTrust();
    }

    protected LibertySignatureValidator setValidatorRequirements(LibertySignatureValidator validator) {
        validator.requireWSAddressingRelatesTo();
        validator.requireWSAddressingAction();
        validator.requireWSSecurityUtilityTimestamp();
        validator.requireFrameworkHeader();

        return validator;
    }

    protected LibertySignatureValidator createLibertySignatureValidator(CredentialVault vault, Federation federation, Element dom) {
        if (vault != null) {
            return new LibertySignatureValidator(vault, super.dom);
        } else if (federation != null) {
            return new LibertySignatureValidator(federation, super.dom);
        } else {
            throw new ModelException("Cannot create LibertySignatureValidator without either CredentialVault or Federation");
        }
    }

    public Date getCreatedTimestamp() {
        final Element createdTimestamp = getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security, WSUTags.timestamp, WSUTags.created);
        if (createdTimestamp == null) {
            throw new ModelBuildException("Could not find wsu:Created element");
        }
        try {
            return XmlUtil.fromXMLTimeStamp(createdTimestamp.getTextContent());
        } catch (ParseException e) {
            throw new ModelBuildException("Could not parse wsu:Created element", e);
        }
    }

    public String getMessageID() {
        final Element messageID = getTag(SOAPTags.envelope, SOAPTags.header, WSATags.messageID);
        if (messageID == null) {
            throw new ModelBuildException("Could not find " + NameSpaces.WSA_1_0_SCHEMA + "#MessageID element");
        }
        return messageID.getTextContent();
    }

    public String getAction() {
        final Element elm = getTag(SOAPTags.envelope, SOAPTags.header, WSATags.action);
        return elm != null ? elm.getTextContent() : null;
    }

    public String getTo() {
        final Element elm = getTag(SOAPTags.envelope, SOAPTags.header, WSATags.to);
        return elm != null ? elm.getTextContent() : null;
    }

    public String getRelatesTo() {
        final Element elm = getTag(SOAPTags.envelope, SOAPTags.header, WSATags.relatesTo);
        return elm != null ? elm.getTextContent() : null;
    }

    public X509Certificate getSigningCertificate() {
        X509Certificate result = null;

        Element keyInfoElm = getKeyInfoElement();

        Element certificateElm = null;
        final KeyInfoType keyInfoType = getKeyInfoType();

        if(keyInfoType == KeyInfoType.XMLDSIG_X509DATA) {
            certificateElm = getTag(keyInfoElm,  DSTags.keyInfo, DSTags.x509Data, DSTags.x509Certificate);
        } else if(keyInfoType == KeyInfoType.WSS_KEYIDENTIFIER || keyInfoType == KeyInfoType.WSS_REFERENCE) {
            Document ownerDocument = dom.getOwnerDocument();
            WSDocInfo wsDocInfo = new WSDocInfo(ownerDocument);
            try {
                WSDocInfoStore.store(wsDocInfo);
                //Using SecurityTokenReference to lookup Certificate using getTokenElement    
                SecurityTokenReference securityTokenReference = new SecurityTokenReference(getTag(keyInfoElm, DSTags.keyInfo, WSSETags.securityTokenReference));

                if(keyInfoType == KeyInfoType.WSS_REFERENCE) {
                    certificateElm = securityTokenReference.getTokenElement(ownerDocument, wsDocInfo, null);
                } else if(keyInfoType == KeyInfoType.WSS_KEYIDENTIFIER) {
                    Element assertionElm = securityTokenReference.getKeyIdentifierTokenElement(ownerDocument, wsDocInfo, null);
                    certificateElm = getTag(assertionElm, SAMLTags.assertion, DSTags.signature, DSTags.keyInfo, DSTags.x509Data, DSTags.x509Certificate);
                }

            } catch (WSSecurityException e) {
                throw new ModelException("Unable to extract Certificate from SecurityTokenReference.", e);
            } finally {
                WSDocInfoStore.delete(wsDocInfo);
            }
        }

        if(certificateElm != null) {
            result = CertificateParser.asCertificate(XmlUtil.fromBase64(certificateElm.getTextContent())); 
        }

        return result;
    }



    public KeyInfoType getKeyInfoType() {
        Element keyInfoElement = getKeyInfoElement();
        return KeyInfoType.fromKeyInfoElement(keyInfoElement);
    }


    private Element getKeyInfoElement() {
        return getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security, DSTags.signature, DSTags.keyInfo);
    }
}
