/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import java.security.cert.X509Certificate;
import java.util.List;

import org.apache.ws.security.WSDocInfo;
import org.apache.ws.security.WSDocInfoStore;
import org.apache.ws.security.WSSecurityException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import dk.sosi.seal.model.SignatureConfiguration.KeyInfoType;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.LibertyTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.Tag;
import dk.sosi.seal.model.constants.WSATags;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.constants.WSUTags;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.pki.CertificateStatus;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.transform.internal.SecurityTokenReference;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;

/**
 */
public class LibertySignatureValidator extends AbstractDOMInfoExtractor {

	private CredentialVault vault;
	private Federation federation;

	private boolean requireFrameworkHeader;
	private boolean requireWSAddressingRelatesTo;
	private boolean requireReferenceToIdentityToken;
	private boolean requireWSAddressingAction;
	private boolean requireWSSecurityUtilityTimestamp;

	private KeyInfoType identifiedKeyInfoType;

	public LibertySignatureValidator(CredentialVault vault, Element dom) {
		super(dom);
		this.vault = vault;
	}

	public LibertySignatureValidator(Federation federation, Element dom) {
		super(dom);
		this.federation = federation;
	}

	public LibertySignatureValidator requireFrameworkHeader() {
		requireFrameworkHeader = true;
		return this;
	}

	public LibertySignatureValidator requireWSAddressingRelatesTo() {
		requireWSAddressingRelatesTo = true;
		return this;
	}

	public LibertySignatureValidator requireWSAddressingAction() {
		requireWSAddressingAction = true;
		return this;
	}

	public LibertySignatureValidator requireReferenceToIdentityToken() {
		requireReferenceToIdentityToken = true;
		return this;
	}

	public LibertySignatureValidator requireWSSecurityUtilityTimestamp() {
		requireWSSecurityUtilityTimestamp = true;
		return this;
	}

	public LibertySignatureValidator validateSignature() throws ModelBuildException {
		internalValidate(false);
		return this;
	}

	public LibertySignatureValidator validateSignatureAndTrust() throws ModelBuildException {
		internalValidate(true);
		return this;
	}

	public KeyInfoType getKeyInfoType() {
		if (identifiedKeyInfoType == null) {
			Element keyInfoElm = getKeyInfo();
			identifiedKeyInfoType = KeyInfoType.fromKeyInfoElement(keyInfoElm);
		}
		return identifiedKeyInfoType;
	}

	private void internalValidate(boolean checkTrust) {
		final Element signatureElement = getLibertySignatureElement();
		if (!SignatureUtil.validate(signatureElement, federation, vault, checkTrust)) {
			throw new ModelBuildException("Liberty signature could not be validated");
		}

		final List<Element> dereferencedSignedElements = SignatureUtil.dereferenceSignedElements(signatureElement);

		/* Required */
		final Element messageID = getWSAddressingMessageIDElement();
		validateIsReferenced(messageID, dereferencedSignedElements);

		/* If required call - requireWSAddressingRelatesTo() */
		final Element relatesTo = getWSAddressingRelatesToElement(requireWSAddressingRelatesTo);
		if (requireWSAddressingRelatesTo || relatesTo != null) {
			validateIsReferenced(relatesTo, dereferencedSignedElements, WSATags.relatesTo);
		}

		/* If required call - requireWSAddressingAction() */
		final Element action = getWSAddressingActionElement(requireWSAddressingAction);
		if (requireWSAddressingAction || action != null) {
			validateIsReferenced(action, dereferencedSignedElements, WSATags.action);
		}

		final Element timestamp = getTimestampElement();
		if (requireWSSecurityUtilityTimestamp || timestamp != null) {
			validateIsReferenced(timestamp, dereferencedSignedElements);
		}

		final Element framework = getLibertyFrameworkElement(requireFrameworkHeader);
		if (requireFrameworkHeader || framework != null) {
			validateIsReferenced(framework, dereferencedSignedElements);
		}

		final Element bodyElement = getBodyElement();
		validateIsReferenced(bodyElement, dereferencedSignedElements);

		final Element to = getWSAddressingToElement();
		if (to != null) {
			validateIsReferenced(to, dereferencedSignedElements);
		}

		if (requireReferenceToIdentityToken) {
			final Element securityElement = getSecurityElement();
			final Element securityTokenReference = getSecurityTokenReferenceElement(securityElement);
			validateIsReferenced(securityTokenReference, dereferencedSignedElements);
			validateReferenceToIdentityToken(securityTokenReference);
		}

		if (getSignatureCertificate() == null) {
			throw new ModelException("Could not validate Signing certificate");
		}
	}

	private void validateReferenceToIdentityToken(Element securityTokenReference) {
		final Element keyIdentifier = XmlUtil.getFirstChildElementNS(securityTokenReference, NameSpaces.WSSE_SCHEMA, WSSETags.KEY_IDENTIFIER);

		if (keyIdentifier == null) {
			throw new ModelBuildException("Could not find KeyIdentifier element in SecurityTokenReference");
		}
		final Node token = XmlUtil.getElementByIdExtended(dom, keyIdentifier.getTextContent());

		if (!getIdentityTokenElement().equals(token)) {
			throw new ModelBuildException("SecurityTokenReference is not referencing IdentityToken");
		}
	}

	protected Element getIdentityTokenElement() {
		final Element assertion = getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security, SAMLTags.assertion);
		if (assertion == null) {
			throw new ModelBuildException("Could not find Identity Token element (Assertion)");
		}
		return assertion;
	}

	protected void validateIsReferenced(Element element, List<Element> references) {
		if (!references.contains(element)) {
			throwMissingLibertySignatureException(element.getNamespaceURI(), element.getLocalName());
		}
	}

	protected void validateIsReferenced(Element element, List<Element> references, Tag tag) {
		if (!references.contains(element)) {
			if (element != null) {
				throwMissingLibertySignatureException(element.getNamespaceURI(), element.getLocalName());
			} else {
				throwMissingLibertySignatureException(tag.getNS(), tag.getTagName());
			}
		}
	}

	private void throwMissingLibertySignatureException(String namespaceURI, String localName) {
		throw new ModelBuildException("Missing Liberty signature on element " + namespaceURI + "#" + localName);
	}

	private Element getLibertySignatureElement() {
		final Element signature = getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security, DSTags.signature);
		if (signature == null) {
			throw new ModelBuildException("Could not find Liberty signature element");
		}
		return signature;
	}

	protected Element getWSAddressingMessageIDElement() {
		final Element messageID = getTag(SOAPTags.envelope, SOAPTags.header, WSATags.messageID);
		if (messageID == null) {
			throw new ModelBuildException("Could not find WS-Addressing 1.0 MessageID element");
		}
		return messageID;
	}

	protected Element getWSAddressingActionElement() {
		return getWSAddressingActionElement(requireWSAddressingAction);
	}

	protected Element getWSAddressingActionElement(boolean required) {
		final Element action = getTag(SOAPTags.envelope, SOAPTags.header, WSATags.action);
		if (required && action == null) {
			throw new ModelBuildException("Could not find WS-Addressing 1.0 Action element");
		}
		return action;
	}

	protected Element getWSAddressingToElement() {
		return getTag(SOAPTags.envelope, SOAPTags.header, WSATags.to);
	}

	protected Element getWSAddressingRelatesToElement() {
		return getWSAddressingRelatesToElement(requireWSAddressingRelatesTo);
	}

	protected Element getWSAddressingRelatesToElement(boolean required) {
		final Element relatesTo = getTag(SOAPTags.envelope, SOAPTags.header, WSATags.relatesTo);
		if (required && relatesTo == null) {
			throw new ModelBuildException("Could not find WS-Addressing 1.0 RelatesTo element");
		}
		return relatesTo;
	}

	private Element getLibertyFrameworkElement(boolean required) {
		final Element framework = getTag(SOAPTags.envelope, SOAPTags.header, LibertyTags.framework);
		if (required && framework == null) {
			throw new ModelBuildException("Could not find Liberty Framework element");
		}
		return framework;
	}

	protected Element getTimestampElement() {
		return getTimestampElement(requireWSSecurityUtilityTimestamp);
	}

	protected Element getTimestampElement(boolean required) {
		final Element timestamp = getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security, WSUTags.timestamp);
		if (required && timestamp == null) {
			throw new ModelBuildException("Could not find Timestamp element");
		}
		return timestamp;
	}

	private Element getSecurityTokenReferenceElement(Element root) {

		final Element securityTokenReference = getDescendant(root, WSSETags.securityTokenReference);

		if (securityTokenReference == null) {
			throw new ModelBuildException("Could not find SecurityTokenReference element");
		}
		return securityTokenReference;
	}

	private Element getBodyElement() {
		final Element body = getTag(SOAPTags.envelope, SOAPTags.body);
		if (body == null) {
			throw new ModelBuildException("Could not find Body element");
		}
		return body;
	}

	private Element getSecurityElement() {
		final Element security = getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security);
		if (security == null) {
			throw new ModelBuildException("Could not find Security element");
		}
		return security;
	}

	private Element getKeyInfo() {
		final Element keyInfo = getTag(getSecurityElement(), WSSETags.security, DSTags.signature, DSTags.keyInfo);
		if (keyInfo == null) {
			throw new ModelBuildException("Could not find KeyInfo element");
		}
		return keyInfo;
	}

	public LibertySignatureValidator validateAndTrustHolderOfKeyRelation() throws ModelBuildException {
		Element signatureCertificateElm = getSignatureCertificate();
		Element holderOfKeyCertificateElm = getHolderOfKeyCertificate();

		byte[] signatureCertificateTextContent = XmlUtil.fromBase64(signatureCertificateElm.getTextContent());
		X509Certificate signatureCertificate = CertificateParser.asCertificate(signatureCertificateTextContent);

		byte[] holderOfKeyCertificateTextContent = XmlUtil.fromBase64(holderOfKeyCertificateElm.getTextContent());
		X509Certificate holderOfKeyCertificate = CertificateParser.asCertificate(holderOfKeyCertificateTextContent);

		if (!signatureCertificate.equals(holderOfKeyCertificate)) {
			throw new ModelBuildException("Signing certificate and holder-of-key certificate relation do not match.");
		}

		CertificateStatus certificateStatus = federation.getCertificateStatus(holderOfKeyCertificate);
		if (!certificateStatus.isValid()) {
			throw new ModelException("The holder-of-key certificate is not trusted!" + (certificateStatus.hasMessage() ? " " + certificateStatus.getMessage() : ""));
		}

		return this;
	}

	private Element getSignatureCertificate() {
		final Element keyInfo = getKeyInfo();
		final KeyInfoType keyInfoType = getKeyInfoType();
		Element certificate = null;
		switch (keyInfoType) {
			case XMLDSIG_X509DATA :
				certificate = getCertificateFromX509DATA(keyInfo);
				break;
			case WSS_REFERENCE :
				certificate = getCertificateFromSTRReference(keyInfo);
				break;
			case WSS_KEYIDENTIFIER :
				certificate = getCertificateFromSTRKeyIdentifier(keyInfo);
				break;
			default :
				certificate = null;
				break;
		}

		if (certificate == null) {
			throw new ModelBuildException("Could not find certificate element");
		}
		return certificate;
	}

	private Element getCertificateFromSTRKeyIdentifier(Element keyInfo) {
		Element certificate = null;
		if (getTag(keyInfo, DSTags.keyInfo, WSSETags.securityTokenReference) != null) {
			Element securityTokenReference = getTag(keyInfo, DSTags.keyInfo, WSSETags.securityTokenReference);
			Document ownerDocument = dom.getOwnerDocument();
			WSDocInfo wsDocInfo = new WSDocInfo(ownerDocument);
			try {
				WSDocInfoStore.store(wsDocInfo);
				Element assertion = new SecurityTokenReference(securityTokenReference).getKeyIdentifierTokenElement(ownerDocument, wsDocInfo, null);
				certificate = getCertificateFromX509DATA(getTag(assertion, SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData, DSTags.keyInfo));
			} catch (WSSecurityException e) {
				throw new ModelBuildException("Could not load certificate element using SecurityTokenReference");
			} finally {
				WSDocInfoStore.delete(wsDocInfo);
			}
		}
		return certificate;
	}

	private Element getCertificateFromSTRReference(Element keyInfo) {
		Element certificate = null;
		if (getTag(keyInfo, DSTags.keyInfo, WSSETags.securityTokenReference) != null) {
			Element securityTokenReference = getTag(keyInfo, DSTags.keyInfo, WSSETags.securityTokenReference);
			Document ownerDocument = dom.getOwnerDocument();
			WSDocInfo wsDocInfo = new WSDocInfo(ownerDocument);
			try {
				WSDocInfoStore.store(wsDocInfo);
				certificate = new SecurityTokenReference(securityTokenReference).getTokenElement(ownerDocument, wsDocInfo, null);
			} catch (WSSecurityException e) {
				throw new ModelBuildException("Could not load certificate element using SecurityTokenReference");
			} finally {
				WSDocInfoStore.delete(wsDocInfo);
			}
		}
		return certificate;
	}

	private Element getCertificateFromX509DATA(Element keyInfo) {
		Element certificate = null;
		if (getTag(keyInfo, DSTags.keyInfo, DSTags.x509Data) != null) {
			certificate = getTag(keyInfo, DSTags.keyInfo, DSTags.x509Data, DSTags.x509Certificate);
		} else {
			String keyNameVal = XmlUtil.getTextNodeValue(TagUtil.getFirstChildElementNS(keyInfo, DSTags.keyName));
			Element elementByIdAndTagNameNS = XmlUtil.getElementByIdAndTagNameNS(DSTags.keyInfo.getTagName(), DSTags.keyInfo.getNS(), keyNameVal, dom.getOwnerDocument());
			certificate = TagUtil.getFirstChildElementNS(elementByIdAndTagNameNS, DSTags.x509Data, DSTags.x509Certificate);
		}
		return certificate;
	}

	private Element getHolderOfKeyCertificate() {
		final Element subjectConfirmation = getTag(SOAPTags.envelope, SOAPTags.header, WSSETags.security, SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation);

		final String method = subjectConfirmation != null ? subjectConfirmation.getAttribute("Method") : null;

		final boolean isHolderOfKeyMethod = SAMLValues.CONFIRMATION_METHOD_HOLDER_OF_KEY.equals(method);

		Element certificate = null;
		if(isHolderOfKeyMethod) {
			Element keyInfoElm = getTag(subjectConfirmation, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData, DSTags.keyInfo);
			
			if(TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.x509Data, DSTags.x509Certificate) != null) {
				certificate = TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.x509Data, DSTags.x509Certificate);
			} else if(TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.keyName) != null) {
				String keyNameVal = XmlUtil.getTextNodeValue(TagUtil.getFirstChildElementNS(keyInfoElm, DSTags.keyName));
				Element elementByIdAndTagNameNS = XmlUtil.getElementByIdAndTagNameNS(DSTags.keyInfo.getTagName(), DSTags.keyInfo.getNS(), keyNameVal, dom.getOwnerDocument());
				certificate = TagUtil.getFirstChildElementNS(elementByIdAndTagNameNS, DSTags.x509Data, DSTags.x509Certificate);
				
			}
			//, DSTags.x509Data, DSTags.x509Certificate) : null;
			getTag(subjectConfirmation, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData, DSTags.keyInfo);
		}

		if (certificate == null) {
			throw new ModelBuildException("Could not find holder-of-key certificate element");
		}
		return certificate;
	}
}
