/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.OIOSAML3Attributes;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.constants.SAMLValues;
import org.w3c.dom.Element;

/**
 * Representation of a OIO3BST assertion used for citizens
 */
public class OIO3BSTCitizenSAMLAssertion extends OIOBSTSAMLAssertion {

	public enum Attribute implements SAMLAttribute {
		// Specifies the version of the OIOSAML profile specification
		SPEC_VER("specVer", OIOSAML3Attributes.SPEC_VERSION, true),
		// Contains the overall level of assuranceof the authentication as defined by the Danish [NIST] standard.
		// The allowed values are 1 to 5.
		ASSURANCE_LEVEL("assuranceLevel", OIOSAML3Attributes.ASSURANCE_LEVEL, true),
		// Contains the Danish CPR number for the identity represented by 10 digits.
		CPR("cpr", OIOSAML3Attributes.CPR, true);

		private final String id, name, nameFormat;
		private final boolean mandatory;

		Attribute(String id, String name, boolean mandatory) {
			this(id, name, SAMLValues.NAME_FORMAT_URI, mandatory);
		}

		Attribute(String id, String name, String nameFormat, boolean mandatory) {
			this.id = id;
			this.name = name;
			this.mandatory = mandatory;
			this.nameFormat = nameFormat;
		}

		public String getId() {
			return id;
		}

		public String getName() {
			return name;
		}

		public String getNameFormat() {
			return nameFormat;
		}

		public boolean isMandatory() {
			return mandatory;
		}
	}

	public OIO3BSTCitizenSAMLAssertion(Element element) {
		super(element);
	}

	public OIO3BSTCitizenSAMLAssertion(Element dom, boolean validate) {
		super(dom, validate);
	}

	@Override
	public Type getType() {
		return Type.OIO3BST_CITIZEN;
	}

	/**
	 * Extract the <code>https://data.gov.dk/concept/core/nsis/loa</code> value from the
	 * token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic" Name="https://data.gov.dk/concept/core/nsis/loa"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;Substantial&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the <code>https://data.gov.dk/concept/core/nsis/loa</code>
	 *         tag.
	 */
	@Override
	public String getAssuranceLevel() {
		return getAttributeValue(Attribute.ASSURANCE_LEVEL);
	}

	/**
	 * Extract the <code>https://data.gov.dk/model/core/eid/cprNumber</code> value from
	 * the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic"
	 *       Name="https://data.gov.dk/model/core/eid/cprNumber"&gt;
	 *       &lt;saml:AttributeValue&gt;2702681273&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/model/core/eid/cprNumber</code> tag.
	 */
	public String getCpr() {
		return getAttributeValue(Attribute.CPR);
	}

	/**
	 * Extract the <code>https://data.gov.dk/model/core/specVersion</code> value from the
	 * DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic" Name="https://data.gov.dk/model/core/specVersion"&gt;
	 *       &lt;saml:AttributeValue&gt;OIO-SAML-3.0&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the <code>https://data.gov.dk/model/core/specVersion</code> tag.
	 */
	public String getSpecVersion() {
		return getAttributeValue(Attribute.SPEC_VER);
	}
}
