/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import org.w3c.dom.Element;

import dk.sosi.seal.model.BasicPrivileges.Profile;
import dk.sosi.seal.model.constants.OIOSAML3Attributes;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;

/**
 * 
 *          Based on
 *          https://digst.dk/media/23391/oiosaml-web-sso-profile-302.pdf
 */
public class OIO3BSTSAMLAssertion extends OIOBSTSAMLAssertion {
	public enum Attribute implements SAMLAttribute {
		// Specifies the version of the OIOSAML profile specification
		SPEC_VER("specVer", OIOSAML3Attributes.SPEC_VERSION, true),
		// Contains the overall level of assuranceof the authentication as defined by
		// the Danish [NSIS] standard.
		// The allowedvalues are ‘Low’, ‘Substantial’ and ‘High’.
		ASSURANCE_LEVEL_NSIS("NSIS assuranceLevel", OIOSAML3Attributes.ASSURANCE_LEVEL, false),
		ASSURANCE_LEVEL_NIST("NIST assuranceLevel", OIOSAMLAttributes.ASSURANCE_LEVEL, false),
		// Contains the Danish CPR number for the identity represented by 10 digits.
		CPR("cpr", OIOSAML3Attributes.CPR, false),
		// Contains a UUID for the professional identity which is shared across all
		// public sector SPs. The identifier is specific to the professional role and is
		// not related to the associated natural person. The UUID MUST follow RFC 4122.
		// This attribute is the successor to the RIDattribute but is globally unique.
		UUID_PROFESSIONAL("professionalUuid", OIOSAML3Attributes.PROFESSIONAL_UUID /*"https://data.gov.dk/model/core/eid/professional/uuid/persistent"*/, false),
		// Contains the CVRnumber (8 digits) of the organization related to the
		// authentication context. Note that a professional may be associated with
		// several organizations but only one organization is allowed per authentication
		// context.
		CVR("cvr", "https://data.gov.dk/model/core/eid/professional/cvr", true),
		// Contains the name of the organization related to the authentication context.
		// Note that a professional may be associated with several organizationscbut
		// only one organization is allowed per authentication context.
		NAME_ORGANIZATION("organizationName", "https://data.gov.dk/model/core/eid/professional/orgName", true),
		// Contains the email address of the identity. In cases there are multiple
		// addresses known this attribute can be multi-valued (i.e. using multiple
		// <AttributeValue> elements).
		EMAIL("email", "https://data.gov.dk/model/core/eid/email", false),
		// Contains analias of the identity. This attribute can be used as a display
		// name selected by the user as an alternative to the above name attributes.
		ALIAS("alias", "https://data.gov.dk/model/core/eid/alias", false),
		// Contains a base64-encoded value describing privileges assigned to the
		// identity (see OIO Basic Privilege Profile specification [OIOBPP] for
		// details).
		PRIVILEGES("privileges", "https://data.gov.dk/model/core/eid/privilegesIntermediate", false);

		private final String id, name;
		private final boolean mandatory;

		private Attribute(String id, String name, boolean mandatory) {
			this.id = id;
			this.name = name;
			this.mandatory = mandatory;
		}

		public String getId() {
			return id;
		}

		public String getName() {
			return name;
		}

		public String getNameFormat() {
			return "urn:oasis:names:tc:SAML:2.0:attrname-format:uri";
		}

		public boolean isMandatory() {
			return mandatory;
		}
	}

	public OIO3BSTSAMLAssertion(Element element) {
		super(element);
	}

	public OIO3BSTSAMLAssertion(Element dom, boolean validate) {
		super(dom, validate);
	}

	/**
	 * Extract the <code>https://data.gov.dk/concept/core/nsis/loa</code> value from
	 * the token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/concept/core/nsis/loa"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;Low/Substantial/High&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/concept/core/nsis/loa</code> tag.
	 */
	@Override
	public String getAssuranceLevel() {
		String attributeValue = getAttributeValue(Attribute.ASSURANCE_LEVEL_NIST);
		if(attributeValue != null) {
			return attributeValue;
		} else {
			attributeValue = getAttributeValue(Attribute.ASSURANCE_LEVEL_NSIS);
			return attributeValue;
		}
		
	}

	/**
	 * Extract the <code>https://data.gov.dk/concept/core/eid/cprNumber</code> value
	 * from the token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/concept/core/eid/cprNumber"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;1234567890&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/concept/core/eid/cprNumber</code> tag.
	 */
	public String getCpr() {
		return getAttributeValue(Attribute.CPR);
	}

	/**
	 * Extract the
	 * <code>https://data.gov.dk/model/core/eid/professional/orgName</code>
	 * (organizationName) value from the token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/model/core/eid/professional/orgName" &gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;Lægehuset på bakken&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/model/core/eid/professional/orgName</code>
	 *         (organizationName) tag.
	 */
	public String getOrganizationName() {
		return getAttributeValue(Attribute.NAME_ORGANIZATION);
	}

	/**
	 * Extract the <code>dk:gov:saml:attribute:CvrNumberIdentifier</code> value from
	 * the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri"
	 *         Name="dk:gov:saml:attribute:CvrNumberIdentifier"&gt;
	 *         &lt;saml:AttributeValue xsi:type="xs:string"&gt;20688092&lt;/saml:AttributeValue&gt;
	 *       &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>dk:gov:saml:attribute:CvrNumberIdentifier</code> tag.
	 */
	public String getCvrNumberIdentifier() {
		return getAttributeValue(Attribute.CVR);
	}

	/**
	 * Extract the <code>https://data.gov.dk/model/core/specVersion</code> value
	 * from the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/model/core/specVersion"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;DK-SAML-3.0&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/model/core/specVersion</code> tag.
	 */
	public String getSpecVersion() {
		return getAttributeValue(Attribute.SPEC_VER);
	}
	
	/**
	 * Extract the <code>https://data.gov.dk/model/core/specVersion</code> value
	 * from the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/model/core/eid/professional/uuid/persistent"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;URN:xyz:123&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/model/core/eid/professional/uuid/persistent</code> tag.
	 */
	@Override
	public String getUuid() {
		return getAttributeValue(Attribute.UUID_PROFESSIONAL);
	}
	
	@Override
	public Type getType() {
		return Type.OIO3BST;
	}

	@Override
	public String getRidNumberIdentifier() {
		return null;
	}

	/**
	 * En (frivillig) angivelse af brugerens 'nationale roller' formatteret som
	 * OIOBPP 1.2 XML struktur og indlejret i Base64 encoded form <br>
	 *
	 * <pre>
	 *     &lt;Attribute Name="https://data.gov.dk/model/core/eid/privilegesIntermediate"
	 *       NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri"&gt;
	 *       &lt;AttributeValue>Bx23z ....&lt;/AttributeValue&gt;
	 *     &lt;/Attribute&gt;
	 * </pre>
	 *
	 * I base64 decoded form <br>
	 *
	 * <pre>
	 *     &lt;bpp:PrivilegeList
	 *     	xmlns:bpp="http://digst.dk/oiosaml/basic_privilege_profile"&gt;
	 *     	&lt;PrivilegeGroup Scope="urn:dk:gov:saml:cvrNumberIdentifier:20301823"&gt;
	 *     	  &lt;Privilege>urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3&lt;/Privilege&gt;
	 *     	&lt;/PrivilegeGroup&gt;
	 *     &lt;/bpp:PrivilegeList&gt;
	 * </pre>
	 *
	 */
	@Override
	public BasicPrivileges getBasicPrivileges() {
		String privileges = getAttributeValue(Attribute.PRIVILEGES);
		return privileges != null ? BasicPrivileges.decode(privileges, Profile.V12) : null;
	}
}
