/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.*;
import dk.sosi.seal.security.HolderOfKey;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Element;

import java.math.BigInteger;
import java.util.Date;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;

import static java.util.Optional.*;

public abstract class OIOBSTSAMLAssertion extends AbstractOIOSAMLAssertion {
    /**
     * Danner rammen for de enums skabt i sub-klasserne til at opbygge
     * <AttributeStatement> strukturen i Assertion
     */
    interface SAMLAttribute {
        String getId();

        String getName();

        String getNameFormat();

        boolean isMandatory();
    }

    /**
     * Typen af OIOBSTSAMLAssertion
     */
    public enum Type {
        OIOH2BST, OIOH3BST, OIO3BST, OIO2BST_LEGACY, OIO2BST_CITIZEN, OIO3BST_CITIZEN;
    }

    public OIOBSTSAMLAssertion(Element dom) {
        this(dom, true);
    }

    public OIOBSTSAMLAssertion(Element dom, boolean validate) {
        super(dom, validate);
    }

    protected String getAttributeValue(SAMLAttribute attr) {
        return getAttributeValue(getNamesSpaces(), dom, attr);
    }


    private static String getAttributeValue(Map<String, String> ns, Element assertion, SAMLAttribute attr) {
        Element elm = getAttributeElement(ns, assertion, attr.getName());

        if (elm == null) {
            if (attr.isMandatory()) {
                throw new MandatoryAttributeException(
                        "Mandatory '" + attr.getId() + "' SAML attribute (" + attr.getName() + ") is missing");
            }
            return null;
        }

        String nameFormat = elm.getAttribute(SAMLAttributes.NAME_FORMAT);
        assert nameFormat != null && nameFormat.equals(attr.getNameFormat());

        return elm.getTextContent().trim();
    }

    /**
     * Get the value of an attribute matching one of the given attribute names, starting from the first.
     */
    public static String getAttributeValue(Element assertionElm, String... attrNames) {
        Map<String, String> namesSpaces = getNamesSpaces(assertionElm);
        for (String attrName : attrNames) {
            Element elm = getAttributeElement(namesSpaces, assertionElm, attrName);
            if (elm != null) {
                return elm.getTextContent().trim();
            }
        }
        return null;
    }

    @Override
    public Date getNotBefore() {
        try {
            return super.getNotBefore();
        } catch (ModelException e) {
            return new Date(0); // NotBefore is not required, but is used to validate timestamp
        }
    }

    public Optional<HolderOfKey> getHolderOfKeyCertificate() {
        final Element subjectConfirmation = getTag(SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation);
        final String method = subjectConfirmation != null ? subjectConfirmation.getAttribute("Method") : null;
        final boolean isHolderOfKeyMethod = SAMLValues.CONFIRMATION_METHOD_HOLDER_OF_KEY.equals(method);
        return isHolderOfKeyMethod
                ? parseCertficateFromSubjectToHok(subjectConfirmation)
                .map(Optional::of) // A 'hack' because Java 8 does not have the 'or' method introduced in Java 9 on Optional.
                .orElseGet(() -> parseIssuerNameAndSerialFromSubjectToHok(subjectConfirmation))
                : empty();
    }

    private Optional<HolderOfKey> parseCertficateFromSubjectToHok(Element subjectConfirmation) {
        return safeGetTag(subjectConfirmation, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData,
                DSTags.keyInfo, DSTags.x509Data, DSTags.x509Certificate)
                .map(parseCertificateElementToHok);
    }

    private final Function<Element, HolderOfKey> parseCertificateElementToHok =
            elm -> new HolderOfKey(CertificateParser.asCertificate(XmlUtil.fromBase64(elm.getTextContent())));

    private Optional<HolderOfKey> parseIssuerNameAndSerialFromSubjectToHok(Element subjectConfirmation) {

        Optional<Element> issuerSerialOpt =
                safeGetTag(subjectConfirmation, SAMLTags.subjectConfirmation, SAMLTags.subjectConfirmationData,
                        DSTags.keyInfo, DSTags.x509Data, DSTags.x509IssuerSerial)
                        .map(Optional::of) // A 'hack' because Java 8 does not have the 'or' method introduced in Java 9 on Optional.
                        .orElse(safeGetTag(subjectConfirmation, SAMLTags.subjectConfirmation,
                                SAMLTags.subjectConfirmationData, DSTags.keyInfo, EncTags.encryptedKey, DSTags.keyInfo,
                                DSTags.x509Data, DSTags.x509IssuerSerial));

        return issuerSerialOpt
                .flatMap(issuerSerial -> getIssuerName(issuerSerial).map(Element::getTextContent)
                .flatMap(issuerName -> getSerialNumber(issuerSerial)
                        .map(Element::getTextContent)
                        .map(BigInteger::new)
                        .map(serialNumber -> new HolderOfKey(issuerName, serialNumber))
                ));
    }

    private Optional<Element> getIssuerName(Element elm) {
        return safeGetTag(elm, DSTags.x509IssuerSerial, DSTags.x509IssuerName);
    }

    private Optional<Element> getSerialNumber(Element elm) {
        return safeGetTag(elm, DSTags.x509IssuerSerial, DSTags.x509SerialNumber);
    }

    // Always required
    public abstract Type getType();

    // Attributes - to be overwritten
    public String getSpecVersion() {
        return null;
    }

    public String getOrganizationName() {
        return null;
    }

    public String getRidNumberIdentifier() {
        return null;
    }

    public String getCpr() {
        return null;
    }

    public String getCvrNumberIdentifier() {
        return null;
    }

    public String getUuid() {
        return null;
    }

    public BasicPrivileges getBasicPrivileges() {
        return null;
    }
}
