/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.WSFAuthAttributes;
import dk.sosi.seal.model.constants.WST14Tags;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.security.HolderOfKey;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.security.PrivateKey;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public abstract class OIOBSTSAMLAssertionRequest extends OIOWSTrustRequest {

    private OIOBSTSAMLAssertion assertion;

    public OIOBSTSAMLAssertionRequest(Document doc) {
        super(doc);
    }

    protected Element getEncryptedAssertionElement() {
        return getTag(getActAsElement(), WST14Tags.actAs, SAMLTags.encryptedAssertion);
    }

    protected Element getAssertionElement() {
        return getTag(getActAsElement(), WST14Tags.actAs, SAMLTags.assertion);
    }

    public boolean isAssertionEncrypted() {
        return getEncryptedAssertionElement() != null;
    }

    public void decryptAssertion(PrivateKey key) {
        if (isAssertionEncrypted()) {
            Element decryptedAssertionElement = EncryptionUtil.decryptAndDetach(getEncryptedAssertionElement(), key);
            assertion = getOIOBSTSAMLAssertion(decryptedAssertionElement);
        } else {
            throw new ModelException(
                    "Assertion is not encrypted and should be fetched without calling decrypt method.");
        }
    }

    public OIOBSTSAMLAssertion getOIOBSTSAMLAssertion(PrivateKey key) {
        decryptAssertion(key);
        return getOIOBSTSAMLAssertion();
    }

    public OIOBSTSAMLAssertion getOIOBSTSAMLAssertion() {
        if (assertion == null) {
            if (isAssertionEncrypted()) {
                throw new ModelException(
                        "Assertion is encrypted. Call method decryptAssertion with private key before fetching " +
                                "Assertion.");
            } else {
                Element assertionElm = getAssertionElement();
                assertion = assertionElm != null ? getOIOBSTSAMLAssertion(assertionElm) : null;
            }
        }
        return assertion;
    }

    public abstract OIOBSTSAMLAssertion getOIOBSTSAMLAssertion(Element assertionElement);

    public void validateHolderOfKeyRelation() throws ModelBuildException {
        OIOBSTSAMLAssertion oiosamlAssertion = getOIOBSTSAMLAssertion();
        if (oiosamlAssertion == null) {
            throw new ModelBuildException("No assertion element found in request.");
        }

        HolderOfKey hok = oiosamlAssertion
                .getHolderOfKeyCertificate()
                .orElseThrow(() -> new ModelBuildException("Could not find holder-of-key certificate"));

        if (!hok.matches(getSigningCertificate())) {
            throw new ModelBuildException("Signing certificate and holder-of-key certificate relation do not match.");
        }
    }
    public Map<String, String> getClaimMap() {
        Map<String, String> result = new HashMap<>();
        List<Element> claims = super.getClaims();
        for (Element claim : claims) {
            String type = claim.getAttribute(WSFAuthAttributes.URI);
            type = result.containsKey(type) ? type + "(duplicate)" : type;
            String value = claim.getTextContent();
            result.put(type, value);
        }
        return result;
    }

    protected String getClaimValueByUri(String uri) {
        return getClaimMap().get(uri);
    }
}
