/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import static dk.sosi.seal.model.constants.OIOSAMLAttributes.ASSURANCE_LEVEL;
import static dk.sosi.seal.model.constants.OIOSAMLAttributes.COMMON_NAME;
import static dk.sosi.seal.model.constants.OIOSAMLAttributes.COMMON_NAME_FRIENDLY;
import static dk.sosi.seal.model.constants.OIOSAMLAttributes.CPR_NUMBER;

import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.model.dombuilders.AbstractSAMLBuilder;
import dk.sosi.seal.pki.DistinguishedName;
import dk.sosi.seal.pki.SignatureProviderFactory;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;

/**
 * Used for building response tokens for the JWT2OIOSaml service in STS
 */
public class OIOBootstrapTokenBuilder extends AbstractSAMLBuilder<OIOBootstrapTokenBuilder, OIOBootstrapToken> {
    private DistinguishedName subjectDN;

    private final CredentialVault signingVault;

    private Node subjectNode;

    private String audienceRestriction;

    private final Instant now = Instant.now();
    private final Date created = Date.from(now);
    private Date validFrom = created;
    private Date expires = Date.from(now.plus(1, ChronoUnit.HOURS));


    private String cpr;

    private X509Certificate holderOfKeyCertificate;


    /**
     */
    public OIOBootstrapTokenBuilder(final CredentialVault signingVault, final String issuer) {
        super();
        super.setIssuer(issuer);
        this.signingVault = signingVault;

        //this(signingVault, formatDN(certInfo.getSubjectDN()));
    }

    public OIOBootstrapTokenBuilder setCpr(String cprNumber) {
        this.cpr = cprNumber;
        return this;
    }

    public OIOBootstrapTokenBuilder setSubjectName(String sn) {
        this.subjectDN = new DistinguishedName(sn);
        return this;
    }

    public OIOBootstrapTokenBuilder setAudienceRestriction(String audienceRestriction) {
        this.audienceRestriction = audienceRestriction;
        return this;
    }

    public OIOBootstrapTokenBuilder expires(Date expires) {
        this.expires = expires;
        return this;
    }

    public OIOBootstrapTokenBuilder validFrom(Date validFrom) {
        this.validFrom = validFrom;
        return this;
    }

    public OIOBootstrapTokenBuilder setHolderOfKeyCertificate(X509Certificate holderOfKeyCertificate) {
        this.holderOfKeyCertificate = holderOfKeyCertificate;
        return this;
    }

    @Override
    public OIOBootstrapToken build() throws ModelException {
        Document document = createDocument();       
        signAssertion(document);
        return new OIOBootstrapToken(document.getDocumentElement());
    }

    private void signAssertion(Document document) {
        SignatureConfiguration signatureConfiguration = new SignatureConfiguration(new String[]{assertionID}, assertionID, IDValues.Id);
        signatureConfiguration.setSignatureSiblingNode(subjectNode);
        SignatureUtil.sign(SignatureProviderFactory.fromCredentialVault(signingVault), document, signatureConfiguration);
    }

    @Override
    protected final void appendToRoot(Document doc, Element root) {
        super.appendToRoot(doc, root);
        subjectNode = createSubject();
        root.appendChild(subjectNode);
        root.appendChild(createConditions());
        root.appendChild(createAuthnStatement());
        root.appendChild(createAttributeStatement());
    }

    private Node createSubject() {
        Element subjectElm = createElement(SAMLTags.subject);

        Element nameIdElm = createElement(SAMLTags.nameID);
        nameIdElm.setAttributeNS(null, SAMLAttributes.FORMAT, SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME);
        nameIdElm.setTextContent(formatDN(subjectDN));
        subjectElm.appendChild(nameIdElm);

        Element subjectConfirmationElm = createElement(SAMLTags.subjectConfirmation);
        subjectConfirmationElm.setAttributeNS(null, SAMLAttributes.METHOD, SAMLValues.CONFIRMATION_METHOD_HOLDER_OF_KEY);
        subjectElm.appendChild(subjectConfirmationElm);

        Element subjectConfirmationDataElm = createElement(SAMLTags.subjectConfirmationData);
        subjectConfirmationElm.appendChild(subjectConfirmationDataElm);

        Element keyInfoElm = createElement(DSTags.keyInfo);
        Element x509DataElm = createElement(DSTags.x509Data);
        keyInfoElm.appendChild(x509DataElm);
        Element x509CertificateElm = createElement(DSTags.x509Certificate);
        x509DataElm.appendChild(x509CertificateElm);
        try {
            String encodedCert = XmlUtil.toBase64(holderOfKeyCertificate.getEncoded());
            x509CertificateElm.setTextContent(encodedCert);
        } catch (CertificateEncodingException e) {
            throw new ModelException("Unable to encode certificate", e);
        }
        subjectConfirmationDataElm.appendChild(keyInfoElm);

        return subjectElm;
    }

    private Node createConditions() {
        Element conditionsElm = createElement(SAMLTags.conditions);
        conditionsElm.setAttributeNS(null, SAMLAttributes.NOT_BEFORE, XmlUtil.getDateFormat(true).format(validFrom));
        conditionsElm.setAttributeNS(null, SAMLAttributes.NOT_ON_OR_AFTER, XmlUtil.getDateFormat(true).format(expires));

        Element audienceRestrictionElm = createElement(SAMLTags.audienceRestriction);
        conditionsElm.appendChild(audienceRestrictionElm);

        Element audienceElm = createElement(SAMLTags.audience);
        audienceElm.setTextContent(audienceRestriction);
        audienceRestrictionElm.appendChild(audienceElm);

        return conditionsElm;
    }

    private Node createAuthnStatement() {
        Element authnStatementElm = createElement(SAMLTags.authnStatement);
        authnStatementElm.setAttributeNS(null, SAMLAttributes.AUTHN_INSTANT, XmlUtil.getDateFormat(true).format(this.created));

        Element authnContextElm = createElement(SAMLTags.authnContext);
        Element authContextClassRefElm = createElement(SAMLTags.authnContextClassRef);
        authContextClassRefElm.setTextContent(SAMLValues.AUTHN_CONTEXT_CLASS_REF);
        authnContextElm.appendChild(authContextClassRefElm);
        authnStatementElm.appendChild(authnContextElm);
        // SessionIndex is mandatory in OIOSAML (to support per session logout), but is meaningless
        // for STS issued assertions
        authnStatementElm.setAttributeNS(null, SAMLAttributes.SESSION_INDEX, assertionID);
        return authnStatementElm;
    }

    private Element createAttributeStatement() {
        Element attributeStatementElm = createElement(SAMLTags.attributeStatement);

        addAttribute(attributeStatementElm, OIOSAMLAttributes.SPEC_VERSION, SAMLValues.DK_SPEC_VERSION);
        addAttribute(attributeStatementElm, CPR_NUMBER, cpr);
        addAttribute(attributeStatementElm, ASSURANCE_LEVEL, "3");
        addAttribute(attributeStatementElm, COMMON_NAME, COMMON_NAME_FRIENDLY, subjectDN.getCommonName());
        return attributeStatementElm;
    }

    private static String formatDN(DistinguishedName name) {
        StringBuilder builder = new StringBuilder("C=");
        builder.append(name.getCountry());
        builder.append(",O=");
        builder.append(name.getOrganization());
        builder.append(",CN=");
        builder.append(name.getCommonName());
        builder.append(",Serial=");
        builder.append(name.getSubjectSerialNumber());
        return builder.toString();
    }
}
