/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import java.security.cert.X509Certificate;

import org.w3c.dom.Element;

import dk.sosi.seal.model.BasicPrivileges.Profile;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.OIOSAML3Attributes;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

/**
 */
public class OIOH2BSTSAMLAssertion extends OIOBSTSAMLAssertion {

	public enum Attribute implements SAMLAttribute {
		// Specifies the version of the OIOSAML profile specification
		SPEC_VER("specVer", OIOSAMLAttributes.SPEC_VERSION, true),
		// Contains the overall level of assuranceof the authentication as defined by
		// the Danish [NSIT] standard.
		// The allowedvalues are 1 to 5.
		ASSURANCE_LEVEL("assuranceLevel", OIOSAMLAttributes.ASSURANCE_LEVEL, true),
		// Contains the Danish CPR number for the identity represented by 10 digits.
		CPR("cpr", OIOSAMLAttributes.CPR_NUMBER, true),
		// Contains a RID for the professional identity which is shared across all
		// public sector SPs.
		RID("rid", OIOSAMLAttributes.RID_NUMBER, false),
		// Contains the CVR number (8 digits) of the organization related to the
		// authentication context. Note that a professional may be associated with
		// several organizations but only one organization is allowed per authentication
		// context.
		CVR("cvr", OIOSAMLAttributes.CVR_NUMBER, true),
		// Contains the name of the organization related to the authentication context.
		// Note that a professional may be associated with several organizationscbut
		// only one organization is allowed per authentication context.
		NAME_ORGANIZATION("organizationName", OIOSAMLAttributes.ORGANIZATION_NAME, true),
		// Ervhvervspersonens global unikke ID, som tildelt i den fællesoffentlige
		// erhvervsadministration
		UUID("uuid", OIOSAML3Attributes.PROFESSIONAL_UUID, SAMLValues.NAME_FORMAT_URI, false),
		// En (frivillig) angivelse af brugerens centralt trustede og lokalt
		// administrerede 'nationale roller' formatteret som OIOBPP 1.2 XML struktur og
		// indlejret i Base64 encoded form
		PRIVILEGES("privileges", OIOSAMLAttributes.PRIVILEGES_INTERMEDIATE, SAMLValues.NAME_FORMAT_URI, false);

		private final String id, name, nameFormat;
		private final boolean mandatory;

		private Attribute(String id, String name, boolean mandatory) {
			this(id, name, SAMLValues.NAME_FORMAT_BASIC, mandatory);
		}

		private Attribute(String id, String name, String nameFormat, boolean mandatory) {
			this.id = id;
			this.name = name;
			this.mandatory = mandatory;
			this.nameFormat = nameFormat;
		}

		public String getId() {
			return id;
		}

		public String getName() {
			return name;
		}

		public String getNameFormat() {
			return nameFormat;
		}

		public boolean isMandatory() {
			return mandatory;
		}
	}

	public OIOH2BSTSAMLAssertion(Element element) {
		super(element);
	}

	public OIOH2BSTSAMLAssertion(Element dom, boolean validate) {
		super(dom, validate);
	}

	@Override
	public Type getType() {
		return Type.OIOH2BST;
	}

	/**
	 * Extract the <code>dk:gov:saml:attribute:AssuranceLevel</code> value from the
	 * token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic" Name="dk:gov:saml:attribute:AssuranceLevel"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;3&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the <code>dk:gov:saml:attribute:AssuranceLevel</code>
	 *         tag.
	 */
	@Override
	public String getAssuranceLevel() {
		return getAttributeValue(Attribute.ASSURANCE_LEVEL);
	}

	/**
	 * Extract the <code>dk:gov:saml:attribute:CprNumberIdentifier</code> value from
	 * the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic"
	 *       Name="dk:gov:saml:attribute:CprNumberIdentifier"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;2702681273&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>dk:gov:saml:attribute:CprNumberIdentifier</code> tag.
	 */
	public String getCpr() {
		return getAttributeValue(Attribute.CPR);
	}

	/**
	 * Extract the <code>dk:gov:saml:attribute:CvrNumberIdentifier</code> value from
	 * the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic"
	 *         Name="dk:gov:saml:attribute:CvrNumberIdentifier"&gt;
	 *         &lt;saml:AttributeValue xsi:type="xs:string"&gt;20688092&lt;/saml:AttributeValue&gt;
	 *       &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>dk:gov:saml:attribute:CvrNumberIdentifier</code> tag.
	 */
	public String getCvrNumberIdentifier() {
		return getAttributeValue(Attribute.CVR);
	}

	/**
	 * Extract the <code>urn:oid:2.5.4.10</code>/organizationName value from the
	 * token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic" Name="urn:oid:2.5.4.10"
	 *       FriendlyName="organizationName"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;Lægehuset på bakken&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the <code>urn:oid:2.5.4.10</code>/organizationName tag.
	 */
	public String getOrganizationName() {
		return getAttributeValue(Attribute.NAME_ORGANIZATION);
	}

	/**
	 * Extract the <code>dk:gov:saml:attribute:RidNumberIdentifier</code> value from
	 * the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic"
	 *         Name="dk:gov:saml:attribute:RidNumberIdentifier"&gt;
	 *         &lt;saml:AttributeValue xsi:type="xs:string"&gt;1118061020235&lt;/saml:AttributeValue&gt;
	 *       &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>dk:gov:saml:attribute:RidNumberIdentifier</code> tag.
	 */
	public String getRidNumberIdentifier() {
		return getAttributeValue(Attribute.RID);
	}

	/**
	 * Extract the <code>dk:gov:saml:attribute:SpecVer</code> value from the
	 * DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic" Name="dk:gov:saml:attribute:SpecVer"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;DK-SAML-2.0&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the <code>dk:gov:saml:attribute:SpecVer</code> tag.
	 */
	public String getSpecVersion() {
		Element specVersionElm = getAttributeElement(OIOSAMLAttributes.SPEC_VERSION);
		if (specVersionElm == null) {
			throw new ModelException("Mandatory 'specVersion' SAML attribute (dk:gov:saml:attribute:SpecVer) is missing");
		}
		return specVersionElm.getTextContent().trim();
	}

	/**
	 * Gets the signing certificate
	 *
	 * @return Returns the certificate contained in assertions XML signature or
	 *         <code>null</code> if the assertion was not signed
	 */
	public X509Certificate getSigningCertificate() {
		Element certificateElm = getTag(SAMLTags.assertion, DSTags.signature, DSTags.keyInfo, DSTags.x509Data, DSTags.x509Certificate);
		if (certificateElm == null) {
			return null;
		} else {
			return CertificateParser.asCertificate(XmlUtil.fromBase64(certificateElm.getTextContent()));
		}
	}

	@Override
	public String getUuid() {
		return getAttributeValue(Attribute.UUID);
	}
	
	/**
	 * En (frivillig) angivelse af brugerens 'nationale roller' formatteret som
	 * OIOBPP 1.2 XML struktur og indlejret i Base64 encoded form <br>
	 * 
	 * <pre>
	 *     &lt;saml:Attribute 
	 *     	NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" 
	 *     	Name="dk:gov:saml:attribute:Privileges_intermediate"&gt;
	 *     	&lt;saml:AttributeValue xsi:type="xs:string"&gt;Bx23z .... &lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 * 
	 * I base64 decoded form <br>
	 * 
	 * <pre>
	 *     &lt;bpp:PrivilegeList 
	 *     	xmlns:bpp="http://digst.dk/oiosaml/basic_privilege_profile"&gt;
	 *     	&lt;PrivilegeGroup Scope="urn:dk:gov:saml:cvrNumberIdentifier:20301823"&gt;
	 *     	  &lt;Privilege>urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3&lt;/Privilege&gt; 
	 *     	&lt;/PrivilegeGroup&gt; 
	 *     &lt;/bpp:PrivilegeList&gt;
	 * </pre>
	 * 
	 */
	public BasicPrivileges getBasicPrivileges() {
		String privileges = getAttributeValue(Attribute.PRIVILEGES);
		return privileges != null ? BasicPrivileges.decode(privileges, Profile.V12) : null;
	}
}
