/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import dk.sosi.seal.model.AssuranceLevel.NIST;
import dk.sosi.seal.model.OIOBSTSAMLAssertion.SAMLAttribute;
import dk.sosi.seal.model.OIOH2BSTSAMLAssertion.Attribute;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.xml.XmlUtil;

/**
 */
public class OIOH2BSTSAMLAssertionBuilder extends AbstractOIOBSTSAMLAssertionBuilder<OIOH2BSTSAMLAssertionBuilder, OIOH2BSTSAMLAssertion> {
    //OIOH2BST og OIOH3BST
	private BasicPrivilegesDOMBuilder privilegesBuilder;
	
	public OIOH2BSTSAMLAssertionBuilder() {
		addAttribute(Attribute.SPEC_VER, SAMLValues.DK_SPEC_VERSION);
	}

	@Override
	public void validateBeforeBuild() {
		//OIOH2BST og OIOH3BST
		if(privilegesBuilder != null) {
			privilegesBuilder.validateBeforeBuild();
			//We didn't fail validation
			addPrivilegesAsAttribute();
		}
		
		super.validateBeforeBuild();
		
		/*OIOH2BST kan indeholde enten en RID eller en global UUID, men ikke begge samtidigt*/
		String rid = getAttribute(OIOH2BSTSAMLAssertion.Attribute.RID);
		String uuid = getAttribute(OIOH2BSTSAMLAssertion.Attribute.UUID);
		if(rid != null && uuid != null) {
			throw new ModelException("Only one of RID or UUID can be present.");
		} else if(rid == null && uuid == null) {
			throw new ModelException("One of RID or UUID must be present.");
		}
		
	}
	
	@Override
	protected SAMLAttribute[] getAttributes() {
		return OIOH2BSTSAMLAssertion.Attribute.values();
	}
	
	@Override
	protected String getNameIDFormat() {
		return SAMLValues.NAMEID_FORMAT_UNSPECIFIED;
	}
	
	/**
	 * Hvis Privilege Builder har valideret indhold, så tilføjer vi privileger som base64 kodet attribut.
	 */
	private void addPrivilegesAsAttribute() {
		Document priviliges = privilegesBuilder.build();
		String privilegesStr = XmlUtil.node2String(priviliges, true, false);
		String privilegesB64 = XmlUtil.toBase64(privilegesStr.getBytes());
		addAttribute(OIOH2BSTSAMLAssertion.Attribute.PRIVILEGES, privilegesB64);
	}
	
    public OIOH2BSTSAMLAssertionBuilder setAssuranceLevel(NIST nist) {
    	return addAttribute(OIOH2BSTSAMLAssertion.Attribute.ASSURANCE_LEVEL, nist);
    }
    
    public OIOH2BSTSAMLAssertionBuilder setCpr(String cpr) {
    	return addAttribute(OIOH2BSTSAMLAssertion.Attribute.CPR, cpr);
    }
    
    public OIOH2BSTSAMLAssertionBuilder setRid(String rid) {
    	return addAttribute(OIOH2BSTSAMLAssertion.Attribute.RID, rid);
    }
    
    public OIOH2BSTSAMLAssertionBuilder setUuid(String uuid) {
    	return addAttribute(OIOH2BSTSAMLAssertion.Attribute.UUID, uuid);
    }
    
    public OIOH2BSTSAMLAssertionBuilder setCvr(String cvr) {
    	return addAttribute(OIOH2BSTSAMLAssertion.Attribute.CVR, cvr);
    }
    
    public OIOH2BSTSAMLAssertionBuilder setOrganizationName(String name) {
    	return addAttribute(OIOH2BSTSAMLAssertion.Attribute.NAME_ORGANIZATION, name);
    }

	public OIOH2BSTSAMLAssertionBuilder setPrivileges(BasicPrivilegesDOMBuilder builder) {
		this.privilegesBuilder = builder;
		return this;
	}
	
	protected OIOH2BSTSAMLAssertion createAssertion(Element elm) {
		return new OIOH2BSTSAMLAssertion(elm);
	}
}
