/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Element;

import dk.sosi.seal.model.BasicPrivileges.Profile;
import dk.sosi.seal.model.constants.OIOSAML3Attributes;
import dk.sosi.seal.model.constants.SAMLValues;

import java.util.Date;

/**
 * 
 *          Based on
 *          https://digst.dk/media/23391/oiosaml-web-sso-profile-302.pdf
 */
public class OIOH3BSTSAMLAssertion extends OIOBSTSAMLAssertion {
	public enum Attribute implements SAMLAttribute {
		// Specifies the version of the OIOSAML profile specification
		SPEC_VER("specVer", OIOSAML3Attributes.SPEC_VERSION, true),
		// Specifies an additional version of the OIOSAML profile specification (to separate it from OIO3BST)
		SPEC_VER_ADDITIONAL("specVerAdditional", OIOSAML3Attributes.SPEC_VERSION_ADDITIONAL, true),
		// Contains the overall level of assuranceof the authentication as defined by
		// the Danish [NSIS] standard.
		// The allowedvalues are ‘Low’, ‘Substantial’ and ‘High’.
		ASSURANCE_LEVEL("assuranceLevel", OIOSAML3Attributes.ASSURANCE_LEVEL, true),
		// Contains a UUID for the professional identity which is shared across all
		// public sector SPs. The identifier is specific to the professional role and is
		// not related to the associated natural person. The UUID MUST follow RFC 4122.
		// This attribute is the successor to the RIDattribute but is globally unique.
		UUID_PROFESSIONAL("professionalUuid", OIOSAML3Attributes.PROFESSIONAL_UUID, false),
		// Contains the CVR number (8 digits) of the organization related to the
		// authentication context. Note that a professional may be associated with
		// several organizations but only one organization is allowed per authentication
		// context.
		CVR("cvr", OIOSAML3Attributes.CVR, true),
		// Contains the name of the organization related to the authentication context.
		// Note that a professional may be associated with several organizationscbut
		// only one organization is allowed per authentication context.
		NAME_ORGANIZATION("organizationName", OIOSAML3Attributes.ORGANIZATION_NAME, true),
		// Contains the email address of the identity. In cases there are multiple
		// addresses known this attribute can be multi-valued (i.e. using multiple
		// <AttributeValue> elements).
		EMAIL("email", OIOSAML3Attributes.EMAIL, false),
		// Contains analias of the identity. This attribute can be used as a display
		// name selected by the user as an alternative to the above name attributes.
		ALIAS("alias", OIOSAML3Attributes.ALIAS, false),
		// Contains a base64-encoded value describing privileges assigned to the
		// identity (see OIO Basic Privilege Profile specification [OIOBPP] for
		// details).
		PRIVILEGES("privileges", OIOSAML3Attributes.PRIVILEGES_INTERMEDIATE, false);
		private final String id, name;
		private final boolean mandatory;

		private Attribute(String id, String name, boolean mandatory) {
			this.id = id;
			this.name = name;
			this.mandatory = mandatory;
		}

		public String getId() {
			return id;
		}

		public String getName() {
			return name;
		}

		public String getNameFormat() {
			return SAMLValues.NAME_FORMAT_URI;
		}

		public boolean isMandatory() {
			return mandatory;
		}
	}

	public OIOH3BSTSAMLAssertion(Element element) {
		super(element);
	}

	public OIOH3BSTSAMLAssertion(Element dom, boolean validate) {
		super(dom, validate);
	}

	/**
	 * Extract the <code>https://data.gov.dk/concept/core/nsis/loa</code> value from
	 * the token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/concept/core/nsis/loa"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;Low/Substantial/High&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/concept/core/nsis/loa</code> tag.
	 */
	@Override
	public String getAssuranceLevel() {
		return getAttributeValue(Attribute.ASSURANCE_LEVEL);
	}

	/**
	 * Extract the
	 * <code>https://data.gov.dk/model/core/eid/professional/orgName</code>
	 * (organizationName) value from the token.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/model/core/eid/professional/orgName" &gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;Lægehuset på bakken&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/model/core/eid/professional/orgName</code>
	 *         (organizationName) tag.
	 */
	public String getOrganizationName() {
		return getAttributeValue(Attribute.NAME_ORGANIZATION);
	}

	/**
	 * Extract the <code>dk:gov:saml:attribute:CvrNumberIdentifier</code> value from
	 * the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri"
	 *         Name="dk:gov:saml:attribute:CvrNumberIdentifier"&gt;
	 *         &lt;saml:AttributeValue xsi:type="xs:string"&gt;20688092&lt;/saml:AttributeValue&gt;
	 *       &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>dk:gov:saml:attribute:CvrNumberIdentifier</code> tag.
	 */
	public String getCvrNumberIdentifier() {
		return getAttributeValue(Attribute.CVR);
	}

	/**
	 * Extract the <code>https://data.gov.dk/model/core/specVersion</code> value
	 * from the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/model/core/specVersion"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;OIO-SAML-3.0&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/model/core/specVersion</code> tag.
	 */
	public String getSpecVersion() {
		return getAttributeValue(Attribute.SPEC_VER);
	}

	/**
	 * Extract the <code>https://healthcare.data.gov.dk/model/core/specVersion</code> value
	 * from the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://healthcare.data.gov.dk/model/core/specVersion"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;OIO-SAML-H-3.0&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://healthcare.data.gov.dk/model/core/specVersion</code> tag.
	 */
	public String getSpecVersionAdditional() {
		return getAttributeValue(Attribute.SPEC_VER_ADDITIONAL);
	}
	
	/**
	 * Extract the <code>https://data.gov.dk/model/core/specVersion</code> value
	 * from the DOM.<br />
	 *
	 * <pre>
	 *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri" Name="https://data.gov.dk/model/core/eid/professional/uuid/persistent"&gt;
	 *       &lt;saml:AttributeValue xsi:type="xs:string"&gt;URN:xyz:123&lt;/saml:AttributeValue&gt;
	 *     &lt;/saml:Attribute&gt;
	 * </pre>
	 *
	 * @return The value of the
	 *         <code>https://data.gov.dk/model/core/eid/professional/uuid/persistent</code> tag.
	 */
	@Override
	public String getUuid() {
		return getAttributeValue(Attribute.UUID_PROFESSIONAL);
	}
	
	@Override
	public Type getType() {
		return Type.OIOH3BST;
	}

	@Override
	public String getRidNumberIdentifier() {
		return null;
	}
	
	/**
	 * CPR er ikke med i OIOH3BST
	 */
	public String getCpr() {
		return null;
	}
	
	/**
	 * En (frivillig) angivelse af brugerens 'nationale roller' formatteret som
	 * OIOBPP 1.2 XML struktur og indlejret i Base64 encoded form <br>
	 * 
	 * <pre>
	 *     &lt;Attribute Name="https://data.gov.dk/model/core/eid/privilegesIntermediate"
	 *       NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:uri"&gt;
	 *       &lt;AttributeValue>Bx23z ....&lt;/AttributeValue&gt;
	 *     &lt;/Attribute&gt;
	 * </pre>
	 * 
	 * I base64 decoded form <br>
	 * 
	 * <pre>
	 *     &lt;bpp:PrivilegeList 
	 *     	xmlns:bpp="http://digst.dk/oiosaml/basic_privilege_profile"&gt;
	 *     	&lt;PrivilegeGroup Scope="urn:dk:gov:saml:cvrNumberIdentifier:20301823"&gt;
	 *     	  &lt;Privilege>urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3&lt;/Privilege&gt; 
	 *     	&lt;/PrivilegeGroup&gt; 
	 *     &lt;/bpp:PrivilegeList&gt;
	 * </pre>
	 * 
	 */
	@Override
	public BasicPrivileges getBasicPrivileges() {
		String privileges = getAttributeValue(Attribute.PRIVILEGES);
		return privileges != null ? BasicPrivileges.decode(privileges, Profile.V12) : null;
	}

	/**
	 * Extract the <code>saml:AuthnStatement#AuthnInstant</code> value from the DOM.<br />
	 *
	 * <pre>
	 *    &lt;saml:AuthnStatement AuthnInstant="2022-10-04T11:09:56.904Z"&gt;
	 *      &lt;saml:AuthnContext&gt;
	 *         &lt;saml:AuthnContextClassRef&gt;https://data.gov.dk/concept/core/nsis&lt;/saml:AuthnContextClassRef&gt;
	 *      &lt;/saml:AuthnContext&gt;
	 *    &lt;/saml:AuthnStatement&gt;
	 * </pre>
	 *
	 * @return The value of the <code>saml:AuthnStatement#AuthnInstant</code> tag.
	 */
	public Date getIssuerInstant() {
		Element ac = getTag(SAMLTags.assertion, SAMLTags.authnStatement);
		return convertToDate(ac, SAMLAttributes.AUTHN_INSTANT);
	}
}
