/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.OIOSAML3Attributes;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import dk.sosi.seal.model.AssuranceLevel.NSIS;
import dk.sosi.seal.model.OIOBSTSAMLAssertion.SAMLAttribute;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Node;

import java.util.Date;

/**
 */
public class OIOH3BSTSAMLAssertionBuilder extends AbstractOIOBSTSAMLAssertionBuilder<OIOH3BSTSAMLAssertionBuilder, OIOH3BSTSAMLAssertion>{
    //OIOH2BST og OIOH3BST
	private BasicPrivilegesDOMBuilder privilegesBuilder;

	private Date issuerInstant;

    public OIOH3BSTSAMLAssertionBuilder() {
    	addAttribute(OIOH3BSTSAMLAssertion.Attribute.SPEC_VER, "OIO-SAML-3.0"); 
    	addAttribute(OIOH3BSTSAMLAssertion.Attribute.SPEC_VER_ADDITIONAL, "OIO-SAML-H-3.0");
    }
    
    @Override
    public void validateBeforeBuild() {   
        //OIOH2BST og OIOH3BST
		if(privilegesBuilder != null) {
			privilegesBuilder.validateBeforeBuild();
			//We didn't fail validation
			addPrivilegesAsAttribute();
		}
		
		 super.validateBeforeBuild();
    }
    
	@Override
	protected SAMLAttribute[] getAttributes() {
		return OIOH3BSTSAMLAssertion.Attribute.values();
	}
	
	@Override
	protected String getNameIDFormat() {
		return SAMLValues.NAMEID_FORMAT_PERSISTENT;
	}
    
	/**
	 * Hvis Privilege Builder har valideret indhold, så tilføjer vi privileger som base64 kodet attribut.
	 */
	private void addPrivilegesAsAttribute() {
		Document priviliges = privilegesBuilder.build();
		String privilegesStr = XmlUtil.node2String(priviliges, true, false);
		String privilegesB64 = XmlUtil.toBase64(privilegesStr.getBytes());
		addAttribute(OIOH3BSTSAMLAssertion.Attribute.PRIVILEGES, privilegesB64);
	}

    @Override
    protected void addRootAttributes(Element root) {
        super.addRootAttributes(root);
    }
    
    public OIOH3BSTSAMLAssertionBuilder setAssuranceLevel(NSIS nsis) {
    	return addAttribute(OIOH3BSTSAMLAssertion.Attribute.ASSURANCE_LEVEL, nsis);
    }
    
    public OIOH3BSTSAMLAssertionBuilder setUuid(String uuid) {
    	return addAttribute(OIOH3BSTSAMLAssertion.Attribute.UUID_PROFESSIONAL, uuid);
    }
    
    public OIOH3BSTSAMLAssertionBuilder setCvr(String cvr) {
    	return addAttribute(OIOH3BSTSAMLAssertion.Attribute.CVR, cvr);
    }
    
    public OIOH3BSTSAMLAssertionBuilder setOrganizationName(String name) {
    	return addAttribute(OIOH3BSTSAMLAssertion.Attribute.NAME_ORGANIZATION, name);
    }
    
	public OIOH3BSTSAMLAssertionBuilder setPrivileges(BasicPrivilegesDOMBuilder builder) {
		this.privilegesBuilder = builder;
		return this;
	}

	public OIOH3BSTSAMLAssertionBuilder setIssuerInstant(Date issuerInstant) {
		this.issuerInstant = issuerInstant;
		return this;
	}

	protected OIOH3BSTSAMLAssertion createAssertion(Element elm) {
		return new OIOH3BSTSAMLAssertion(elm);
	}

	@Override
	protected Node createAuthnStatement() {
		if (issuerInstant != null) {
			Element authnStatementElm = createElement(SAMLTags.authnStatement);
			authnStatementElm.setAttributeNS(null, SAMLAttributes.AUTHN_INSTANT, XmlUtil.getDateFormat(true).format(this.issuerInstant));

			Element authnContextElm = createElement(SAMLTags.authnContext);
			Element authContextClassRefElm = createElement(SAMLTags.authnContextClassRef);
			authContextClassRefElm.setTextContent(OIOSAML3Attributes.AUTHNINSTANT);
			authnContextElm.appendChild(authContextClassRefElm);
			authnStatementElm.appendChild(authnContextElm);

			return authnStatementElm;
		}
		return null;
	}
}
