/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.*;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Element;

import java.security.cert.X509Certificate;

/**
 */
public class OIOSAMLAssertion extends AbstractOIOSAMLAssertion {

    /*pp*/ OIOSAMLAssertion(Element element, boolean validate) {
        super(element, validate);
    }

    public OIOSAMLAssertion(Element element) {
        this(element, true);
    }


    /**
     * Extract the <code>dk:gov:saml:attribute:RidNumberIdentifier</code> value from the DOM.<br />
     *
     * <pre>
     *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic"
     *         Name="dk:gov:saml:attribute:RidNumberIdentifier"&gt;
     *         &lt;saml:AttributeValue xsi:type="xs:string"&gt;1118061020235&lt;/saml:AttributeValue&gt;
     *       &lt;/saml:Attribute&gt;
     * </pre>
     *
     * @return The value of the <code>dk:gov:saml:attribute:RidNumberIdentifier</code> tag.
     */
    public String getRidNumberIdentifier() {
        return getAttribute(OIOSAMLAttributes.RID_NUMBER);
    }

    public String getCertificateIssuer() {
        return getAttribute(OIOSAMLAttributes.CERTIFICATE_ISSUER);
    }

    /**
     * Extract the <code>urn:oid:1.3.6.1.4.1.1466.115.121.1.8</code>/userCertificate value from the DOM.<br />
     *
     * <pre>
     *     &lt;saml:Attribute NameFormat="urn:oasis:names:tc:SAML:2.0:attrname-format:basic"
     *         Name="urn:oid:1.3.6.1.4.1.1466.115.121.1.8"
     *         FriendlyName="userCertificate"&gt;
     *         &lt;saml:AttributeValue xsi:type="xs:string"&gt;MIIB5DCCAU0CBAJQodoZIhvcNAQ....&lt;/saml:AttributeValue&gt;
     *       &lt;/saml:Attribute&gt;
     * </pre>
     *
     * @return The value of the <code>urn:oid:1.3.6.1.4.1.1466.115.121.1.8</code>/userCertificate tag.
     */
    public X509Certificate getUserCertificate() {
        String certString = getAttribute(OIOSAMLAttributes.USER_CERTIFICATE);
        if (certString == null) {
            return null;
        } else {
            return CertificateParser.asCertificate(XmlUtil.fromBase64(certString));
        }
    }

    public boolean isYouthCertificate() {
        String isYouthCert = getAttribute(OIOSAMLAttributes.IS_YOUTH_CERT);
        return Boolean.parseBoolean(isYouthCert);
    }

    public String getCertificateSerial() {
        return getAttribute(OIOSAMLAttributes.CERTIFICATE_SERIAL);
    }

    public String getUID() {
        return getAttribute(OIOSAMLAttributes.UID);
    }

    public UserIDCard getUserIDCard() {
        Element discoveryEPR = getAttributeElement(OIOSAMLAttributes.DISCOVERY_EPR);
        Element metadata = getTag(discoveryEPR, SAMLTags.attribute, SAMLTags.attributeValue, WSATags.endpointReference, WSATags.metadata);
        Element serviceType = getTag(metadata, WSATags.metadata, LibertyDiscoveryTags.serviceType);
        if (serviceType != null && LibertyValues.SOSI_URN.equals(serviceType.getTextContent())) {
            Element idCardElm = getTag(metadata, WSATags.metadata, LibertyDiscoveryTags.securityContext, LibertySecurityTags.token, SAMLTags.assertion);
            return (UserIDCard) IDCard.from(idCardElm);
        } else {
            return null;
        }
    }
}
