/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import org.apache.ws.security.WSSecurityException;
import org.apache.xml.security.exceptions.XMLSecurityException;
import org.apache.xml.security.keys.KeyInfo;
import org.apache.xml.security.signature.XMLSignature;
import org.apache.xml.security.utils.Constants;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import dk.sosi.seal.transform.internal.SecurityTokenReference;

/**
 *
 * A simple class holding configuration parameters when generating XML
 * signatures using @link{SignatureUtil}
 *
 */
public class SignatureConfiguration {

	/*
	 * This type is the Reference type used in SignedInfo
	 */
	public static enum Type {
		DIRECT_REFERENCE_NOT_ENVELOPED, DIRECT_REFERENCE, SECURITY_TOKEN_REFERENCE
	}

	/*
	 * This type is the values used in KeyInfo - WSS_XXX is the values used in
	 * KeyInfo/SecurityTokenReference
	 */
	public static enum KeyInfoType {
		XMLDSIG_KEYNAME, XMLDSIG_X509DATA, WSS_KEYIDENTIFIER, WSS_REFERENCE;
		public static KeyInfoType fromKeyInfoElement(Element kiElm) {
			try {
				KeyInfo ki = new KeyInfo(kiElm, null);
				return fromKeyInfo(ki);
			} catch (XMLSecurityException e) {
				return null;
			}
		}

		public static KeyInfoType fromKeyInfo(KeyInfo ki) {
			KeyInfoType result = null;
			if (ki != null) {
				if (ki.containsKeyName()) {
					result = XMLDSIG_KEYNAME;
				} else if (ki.containsX509Data()) {
					result = XMLDSIG_X509DATA;
				} else {
					result = fromKeyInfoChildNodes(ki.getElement().getChildNodes());
				}
			}
			// We could not find any KeyInfoType
			return result;
		}
		
		private static KeyInfoType fromKeyInfoChildNodes(NodeList nl) {
			KeyInfoType result = null;
			for (int i = 0; i < nl.getLength(); i++) {
				Node node = nl.item(i);
				if (node.getNodeType() == Node.ELEMENT_NODE) {
					result = fromSecurityTokenReferenceElement((Element) node);
				}
				if(result != null) {
					break;
				}
			}
			return result;
		}
		
		public static KeyInfoType fromSecurityTokenReferenceElement(Element strElm) {
			try {
				SecurityTokenReference str = new SecurityTokenReference(strElm);
				return fromSecurityTokenReference(str);
			} catch (WSSecurityException e) {
				// This was not a SecurityTokenReference element. Continue to do lookup!
				return null;
			}
		}
		
		public static KeyInfoType fromSecurityTokenReference(SecurityTokenReference str) {
			KeyInfoType result = null;
			if(str != null) {
				if (str.containsKeyIdentifier()) {
					result = WSS_KEYIDENTIFIER;
				} else if (str.containsReference()) {
					result = WSS_REFERENCE;
				}			
			}
			return result;
		}
	};

	public static class Reference {

		private final String uri;
		private final Type type;

		public Reference(String uri, Type type) {
			this.uri = uri;
			this.type = type;
		}

		private static Reference[] fromDirectReferenceStringArray(String[] referenceURIs) {
			Reference[] references = new Reference[referenceURIs.length];
			for (int i = 0; i < referenceURIs.length; i++) {
				references[i] = new Reference(referenceURIs[i], Type.DIRECT_REFERENCE);
			}
			return references;
		}

		public String getURI() {
			return uri;
		}

		public Type getType() {
			return type;
		}
	}

	private final Reference[] references;
	private final String signatureParentID;
	private final String idAttributeName;

	private Node signatureSiblingNode;
	private String keyInfoId;
	private String identityTokenId;

	public static final KeyInfoType DEFAULT_KEY_INFO_TYPE = KeyInfoType.XMLDSIG_X509DATA;
	private KeyInfoType selectedKeyInfoType = DEFAULT_KEY_INFO_TYPE;
	
	private String signatureMethodAlgorithmUri = XMLSignature.ALGO_ID_SIGNATURE_RSA_SHA1;
	private String digestMethodAlgorithmUri = Constants.ALGO_ID_DIGEST_SHA1;

	public SignatureConfiguration(Reference[] references, String signatureParentID, String idAttributeName) {
		this.references = references;
		this.signatureParentID = signatureParentID;
		this.idAttributeName = idAttributeName;
	}

	public SignatureConfiguration(String[] referenceURIs, String signatureParentID, String idAttributeName) {
		this(Reference.fromDirectReferenceStringArray(referenceURIs), signatureParentID, idAttributeName);
	}

	public void setSignatureSiblingNode(Node signatureSiblingNode) {
		this.signatureSiblingNode = signatureSiblingNode;
	}

	public void setKeyInfoId(String keyInfoId) {
		this.keyInfoId = keyInfoId;
	}
	
	public void setIdentityTokenId(String identityTokenId) {
		this.identityTokenId = identityTokenId;
	}

	public Reference[] getReferences() {
		return references;
	}

	public String getSignatureParentID() {
		return signatureParentID;
	}

	public Node getSignatureSiblingNode() {
		return signatureSiblingNode;
	}

	public String getIdAttributeName() {
		return idAttributeName;
	}

	public String getKeyInfoId() {
		return keyInfoId;
	}
	
	public String getIdentityTokenId() {
		return identityTokenId;
	}

	//
	// Security Section
	//

	/**
	 * 
	 * @param addCertAsRef
	 * @deprecated Use {@link #useAddCertificateAsReference()} or
	 *             {@link #useDefaultKeyInfoType()}
	 */
	public void setAddCertificateAsReference(boolean addCertAsRef) {
		if (addCertAsRef) {
			useAddCertificateAsReference();
		} else if (isAddCertificateAsReference()) {
			useDefaultKeyInfoType();
		}
	}

	public void useAddCertificateAsReference() {
		this.selectedKeyInfoType = KeyInfoType.XMLDSIG_KEYNAME;
	}

	public boolean isAddCertificateAsReference() {
		return this.selectedKeyInfoType == KeyInfoType.XMLDSIG_KEYNAME;
	}

	/**
	 * JIRA: https://jira.nspop.dk/browse/SDS-2282 Til implementering af
	 * signerings-certificat som SecurityTokenReference i stedet for som indlejret
	 * X509Certificate.
	 * 
	 * @deprecated Use {@link #useSecurityTokenReferenceToBinarySecurityToken()} or
	 *             {@link #useDefaultKeyInfoType()}
	 */
	public void setUseSecurityTokenReference(boolean useSecurityTokenReference) {
		if (useSecurityTokenReference) {
			useSecurityTokenReferenceToBinarySecurityToken();
		} else if (isUseSecurityTokenReferenceToBinarySecurityToken()) {
			useDefaultKeyInfoType();
		}
	}
	
	@Deprecated
	public boolean isUseSecurityTokenReference() {
		return isUseSecurityTokenReferenceToBinarySecurityToken(); 
	}

	public void useSecurityTokenReferenceToBinarySecurityToken() {
		this.selectedKeyInfoType = KeyInfoType.WSS_REFERENCE;
	}

	public boolean isUseSecurityTokenReferenceToBinarySecurityToken() {
		return selectedKeyInfoType == KeyInfoType.WSS_REFERENCE;
	}

	/**
	 */
	public void useSecurityTokenReferenceToSamlAssertion() {
		this.selectedKeyInfoType = KeyInfoType.WSS_KEYIDENTIFIER;
	}

	public boolean isUseSecurityTokenReferenceToSamlAssertion() {
		return selectedKeyInfoType == KeyInfoType.WSS_KEYIDENTIFIER;
	}

	public void useDefaultKeyInfoType() {
		this.selectedKeyInfoType = DEFAULT_KEY_INFO_TYPE;
	}

	public void setKeyInfoType(KeyInfoType keyInfoType) {
		this.selectedKeyInfoType = keyInfoType;
	}
	
	/*SignatureMethodUri*/
	public void setSignatureMethodAlgorithmUri(String signatureMethodUri) {
		this.signatureMethodAlgorithmUri = signatureMethodUri;
	}

	public String getSignatureMethodAlgorithmUri() {
		return signatureMethodAlgorithmUri;
	}
	
	/*Digest Method Algorithm*/
	public void setDigestMethodAlgorithmUri(String digestMethodAlgorithmUri) {
		this.digestMethodAlgorithmUri = digestMethodAlgorithmUri;
	}
	
	public String getDigestMethodAlgorithmUri() {
		return digestMethodAlgorithmUri;
	}
}
