package dk.sosi.seal.model;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

import static dk.sosi.seal.xml.XmlUtil.getChildElements;

public class SubjectRelations11Impl implements SubjectRelations {

    private final List<VerifiedRelation> verifiedRelations = new ArrayList<>();

    public SubjectRelations11Impl(Element root) {
        validate(root);
        parse(root);
    }

    public SubjectRelations11Impl() {
    }

    public void addVerifiedRelation(VerifiedRelation verifiedRelation) {
        verifiedRelations.add(verifiedRelation);
    }

    public List<VerifiedRelation> getVerifiedRelations() {
        return verifiedRelations;
    }

    protected void parse(Element root) {
        List<Element> verifiedRelations = getChildElements(root, "VerifiedRelation");
        for (Element verifiedRelation : verifiedRelations) {
            String relationType = verifiedRelation.getAttribute("relationType");
            String relatedPersonID = verifiedRelation.getAttribute("relatedPersonID");
            String relatedPersonIDType = verifiedRelation.getAttribute("relatedPersonIDType");

            String verifiedRelationAge = verifiedRelation.getAttribute("relatedPersonAge");

            // Create a new VerifiedRelation object and add it to the list
            if (!verifiedRelationAge.isEmpty()) {
                int relatedPersonAge = Integer.parseInt(verifiedRelation.getAttribute("relatedPersonAge"));
                this.verifiedRelations.add(new VerifiedRelation(
                        RelationType.from(relationType),
                        relatedPersonID,
                        PersonIDType.from(relatedPersonIDType),
                        relatedPersonAge
                ));
            } else {
                this.verifiedRelations.add(new VerifiedRelation(
                        RelationType.from(relationType),
                        relatedPersonID,
                        PersonIDType.from(relatedPersonIDType)
                ));
            }
        }
    }

    private static void validate(Node node) {
        try {
            Schema schema = SchemaUtil.loadSchema("/OIOITP_Subject_relations_profile-v11.xsd");
            schema.newValidator().validate(new DOMSource(node));
        } catch (SAXException | IOException e) {
            throw new ModelException("Error validating subject relations: " + e.getMessage(), e);
        }
    }

    public enum PersonIDType {
        CPR("URN:OID:1.2.208.176.1.2");

        private final String value;

        PersonIDType(String value) {
            this.value = value;
        }

        public String getValue() {
            return value;
        }

        // Return Enum based on its value
        public static PersonIDType from(String value) {
            for (PersonIDType idType : values()) {
                if (idType.value.equals(value)) {
                    return idType;
                }
            }
            throw new IllegalArgumentException("No enum constant " + PersonIDType.class.getCanonicalName() + " with value " + value);
        }
    }

    public enum RelationType {
        WARD_CUSTODY_HOLDER("wardCustodyHolder"),
        PARTLY_WARD_CUSTODY_HOLDER("partlyWardCustodyHolder"),
        PARENTAL_CUSTODY_HOLDER("parentalCustodyHolder");

        private final String value;

        RelationType(String value) {
            this.value = value;
        }

        public String getValue() {
            return value;
        }

        // Return Enum based on its value
        public static RelationType from(String value) {
            for (RelationType relationType : values()) {
                if (relationType.value.equals(value)) {
                    return relationType;
                }
            }
            throw new IllegalArgumentException("No enum constant " + RelationType.class.getCanonicalName() + " with value " + value);
        }
    }

    public static class VerifiedRelation {
        private final RelationType relationType;
        private final String relatedPersonID;
        private final PersonIDType relatedPersonIDType;
        private final Integer relatedPersonAge;

        public VerifiedRelation(RelationType relationType, String relatedPersonID, PersonIDType relatedPersonIDType) {
            this.relationType = relationType;
            this.relatedPersonID = relatedPersonID;
            this.relatedPersonIDType = relatedPersonIDType;
            this.relatedPersonAge = null;
        }

        public VerifiedRelation(RelationType relationType, String relatedPersonID, PersonIDType relatedPersonIDType, int relatedPersonAge) {
            this.relationType = relationType;
            this.relatedPersonID = relatedPersonID;
            this.relatedPersonIDType = relatedPersonIDType;
            this.relatedPersonAge = relatedPersonAge;
        }

        public RelationType getRelationType() {
            return relationType;
        }

        public String getRelatedPersonID() {
            return relatedPersonID;
        }

        public PersonIDType getRelatedPersonIDType() {
            return relatedPersonIDType;
        }

        public Optional<Integer> getRelatedPersonAge() {
            return Optional.ofNullable(relatedPersonAge);
        }
    }
}
