/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model.dombuilders;

import dk.sosi.seal.model.OnBehalfOfClaim;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.OIOSAMLAttributes;
import dk.sosi.seal.model.constants.WSFAuthAttributes;
import dk.sosi.seal.model.constants.WSFAuthTags;
import dk.sosi.seal.model.constants.WSFAuthValues;
import dk.sosi.seal.model.constants.WSTTags;
import dk.sosi.seal.model.constants.WSTrustAttributes;

import java.util.Optional;

import static java.util.Optional.*;
import static dk.sosi.seal.model.constants.HealthcareSAMLAttributes.*;
import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType;
import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.*;

/**
 * Created by chg on 22/02/2017.
 */
public abstract class AbstractOIOToIdentityTokenRequestDOMBuilder<T extends AbstractOIOToIdentityTokenRequestDOMBuilder<T>> extends OIOWSTrustRequestDOMBuilder<T> {

    protected String cprNumberClaim;
    private Optional<OnBehalfOfClaim> onBehalfOfClaim = empty();

    @SuppressWarnings("unchecked")
	public T setCPRNumberClaim(String cprNumberClaim) {
        this.cprNumberClaim = cprNumberClaim;
        return (T) this;
    }

    /**
     * @deprecated
     * Use 'setOnBehalfOfClaim' instead.
     */
    @Deprecated()
    @SuppressWarnings("unchecked")
	public T setProcurationByCprClaim(String onBehalfOf) {
        this.onBehalfOfClaim = ofNullable(onBehalfOf).map(obo -> new OnBehalfOfClaim(obo, Procuration));
        return (T) this;
    }

    @SuppressWarnings("unchecked")
    public T setOnBehalfOfClaim(String value, OnBehalfOfClaimType type) {
        this.onBehalfOfClaim = of(new OnBehalfOfClaim(value, type));
        return (T) this;
    }

    /**
     * Denne metode er magen til 'setOnBehalfOfClaim(String, OnBehalfOfClaimType)',
     * men tillader nemmere at give et instans af 'OnBehalfOfClaim',
     * eller at sætte claim til ingenting med Optional::empty
     */
    public T setOnBehalfOfClaim(Optional<OnBehalfOfClaim> claim) { this.onBehalfOfClaim = claim; return (T) this; }

    @Override
    protected void validateBeforeBuild() throws ModelException {
        super.validateBeforeBuild();
        validate("signingVault", signingVault);
    }

    @Override
    protected void addExtraNamespaceDeclarations(Element envelope) {
        super.addExtraNamespaceDeclarations(envelope);
        envelope.setAttributeNS(NameSpaces.XMLNS_SCHEMA, NameSpaces.NS_XMLNS + ":" + NameSpaces.NS_WSF_AUTH, NameSpaces.WSF_AUTH_SCHEMA);
    }

    @Override
    protected void addClaims(Document doc, Element requestSecurityToken) {
        Element claimsElm = createElement(WSTTags.claims);
        claimsElm.setAttributeNS(null, WSTrustAttributes.DIALECT, WSFAuthValues.CLAIMS_DIALECT);
        requestSecurityToken.appendChild(claimsElm);

        Element cprClaimElm = buildClaimElement(OIOSAMLAttributes.CPR_NUMBER, this.cprNumberClaim);
        claimsElm.appendChild(cprClaimElm);

        onBehalfOfClaim
                .map(claim -> buildClaimElement(ON_BEHALF_OF, toSamlAttribute(claim.type) + ':' + claim.value))
                .ifPresent(claimsElm::appendChild);
    }

    private static String toSamlAttribute(OnBehalfOfClaimType t) {
        Optional<String> o =
                t == Procuration ? of(THROUGH_PROCURATION_BY_CPR) :
                t == WardCustody ? of(THROUGH_WARD_CUSTODY_BY_CPR) :
                t == PartlyWardCustody ? of(THROUGH_PARTLY_WARD_CUSTODY_BY_CPR) :
                t == ParentalCustody ? of(THROUGH_PARENTAL_CUSTODY_BY_CPR) :
                empty();
        return o.orElseThrow(() -> new RuntimeException("'" + t + "' has no defined conversion to SAML attribute"));
    }

    private Element buildClaimElement(String attributeName, String value) {
        Element cprClaimElm = createElement(WSFAuthTags.claimType);
        cprClaimElm.setAttributeNS(null, WSFAuthAttributes.URI, attributeName);

        Element valueElm = createElement(WSFAuthTags.value);
        valueElm.setTextContent(value);
        cprClaimElm.appendChild(valueElm);
        return cprClaimElm;
    }
}