/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model.dombuilders;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import dk.sosi.seal.model.LibertyMessageDOMEnhancer;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.SignatureConfiguration;
import dk.sosi.seal.model.SignatureConfiguration.KeyInfoType;
import dk.sosi.seal.model.constants.IDValues;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.WSATags;
import dk.sosi.seal.model.constants.WSSEAttributes;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.constants.WSTrustConstants;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;

/**
 * Abstract implementation of the <code>AbstractDOMBuilder</code> class.<br />
 * This class extends the <code>AbstractDOMBuilder</code> with functionality
 * used when creates a SOAP message.
 */
public abstract class OIOWSTrustDOMBuilder<T extends OIOWSTrustDOMBuilder<T>> extends AbstractSOAPBuilder<Document> {

	protected CredentialVault signingVault;

	protected KeyInfoType selectedKeyInfoType = SignatureConfiguration.DEFAULT_KEY_INFO_TYPE;

	protected abstract void addExtraHeaders(Document doc, Element header);

	protected abstract void addExtraNamespaceDeclarations(Element envelope);

	protected final void addHeaderContent(Document doc, Element header) {
		final Element action = TagUtil.createElementNS(doc, WSATags.action);
		header.appendChild(action);
		action.setTextContent(WSTrustConstants.WST_1_3_ISSUE_ACTION);

		final Element messageID = TagUtil.createElementNS(doc, WSATags.messageID);
		header.appendChild(messageID);
		messageID.setTextContent(XmlUtil.generateUUID());

		addExtraHeaders(doc, header);
	}

	@Override
	protected void addRootAttributes(Element envelope) {
		addNS(envelope, NameSpaces.NS_SOAP, NameSpaces.SOAP_SCHEMA);
		addNS(envelope, NameSpaces.NS_SAML, NameSpaces.SAML2ASSERTION_SCHEMA);
		addNS(envelope, NameSpaces.NS_WSA, NameSpaces.WSA_1_0_SCHEMA);
		addNS(envelope, NameSpaces.NS_WSP, NameSpaces.WSP_SCHEMA);
		addNS(envelope, NameSpaces.NS_WST, NameSpaces.WST_1_3_SCHEMA);

		addExtraNamespaceDeclarations(envelope);
	}

	/**
	 * Build the final response <code>Document</code>.<br />
	 * Before the <code>Document</code> is generated all attributes will be
	 * validated.<br />
	 * <br />
	 * A <code>Document</code> is generated each time this method is called. Calling
	 * this method multiple times will therefore return multiple objects.
	 *
	 * @return DOM representation of the Identity token response.
	 * @throws dk.sosi.seal.model.ModelException Thrown if the builder finds a
	 *         validation problem.
	 */
	@Override
	public final Document build() throws ModelException {
		Document document = createDocument();
		enhanceAndSign(document);
		return document;
	}
	
	protected void enhanceAndSign(Document document) {
		if (signingVault != null) {
			new LibertyMessageDOMEnhancer(signingVault, selectedKeyInfoType, document, false).enhanceAndSign();
		}
	}

	protected void addSecurityTokenReferenceToIDCard(Document doc, Element parent) {
		final Element securityTokenReference = doc.createElementNS(NameSpaces.WSSE_SCHEMA, WSSETags.SECURITY_TOKEN_REFERENCE_PREFIXED);
		parent.appendChild(securityTokenReference);
		final Element reference = doc.createElementNS(NameSpaces.WSSE_SCHEMA, WSSETags.REFERENCE_PREFIXED);
		reference.setAttributeNS(null, WSSEAttributes.URI, IDValues.IDCARD_REFERENCE);
		securityTokenReference.appendChild(reference);
	}

	public T setSigningVault(CredentialVault signingVault) {
		return setSigningVault(signingVault, false);
	}

	
	/**
	 * 
	 * @param signingVault
	 * @param useSecurityTokenReference
	 * @return
	 * @deprecated Use {@link #setSigningVault(CredentialVault, KeyInfoType)}
	 */
	public T setSigningVault(CredentialVault signingVault, boolean useSecurityTokenReference) {
		return setSigningVault(signingVault, useSecurityTokenReference ? KeyInfoType.WSS_REFERENCE : null);
	}

	@SuppressWarnings("unchecked")
	public T setSigningVault(CredentialVault signingVault, KeyInfoType keyInfoType) {
		this.signingVault = signingVault;
		this.selectedKeyInfoType = keyInfoType != null ? keyInfoType : SignatureConfiguration.DEFAULT_KEY_INFO_TYPE;
		
		return (T) this;
	}
}