/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.pki;

import java.security.cert.X509Certificate;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

/**
 * Abstract implementation of the CertificationAuthority interface.
 *
 * @author ads@lakeside.dk
 * @since 2.0
 */
public abstract class AbstractOCESCertificationAuthority implements CertificationAuthority {

    private final CertificateStatusChecker certificateStatusChecker;
    private final IntermediateCertificateCache intermediateCertificateCache;
    private final FederationCertificateResolver federationCertificateResolver;
    private final AuditEventHandler eventHandler;

    protected AbstractOCESCertificationAuthority(SOSIConfiguration configuration, CertificateStatusChecker certificateStatusChecker, IntermediateCertificateCache intermediateCertificateCache, FederationCertificateResolver federationCertificateResolver) {
        configuration.verify();
        if (certificateStatusChecker == null) throw new IllegalArgumentException("'certificateStatusChecker' cannot be null");
        if (intermediateCertificateCache == null) throw new IllegalArgumentException("'intermediateCertificateCache' cannot be null");

        this.eventHandler = configuration.getAuditEventHandler();

        this.certificateStatusChecker = certificateStatusChecker;
        this.intermediateCertificateCache = intermediateCertificateCache;
        this.federationCertificateResolver = federationCertificateResolver;
    }

    public final boolean isValid(X509Certificate certificate) throws PKIException {
        return getCertificateStatus(certificate).isValid();
    }

    public CertificateStatus getCertificateStatus(final X509Certificate cert) throws PKIException {
        CertificateStatus certificateStatus = getStatusOnDates(cert);
        if (certificateStatus.isValid()) {
            if (OCESUtil.isProbableOCES1Certificate(cert) && OCESUtil.isIssuerOf(cert, getOCES1RootCertificate())) {
                throw new PKIException("The supplied certificate with DN '" + new DistinguishedName(cert.getSubjectX500Principal()) + "' is an OCES1 certificate. OCES1 certificates are no longer supported.");
            }
            boolean issuerOfOCES2 = false, issuerOfOCES3 = false;

            /*OCES2*/
            issuerOfOCES2 = OCESUtil.isProbableOCES2Certificate(cert) && OCESUtil.isIssuerOf(cert, getAndValidateIntermediateOCES2Certificate(cert));


            /*OCES3*/
            if(!issuerOfOCES2) {
                issuerOfOCES3 = OCESUtil.isProbableOCES3Certificate(cert) && OCESUtil.isIssuerOf(cert, getAndValidateIntermediateOCES3Certificate(cert));	
            }

            if(!(issuerOfOCES2 || issuerOfOCES3)) {
                DistinguishedName dn  = new DistinguishedName(cert.getSubjectX500Principal()) ;
                throw new PKIException("The supplied certificate  with DN '" + dn + "' is not a " + getCertificationAuthorityName() + " certificate");            	
            }

            certificateStatus = checkRevocation(cert);
        }
        auditLogCertificationStatus(certificateStatus, cert);
        return certificateStatus;
    }

    private void auditLogCertificationStatus(CertificateStatus certificateStatus, Object certificate) {
        if (certificateStatus.isValid()) {
            eventHandler.onInformationalAuditingEvent(AuditEventHandler.EVENT_TYPE_INFO_CERTIFICATE_VALIDATED, new Object[]{certificate});
        } else {
            eventHandler.onErrorAuditingEvent(AuditEventHandler.EVENT_TYPE_ERROR_VALIDATING_CERTIFICATE, new Object[]{certificate});
        }
    }

    protected abstract X509Certificate getOCES1RootCertificate();

    protected abstract X509Certificate getOCES2RootCertificate();

    protected abstract List<X509Certificate> getOCES3RootCertificates();

    protected abstract String getCertificationAuthorityName();

    // visible for unit-testing
    /*pp*/ boolean checkDates(X509Certificate certificate) {
        return getStatusOnDates(certificate).isValid();
    }

    /*pp*/ CertificateStatus getStatusOnDates(X509Certificate certificate) {
        if (certificate.getNotAfter().getTime() < System.currentTimeMillis()) {
            return new CertificateStatus(false, null, "Certificate is expired.");
        } else if (certificate.getNotBefore().getTime() > System.currentTimeMillis()) {
            return new CertificateStatus(false, null, "Certificate is not yet valid.");
        }
        return new CertificateStatus(true, null, null);
    }

    // visible for unit-testing
    /* pp */ X509Certificate getAndValidateIntermediateOCES2Certificate(X509Certificate certificate) throws PKIException {
        X509Certificate intermediateCertificate = intermediateCertificateCache.getCertificate(OCESUtil.retrieveIntermediateCertificateURI(certificate));

        if (!checkDates(intermediateCertificate)) {
            throw new PKIException("Intermediate certificate not valid in time");
        }
        if (!OCESUtil.isIssuerOf(intermediateCertificate, getOCES2RootCertificate())) {
            throw new PKIException("Intermediate certificate not issued by " + getCertificationAuthorityName() + " root certificate");
        }
        if (!checkRevocation(intermediateCertificate).isValid()) {
            throw new PKIException("Intermediate certificate is revoked");
        }
        return intermediateCertificate;
    }

    // visible for unit-testing
    /* pp */ X509Certificate getAndValidateIntermediateOCES3Certificate(X509Certificate certificate) throws PKIException {
        X509Certificate intermediateCertificate = intermediateCertificateCache.getCertificate(OCESUtil.retrieveIntermediateCertificateURI(certificate));

        if (!checkDates(intermediateCertificate)) {
            throw new PKIException("Intermediate certificate not valid in time");
        }
        if (!OCESUtil.isIssuerOf(intermediateCertificate, getOCES3RootCertificates())) {
            throw new PKIException("Intermediate certificate not issued by " + getCertificationAuthorityName() + " root certificate");
        }
        if (!checkRevocation(intermediateCertificate).isValid()) {
            throw new PKIException("Intermediate certificate is revoked");
        }
        return intermediateCertificate;
    }

    private CertificateStatus checkRevocation(X509Certificate certificate) {
        try {
            return certificateStatusChecker.getRevocationStatus(certificate);
        } catch (PKIException e) {
            eventHandler.onWarningAuditingEvent(e.getMessage(), null);
            throw e;
        }
    }

    public X509Certificate getFederationCertificate(FederationCertificateReference reference) {
        return federationCertificateResolver.getFederationCertificate(reference);
    }
    
    public static List<X509Certificate> asCertificates(String... base64data){
    	return Stream.of(base64data).map(e -> asCertificate(e)).collect(Collectors.toList());
    }
    
    public static X509Certificate asCertificate(String base64data){
    	return CertificateParser.asCertificate(XmlUtil.fromBase64(base64data));
    }
}