/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.pki;

import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

import dk.sosi.seal.pki.impl.HashMapCertificateCache;
import dk.sosi.seal.pki.impl.PropertiesSOSIConfiguration;
import dk.sosi.seal.pki.impl.federationcert.FederationCertificateStoreAdapter;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

/**
 * Implementation of CertificationAuthority representing the OCES test CA.
 *
 * @author ads@lakeside.dk
 * @since 2.0
 */
public class OCESTestCertificationAuthority extends AbstractOCESCertificationAuthority {

	private static final String OCES1_TEST_ROOT_CERTIFICATE_BASE_64 =
			"MIIEXTCCA8agAwIBAgIEQDYX/DANBgkqhkiG9w0BAQUFADA/MQswCQYDVQQGEwJE" +
					"SzEMMAoGA1UEChMDVERDMSIwIAYDVQQDExlUREMgT0NFUyBTeXN0ZW10ZXN0IENB" +
					"IElJMB4XDTA0MDIyMDEzNTE0OVoXDTM3MDYyMDE0MjE0OVowPzELMAkGA1UEBhMC" +
					"REsxDDAKBgNVBAoTA1REQzEiMCAGA1UEAxMZVERDIE9DRVMgU3lzdGVtdGVzdCBD" +
					"QSBJSTCBnzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEArawANI56sljDsnosDU+M" +
					"p4r+RKFys9c5qy8jWZyA+7PYFs4+IZcFxnbNuHi8aAcbSFOUJF0PGpNgPEtNc+XA" +
					"K7p16iawNTYpMkHm2VoInNfwWEj/wGmtb4rKDT2a7auGk76q+Xdqnno4PRO8e7AK" +
					"EHw7pN3kiHmZCI48PTRpRx8CAwEAAaOCAmQwggJgMA8GA1UdEwEB/wQFMAMBAf8w" +
					"DgYDVR0PAQH/BAQDAgEGMIIBAwYDVR0gBIH7MIH4MIH1BgkpAQEBAQEBAQEwgecw" +
					"LwYIKwYBBQUHAgEWI2h0dHA6Ly93d3cuY2VydGlmaWthdC5kay9yZXBvc2l0b3J5" +
					"MIGzBggrBgEFBQcCAjCBpjAKFgNUREMwAwIBARqBl1REQyBUZXN0IENlcnRpZmlr" +
					"YXRlciBmcmEgZGVubmUgQ0EgdWRzdGVkZXMgdW5kZXIgT0lEIDEuMS4xLjEuMS4x" +
					"LjEuMS4xLjEuIFREQyBUZXN0IENlcnRpZmljYXRlcyBmcm9tIHRoaXMgQ0EgYXJl" +
					"IGlzc3VlZCB1bmRlciBPSUQgMS4xLjEuMS4xLjEuMS4xLjEuMS4wEQYJYIZIAYb4" +
					"QgEBBAQDAgAHMIGWBgNVHR8EgY4wgYswVqBUoFKkUDBOMQswCQYDVQQGEwJESzEM" +
					"MAoGA1UEChMDVERDMSIwIAYDVQQDExlUREMgT0NFUyBTeXN0ZW10ZXN0IENBIElJ" +
					"MQ0wCwYDVQQDEwRDUkwxMDGgL6AthitodHRwOi8vdGVzdC5jcmwub2Nlcy5jZXJ0" +
					"aWZpa2F0LmRrL29jZXMuY3JsMCsGA1UdEAQkMCKADzIwMDQwMjIwMTM1MTQ5WoEP" +
					"MjAzNzA2MjAxNDIxNDlaMB8GA1UdIwQYMBaAFByYCUcaTDi5EMUEKVvx9E6Aasx+" +
					"MB0GA1UdDgQWBBQcmAlHGkw4uRDFBClb8fROgGrMfjAdBgkqhkiG9n0HQQAEEDAO" +
					"GwhWNi4wOjQuMAMCBJAwDQYJKoZIhvcNAQEFBQADgYEApyoAjiKq6WK5XaKWUpVs" +
					"kutzohv1VcCke/3JeUVtmB+byexJMC171s4RHoqcbufcI2ASVWwu84i45MaKg/nx" +
					"oqojMyY19/W2wbQFEdsxUCnLa9e9tlWj0xS/AaKeUhk2MBOqv+hMdc71jOqc5JN7" +
					"T2Ba6ZRIY5uXkO3IGZ3XUsw=";

	private static final String OCES2_TEST_IG_ROOT_CERTIFICATE_BASE_64 =
			"MIIGRTCCBC2gAwIBAgIETHO9tTANBgkqhkiG9w0BAQsFADBOMQswCQYDVQQGEwJE" +
					"SzESMBAGA1UEChMJVFJVU1QyNDA4MSswKQYDVQQDEyJUUlVTVDI0MDggU3lzdGVt" +
					"dGVzdCBJWCBQcmltYXJ5IENBMB4XDTEwMDgyNDEyMTAyM1oXDTM3MTIyNDEyNDAy" +
					"M1owTjELMAkGA1UEBhMCREsxEjAQBgNVBAoTCVRSVVNUMjQwODErMCkGA1UEAxMi" +
					"VFJVU1QyNDA4IFN5c3RlbXRlc3QgSVggUHJpbWFyeSBDQTCCAiIwDQYJKoZIhvcN" +
					"AQEBBQADggIPADCCAgoCggIBAMgd7UdslIik/4S2EF+i37FaxHOD+tvtJQgeMAei" +
					"0kOBFtCuu+tz6uJGWOVDRvh6SyTncdZGAlRKNZAK+ZULUnU1pdB2fbV9rhLF4q0M" +
					"BGSgjUd+DpQhUmLi2QLaZvfmmTz4melVewCtYjqCRzPULHetHQKCQIduIhMfR0EE" +
					"e38Ooy6PwLEUrYbKyq6rd0Xf2jcSV0srM3INfEULmeWld/kYPI8SH6M/XXiyvhFv" +
					"ymAYY3v9XlAWUtTSnJmqs1yU6xpQG1VwRsHQSDvyWmPluGKwELCLWKXK2sNco6Yy" +
					"RwNGcnhsjM2kPZ8nhgDJNVFFdd9AjD/qAeex54n+sJHMH/WtmOz9HWeQYrbGO+lW" +
					"W/ZXss8Z+KlMzje3pWgxYIhK8OZoRvoUKoLQ1JJH/KjgwcZxuxKzGm7uwoLGHUjg" +
					"Yr/1TzJT+sddLTK9TNL2SOwATbg+ueZ7kqIt7Uxih9203b4Y1x1rtIxa7zxtZ4Fc" +
					"MvOc8rVfEnanBdhC1nUCThPivf6HrsybD3FG/22FQdq/7ZmcOB2avn4Z1F983Wlc" +
					"o6etLHsHfqDy771bMO83aLp/bHBHqOUG7bnNaSegmK5blfEBmYkzAXFaxQnr02LK" +
					"7v54dCO8lzBya/06erErdTywSRGLN/+We/h2NVGDokv6remDdAC0XFIs4WrTSvYg" +
					"oiP/AgMBAAGjggEpMIIBJTAPBgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIB" +
					"BjARBgNVHSAECjAIMAYGBFUdIAAwga4GA1UdHwSBpjCBozA6oDigNoY0aHR0cDov" +
					"L2NybC5zeXN0ZW10ZXN0OS50cnVzdDI0MDguY29tL3N5c3RlbXRlc3Q5LmNybDBl" +
					"oGOgYaRfMF0xCzAJBgNVBAYTAkRLMRIwEAYDVQQKEwlUUlVTVDI0MDgxKzApBgNV" +
					"BAMTIlRSVVNUMjQwOCBTeXN0ZW10ZXN0IElYIFByaW1hcnkgQ0ExDTALBgNVBAMT" +
					"BENSTDEwHwYDVR0jBBgwFoAUAMhRPjg1v23MAbpjBIk5L7AlcdowHQYDVR0OBBYE" +
					"FADIUT44Nb9tzAG6YwSJOS+wJXHaMA0GCSqGSIb3DQEBCwUAA4ICAQAk/ghXxPKM" +
					"5E/VwViE0UtJQKBzsaCT33Jzqx081Cmt8mfQTEhpVhiE3jMkYYj5kaN0qqHfuvip" +
					"mcpjs4qs38lpZGR13XeuHKY5QLEKo7L14DxhmJi3nfBIUMdcplQpvGZFr9zmyWZ3" +
					"DUXNdLfKLwXXZHJB5+N3TrOk/11yksibNLEDLpS/tCjYKZI3VKL/6QDdFbR1JjCy" +
					"t6hUeCG4Do2SIggst3oiKRcuPYkX6kukm1V5+vY8i0zRd48jKh3oPQFyi5StD1+o" +
					"uHYLHDr5UgueC77xJ3ZcVpyToxJjc2mxqovB5r2Zrfs9JdT/iLQDs5kvpkOuZL8F" +
					"4yPj3PgNvz1WZkQq/QwlO6EdwoAiLTzWxlnTSQ2XGYEjREkOglrLuRoBWz89ZgMC" +
					"xrMfPWbCRyTC6i5MRNmdRKUtqhe/KO2oSuO1RioIO0sTe2tnkiEmIN7kXD92R1KL" +
					"JCZB2NFaWOv+yU1GvpER2gXrlvq/yoFuU8g+72BT6UiaCsmr7L1iK7poJKDClS+A" +
					"t+5/+gvQRq9BjGtR/q4d3B8xL8Mg58rZbf6FHas6cb0c3e9iVtqSQviXO6VYPQch" +
					"X8rjBrXViDvlKXa3fwu6pzhJhJQnsM0jgSV7wEQfoRoTvkXPxwik1xyroV3qKIhx" +
					"y3pgq7fDfTxMgVDvMIhjU0+ZQ/DP4ska2g==";

	private static final String OCES2_TEST_PP_ROOT_CERTIFICATE_BASE_64 =
			"MIIGSDCCBDCgAwIBAgIES+pulDANBgkqhkiG9w0BAQsFADBPMQswCQYDVQQGEwJE" +
					"SzESMBAGA1UEChMJVFJVU1QyNDA4MSwwKgYDVQQDEyNUUlVTVDI0MDggU3lzdGVt" +
					"dGVzdCBWSUkgUHJpbWFyeSBDQTAeFw0xMDA1MTIwODMyMTRaFw0zNzAxMTIwOTAy" +
					"MTRaME8xCzAJBgNVBAYTAkRLMRIwEAYDVQQKEwlUUlVTVDI0MDgxLDAqBgNVBAMT" +
					"I1RSVVNUMjQwOCBTeXN0ZW10ZXN0IFZJSSBQcmltYXJ5IENBMIICIjANBgkqhkiG" +
					"9w0BAQEFAAOCAg8AMIICCgKCAgEApuuMpdHu/lXhQ+9TyecthOxrg5hPgxlK1rpj" +
					"syBNDEmOEpmOlK8ghyZ7MnSF3ffsiY+0jA51p+AQfYYuarGgUQVO+VM6E3VUdDpg" +
					"WEksetCYY8L7UrpyDeYx9oywT7E+YXH0vCoug5F9vBPnky7PlfVNaXPfgjh1+66m" +
					"lUD9sV3fiTjDL12GkwOLt35S5BkcqAEYc37HT69N88QugxtaRl8eFBRumj1Mw0LB" +
					"xCwl21GdVY4EjqH1Us7YtRMRJ2nEFTCRWHzm2ryf7BGd80YmtJeL6RoiidwlIgzv" +
					"hoFhv4XdLHwzaQbdb9s141q2s9KDPZCGcgIgeXZdqY1Vz7UBCMiBDG7q2S2ni7wp" +
					"UMBye+iYVkvJD32srGCzpWqG7203cLyZCjq2oWuLkL807/Sk4sYleMA4YFqsazIf" +
					"V+M0OVrJCCCkPysS10n/+ioleM0hnoxQiupujIGPcJMA8anqWueGIaKNZFA/m1IK" +
					"wnn0CTkEm2aGTTEwpzb0+dCATlLyv6Ss3w+D7pqWCXsAVAZmD4pncX+/ASRZQd3o" +
					"SvNQxUQr8EoxEULxSae0CPRyGwQwswGpqmGm8kNPHjIC5ks2mzHZAMyTz3zoU3h/" +
					"QW2T2U2+pZjUeMjYhyrReWRbOIBCizoOaoaNcSnPGUEohGUyLPTbZLpWsm3vjbyk" +
					"7yvPqoUCAwEAAaOCASowggEmMA8GA1UdEwEB/wQFMAMBAf8wDgYDVR0PAQH/BAQD" +
					"AgEGMBEGA1UdIAQKMAgwBgYEVR0gADCBrwYDVR0fBIGnMIGkMDqgOKA2hjRodHRw" +
					"Oi8vY3JsLnN5c3RlbXRlc3Q3LnRydXN0MjQwOC5jb20vc3lzdGVtdGVzdDcuY3Js" +
					"MGagZKBipGAwXjELMAkGA1UEBhMCREsxEjAQBgNVBAoTCVRSVVNUMjQwODEsMCoG" +
					"A1UEAxMjVFJVU1QyNDA4IFN5c3RlbXRlc3QgVklJIFByaW1hcnkgQ0ExDTALBgNV" +
					"BAMTBENSTDEwHwYDVR0jBBgwFoAUI7pMMZDh08zTG7MbWrbIRc3Tg5cwHQYDVR0O" +
					"BBYEFCO6TDGQ4dPM0xuzG1q2yEXN04OXMA0GCSqGSIb3DQEBCwUAA4ICAQCRJ9TM" +
					"7sISJBHQwN8xdey4rxA0qT7NZdKICcIxyIC82HIOGAouKb3oHjIoMgxIUhA3xbU3" +
					"Putr4+Smnc1Ldrw8AofLGlFYG2ypg3cpF9pdHrVdh8QiERozLwfNPDgVeCAnjKPN" +
					"t8mu0FWBS32tiVM5DEOUwDpoDDRF27Ku9qTFH4IYg90wLHfLi+nqc2HwVBUgDt3t" +
					"XU6zK4pzM0CpbrbOXPJOYHMvaw/4Em2r0PZD+QOagcecxPMWI65t2h/USbyO/ah3" +
					"VKnBWDkPsMKjj5jEbBVRnGZdv5rcJb0cHqQ802eztziA4HTbSzBE4oRaVCrhXg/g" +
					"6Jj8/tZlgxRI0JGgAX2dvWQyP4xhbxLNCVXPdvRV0g0ehKvhom1FGjIz975/DMav" +
					"kybh0gzygq4sY9Fykl4oT4rDkDvZLYIxS4u1BrUJJJaDzHCeXmZqOhx8She+Fj9Y" +
					"wVVRGfxT4FL0Qd3WAtaCVyhSQ6SkZgrPvzAmxOUruI6XhEhYGlP5O8WFETiATxuZ" +
					"AJNuKMJtibfRhMNsQ+TVv/ZPr5Swe+3DIQtmt1MIlGlTn4k40z4s6gDGKiFwAYXj" +
					"d/kID32R/hJPE41o9+3nd8aHZhBy2lF0jKAmr5a6Lbhg2O7zjGq7mQ3MceNeebuW" +
					"XD44AxIinryzhqnEWI+BxdlFaia3U7o2+HYdHw==";

	private static final String OCES3_TEST_ROOT_CERTIFICATE_BASE_64_CTI =
			"MIIGsjCCBGagAwIBAgIUVz9X5nUw8aB3ffvGnwkEONM2AlYwQQYJKoZIhvcNAQEK" +
					"MDSgDzANBglghkgBZQMEAgMFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgMF" +
					"AKIDAgFAMGIxJDAiBgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTET" +
					"MBEGA1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQsw" +
					"CQYDVQQGEwJESzAeFw0yMTAxMjgwOTQ5MjVaFw00NjAxMjIwOTQ5MjRaMGIxJDAi" +
					"BgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTETMBEGA1UECwwKVGVz" +
					"dCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQswCQYDVQQGEwJESzCC" +
					"AiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAL1pnHOstgHrlMGwwnUMP7I6" +
					"uEgZQkzvNcdOF3sjtJh3/i/DwdpAYTEylpf0JQrJl6lJogiukL0wDe08Xvk/NGtc" +
					"0ISWV5nLnD1Vb16RWdInyuz+g7QQepMQljFfsXHixmbjA/ICbmA+/5qDRFHK49zI" +
					"UDRnhDKxgLYGBTFoIvL7mJ3SQJfzoD8Ge0vjUbNvH479k3zmnbOZaht/l4cZcPZH" +
					"0hjdLFqgIaukwNt8Yk14KeMxqNm4KzPpikGnHtZ30xqVNdCGOetSClRB9d8Qx9Hw" +
					"55/vH5/5L9CPAlYH+2TX1l9O+i6kwleE0T7NgVRTTyJitXB7KSPbnDTKH/LVHJBH" +
					"piCQBsqxJguA9k6ujWV2G/Ko2NYkKpMvxc8lQcroMx7LxknEEX/juwkO4aRFaLWQ" +
					"AAJbeo7Av/rPZpYoCN2ri8jVtPq14Aqwx1Y66heYYZPCYMPYD6F22jNcxQE1Y1NI" +
					"GcdIkx8y7j8gP0XgpioRtbz1ej+fw+vGPW1mNQ2Cucg4ZL/o40jwTyigR3qiUa2D" +
					"pDEd1nWYN3X6N1CBz1EYCGlQPNcuQ2IponQlyRCmBQ+e9N9ouPajdcq7QjAZQE78" +
					"mhSwnF996NRnjdg5sMBqRvVR7DpNRb51LitiLJRiC+gOQy94bI4Qu5ewG5ZHHNf6" +
					"PTpZ3g5GV70DsBmnAtrhAgMBAAGjgfcwgfQwDwYDVR0TAQH/BAUwAwEB/zBzBggr" +
					"BgEFBQcBAQRnMGUwOwYIKwYBBQUHMAKGL2h0dHA6Ly9jYTEuY3RpLWdvdi5kay9v" +
					"Y2VzL3Jvb3QvY2FjZXJ0L3Jvb3QuY2VyMCYGCCsGAQUFBzABhhpodHRwOi8vY2Ex" +
					"LmN0aS1nb3YuZGsvb2NzcDA9BgNVHR8ENjA0MDKgMKAuhixodHRwOi8vY2ExLmN0" +
					"aS1nb3YuZGsvb2Nlcy9yb290L2NybC9yb290LmNybDAdBgNVHQ4EFgQUOdze3tCQ" +
					"Jkeg4MZuSX8m8qkv9hswDgYDVR0PAQH/BAQDAgEGMEEGCSqGSIb3DQEBCjA0oA8w" +
					"DQYJYIZIAWUDBAIDBQChHDAaBgkqhkiG9w0BAQgwDQYJYIZIAWUDBAIDBQCiAwIB" +
					"QAOCAgEABzNBub3P/ND9hpsevGZSm0mz0OhvkRUGU1JLKpCmaXMx4lb86gxzpj4C" +
					"6wjxWfofNCsRWlLEN52aJslFjzFm7LYvi+S9LTp+yxas7JSFXXkViaDfi2y+FiIZ" +
					"IJhkQWOV1gzkGZwjTEqhpzXnYJYTyLk/MZ1E1o0EcuxDFk2izpUFDR2TdYrCy8xQ" +
					"YRTOVdB/oW7nljdQYEzaDrrnQwanAnNkqU5Lse5DQ8P5aATF+BBL7QFQwqnMiFeY" +
					"2THu948ggl7oTuaPrFysWv9JrxWRQD9NX4pC1mgUkgJnejGOihaRKjVDcJMxCe4Y" +
					"B5tpX+JhaNO+gW4YlwxSgFAb7UG+nF8qtGCRoAZQaPzWVokYyYHSW/qLMCPySiWA" +
					"9GlMEGDuiZF/xkO96vtNA2MCJnPQCW1OJw+v5IQWWFuYXCYMNzepfiGUNB+ExYP2" +
					"wuRkMOm2EBCWt8vay2/aUEQSrWAd2BuT/nUWt7vkT7cQRtiu89axJ3SZ8P+QwDr+" +
					"zoEF82tKZJ2hXuVjVaiEwxnDmv+UuQMkl52LY3XRfa6F7zO9CoTbPyCazkp+6YWg" +
					"OdttM00ic5kscFT5kOH/0Pb6pILDWDnJ5L8XqVxb3zRXO/DM/IAwF2uV1PbLq7ly" +
					"LTiVhleQahXJ2SeRnon8KdjMiCdjlpOxHZZgEENPf08RSOQqEiI=";

	private static final String OCES3_TEST_ROOT_CERTIFICATE_BASE_64 =
			"MIIGzDCCBICgAwIBAgIUL1kv/d9AJgiDja3yxq3JMLWq5ZwwQQYJKoZIhvcNAQEK" +
					"MDSgDzANBglghkgBZQMEAgMFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgMF" +
					"AKIDAgFAMG8xJDAiBgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTEg" +
					"MB4GA1UECwwXVGVzdCAtIGNlcnRpZmlrYXQubG9jYWwxGDAWBgNVBAoMD0RlbiBE" +
					"YW5za2UgU3RhdDELMAkGA1UEBhMCREswHhcNMjAwOTAzMDgwODA0WhcNNDUwODI4" +
					"MDgwODA0WjBvMSQwIgYDVQQDDBtEZW4gRGFuc2tlIFN0YXQgT0NFUyByb2QtQ0Ex" +
					"IDAeBgNVBAsMF1Rlc3QgLSBjZXJ0aWZpa2F0LmxvY2FsMRgwFgYDVQQKDA9EZW4g" +
					"RGFuc2tlIFN0YXQxCzAJBgNVBAYTAkRLMIICIjANBgkqhkiG9w0BAQEFAAOCAg8A" +
					"MIICCgKCAgEAtCAGfg+ad/fEftkUn4D0WtaRLEBKIwfjEVUR+Q+rYzd/fg2rH12+" +
					"HfIEJd1ouAOQiMuFTpnT0Wg8j3GotB5Z7ClUaGWWKBtLq2XZFj0MpFs29nzM0P5k" +
					"+1i1Nr7EdPthBKPCRPZfgVAGLawFiHEXB38VL+ufB/2g8mxacrjMBVW0svx6MBIi" +
					"vK9caoJLFFgXElvV5l/ZPh4bnZU9wkD+bQsARRovTJl762RIG2T+I+nwpLqAPwSJ" +
					"qgZ23hjw/5hTygRgclsrG0DJZbPF70pJR+Gzptle99/fB714bS9bOqVFueddshmE" +
					"/jvG6DRK+vtL2RWPHFEnIn3v8lcg9S7eqJkh6OoUJJk8iyPxGJYfIFOuy8I45ZCL" +
					"LyS/YZg19nnHZ4HfULCvnmD/VPBpaBAFnJD6f2LPw2aygJ1GZCBXzoUn7GgVzzhX" +
					"DIerVIYVyGALEQfeIL9tR+IMi0eqF7mDlJNeXmlLf5cZ5NS/YlK+9Rt+DAp29V2z" +
					"RjfbN6LH+2GnuLOgYNMnMzBjqJ9Ycm7a3a3KNYeNXtc1sOwluP+g++XjQxQvHuPc" +
					"W2TGOxAPlmR1LbVqOg/enxJVNq5XLtmKkK/uwcFy4haDxke6zHVw3WFxngWlF/9G" +
					"IWRYsnBKfCHtSdxOtyg5juiSUhXxc0PuAhRmtiynaNyWjSHS0542x9ECAwEAAaOB" +
					"9zCB9DAPBgNVHRMBAf8EBTADAQH/MHMGCCsGAQUFBwEBBGcwZTA7BggrBgEFBQcw" +
					"AoYvaHR0cDovL2NhMS5jdGktZ292LmRrL29jZXMvcm9vdC9jYWNlcnQvcm9vdC5j" +
					"ZXIwJgYIKwYBBQUHMAGGGmh0dHA6Ly9jYTEuY3RpLWdvdi5kay9vY3NwMD0GA1Ud" +
					"HwQ2MDQwMqAwoC6GLGh0dHA6Ly9jYTEuY3RpLWdvdi5kay9vY2VzL3Jvb3QvY3Js" +
					"L3Jvb3QuY3JsMB0GA1UdDgQWBBSRMCad0YUYe/4owiB6zt8NjDeNQjAOBgNVHQ8B" +
					"Af8EBAMCAQYwQQYJKoZIhvcNAQEKMDSgDzANBglghkgBZQMEAgMFAKEcMBoGCSqG" +
					"SIb3DQEBCDANBglghkgBZQMEAgMFAKIDAgFAA4ICAQBPVxY+WGnuOvC254XUDH3B" +
					"6GGKPo8wEsSomN9ofz5zYQbq8fuJlGgJLWBFRQOEifqChHXTfwrNEP75zBjPr7/S" +
					"zVvD0sO++sMhsgWRtoIvjI8c6XHG3I9XEncaVcVk7X2QoxkUgExa2y3UagGCa7WW" +
					"wGfyKiZqeJ4OBXBsk0ElZtu27Jdm04D6KXoLlp1OwFu+jc8J5d2W0u1d6+gTjwuG" +
					"pSYQJFHUjfP4oQSqevzncxnDA5Hb4p2UO6ACPj1LQpH+XVZpsWnAVQU0fAST/t3e" +
					"5Rff+Oj1ftp6B6bbM6JQAiW1zOLWp09milnEjJbalu5CDmBkMUmlGrRcueK4BDHm" +
					"Rv/cayKYV3QyOpsPVd3iBp3ITKIetr49HE/+27VDeSEFFkBiKLtQT4+Jexm032ba" +
					"XPUFu91mt58Y7ZfyIZBj7fHGjO73/YMd5cGukUDveYivX9Z+7IJOz37S55pz+fQz" +
					"aUSttu4E5arKlVnDcFYXNQs3O+MdXBRjoO+suyxx6saV4eSXpCDSNtqxSyb8I/7P" +
					"pqwlbMmG/YwtrAvJw3ap/Bs7EOF8Z2b9dw1Ealq4cEOaIO6AD//C2KeGiIHLD+d/" +
					"I6qcTPa11bxUlLO8jxmh9W8r9n+oAKUEGj9wOngsBDGKcBhOpNpfuHX+htPnJcGA" +
					"AIq8F/rdTvfewR4Jv4se5A==";

	private static final String OCES3_TEST_ROOT_CERTIFICATE_BASE_64_NEW =
			"MIIGsjCCBGagAwIBAgIUVz9X5nUw8aB3ffvGnwkEONM2AlYwQQYJKoZIhvcNAQEK"+
					"MDSgDzANBglghkgBZQMEAgMFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgMF"+
					"AKIDAgFAMGIxJDAiBgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTET"+
					"MBEGA1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQsw"+
					"CQYDVQQGEwJESzAeFw0yMTAxMjgwOTQ5MjVaFw00NjAxMjIwOTQ5MjRaMGIxJDAi"+
					"BgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTETMBEGA1UECwwKVGVz"+
					"dCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQswCQYDVQQGEwJESzCC"+
					"AiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBAL1pnHOstgHrlMGwwnUMP7I6"+
					"uEgZQkzvNcdOF3sjtJh3/i/DwdpAYTEylpf0JQrJl6lJogiukL0wDe08Xvk/NGtc"+
					"0ISWV5nLnD1Vb16RWdInyuz+g7QQepMQljFfsXHixmbjA/ICbmA+/5qDRFHK49zI"+
					"UDRnhDKxgLYGBTFoIvL7mJ3SQJfzoD8Ge0vjUbNvH479k3zmnbOZaht/l4cZcPZH"+
					"0hjdLFqgIaukwNt8Yk14KeMxqNm4KzPpikGnHtZ30xqVNdCGOetSClRB9d8Qx9Hw"+
					"55/vH5/5L9CPAlYH+2TX1l9O+i6kwleE0T7NgVRTTyJitXB7KSPbnDTKH/LVHJBH"+
					"piCQBsqxJguA9k6ujWV2G/Ko2NYkKpMvxc8lQcroMx7LxknEEX/juwkO4aRFaLWQ"+
					"AAJbeo7Av/rPZpYoCN2ri8jVtPq14Aqwx1Y66heYYZPCYMPYD6F22jNcxQE1Y1NI"+
					"GcdIkx8y7j8gP0XgpioRtbz1ej+fw+vGPW1mNQ2Cucg4ZL/o40jwTyigR3qiUa2D"+
					"pDEd1nWYN3X6N1CBz1EYCGlQPNcuQ2IponQlyRCmBQ+e9N9ouPajdcq7QjAZQE78"+
					"mhSwnF996NRnjdg5sMBqRvVR7DpNRb51LitiLJRiC+gOQy94bI4Qu5ewG5ZHHNf6"+
					"PTpZ3g5GV70DsBmnAtrhAgMBAAGjgfcwgfQwDwYDVR0TAQH/BAUwAwEB/zBzBggr"+
					"BgEFBQcBAQRnMGUwOwYIKwYBBQUHMAKGL2h0dHA6Ly9jYTEuY3RpLWdvdi5kay9v"+
					"Y2VzL3Jvb3QvY2FjZXJ0L3Jvb3QuY2VyMCYGCCsGAQUFBzABhhpodHRwOi8vY2Ex"+
					"LmN0aS1nb3YuZGsvb2NzcDA9BgNVHR8ENjA0MDKgMKAuhixodHRwOi8vY2ExLmN0"+
					"aS1nb3YuZGsvb2Nlcy9yb290L2NybC9yb290LmNybDAdBgNVHQ4EFgQUOdze3tCQ"+
					"Jkeg4MZuSX8m8qkv9hswDgYDVR0PAQH/BAQDAgEGMEEGCSqGSIb3DQEBCjA0oA8w"+
					"DQYJYIZIAWUDBAIDBQChHDAaBgkqhkiG9w0BAQgwDQYJYIZIAWUDBAIDBQCiAwIB"+
					"QAOCAgEABzNBub3P/ND9hpsevGZSm0mz0OhvkRUGU1JLKpCmaXMx4lb86gxzpj4C"+
					"6wjxWfofNCsRWlLEN52aJslFjzFm7LYvi+S9LTp+yxas7JSFXXkViaDfi2y+FiIZ"+
					"IJhkQWOV1gzkGZwjTEqhpzXnYJYTyLk/MZ1E1o0EcuxDFk2izpUFDR2TdYrCy8xQ"+
					"YRTOVdB/oW7nljdQYEzaDrrnQwanAnNkqU5Lse5DQ8P5aATF+BBL7QFQwqnMiFeY"+
					"2THu948ggl7oTuaPrFysWv9JrxWRQD9NX4pC1mgUkgJnejGOihaRKjVDcJMxCe4Y"+
					"B5tpX+JhaNO+gW4YlwxSgFAb7UG+nF8qtGCRoAZQaPzWVokYyYHSW/qLMCPySiWA"+
					"9GlMEGDuiZF/xkO96vtNA2MCJnPQCW1OJw+v5IQWWFuYXCYMNzepfiGUNB+ExYP2"+
					"wuRkMOm2EBCWt8vay2/aUEQSrWAd2BuT/nUWt7vkT7cQRtiu89axJ3SZ8P+QwDr+"+
					"zoEF82tKZJ2hXuVjVaiEwxnDmv+UuQMkl52LY3XRfa6F7zO9CoTbPyCazkp+6YWg"+
					"OdttM00ic5kscFT5kOH/0Pb6pILDWDnJ5L8XqVxb3zRXO/DM/IAwF2uV1PbLq7ly"+
					"LTiVhleQahXJ2SeRnon8KdjMiCdjlpOxHZZgEENPf08RSOQqEiI=";

	private static final String OCES3_TEST_INTERMEDIATE_CERTIFICATE_BASE_64 =
			"MIIGezCCBC+gAwIBAgIUfmDIhrulaeIVdxkaXYcDvQO4eEcwQQYJKoZIhvcNAQEK" +
					"MDSgDzANBglghkgBZQMEAgMFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgMF" +
					"AKIDAgFAMG8xJDAiBgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTEg" +
					"MB4GA1UECwwXVGVzdCAtIGNlcnRpZmlrYXQubG9jYWwxGDAWBgNVBAoMD0RlbiBE" +
					"YW5za2UgU3RhdDELMAkGA1UEBhMCREswHhcNMjAwOTAzMDgwODQ4WhcNMzAwOTAx" +
					"MDgwODQ4WjB4MS0wKwYDVQQDDCREZW4gRGFuc2tlIFN0YXQgT0NFUyB1ZHN0ZWRl" +
					"bmRlLUNBIDExIDAeBgNVBAsMF1Rlc3QgLSBjZXJ0aWZpa2F0LmxvY2FsMRgwFgYD" +
					"VQQKDA9EZW4gRGFuc2tlIFN0YXQxCzAJBgNVBAYTAkRLMIIBojANBgkqhkiG9w0B" +
					"AQEFAAOCAY8AMIIBigKCAYEAstpKXcp2IronIPJYAiaQGbpB1zDUFGHwZvcyEoch" +
					"UT0/siGbJUtJ2vkxDXmPTcowlj//bPoaU8ldp7Nh6GKaIIM0KM8Io1j+cztW0TrQ" +
					"MVwyl6SDLX/n0/NY8IAgZwqB6DIWB9iUeEghRJM33PLQ3ZHy32p00kKBp9zD+hdP" +
					"T8PtLCRGzh7RVaYzVDdk8/4VKd2/kCwCoRc+OxkfcGYIxN9OrXzgKjSoEEYLJmhN" +
					"2bLtxiAMlFIGXJvGanVVyhBeb3vxBjwCCxXSB3+fJ8kuFnkju0AwIV9e+bD+gojS" +
					"Z8sUlMh+Ms2mJzk0PnoAoEnaItI2mO6EmURaD1ZWlLwLyJ2JsrnOHuz3IJIVUtZD" +
					"vAHgfJL1+J6ZMK6ZkZNqewdmfZzaEJ+edB9IzBhZjRoZL79Fjw5R5sMyFsUeYftx" +
					"20RINmqbwKFll2v0PEFHytm+s0doA4hh2UPK9fhXCXRpp5ezbelt+uze8aHk8A0r" +
					"hTowKrdm80QmcJc//Lb1h2i9AgMBAAGjggEcMIIBGDASBgNVHRMBAf8ECDAGAQH/" +
					"AgEAMB8GA1UdIwQYMBaAFJEwJp3RhRh7/ijCIHrO3w2MN41CMHMGCCsGAQUFBwEB" +
					"BGcwZTA7BggrBgEFBQcwAoYvaHR0cDovL2NhMS5jdGktZ292LmRrL29jZXMvcm9v" +
					"dC9jYWNlcnQvcm9vdC5jZXIwJgYIKwYBBQUHMAGGGmh0dHA6Ly9jYTEuY3RpLWdv" +
					"di5kay9vY3NwMD0GA1UdHwQ2MDQwMqAwoC6GLGh0dHA6Ly9jYTEuY3RpLWdvdi5k" +
					"ay9vY2VzL3Jvb3QvY3JsL3Jvb3QuY3JsMB0GA1UdDgQWBBQrmgjrvOybyA48bdMz" +
					"1dYWJGbhBTAOBgNVHQ8BAf8EBAMCAQYwQQYJKoZIhvcNAQEKMDSgDzANBglghkgB" +
					"ZQMEAgMFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgMFAKIDAgFAA4ICAQAa" +
					"5UEls+uE3JOA9AXJ0lT8uGe7EFSGscTD5YeApZZZ5XTUSbSvZMIuB1elt7QWQw/a" +
					"MjCoojH690P/vAzvnDAKR67/0PgjN9pBB+SaFx4CovprHc2YH1MjVZixVPtuYaJz" +
					"iN/9HJEAtwYmZnvJp9hkAqkjr9w2iFOo4an5zK9U0VwFTq1oCGFgFpNpke2Wb/f2" +
					"fQGRWy/sv1p28p+VrTNXTHyDeo+fHg4dYUur0jN70ADbrAHKmgwG8VAmXY8SG1j7" +
					"weUHlX/82GlQx9DVIJISKFhn7xkD4lKR6ej3UmeyuKob37M+kc21k7ncYd95aUVq" +
					"wstUvBg1HbUUxMxpBSJMWeqAgE6oOLoE1NFH99G8/Z9JpVCqNJNIM0cP97iXjqXR" +
					"6ITXejIfBmj00s67mUaZ2iy8RTZ7Caq2h9sOJG6PK50pUMxiTh0CMPRcyP2+cmsH" +
					"UoFU+agL9YHEP+9WQETokpDqN+cGybhGSMYLR+jyzx8pU9RSKUKEAHiGCHJx6yeP" +
					"wa2VyMli9eCL2qDkumGtBUU5Dumeso3cY+57VwQXdwCRjru92UbZor49lzQndDwU" +
					"HUOdJgiiZ8BazH1jjr+AhUOW/YXIAlfv6uZAqdPy3E1proehpVPcVzvO33c50ViP" +
					"AuLD3rfLfy0wIqYi3tt4JA/ruGCNWiDnpO6xgAP6kg==";

	private static final String OCES3_TEST_INTERMEDIATE_CERTIFICATE_BASE_64_NEW =
					"MIIGYTCCBBWgAwIBAgIUc01wpi5SHNF8ZEyZoJAOHVngYa4wQQYJKoZIhvcNAQEK"+
					"MDSgDzANBglghkgBZQMEAgMFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgMF"+
					"AKIDAgFAMGIxJDAiBgNVBAMMG0RlbiBEYW5za2UgU3RhdCBPQ0VTIHJvZC1DQTET"+
					"MBEGA1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQsw"+
					"CQYDVQQGEwJESzAeFw0yMTAzMTIwNzQyNTlaFw0zMTAzMTAwNzQyNThaMGsxLTAr"+
					"BgNVBAMMJERlbiBEYW5za2UgU3RhdCBPQ0VTIHVkc3RlZGVuZGUtQ0EgMTETMBEG"+
					"A1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQswCQYD"+
					"VQQGEwJESzCCAaIwDQYJKoZIhvcNAQEBBQADggGPADCCAYoCggGBANOLofW5cEL4"+
					"GKGtDfVb75BI+UXI5nl/gnoi1CJcsirR+07jFP4x7mSZe/Olr0z1M3boTVaLCtdZ"+
					"urqR6AysVI9z9ymzE8zCLxE5U1x3ay/x/0J0iTxpqHqDhSF0OH+Ymb1vgp+9ta6k"+
					"PweSdG+WjJL+5xkeynGP2Zd80O2RusLAjx7s7b8sR73ImevflENJXNAJwL6MQrx4"+
					"wiTZYdPOVOSyWKTMHnBJ0MTinPyFxIv8zfbpJ2w8I+/YNTzI9EsXDDdBpQ4+asKx"+
					"A4qZY/1+OQbVTGgMTNQKl7GireR2R+cTobGiD4zzd/7mt12/gQDkjOFtSVCuOfCU"+
					"82EKilEysrnKhqFgtyGzUgRQ0PFePdI8DS50popRVtNlKKo17eE3P2GlFr9SIm9U"+
					"48GN10Eb1q+X+K1OjmNJBUVYVHF3fxC2iPySLbwkS2DSj/A6fBznj/FkIGL1BzcG"+
					"+DQPimsqKDZoIBm5SBOmi4ujoTxi+4w7h6uLyUt3D3vsPL1yUkl+qQIDAQABo4IB"+
					"HDCCARgwEgYDVR0TAQH/BAgwBgEB/wIBADAfBgNVHSMEGDAWgBQ53N7e0JAmR6Dg"+
					"xm5JfybyqS/2GzBzBggrBgEFBQcBAQRnMGUwOwYIKwYBBQUHMAKGL2h0dHA6Ly9j"+
					"YTEuY3RpLWdvdi5kay9vY2VzL3Jvb3QvY2FjZXJ0L3Jvb3QuY2VyMCYGCCsGAQUF"+
					"BzABhhpodHRwOi8vY2ExLmN0aS1nb3YuZGsvb2NzcDA9BgNVHR8ENjA0MDKgMKAu"+
					"hixodHRwOi8vY2ExLmN0aS1nb3YuZGsvb2Nlcy9yb290L2NybC9yb290LmNybDAd"+
					"BgNVHQ4EFgQUfyif2XGZQuJ159c1di5NCCVtdl4wDgYDVR0PAQH/BAQDAgEGMEEG"+
					"CSqGSIb3DQEBCjA0oA8wDQYJYIZIAWUDBAIDBQChHDAaBgkqhkiG9w0BAQgwDQYJ"+
					"YIZIAWUDBAIDBQCiAwIBQAOCAgEAuJcPSd3E6SqABGEckwvdF0W0TrScGglvOyyX"+
					"5Sx7jChPy1FFJS8vxgoqd8CYoXhzzRzuaTLLWE0P7sz51sZJKGMMA4p+f7/h4ftz"+
					"fEON1359fckE/SK5RA9IDdzVmiKxDoi8bAO/CyO/mXnZm0WFdtsU7OGgHCwAItIj"+
					"NLanSnRijazcwsNfxNJoh99S6dAV/GGzo9NMPIRrLNG3IZboPbQe07MT/DFeHP7/"+
					"xA5KxB2SdnPTfFevulJOARq4Fv+rThaSp+j67AhmzI6qrIj8r4GlCyr3+EI474ie"+
					"ZeiiClMtT+DbHpPPfDMvNk1DYOqWNRi4warsDXhVkhXdGZrAP6OE3hnVbS94q5B7"+
					"FhwugFl5XzSu06F3s3BLHrEbMCff2QZuGtf2xZIMRlWO8tivguxM2KLJYewGbOuM"+
					"uk1SQcB22fF0NQH2yzqnW6DFTGORDgYfy8IJItv82IQxKiS7/FMaoMp3mLyqb9dN"+
					"rgRyvnxwMuzLiVBDSJX0Hf78Nx/MwFvdbLawzdi9Fro9EWJ1hzjfJMZTiCP7my/E"+
					"5UJMj/Sjce3jLep25c4pu8qrY7PC6o4OqICOkLJAVNL1TU+OVpKnUd7flWBOA4rH"+
					"aR0dZY+OC/ZQhymMgT6Zg+Yj6d3cNwh8tiJTlCUYSlX27NP2HGhjZ5OlyHcOXtvQ"+
					"3VlJhI4=";

	public static final X509Certificate OCES_1_TEST_ROOT_CERTIFICATE = asCertificate(OCES1_TEST_ROOT_CERTIFICATE_BASE_64);

	public static final X509Certificate OCES_2_TEST_ROOT_CERTIFICATE = asCertificate(OCES2_TEST_PP_ROOT_CERTIFICATE_BASE_64);
	public static final List<X509Certificate> OCES_2_TEST_ROOT_CERTIFICATES = Collections.singletonList(OCES_2_TEST_ROOT_CERTIFICATE);

	public static final List<X509Certificate> OCES_3_TEST_ROOT_CERTIFICATES = asCertificates(OCES3_TEST_ROOT_CERTIFICATE_BASE_64_NEW, OCES3_TEST_ROOT_CERTIFICATE_BASE_64);

	public static final X509Certificate OCES_3_TEST_INTERMEDIATE_CERTIFICATE = asCertificate(OCES3_TEST_INTERMEDIATE_CERTIFICATE_BASE_64_NEW);
	/**
	 * @deprecated will be removed in future Seal release
	 */
	@Deprecated
	public static final X509Certificate OCES_2_TEST_IG_ROOT_CERTIFICATE = CertificateParser.asCertificate(XmlUtil.fromBase64(OCES2_TEST_IG_ROOT_CERTIFICATE_BASE_64));

	/**
	 * @deprecated use {@link #OCES_2_TEST_ROOT_CERTIFICATE}
	 */
	@Deprecated
	public static final X509Certificate OCES_2_TEST_PP_ROOT_CERTIFICATE = OCES_2_TEST_ROOT_CERTIFICATE;

	/**
	 * Constructor for the <code>OCESTestCertificationAuthority</code> class.
	 *
	 * @param properties
	 *            The initialization <code>Properties</code> of the system
	 * @deprecated will be removed in future Seal release
	 */
	@Deprecated
	public OCESTestCertificationAuthority(Properties properties, CertificateStatusChecker certificateStatusChecker, IntermediateCertificateCache intermediateCertificateCache) {
		// This is not nice at all! To be removed in next Seal release
		this(PropertiesSOSIConfiguration.createWithDefaultOcesTestProperties(properties), certificateStatusChecker, intermediateCertificateCache,
				new FederationCertificateStoreAdapter(PropertiesSOSIConfiguration.createWithDefaultOcesTestProperties(properties), new HashMapCertificateCache()));
	}

	public OCESTestCertificationAuthority(SOSIConfiguration configuration, CertificateStatusChecker certificateStatusChecker, IntermediateCertificateCache intermediateCertificateCache,
			FederationCertificateResolver federationCertificateResolver) {
		super(configuration, certificateStatusChecker, intermediateCertificateCache, federationCertificateResolver);
	}

	protected X509Certificate getOCES1RootCertificate() {
		return OCES_1_TEST_ROOT_CERTIFICATE;
	}

	@Override
	protected X509Certificate getOCES2RootCertificate() {
		return OCES_2_TEST_ROOT_CERTIFICATE;
	}

	@Override
	protected List<X509Certificate> getOCES3RootCertificates() {
		return OCES_3_TEST_ROOT_CERTIFICATES;
	}

	protected String getCertificationAuthorityName() {
		return "OCES Test";
	}

}