/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.pki;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Properties;

import org.bouncycastle.asn1.ASN1Primitive;
import org.bouncycastle.asn1.x509.AccessDescription;
import org.bouncycastle.asn1.x509.AuthorityInformationAccess;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.X509ObjectIdentifiers;
import org.bouncycastle.cert.jcajce.JcaX509ExtensionUtils;

/**
 */
public class OCESUtil {

	static URI retrieveIntermediateCertificateURI(X509Certificate certificate) throws PKIException {

		byte[] authorityInfoAccessValue = certificate.getExtensionValue(Extension.authorityInfoAccess.getId());
		if (authorityInfoAccessValue == null) {
			throw new PKIException("Invalid certificate - Authority Information Access ("+Extension.authorityInfoAccess.getId()+") not found.");
		}

		try {
			ASN1Primitive fromExtensionValue = JcaX509ExtensionUtils.parseExtensionValue(authorityInfoAccessValue);
			AuthorityInformationAccess authorityInfoAccess = AuthorityInformationAccess.getInstance(fromExtensionValue);

			AccessDescription[] accessDescriptions = authorityInfoAccess.getAccessDescriptions();
			for (AccessDescription accessDescription : accessDescriptions) {
				if (X509ObjectIdentifiers.id_ad_caIssuers.equals(accessDescription.getAccessMethod())) {
					return new URI(accessDescription.getAccessLocation().getName().toString());
				}
			}
		} catch (IOException | URISyntaxException ex) {
			throw new PKIException(ex);
		}

		throw new PKIException("Invalid certificate - CA Issuers (1.3.6.1.5.5.7.48.2) not found under Authority Information Access.");
	}
	
	private static boolean issuerNameContains(String issuerName, String str) {
		return issuerName != null && issuerName.indexOf(str) != -1;
	}
	
	private static boolean hasIssuerPrincipal(X509Certificate certificate) {
		return certificate != null && certificate.getIssuerX500Principal() != null;
	}

	public static boolean isProbableOCES1Certificate(X509Certificate certificate) {
		return hasIssuerPrincipal(certificate) && isProbableOCES1Certificate(certificate.getIssuerX500Principal().getName());
	}
	
	public static boolean isProbableOCES1Certificate(String issuerName) {
		return issuerNameContains(issuerName, "TDC OCES");
	}

	public static boolean isProbableOCES2Certificate(X509Certificate certificate) {
		return hasIssuerPrincipal(certificate) && isProbableOCES2Certificate(certificate.getIssuerX500Principal().getName());
	}

	public static boolean isProbableOCES2Certificate(String issuerName) {
		return issuerNameContains(issuerName, "TRUST2408");
	}

	public static boolean isProbableOCES3Certificate(X509Certificate certificate) {
		return hasIssuerPrincipal(certificate) && isProbableOCES3Certificate(certificate.getIssuerX500Principal().getName());
	}

	public static boolean isProbableOCES3Certificate(String issuerName) {
		return issuerNameContains(issuerName, "Den Danske Stat OCES");
	}

	public static boolean isProbableIntermediateOrRootCertificate(X509Certificate certificate) {
		return hasIssuerPrincipal(certificate) && isProbableIntermediateOrRootCertificate(certificate.getIssuerX500Principal().getName());
	}

	public static boolean isProbableIntermediateOrRootCertificate(String issuerName) {
		return 	/* OCES2 */(isProbableOCES2Certificate(issuerName) && issuerName.indexOf("Primary") != -1) ||
				/* OCES3 */(isProbableOCES3Certificate(issuerName) && issuerName.indexOf("rod-CA") != -1);
	}
	
	static boolean isIssuerOf(X509Certificate certificate, List<X509Certificate> verifyAgainstList) throws PKIException {
		return getIssuerOf(certificate, verifyAgainstList) != null;
	}

	static X509Certificate getIssuerOf(X509Certificate certificate, List<X509Certificate> verifyAgainstList) throws PKIException {
		PKIException pkiException = null;

		for (X509Certificate verifyAgainst : verifyAgainstList) {
			try {
				if (isIssuerOf(certificate, verifyAgainst)) {
					return verifyAgainst;
				}
			} catch (PKIException e) {
				if (pkiException == null) {
					pkiException = e;
				}
			}
		}

		if (pkiException != null) {
			throw pkiException;
		} else {
			return null;
		}
	}

	static boolean isIssuerOf(X509Certificate certificate, X509Certificate verifyAgainst) throws PKIException {
		if (verifyAgainst == null) {
			return false;
		}

		try {
			certificate.verify(verifyAgainst.getPublicKey());
			return true; // NOPMD
		} catch (InvalidKeyException e) {
			return false; // NOPMD
		} catch (CertificateException e) {
			throw new PKIException("Failed to establish issuer of");
		} catch (NoSuchAlgorithmException e) {
			throw new PKIException("Failed to establish issuer of");
		} catch (NoSuchProviderException e) {
			throw new PKIException("Failed to establish issuer of");
		} catch (SignatureException e) {
			return false;
		}
	}

	@Deprecated
	// Not used
	public static String getPropertyNotNull(String propertyName, Properties props) {
		String value = props.getProperty(propertyName);
		if (value == null) {
			throw new IllegalArgumentException("Property '" + propertyName + "' is not defined.");
		}
		return value;
	}
}
