/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.xml;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.Tag;
import dk.sosi.seal.model.constants.Tag.ID;
import dk.sosi.seal.model.constants.WSUAttributes;

public class TagUtil {

	/**
	 * Like calling createElementNS(doc, tag, true)
	 * 
	 * @param doc
	 * @param tag
	 * @return
	 */
	public static Element createElementNS(Document doc, Tag tag) {
		return createElementNS(doc, tag, true);
	}

	public static Tag getTag(Element elm) {
		return create(elm.getNamespaceURI(), elm.getPrefix(), elm.getLocalName());
	}

	/**
	 * Create Element with tag values using Document
	 * 
	 * @param doc
	 * @param tag
	 * @param usePrefix
	 * @return
	 */
	public static Element createElementNS(Document doc, Tag tag, boolean usePrefix) {
		Element element = doc.createElementNS(tag.getNS(), getQName(tag, usePrefix));
		if (tag.getIdAttribute() != null) {
			element.setAttributeNS(tag.getIdAttribute().getNS(), tag.getIdAttribute().getKey(), tag.getIdAttribute().getValue());
			element.setIdAttribute(tag.getIdAttribute().getKey(), true);
		}

		return element;
	}

	/**
	 * Same as tag.getTagName()
	 * 
	 * @param tag
	 * @param usePrefix
	 * @return
	 */
	public static String getLocalPart(Tag tag) {
		return tag.getTagName();
	}

	/**
	 * Same as tag.getPrefix()
	 * 
	 * @param tag
	 * @param usePrefix
	 * @return
	 */
	public static String getPrefix(Tag tag) {
		return tag.getPrefix();
	}

	/**
	 * Same as calling getName(tag, false)
	 * 
	 * @param tag
	 * @param usePrefix
	 * @return
	 */
	public static String getQName(Tag tag) {
		return getQName(tag, false);
	}

	/**
	 * Get name of tag - if usePrefix is <code>true</code> then name is qualified
	 * name otherwise it is tags local name
	 * 
	 * @param tag
	 * @param usePrefix
	 * @return
	 */
	public static String getQName(Tag tag, boolean usePrefix) {
		return usePrefix && hasPrefix(tag) ? getPrefixedName(tag) : getUnprefixedName(tag);
	}

	/**
	 * Is prefix set on tag
	 * 
	 * @param tag
	 * @return
	 */
	public static boolean hasPrefix(Tag tag) {
		return getPrefix(tag) != null;
	}

	/**
	 * Qualified name of tag is tags prefix ang tags localname - seperated with ':'
	 * This method does not check if either values on tag is <code>null</code>
	 * 
	 * @param tag
	 * @return
	 */
	public static String getUnprefixedName(Tag tag) {
		return getLocalPart(tag);
	}

	/**
	 * Qualified name of tag is tags prefix ang tags localname - seperated with ':'
	 * This method does not check if either values on tag is <code>null</code>
	 * 
	 * @param tag
	 * @return
	 */
	public static String getPrefixedName(Tag tag) {
		return getPrefix(tag) + ":" + getLocalPart(tag);
	}

	/**
	 * Same as calling getFirstChildElementNS(elm, tag, false);
	 * 
	 * @param elm
	 * @param tag
	 * @return
	 */
	public static Element getFirstChildElementNS(Element elm, Tag tag) {
		return getFirstChildElementNS(elm, tag, false);
	}

	/**
	 * Same as calling getFirstChildElementNS(elm, tag, false);
	 * 
	 * @param elm
	 * @param tag
	 * @return
	 */
	public static Element getFirstChildElementNS(Element elm, Tag... tags) {
		Element parent = elm;
		for (Tag tag : tags) {
			parent = getFirstChildElementNS(parent, tag);
		}
		return parent;
	}

	/**
	 * Same as calling XmlUtil.getFirstChildElementNS(elm, tag.getNS(),
	 * tag.getTagName()) If usePrefix is <code>true</code> then same as calling
	 * XmlUtil.getFirstChildElementNS(elm, tag.getNS(),
	 * tag.getPrefix()+":"+tag.getTagName())
	 * 
	 * @param elm
	 * @param tag
	 * @param usePrefix
	 * @return
	 */
	public static Element getFirstChildElementNS(Element elm, Tag tag, boolean usePrefix) {
		String ns = tag.getNS() != null ? tag.getNS() : "*";
		return XmlUtil.getFirstChildElementNS(elm, ns, getQName(tag, usePrefix));
	}

	/**
	 * Same as calling getElementsByTagNameNS(elm, tag, false)
	 * 
	 * @param elm
	 * @param tag
	 * @return
	 */
	public static NodeList getElementsByTagNameNS(Element elm, Tag tag) {
		return getElementsByTagNameNS(elm, tag, false);
	}

	/**
	 * If usepPrefix is true: Same as calling elm.geElementsByTagNameNS(tag.getNS(),
	 * tag.getTagName() ) If usepPrefix is false: Same as calling
	 * elm.geElementsByTagNameNS(tag.getNS(), tag.getPrefix()+":"+tag.getTagName() )
	 * 
	 * @param elm
	 * @param tag
	 * @param usePrefix
	 * @return
	 */
	public static NodeList getElementsByTagNameNS(Element elm, Tag tag, boolean usePrefix) {
		return elm.getElementsByTagNameNS(tag.getNS(), getQName(tag, usePrefix));
	}

	public static List<Element> getChildElementsNS(Element elm, Tag tag) {
		return getChildElementsNS(elm, tag, false);
	}

	public static List<Element> getChildElementsNS(Element elm, Tag tag, boolean usePrefix) {
		return XmlUtil.getChildElementsNS(elm, tag.getNS(), getQName(tag, usePrefix));
	}

	public static Element addTag(Element documentElement, Tag... hierarchy) {
		Queue<Tag> queue = new LinkedList<>(Arrays.asList(hierarchy));
		return addTag(documentElement, queue);
	}

	public static Element setText(Element elm, String text) {
		Text textNode = elm.getOwnerDocument().createTextNode(text);
		elm.appendChild(textNode);
		return elm;
	}

	public static Element setAttribute(Element elm, String attributeName, String attributeValue) {
		elm.setAttribute(attributeName, attributeValue);
		return elm;
	}

	private static Element addTag(Element parent, Queue<Tag> queue) {
		if (parent == null || queue == null) {
			return null;
		}

		Tag tag = queue.poll();
		Tag peek = queue.peek();

		if (tag == null) {
			return null;
		}

		if (getTag(parent).equals(tag)) {
			return addTag(parent, queue);
		}

		Element child = getFirstChildElementNS(parent, tag);
		/* Child do not exist! */
		if (child == null) {
			/* Create it */
			child = TagUtil.createElementNS(parent.getOwnerDocument(), tag);
			parent.appendChild(child);

			if (peek == null) {
				return child;
			}
		}

		return addTag(child, queue);
	}

	public static Tag create(final String name) {
		return create(null, null, name);
	}

	public static Tag create(final String namespace, final String prefix, final String name) {
		return new Tag() {

			@Override
			public String getTagName() {
				return name;
			}

			@Override
			public String getPrefix() {
				return prefix;
			}

			@Override
			public String getNS() {
				return namespace;
			}

			@Override
			public boolean equals(Object obj) {
				boolean eq = false;
				if (obj instanceof Tag) {
					Tag other = (Tag) obj;

					eq = isSame(getNS(), other.getNS());
					eq = eq && isSame(getPrefix(), other.getPrefix());
					eq = eq && isSame(getTagName(), other.getTagName());
				}

				return eq;
			}

			private boolean isSame(String s1, String s2) {
				return s1 != null ? s1.equals(s2) : s2 == null;
			}
		};
	}

	public static Tag.ID createId(final Tag tag) {
		String value = decapitalize(tag.getTagName());
		return createId(NameSpaces.WSU_SCHEMA, WSUAttributes.ID_PREFIXED, value);

	}

	public static Tag.ID createId(final String ns, final String key, final String value) {
		return new ID() {
			@Override
			public String getNS() {
				return ns;
			}

			@Override
			public String getValue() {
				return value;
			}
			
			@Override
			public String getKey() {
				return key;
			}
		};
	}

	protected static String decapitalize(String string) {
		if (string == null || string.length() == 0) {
			return string;
		}

		char c[] = string.toCharArray();
		c[0] = Character.toLowerCase(c[0]);

		return new String(c);
	}
}
