/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal;

import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.*;
import dk.sosi.seal.model.dombuilders.*;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.modelbuilders.OIOSAMLAssertionToIDCardRequestModelBuilder;
import dk.sosi.seal.pki.*;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.util.SOSITestUtils;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.Date;

import static dk.sosi.seal.model.BlurredOrganisation.OrgType.Cvr;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.FromRelatedPerson;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.SpecificForPerson;
import static dk.sosi.seal.model.OnBehalfOfClaim.OnBehalfOfClaimType.*;
import static dk.sosi.seal.model.constants.HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR;
import static dk.sosi.seal.xml.XmlUtil.*;
import static org.junit.Assert.*;

/**
 */
@RunWith(Parameterized.class)
public class OIOSAMLFactoryTest {

	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public CredentialVault vocesVault;

	@Parameter(2)
	public CredentialVault mocesVault;

	@Parameter(3)
	public CredentialVault untrusted;

	@Parameter(4)
	public CredentialVault holderOfKey;

	private OIOSAMLFactory factory;

	/* Setting up test parameters */
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();
		builder.newTest("OCES3").add(/* VOCES */CredentialVaultTestUtil.getVoces3CredentialVault(),
				/* MOCES */CredentialVaultTestUtil.getMoces3CredentialVault(),
				/* UNTRUSTED */ CredentialVaultTestUtil.getVoces2CredentialVault(),
				/* HOK */ CredentialVaultTestUtil.getVoces3CredentialVault());
		return builder.build();
	}

	@Before
	public void setUp() {
		factory = new OIOSAMLFactory();
	}

	@Test
	public void testOIOSAMLToIDCardRequest() {

		OIOSAMLAssertionToIDCardRequestDOMBuilder domBuilder = factory.createOIOSAMLAssertionToIDCardRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setOIOSAMLAssertion(parseOIOSAMLAssertion());
		domBuilder.setITSystemName("EMS");
		domBuilder.setUserAuthorizationCode("2345C");
		domBuilder.setUserEducationCode("7170");
		domBuilder.setUserGivenName("Fritz");
		domBuilder.setUserSurName("Müller");
		Document requestDoc = domBuilder.build();

		OIOSAMLAssertionToIDCardRequest assertionToIDCardRequest = factory.createOIOSAMLAssertionToIDCardRequestModelBuilder().build(requestDoc);
		assertEquals("EMS", assertionToIDCardRequest.getITSystemName());
		assertEquals("2345C", assertionToIDCardRequest.getUserAuthorizationCode());
		assertEquals("7170", assertionToIDCardRequest.getUserEducationCode());
		assertEquals("Fritz", assertionToIDCardRequest.getUserGivenName());
		assertEquals("Müller", assertionToIDCardRequest.getUserSurName());
		assertEquals("http://sosi.dk", assertionToIDCardRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionToIDCardRequest.getAction());

		assertionToIDCardRequest.validateSignature();
		assertionToIDCardRequest.validateSignatureAndTrust(vocesVault);

		try {
			assertionToIDCardRequest.validateSignatureAndTrust(untrusted);
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), assertionToIDCardRequest.getSigningCertificate());

		OIOSAMLAssertion assertion = assertionToIDCardRequest.getOIOSAMLAssertion();
		assertEquals("25450442", assertion.getCvrNumberIdentifier());
		assertEquals("_020a271f-be04-4378-84ce-0a3ccae065c1", assertion.getID());
		assertion.validateSignatureAndTrust(SOSITestUtils.getNewestIdPTrustVault());
		try {
			assertionToIDCardRequest.validateSignatureAndTrust(vocesVault);
		} catch (ModelException e) {
			fail("Forventede ingen fejl. Message:" + e.getMessage());
		}
	}

	@Test
	public void testOIOSAMLToIDCardResponse() {

		OIOSAMLAssertionToIDCardResponseDOMBuilder domBuilder = factory.createOIOSAMLAssertionToIDCardResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		IDCard idCard = createUserIDCard();
		domBuilder.setIDCard(idCard);
		Document responseDoc = domBuilder.build();

		// responseDoc.getDocumentElement().getFirstChild().getFirstChild().setTextContent("FOO");
		// System.out.println(XmlUtil.node2String(responseDoc, true, true));

		OIOSAMLAssertionToIDCardResponse assertionToIDCardResponse = factory.createOIOSAMLAssertionToIDCardResponseModelBuilder().build(responseDoc);
		assertEquals("1234", assertionToIDCardResponse.getRelatesTo());
		assertEquals("2345", assertionToIDCardResponse.getContext());
		assertEquals(idCard, assertionToIDCardResponse.getIDCard());
		assertEquals("http://sosi.dk", assertionToIDCardResponse.getAppliesTo());
		assertNull(assertionToIDCardResponse.getFaultActor());
		assertNull(assertionToIDCardResponse.getFaultCode());
		assertNull(assertionToIDCardResponse.getFaultString());
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), assertionToIDCardResponse.getSigningCertificate());

		assertionToIDCardResponse.validateSignature();
		assertionToIDCardResponse.validateSignatureAndTrust(getMockFederation());
		try {
			assertionToIDCardResponse.validateSignatureAndTrust(new SOSITestFederation(System.getProperties()));
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}
	}

	@Test
	public void testOIOSAMLToIDCardErrorResponse() {

		OIOSAMLAssertionToIDCardResponseDOMBuilder domBuilder = factory.createOIOSAMLAssertionToIDCardResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setFaultActor("Server");
		domBuilder.setFaultCode("666");
		domBuilder.setFaultString("A horrible error occurred!");

		Document errorResponseDoc = domBuilder.build();
		OIOSAMLAssertionToIDCardResponse errorResponse = factory.createOIOSAMLAssertionToIDCardResponseModelBuilder().build(errorResponseDoc);
		assertEquals("1234", errorResponse.getRelatesTo());
		assertEquals("Server", errorResponse.getFaultActor());
		assertEquals("666", errorResponse.getFaultCode());
		assertEquals("A horrible error occurred!", errorResponse.getFaultString());
		assertNull(errorResponse.getIDCard());
		errorResponse.validateSignature();
		errorResponse.validateSignatureAndTrust(getMockFederation());

		try {
			errorResponse.validateSignatureAndTrust(new SOSIFederation(System.getProperties()));
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}

		domBuilder.setIDCard(createUserIDCard());
		OIOSAMLAssertionToIDCardResponse wronglyBuiltErrorResponse = factory.createOIOSAMLAssertionToIDCardResponseModelBuilder().build(domBuilder.build());
		assertNull(wronglyBuiltErrorResponse.getIDCard());

	}

	@Test
	public void illustrateIDCardIssuing() {

		OIOSAMLAssertionToIDCardRequestDOMBuilder domBuilder = factory.createOIOSAMLAssertionToIDCardRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setOIOSAMLAssertion(parseOIOSAMLAssertion());
		domBuilder.setITSystemName("Harmoni/EMS");
		Document requestDoc = domBuilder.build();

		OIOSAMLAssertionToIDCardRequest assertionToIDCardRequest = new OIOSAMLAssertionToIDCardRequestModelBuilder().build(requestDoc);

		// And STS should:
		// Check signature on request and that the signing certificate is issued by the
		// correct OCES CA
		// Check signature and trust on assertion, that is check signing certificate is
		// the IdPs
		// Check validity in time
		// Check assurance level
		// etc.

		UserInfo userInfo = buildUserInfo(assertionToIDCardRequest);
		OIOSAMLAssertion assertion = assertionToIDCardRequest.getOIOSAMLAssertion();
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, assertion.getCvrNumberIdentifier(), assertion.getOrganizationName());
		SystemInfo systemInfo = new SystemInfo(careProvider, assertionToIDCardRequest.getITSystemName());

		String certHash = SignatureUtil.getDigestOfCertificate(assertion.getUserCertificate());

		UserIDCard idCard = new UserIDCard(DGWSConstants.VERSION_1_0_1, AuthenticationLevel.MOCES_TRUSTED_USER, "SOSI-STS", systemInfo, userInfo, certHash, null, null, null);

		SOSIFactory sosiFactory = new SOSIFactory(vocesVault, System.getProperties());

		// The validity gets set to 24 hours - maybe the idcard should have a shorter
		// lifetime?
		IDCard signedIdCard = sosiFactory.copyToVOCESSignedIDCard(idCard);
        signedIdCard.serialize2DOMDocument(sosiFactory, createEmptyDocument());
        // System.out.println(XmlUtil.node2String(idCardDocElement, true, true));
	}

	private Federation getMockFederation() {
		return new SOSITestFederation(System.getProperties()) {
			@Override
			public boolean isValidSTSCertificate(X509Certificate certificate) {
				return vocesVault.getSystemCredentialPair().getCertificate().equals(certificate);
			}
		};
	}

	private UserInfo buildUserInfo(OIOSAMLAssertionToIDCardRequest request) {
		OIOSAMLAssertion assertion = request.getOIOSAMLAssertion();
		String cpr = "XXXXXXXX"; // Perform lookup based on assertion.getCvrNumberIdentifier() and
									// assertion.getRidNumberIdentifier()
		String givenName;
		String surName;
		if (request.getUserGivenName() != null && request.getUserSurName() != null) {
			givenName = request.getUserGivenName();
			surName = request.getUserSurName();
		} else {
			// The IdP cannot split CommonName and neither should we (assertion.getSurName()
			// returns null)
			givenName = assertion.getCommonName();
			surName = "-";
		}
		String email = assertion.getEmail();
		String occupation = null;
		String role = "YYYYY"; // Lookup based on CPR, use request.getUserEducationCode() to pick the right one
								// (or validate)
		String authorizationCode = "ZZZZZ";// Lookup based on CPR, use request.getUserAuthorizationCode() to pick the right
											// one (or validate)
		return new UserInfo(cpr, givenName, surName, email, occupation, role, authorizationCode);
	}

	private UserIDCard createUserIDCard() {
		return createUserIDCard(new CertificateInfo(mocesVault.getSystemCredentialPair().getCertificate()).toString());
	}

	private UserIDCard createUserIDCard(String alternativeIdentifier) {
		SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "30808460", "Lægehuset på bakken");
		UserInfo userInfo = new UserInfo("1111111118", "Hans", "Dampf", null, null, "7170", "341KY");
		UserIDCard userIDCard = sosiFactory.createNewUserIDCard("IT-System", userInfo, careProvider, AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, alternativeIdentifier);
		Request tmpRequest = sosiFactory.createNewRequest(false, "flow");
		tmpRequest.setIDCard(userIDCard);
		userIDCard.sign(tmpRequest.serialize2DOMDocument(), sosiFactory.getSignatureProvider());
		return userIDCard;
	}

	private OIOSAMLAssertion parseOIOSAMLAssertion() {
		String xml = SOSITestUtils.readXMLStreamAndRemoveFormatting(this.getClass().getResourceAsStream("/oiosaml-examples/test-new-nemlogin-authentication-assertion-2.xml"));
		Document document = readXml(System.getProperties(), xml, false);
		return new OIOSAMLAssertion(document.getDocumentElement());
	}

	@Test
	public void testIDCardToOIOSAMLRequest() {

		IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = factory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setAudience("Sundhed.dk");
		UserIDCard idCard = createUserIDCard();
		domBuilder.setUserIDCard(idCard);
		Document requestDoc = domBuilder.build();

		IDCardToOIOSAMLAssertionRequest assertionRequest = factory.createIDCardToOIOSAMLAssertionRequestModelBuilder().build(requestDoc);
		assertEquals("Sundhed.dk", assertionRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionRequest.getAction());
		assertionRequest.validateSignature();
		assertionRequest.validateSignatureAndTrust(vocesVault);
		try {
			assertionRequest.validateSignatureAndTrust(untrusted);
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), assertionRequest.getSigningCertificate());

		assertEquals(idCard, assertionRequest.getUserIDCard());
		assertionRequest.getUserIDCard().validateSignature();
		assertionRequest.getUserIDCard().validateSignatureAndTrust(mocesVault);
		try {
			assertionRequest.getUserIDCard().validateSignatureAndTrust(new SOSIFederation(System.getProperties()));
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}

	}

	@Test
	public void testIDCardToOIOSAMLRequestIDCardInHeader() {

		IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = factory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();
		domBuilder.requireIDCardInSOAPHeader();
		domBuilder.setAudience("Sundhed.dk");
		UserIDCard idCard = createUserIDCard();
		domBuilder.setUserIDCard(idCard);
		Document requestDoc = domBuilder.build();

		// System.out.println(XmlUtil.node2String(requestDoc, true, true));

		IDCardToOIOSAMLAssertionRequest assertionRequest = factory.createIDCardToOIOSAMLAssertionRequestModelBuilder().build(requestDoc);
		assertEquals("Sundhed.dk", assertionRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionRequest.getAction());
		try {
			assertionRequest.validateSignature();
			fail("Forventede en fejl");
		} catch (ModelBuildException e) {
			assertEquals("Could not find Liberty signature element", e.getMessage());
		}
		assertNull(assertionRequest.getSigningCertificate());

		assertEquals(idCard, assertionRequest.getUserIDCard());
		assertionRequest.getUserIDCard().validateSignature();
		assertionRequest.getUserIDCard().validateSignatureAndTrust(mocesVault);
		try {
			assertionRequest.getUserIDCard().validateSignatureAndTrust(new SOSIFederation(System.getProperties()));
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}
	}

	@Test
	public void testIDCardToOIOSAMLRequestMissingIDCard() {

		IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = factory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();
		domBuilder.requireIDCardInSOAPHeader();
		domBuilder.setAudience("Sundhed.dk");
		UserIDCard idCard = createUserIDCard();
		domBuilder.setUserIDCard(idCard);
		Document requestDoc = domBuilder.build();

		Element idcardAssertion = (Element) requestDoc.getElementsByTagNameNS(NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ASSERTION).item(0);
		idcardAssertion.getParentNode().removeChild(idcardAssertion);

		IDCardToOIOSAMLAssertionRequest assertionRequest = factory.createIDCardToOIOSAMLAssertionRequestModelBuilder().build(requestDoc);
		assertEquals("Sundhed.dk", assertionRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionRequest.getAction());
		try {
			assertionRequest.validateSignature();
			fail("Forventede en fejl");
		} catch (ModelBuildException e) {
			assertEquals("Could not find Liberty signature element", e.getMessage());
		}
		assertNull(assertionRequest.getSigningCertificate());

		try {
			assertEquals(idCard, assertionRequest.getUserIDCard());
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertEquals("Malformed request: IDCard could not be found!", e.getMessage());
		}
	}

	@Test
	public void testOIOSAMLAssertionBuilder() {
		UserIDCard idCard = createUserIDCard();
//		SOSIFactory sosiFactory = new SOSIFactory(mocesVault, System.getProperties());
//		Element element = idCard.serialize2DOMDocument(sosiFactory, XmlUtil.createEmptyDocument());
//		System.out.println(XmlUtil.node2String(element));
		OIOSAMLAssertion assertion = createOIOSAMLAssertion(idCard);
//		System.out.println(XmlUtil.node2String(assertion.getDOM(), true, false));

		assertOIOSAMLAssertion(assertion, idCard);
	}

	@Test
	public void testOIOSAMLAssertionBuilderWithoutCertificateInfo() {
		UserIDCard idCard = createUserIDCard("something");
		OIOSAMLAssertion assertion = createOIOSAMLAssertion(idCard);

		assertEquals("NONE", assertion.getRidNumberIdentifier());
		assertEquals("NONE", assertion.getCertificateIssuer());
		assertEquals("NONE", assertion.getCertificateSerial());
		assertEquals("CVR:30808460-RID:NONE", assertion.getUID());
		assertEquals("NONE", assertion.getRidNumberIdentifier());
		assertEquals("C=DK,O=Lægehuset på bakken // CVR: 30808460,CN=Hans Dampf,Serial=CVR:30808460-RID:NONE", assertion.getSubjectNameID());
	}

	private void assertOIOSAMLAssertion(OIOSAMLAssertion assertion, UserIDCard idCard) {
		//OCES2
		if (OCESUtil.isProbableOCES2Certificate(assertion.getCertificateIssuer())) {
			assertEquals("42634739", assertion.getRidNumberIdentifier());
			assertEquals("C=DK,O=TRUST2408,CN=TRUST2408 Systemtest XXXIV CA", assertion.getCertificateIssuer());
			assertEquals("5fa6990f", assertion.getCertificateSerial());
			assertEquals("CVR:30808460-RID:42634739", assertion.getUID());
			assertEquals("C=DK,O=NETS DANID A/S // CVR:30808460,CN=TU GENEREL MOCES M CPR gyldig,Serial=CVR:30808460-RID:42634739", assertion.getSubjectNameID());
		}
		//OCES3
		if (OCESUtil.isProbableOCES3Certificate(assertion.getCertificateIssuer())) {
			assertEquals("NONE", assertion.getRidNumberIdentifier());
			assertEquals("C=DK,O=Den Danske Stat,CN=Den Danske Stat OCES udstedende-CA 1", assertion.getCertificateIssuer());
			assertEquals("229e2b5e296baf572feb5e9b6d83ee14debcac8a", assertion.getCertificateSerial());
			assertEquals("CVR:30808460-RID:NONE", assertion.getUID());
			assertEquals("C=DK,O=Testorganisation nr. 91026150,CN=Thormund Nissen,Serial=UI:DK-E:G:dc7f4a98-d4f9-4d77-84f9-ae83c4e0bf61", assertion.getSubjectNameID());
		}
		//OCES2+OCES3
		assertFalse(assertion.isYouthCertificate());
		assertNotNull(assertion.getDeliveryNotOnOrAfter());
		assertTrue(assertion.getDeliveryNotOnOrAfter() instanceof Date);
		assertEquals("http://sundhed.dk/saml/SAMLAssertionConsumer", assertion.getRecipient());
		assertEquals(idCard, assertion.getUserIDCard());
		assertion.validateSignatureAndTrust(vocesVault);
	}

	private OIOSAMLAssertion createOIOSAMLAssertion(UserIDCard idCard) {
		OIOSAMLAssertionBuilder builder = factory.createOIOSAMLAssertionBuilder();
		builder.setSigningVault(vocesVault);
		builder.setIssuer("Test STS");
		builder.setUserIdCard(idCard);
		Date now = new Date();
		builder.setNotBefore(now);
		builder.setNotOnOrAfter(new Date(now.getTime() + 60 * 60 * 1000));
		builder.setAudienceRestriction("http://sundhed.dk");
		builder.setRecipientURL("http://sundhed.dk/saml/SAMLAssertionConsumer");
		builder.setDeliveryNotOnOrAfter(new Date(now.getTime() + 5 * 60 * 1000));
		builder.includeIDCardAsBootstrapToken();
		builder.setEmptyAttributeValue("NONE");
		return builder.build();
	}

	@Test
	public void testIDCardToOIOSAMLResponse() {

		UserIDCard idCard = createUserIDCard();
		IDCardToOIOSAMLAssertionResponse response = createIdCardToOIOSAMLAssertionResponse(idCard);

		assertEquals("http://sundhed.dk", response.getAppliesTo());
		response.validateSignature();
		response.validateSignatureAndTrust(getMockFederation());

		try {
			response.validateSignatureAndTrust(new SOSIFederation(System.getProperties()));
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}

		Element decryptedAssertionElm = EncryptionUtil.decryptAndDetach(response.getEncryptedOIOSAMLAssertionElement(), vocesVault.getSystemCredentialPair().getPrivateKey());
		OIOSAMLAssertion assertion = new OIOSAMLAssertion(decryptedAssertionElm);
		assertOIOSAMLAssertion(assertion, idCard);
	}

	private IDCardToOIOSAMLAssertionResponse createIdCardToOIOSAMLAssertionResponse(UserIDCard idCard) {
		IDCardToOIOSAMLAssertionResponseDOMBuilder builder = factory.createIDCardToOIOSAMLAssertionResponseDOMBuilder();
		builder.setSigningVault(vocesVault);
		OIOSAMLAssertion assertion = createOIOSAMLAssertion(idCard);
		builder.setOIOSAMLAssertion(assertion);
		builder.setEncryptionKey(vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
		builder.setRelatesTo("1234");
		builder.setContext("5678");
		Document responseDoc = builder.build();

		// System.out.println(XmlUtil.node2String(responseDoc, true, false));

		// responseDoc = XmlUtil.readXml(System.getProperties(),
		// XmlUtil.node2String(responseDoc), false);

		return factory.createIDCardToOIOSAMLAssertionResponseModelBuilder().build(responseDoc);
	}

	@Test
	public void testIDCardToOIOSAMLErrorResponse() {

		IDCardToOIOSAMLAssertionResponseDOMBuilder domBuilder = factory.createIDCardToOIOSAMLAssertionResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setFaultActor("Server");
		domBuilder.setFaultCode("666");
		domBuilder.setFaultString("A horrible error occurred!");

		Document errorResponseDoc = domBuilder.build();
		IDCardToOIOSAMLAssertionResponse errorResponse = factory.createIDCardToOIOSAMLAssertionResponseModelBuilder().build(errorResponseDoc);
		assertEquals("1234", errorResponse.getRelatesTo());
		assertEquals("Server", errorResponse.getFaultActor());
		assertEquals("666", errorResponse.getFaultCode());
		assertEquals("A horrible error occurred!", errorResponse.getFaultString());
		assertNull(errorResponse.getEncryptedOIOSAMLAssertionElement());
		errorResponse.validateSignature();
		errorResponse.validateSignatureAndTrust(getMockFederation());

		try {
			errorResponse.validateSignatureAndTrust(new SOSIFederation(System.getProperties()));
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}

		domBuilder.setOIOSAMLAssertion(createOIOSAMLAssertion(createUserIDCard()));
		IDCardToOIOSAMLAssertionResponse wronglyBuiltErrorResponse = factory.createIDCardToOIOSAMLAssertionResponseModelBuilder().build(domBuilder.build());
		assertNull(wronglyBuiltErrorResponse.getEncryptedOIOSAMLAssertionElement());

	}

	@Test
	public void illustrateOIOSAMLAssertionIssuing() {

		IDCardToOIOSAMLAssertionRequestDOMBuilder domBuilder = factory.createIDCardToOIOSAMLAssertionRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setAudience("Sundhed.dk");
		UserIDCard idCard = createUserIDCard();
		domBuilder.setUserIDCard(idCard);
		Document requestDoc = domBuilder.build();

		IDCardToOIOSAMLAssertionRequest assertionRequest = factory.createIDCardToOIOSAMLAssertionRequestModelBuilder().build(requestDoc);

		assertionRequest.validateSignature();

		UserIDCard userIDCard = assertionRequest.getUserIDCard();

		/*
		 * Validate userIDCard: - userIDCard.validateSignatureAndTrust(federation); -
		 * userIDCard.isValidInTime() == true - etc ...
		 */

		String audience = assertionRequest.getAppliesTo();

		/*
		 * Retrieve from configuration for audience: - PublicKey - recipientURL -
		 * whether IDCard should be included in response
		 */
		PublicKey audiencePublicKey = vocesVault.getSystemCredentialPair().getCertificate().getPublicKey();
		String recipientURL = "http://sundhed.dk/saml/SAMLAssertionConsumer";

		OIOSAMLAssertionBuilder builder = factory.createOIOSAMLAssertionBuilder();
		builder.setSigningVault(vocesVault);
		builder.setIssuer("Test STS");
		builder.setUserIdCard(userIDCard);
		Date now = new Date();
		builder.setNotBefore(now);
		builder.setNotOnOrAfter(new Date(now.getTime() + 60 * 60 * 1000));
		builder.setDeliveryNotOnOrAfter(new Date(now.getTime() + 5 * 60 * 1000));
		builder.setAudienceRestriction(audience);
		builder.setRecipientURL(recipientURL);

        OIOSAMLAssertion assertion = builder.build();

		IDCardToOIOSAMLAssertionResponseDOMBuilder responseBuilder = factory.createIDCardToOIOSAMLAssertionResponseDOMBuilder();
		responseBuilder.setSigningVault(vocesVault);
		responseBuilder.setEncryptionKey(audiencePublicKey);
		responseBuilder.setOIOSAMLAssertion(assertion);
		responseBuilder.setContext(assertionRequest.getContext());
		responseBuilder.setRelatesTo(assertionRequest.getMessageID());

		Document responseDoc = responseBuilder.build();

	}

	@Test
	public void illustrateUnsolicitedResponseConstruction() {

		UserIDCard userIDCard = createUserIDCard();
		IDCardToOIOSAMLAssertionResponse oiosamlAssertionResponse = createIdCardToOIOSAMLAssertionResponse(userIDCard);

		UnsolicitedResponseDOMBuilder builder = factory.createUnsolicitedResponseDOMBuilder();
		builder.setIssuer("Harmoni/EMS");
		builder.setEncryptedAssertion(oiosamlAssertionResponse.getEncryptedOIOSAMLAssertionElement());
		Document unsolicitedResponseDoc = builder.build();

		// System.out.println(XmlUtil.node2String(unsolicitedResponseDoc, true, true));

	}

	@Test
	public void testOIOBootstrapInvalidSignature() {
		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		try {
			assertion.validateSignatureAndTrust(SOSITestUtils.getNewIdPTrustVault());
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertEquals("Signature on OIOSAMLAssertion is invalid", e.getMessage());
		}
	}

	@Test
	public void testOIOBootstrapUntrustedSigner() {
		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_valid_signature.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		try {
			assertion.validateSignatureAndTrust(SOSITestUtils.getNewIdPTrustVault());
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}
	}

	@Test
	public void testOIOBootstrapSchemaViolation() {
		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_fail_schema_validation.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		try {
			new OIOBootstrapToken(document.getDocumentElement());
			fail("Schema validation should fail");
		} catch (Exception e) {
			assertTrue(e.getMessage(), e.getMessage().contains("Invalid content was found starting with element 'Something'"));
		}

		new OIOBootstrapToken(document.getDocumentElement(), false); // This should not fail
	}

	@Test
	public void testOIOBootstrapToIdentityTokenRequestUntrustedSigner() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");

		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_valid_signature.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		assertion.validateSignatureAndTrust(SOSITestUtils.getNewestIdPTrustVault());
		domBuilder.setOIOBootstrapToken(assertion);

		Document requestDoc = domBuilder.build();

		// write to string and parse to XML
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		OIOBootstrapToIdentityTokenRequest assertionRequest = factory.createOIOBootstrapToIdentityTokenRequestModelBuilder().build(requestDoc);

		try {
			assertionRequest.validateSignatureAndTrust(untrusted);
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}
	}

	@Test
	public void testOIOBootstrapToIdentityTokenRequestBasic() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");

		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_valid_signature.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		assertion.validateSignatureAndTrust(SOSITestUtils.getNewestIdPTrustVault());
		domBuilder.setOIOBootstrapToken(assertion);

		Document requestDoc = domBuilder.build();
		// write to string and parse to XML
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		OIOBootstrapToIdentityTokenRequest assertionRequest = factory.createOIOBootstrapToIdentityTokenRequestModelBuilder().build(requestDoc);
		assertEquals("https://fmk", assertionRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionRequest.getAction());
		assertEquals("2512484916", assertionRequest.getCPRNumberClaim());
		assertNull(assertionRequest.getProcurationForCprClaim());

		OIOBSTSAMLAssertion bootstrapToken = assertionRequest.getOIOBSTSAMLAssertion();
		assertEquals(OIOBSTSAMLAssertion.Type.OIO2BST_LEGACY, bootstrapToken.getType());
		assertEquals(assertion.getAudienceRestriction(), bootstrapToken.getAudienceRestriction());
		assertEquals(assertion.getSubjectNameID(), bootstrapToken.getSubjectNameID());
		assertNull(bootstrapToken.getCpr());
		assertNull(bootstrapToken.getAssuranceLevel());
		assertNull(bootstrapToken.getSpecVersion());

		assertionRequest.validateSignature();
		assertionRequest.validateSignatureAndTrust(vocesVault);
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), assertionRequest.getSigningCertificate());
	}

	@Test
	public void testOIOBootstrapToIdentityTokenRequestWithProcuration() {
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");
		domBuilder.setOnBehalfOfClaim("1111111118", Procuration);

		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_valid_signature.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		assertion.validateSignatureAndTrust(SOSITestUtils.getNewestIdPTrustVault());
		domBuilder.setOIOBootstrapToken(assertion);

		Document requestDoc = domBuilder.build();
		// write to string and parse to XML
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		OIOBootstrapToIdentityTokenRequest assertionRequest = factory.createOIOBootstrapToIdentityTokenRequestModelBuilder().build(requestDoc);
		assertEquals("https://fmk", assertionRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionRequest.getAction());
		assertEquals("2512484916", assertionRequest.getCPRNumberClaim());
		assertEquals("1111111118", assertionRequest.getProcurationForCprClaim());
		OIOBSTSAMLAssertion bootstrapToken = assertionRequest.getOIOBSTSAMLAssertion();
		assertEquals(assertion.getAudienceRestriction(), bootstrapToken.getAudienceRestriction());
		assertEquals(assertion.getSubjectNameID(), bootstrapToken.getSubjectNameID());

		assertionRequest.validateSignature();
		assertionRequest.validateSignatureAndTrust(vocesVault);
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), assertionRequest.getSigningCertificate());
	}

	@Test
	public void testOIOBootstrapToIdentityTokenRequestWithWardCustody() {
		// Arrange
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");
		domBuilder.setOnBehalfOfClaim("1111111118", WardCustody);

		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_valid_signature.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		assertion.validateSignatureAndTrust(SOSITestUtils.getNewestIdPTrustVault());
		domBuilder.setOIOBootstrapToken(assertion);

		Document requestDoc = domBuilder.build();
		String requestDocString = node2String(requestDoc);
		requestDoc = readXml(System.getProperties(), requestDocString, false);

		// Act
		OIOBootstrapToIdentityTokenRequest assertionRequest = factory.createOIOBootstrapToIdentityTokenRequestModelBuilder().build(requestDoc);

		// Assert
		assertEquals("1111111118", assertionRequest.getWardCustodyClaim());
	}

	@Test
	public void testOIOBootstrapToIdentityTokenRequestWithPartlyWardCustody() {
		// Arrange
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");
		domBuilder.setOnBehalfOfClaim("1111111118", PartlyWardCustody);

		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_valid_signature.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		assertion.validateSignatureAndTrust(SOSITestUtils.getNewestIdPTrustVault());
		domBuilder.setOIOBootstrapToken(assertion);

		Document requestDoc = domBuilder.build();
		String requestDocString = node2String(requestDoc);
		requestDoc = readXml(System.getProperties(), requestDocString, false);

		// Act
		OIOBootstrapToIdentityTokenRequest assertionRequest = factory.createOIOBootstrapToIdentityTokenRequestModelBuilder().build(requestDoc);

		// Assert
		assertEquals("1111111118", assertionRequest.getPartlyWardCustodyClaim());
	}

	@Test
	public void testOIOBootstrapToIdentityTokenRequestWithParentalCustody() {
		// Arrange
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");
		domBuilder.setOnBehalfOfClaim("1111111118", ParentalCustody);

		InputSource inputSource = new InputSource(this.getClass().getResourceAsStream("/oiosaml-examples/OIOBootstrapToIdentityToken/NemLog-In_bootstrap_valid_signature.xml"));
		Document document = readXml(System.getProperties(), inputSource, false);
		OIOBootstrapToken assertion = new OIOBootstrapToken(document.getDocumentElement());

		assertion.validateSignatureAndTrust(SOSITestUtils.getNewestIdPTrustVault());
		domBuilder.setOIOBootstrapToken(assertion);

		Document requestDoc = domBuilder.build();
		String requestDocString = node2String(requestDoc);
		requestDoc = readXml(System.getProperties(), requestDocString, false);

		// Act
		OIOBootstrapToIdentityTokenRequest assertionRequest = factory.createOIOBootstrapToIdentityTokenRequestModelBuilder().build(requestDoc);

		// Assert
		assertEquals("1111111118", assertionRequest.getParentalCustodyClaim());
	}

	@Test
	public void testOIOBootstrapToUnencryptedIdentityTokenResponse() {

		OIOBootstrapToIdentityTokenResponseDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		domBuilder.setIdentityToken(getCitizenIdentityTokenBuilder().build());
		Document responseDoc = domBuilder.build();

		OIOBootstrapToIdentityTokenResponse bootstrapToIdentityResponse = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
	}

	@Test
	public void testOIOBootstrapIdentityTokenResponseWithPrivileges() {

		OIOBootstrapToIdentityTokenResponseDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		final CitizenIdentityTokenBuilder identityTokenBuilder = getCitizenIdentityTokenBuilder();
		BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder();
		privilegesBuilder.addPrivilege(THROUGH_PROCURATION_BY_CPR + ":" + "1111111118", "urn:dk:healthcare:fmk:read");
		privilegesBuilder.addPrivilege(THROUGH_PROCURATION_BY_CPR + ":" + "1111111118", "urn:dk:healthcare:fmk:write");
		identityTokenBuilder.setPrivileges(privilegesBuilder);
		domBuilder.setIdentityToken(identityTokenBuilder.build());
		Document responseDoc = domBuilder.build();

		OIOBootstrapToIdentityTokenResponse bootstrapToIdentityResponse = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
		assertEquals(2, identityToken.getPrivileges().getPrivileges(THROUGH_PROCURATION_BY_CPR + ":" + "1111111118").size());
		assertEquals("urn:dk:healthcare:fmk:read", identityToken.getPrivileges().getPrivileges(
                THROUGH_PROCURATION_BY_CPR + ":" + "1111111118").get(0));
		assertEquals("urn:dk:healthcare:fmk:write", identityToken.getPrivileges().getPrivileges(
                THROUGH_PROCURATION_BY_CPR + ":" + "1111111118").get(1));
	}

	@Test
	public void testOIOBootstrapIdentityTokenResponseWithSubjectRelations() {
		// Arrange
		OIOBootstrapToIdentityTokenResponseDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		CitizenIdentityTokenBuilder identityTokenBuilder = getCitizenIdentityTokenBuilder();

		SubjectRelations11Impl subjectRelations = new SubjectRelations11Impl();
		subjectRelations.addVerifiedRelation(new SubjectRelations11Impl.VerifiedRelation(
				SubjectRelations11Impl.RelationType.PARENTAL_CUSTODY_HOLDER,
				"1111111118",
				SubjectRelations11Impl.PersonIDType.CPR));
		SubjectRelationsDomBuilder subjectRelationsDomBuilder = new SubjectRelations11DomBuilder(subjectRelations);
		identityTokenBuilder.setSubjectRelations(subjectRelationsDomBuilder);
		domBuilder.setIdentityToken(identityTokenBuilder.build());
		Document responseDoc = domBuilder.build();

		// Act
		OIOBootstrapToIdentityTokenResponse bootstrapToIdentityResponse = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);

		// Assert
		assertBasicResponse(bootstrapToIdentityResponse);
		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
		SubjectRelations11Impl sr = (SubjectRelations11Impl) identityToken.getSubjectRelations();
		assertEquals(1, sr.getVerifiedRelations().size());
		assertEquals("1111111118", sr.getVerifiedRelations().get(0).getRelatedPersonID());
		assertEquals(SubjectRelations11Impl.RelationType.PARENTAL_CUSTODY_HOLDER, sr.getVerifiedRelations().get(0).getRelationType());
	}

	@Test
	public void testOIOBootstrapIdentityTokenResponseWithBlurredInstructions_v10() {
		OIOBootstrapToIdentityTokenResponseDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		final CitizenIdentityTokenBuilder identityTokenBuilder = getCitizenIdentityTokenBuilder();
		BlurringInstructions10DomBuilder blurringBuilder =
				new BlurringInstructions10DomBuilder("5kZZLNQMNIkz1Y7tCDj3GQ==");
		blurringBuilder.addBlurringInstruction("29190925", Cvr, SpecificForPerson);
		blurringBuilder.addBlurringInstruction("29190941", Cvr, FromRelatedPerson);
		identityTokenBuilder.setBlurringInstructions(blurringBuilder);
		domBuilder.setIdentityToken(identityTokenBuilder.build());
		Document responseDoc = domBuilder.build();

		OIOBootstrapToIdentityTokenResponse bootstrapToIdentityResponse = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
		identityToken.getBlurringInstructions();
		BlurringInstructions10 blurringInstructions10 = ((BlurringInstructions10)identityToken.getBlurringInstructions());
		assertEquals(2, blurringInstructions10.getBlurredOrganisations().size());
		assertEquals("29190925", blurringInstructions10.getBlurredOrganisations().get(0).getValue());
		assertEquals("29190941", blurringInstructions10.getBlurredOrganisations().get(1).getValue());
	}

	@Test
	public void testOIOBootstrapIdentityTokenResponseWithBlurredInstructions_v11() {
		OIOBootstrapToIdentityTokenResponseDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		final CitizenIdentityTokenBuilder identityTokenBuilder = getCitizenIdentityTokenBuilder();
		BlurringInstructions11DomBuilder blurringBuilder =
				new BlurringInstructions11DomBuilder("5kZZLNQMNIkz1Y7tCDj3GQ==");
		blurringBuilder.addBlurringInstruction("29190925", Cvr, SpecificForPerson);
		blurringBuilder.addBlurringInstruction("29190941", Cvr, FromRelatedPerson);
		identityTokenBuilder.setBlurringInstructions(blurringBuilder);
		domBuilder.setIdentityToken(identityTokenBuilder.build());
		Document responseDoc = domBuilder.build();

		OIOBootstrapToIdentityTokenResponse bootstrapToIdentityResponse = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
		identityToken.getBlurringInstructions();
		BlurringInstructions11 blurringInstructions11 = ((BlurringInstructions11)identityToken.getBlurringInstructions());
		assertEquals(2, blurringInstructions11.getBlurredOrganisations().size());
		assertEquals("29190925", blurringInstructions11.getBlurredOrganisations().get(0).getValue());
		assertEquals("29190941", blurringInstructions11.getBlurredOrganisations().get(1).getValue());
	}

	@Test
	public void testOIOBootstrapToEncryptedIdentityTokenResponse() {

		OIOBootstrapToIdentityTokenResponseDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		domBuilder.setIdentityToken(getCitizenIdentityTokenBuilder().build());
		domBuilder.setEncryptionKey(vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
		Document responseDoc = domBuilder.build();

		OIOBootstrapToIdentityTokenResponse bootstrapToIdentityResponse = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(true, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
	}

	@Test
	public void testCitizenIdentityTokenBuilder() {

		IdentityToken identityToken = createIdentityToken();
		assertEquals("https://fmk", identityToken.getAudienceRestriction());
        assertNull(identityToken.getITSystemName());
		assertEquals("2512484916", identityToken.getCpr());

		// System.out.println(XmlUtil.node2String(identityToken.getDOM(), true, true));
	}

	@Test
	public void testJWTToIdentityTokenRequestBasic() {
		JWTToIdentityTokenRequestDOMBuilder domBuilder = factory.createJWTToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");

		domBuilder.setJWTToken("jwt");

		Document requestDoc = domBuilder.build();
		// write to string and parse to XML
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		JWTToIdentityTokenRequest assertionRequest = factory.createJWTToIdentityTokenRequestModelBuilder().build(requestDoc);
		assertEquals("https://fmk", assertionRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionRequest.getAction());
		assertEquals("2512484916", assertionRequest.getCPRNumberClaim());
		assertNull(assertionRequest.getProcurationForCprClaim());
		String jsonWebToken = assertionRequest.getJsonWebToken();
		assertEquals("jwt", jsonWebToken);

		assertionRequest.validateSignature();
		assertionRequest.validateSignatureAndTrust(vocesVault);
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), assertionRequest.getSigningCertificate());
	}

	@Test
	public void testJWTToIdentityTokenRequestWithProcuration() {
		JWTToIdentityTokenRequestDOMBuilder domBuilder = factory.createJWTToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");

		domBuilder.setJWTToken("jwt");
		domBuilder.setOnBehalfOfClaim("1111111118", Procuration);

		Document requestDoc = domBuilder.build();
		// write to string and parse to XML
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		JWTToIdentityTokenRequest assertionRequest = factory.createJWTToIdentityTokenRequestModelBuilder().build(requestDoc);
		assertEquals("https://fmk", assertionRequest.getAppliesTo());
		assertEquals("http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue", assertionRequest.getAction());
		assertEquals("2512484916", assertionRequest.getCPRNumberClaim());

		assertEquals("1111111118", assertionRequest.getProcurationForCprClaim());
		String jsonWebToken = assertionRequest.getJsonWebToken();
		assertEquals("jwt", jsonWebToken);

		assertionRequest.validateSignature();
		assertionRequest.validateSignatureAndTrust(vocesVault);
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), assertionRequest.getSigningCertificate());
	}

	@Test
	public void testJWTToIdentityTokenRequestWithParentalCustody() {
		// Arrange
		JWTToIdentityTokenRequestDOMBuilder domBuilder = factory.createJWTToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");
		domBuilder.setJWTToken("jwt");
		domBuilder.setOnBehalfOfClaim("1111111118", ParentalCustody);
		Document requestDoc = domBuilder.build();
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		// Act
		JWTToIdentityTokenRequest assertionRequest = factory.createJWTToIdentityTokenRequestModelBuilder().build(requestDoc);

		// Assert
		assertEquals("1111111118", assertionRequest.getParentalCustodyClaim());
	}

	@Test
	public void testJWTToIdentityTokenRequestWithWardCustody() {
		// Arrange
		JWTToIdentityTokenRequestDOMBuilder domBuilder = factory.createJWTToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");
		domBuilder.setJWTToken("jwt");
		domBuilder.setOnBehalfOfClaim("1111111118", WardCustody);
		Document requestDoc = domBuilder.build();
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		// Act
		JWTToIdentityTokenRequest assertionRequest = factory.createJWTToIdentityTokenRequestModelBuilder().build(requestDoc);

		// Assert
		assertEquals("1111111118", assertionRequest.getWardCustodyClaim());
	}

	@Test
	public void testJWTToIdentityTokenRequestWithPartlyWardCustody() {
		// Arrange
		JWTToIdentityTokenRequestDOMBuilder domBuilder = factory.createJWTToIdentityTokenRequestDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("2512484916");
		domBuilder.setAudience("https://fmk");
		domBuilder.setJWTToken("jwt");
		domBuilder.setOnBehalfOfClaim("1111111118", PartlyWardCustody);
		Document requestDoc = domBuilder.build();
		requestDoc = readXml(System.getProperties(), node2String(requestDoc), false);

		// Act
		JWTToIdentityTokenRequest assertionRequest = factory.createJWTToIdentityTokenRequestModelBuilder().build(requestDoc);

		// Assert
		assertEquals("1111111118", assertionRequest.getPartlyWardCustodyClaim());
	}

	@Test
	public void testJWTToIdentityTokenRequestProcurationClaimIsSet() {
		Document requestDoc = factory.createJWTToIdentityTokenRequestDOMBuilder()
				.setSigningVault(vocesVault)
				.setCPRNumberClaim("2512484916")
				.setAudience("https://fmk")
				.setJWTToken("jwt")
				.setOnBehalfOfClaim("8111111111", Procuration)
				.build();
		JWTToIdentityTokenRequest request = factory.createJWTToIdentityTokenRequestModelBuilder().build(requestDoc);
		String cprClaim = request.getProcurationForCprClaim();
		assertEquals("8111111111", cprClaim);
	}

	@Test
	public void testJWTToUnencryptedIdentityTokenResponse() {

		JWTToIdentityTokenResponseDOMBuilder domBuilder = factory.createJWTToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		domBuilder.setIdentityToken(getCitizenIdentityTokenBuilder().build());
		Document responseDoc = domBuilder.build();

		JWTToIdentityTokenResponse bootstrapToIdentityResponse = factory.createJWTToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
	}

	@Test
	public void testJWTTokenResponseWithPrivileges() {

		JWTToIdentityTokenResponseDOMBuilder domBuilder = factory.createJWTToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		final CitizenIdentityTokenBuilder identityTokenBuilder = getCitizenIdentityTokenBuilder();
		BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder();
		privilegesBuilder.addPrivilege(THROUGH_PROCURATION_BY_CPR + ":" + "1111111118", "urn:dk:healthcare:fmk:read");
		privilegesBuilder.addPrivilege(THROUGH_PROCURATION_BY_CPR + ":" + "1111111118", "urn:dk:healthcare:fmk:write");
		identityTokenBuilder.setPrivileges(privilegesBuilder);
		domBuilder.setIdentityToken(identityTokenBuilder.build());
		Document responseDoc = domBuilder.build();

		JWTToIdentityTokenResponse bootstrapToIdentityResponse = factory.createJWTToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
		assertEquals(2, identityToken.getPrivileges().getPrivileges(THROUGH_PROCURATION_BY_CPR + ":" + "1111111118").size());
		assertEquals("urn:dk:healthcare:fmk:read", identityToken.getPrivileges().getPrivileges(
                THROUGH_PROCURATION_BY_CPR + ":" + "1111111118").get(0));
		assertEquals("urn:dk:healthcare:fmk:write", identityToken.getPrivileges().getPrivileges(
                THROUGH_PROCURATION_BY_CPR + ":" + "1111111118").get(1));
	}

	@Test
	public void testJWTTokenResponseWithBlurredInstructions() {

		JWTToIdentityTokenResponseDOMBuilder domBuilder = factory.createJWTToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		final CitizenIdentityTokenBuilder identityTokenBuilder = getCitizenIdentityTokenBuilder();
		BlurringInstructions10DomBuilder blurringBuilder =
				new BlurringInstructions10DomBuilder("5kZZLNQMNIkz1Y7tCDj3GQ==");
		blurringBuilder.addBlurringInstruction("29190925", Cvr, SpecificForPerson);
		blurringBuilder.addBlurringInstruction("29190941", Cvr, FromRelatedPerson);
		identityTokenBuilder.setBlurringInstructions(blurringBuilder);
		domBuilder.setIdentityToken(identityTokenBuilder.build());
		Document responseDoc = domBuilder.build();

		JWTToIdentityTokenResponse bootstrapToIdentityResponse = factory.createJWTToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(bootstrapToIdentityResponse);

		IdentityToken identityToken = extractIdentityToken(false, bootstrapToIdentityResponse);
		verifyIdentityTokenAttributes(identityToken);
		identityToken.getBlurringInstructions();
		BlurringInstructions10 blurringInstructions10 = ((BlurringInstructions10)identityToken.getBlurringInstructions());
		assertEquals(2, blurringInstructions10.getBlurredOrganisations().size());
		assertEquals("29190925", blurringInstructions10.getBlurredOrganisations().get(0).getValue());
		assertEquals("29190941", blurringInstructions10.getBlurredOrganisations().get(1).getValue());
	}

	private IdentityToken createIdentityToken() {
		return getCitizenIdentityTokenBuilder().build();
	}

	private CitizenIdentityTokenBuilder getCitizenIdentityTokenBuilder() {
		CitizenIdentityTokenBuilder tokenBuilder = factory.createCitizenIdentityTokenBuilder();
		tokenBuilder.setIssuer("http://sosi");
		tokenBuilder.setAudienceRestriction("https://fmk");
		Date now = new Date();
		tokenBuilder.setNotBefore(new Date(now.getTime() - 1000));
		tokenBuilder.setNotOnOrAfter(new Date(now.getTime() + 5 * 60 * 1000));
		tokenBuilder.setCprNumberAttribute("2512484916");
		tokenBuilder.setSubjectNameID("C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503");
		tokenBuilder.setSubjectNameIDFormat(SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME);
		tokenBuilder.setDeliveryNotOnOrAfter(new Date(now.getTime() + 10 * 1000));
		tokenBuilder.setRecipientURL("https://fmk");
		tokenBuilder.setSigningVault(vocesVault);
		// Just to have another cert
		tokenBuilder.setHolderOfKeyCertificate(holderOfKey.getSystemCredentialPair().getCertificate());
		return tokenBuilder;
	}

	@Test
	public void testEncryptedOIOSAMLAssertionToIdentityTokenRequest() {

		EncryptedOIOSAMLAssertionToIdentityTokenRequestDOMBuilder domBuilder = factory.createEncryptedOIOSAMLAssertionToIdentityTokenRequestDOMBuilder();
		domBuilder.setAudience("https://sosi");
		OIOSAMLAssertion assertion = createOIOSAMLAssertion(createUserIDCard());
		Element encryptedAssertionElm = createEncryptedAssertion(assertion);
		domBuilder.setEncryptedOIOSAMLAssertionElement(encryptedAssertionElm);
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setCPRNumberClaim("1111111118");

		Document requestDoc = domBuilder.build();

		// System.out.println(XmlUtil.node2String(requestDoc, true, true));

		EncryptedOIOSAMLAssertionToIdentityTokenRequest tokenRequest = factory.createEncryptedOIOSAMLAssertionToIdentityTokenRequestModelBuilder().build(requestDoc);
		assertEquals("1111111118", tokenRequest.getCPRNumberClaim());
		assertEquals("https://sosi", tokenRequest.getAppliesTo());
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), tokenRequest.getSigningCertificate());
		Element decryptedAssertionElement = EncryptionUtil.decryptAndDetach(tokenRequest.getEncryptedOIOSAMLAssertionElement(), vocesVault.getSystemCredentialPair().getPrivateKey());
		OIOSAMLAssertion decryptedAssertion = new OIOSAMLAssertion(decryptedAssertionElement);
		assertEquals(assertion.getSubjectNameID(), decryptedAssertion.getSubjectNameID());

		tokenRequest.validateSignature();
		tokenRequest.validateSignatureAndTrust(vocesVault);
		try {
			tokenRequest.validateSignatureAndTrust(mocesVault);
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}

	}

	@Test
	public void testEncryptedOIOSAMLAssertionToUnencryptedIdentityTokenResponse() {

		EncryptedOIOSAMLAssertionToIdentityTokenResponseDOMBuilder domBuilder = factory.createEncryptedOIOSAMLAssertionToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		domBuilder.setIdentityToken(createIdentityToken());
		Document responseDoc = domBuilder.build();

		EncryptedOIOSAMLAssertionToIdentityTokenResponse tokenResponse = factory.createEncryptedOIOSAMLAssertionToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(tokenResponse);

		IdentityToken identityToken = extractIdentityToken(false, tokenResponse);
		verifyIdentityTokenAttributes(identityToken);
	}

	@Test
	public void testEncryptedOIOSAMLAssertionToEncryptedIdentityTokenResponse() {
		EncryptedOIOSAMLAssertionToIdentityTokenResponseDOMBuilder domBuilder = factory.createEncryptedOIOSAMLAssertionToIdentityTokenResponseDOMBuilder();
		domBuilder.setSigningVault(vocesVault);
		domBuilder.setRelatesTo("1234");
		domBuilder.setContext("2345");
		domBuilder.setIdentityToken(createIdentityToken());
		domBuilder.setEncryptionKey(vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());
		Document responseDoc = domBuilder.build();

		EncryptedOIOSAMLAssertionToIdentityTokenResponse tokenResponse = factory.createEncryptedOIOSAMLAssertionToIdentityTokenResponseModelBuilder().build(responseDoc);
		assertBasicResponse(tokenResponse);

		IdentityToken identityToken = extractIdentityToken(true, tokenResponse);
		verifyIdentityTokenAttributes(identityToken);
	}

	private void assertBasicResponse(AbstractIdentityTokenResponse tokenResponse) {
		assertEquals("https://fmk", tokenResponse.getAppliesTo());
		assertEquals(vocesVault.getSystemCredentialPair().getCertificate(), tokenResponse.getSigningCertificate());

		tokenResponse.validateSignature();
		tokenResponse.validateSignatureAndTrust(getMockFederation());
		try {
			tokenResponse.validateSignatureAndTrust(new SOSITestFederation(System.getProperties()));
			fail("Forventede en fejl");
		} catch (ModelException e) {
			assertTrue(e.getMessage().startsWith("The certificate that signed the security token is not trusted!"));
		}
	}

	private IdentityToken extractIdentityToken(boolean encrypted, AbstractIdentityTokenResponse tokenResponse) {
		if (encrypted) {
			assertNull(tokenResponse.getIdentityToken());
			Element encryptedIdentityTokenElement = tokenResponse.getEncryptedIdentityTokenElement();
			return new IdentityToken(EncryptionUtil.decryptAndDetach(encryptedIdentityTokenElement, vocesVault.getSystemCredentialPair().getPrivateKey()));
		} else {
			assertNull(tokenResponse.getEncryptedIdentityTokenElement());
			return tokenResponse.getIdentityToken();
		}
	}

	private void verifyIdentityTokenAttributes(IdentityToken identityToken) {
		assertEquals("2512484916", identityToken.getCpr());
        assertNull(identityToken.getUserEducationCode());
		assertEquals("3", identityToken.getAssuranceLevel());
		assertEquals("DK-SAML-2.0", identityToken.getSpecVersion());
	}

	private Element createEncryptedAssertion(OIOSAMLAssertion assertion) {
		String simpleXml = "<dummy-root/>";
		Document tempDoc = readXml(System.getProperties(), simpleXml, false);
		Element dummyRoot = tempDoc.getDocumentElement();

		Element encryptedAssertionElm = tempDoc.createElementNS(NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ENCRYPTED_ASSERTION_PREFIXED);
		dummyRoot.appendChild(encryptedAssertionElm);
		Element assertionElm = (Element) tempDoc.importNode(assertion.getDOM().getDocumentElement(), true);
		encryptedAssertionElm.appendChild(assertionElm);

		EncryptionUtil.encrypt(assertionElm, vocesVault.getSystemCredentialPair().getCertificate().getPublicKey());

		// System.out.println(XmlUtil.node2String(tempDoc, true, true));

		return (Element) dummyRoot.getFirstChild();
	}
}