/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import static dk.sosi.seal.model.BlurredOrganisation.OrgType.Cvr;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.FromRelatedPerson;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.SpecificForPerson;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.security.cert.X509Certificate;
import java.util.List;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.constants.HealthcareSAMLAttributes;
import dk.sosi.seal.model.constants.SAMLValues;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.vault.CredentialVault;

public abstract class AbstractCitizenIdentityTokenTest extends AbstractModelTest {

    protected class TestCitizenIdentityTokenBuilder extends CitizenIdentityTokenBuilder {
        protected TestCitizenIdentityTokenBuilder(CitizenIdentityToken token, X509Certificate holderOfKey, CredentialVault signingVault) {
            setAudienceRestriction(token.getAudienceRestriction());
            setCprNumberAttribute(token.getCpr());
            setDeliveryNotOnOrAfter(token.getDeliveryNotOnOrAfter());
            setHolderOfKeyCertificate(holderOfKey);
            setIssuer(token.getIssuer());
            setNotBefore(token.getNotBefore());
            setNotOnOrAfter(token.getNotOnOrAfter());

            setPrivileges(new BasicPrivilegesDOMBuilder(token.getPrivileges()));
            setBlurringInstructions(new BlurringInstructions10DomBuilder(token.getBlurringInstructions()));
            setRecipientURL(token.getRecipient());
            setSigningVault(signingVault);
            setSubjectNameID(token.getSubjectNameID());
            setSubjectNameIDFormat(token.getSubjectNameIDFormat());
        }
    }

    protected CitizenIdentityTokenBuilder createBuilder(CitizenIdentityTokenBuilder citizenIdentityTokenBuilder) {
        /* Required */
        citizenIdentityTokenBuilder.setAudienceRestriction("http://fmk-online.dk");
        citizenIdentityTokenBuilder.setRecipientURL("https://fmk");
        citizenIdentityTokenBuilder.setNotBefore(notBefore);
        citizenIdentityTokenBuilder.setNotOnOrAfter(notOnOrAfter);
        citizenIdentityTokenBuilder.setIssuer("http://pan.certifikat.dk/sts/services/SecurityTokenService");

        citizenIdentityTokenBuilder.setCprNumberAttribute("2512484916");
        citizenIdentityTokenBuilder.setSubjectNameID("C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503");
        citizenIdentityTokenBuilder.setSubjectNameIDFormat(SAMLValues.NAMEID_FORMAT_X509_SUBJECT_NAME);

        citizenIdentityTokenBuilder.setDeliveryNotOnOrAfter(d(10000L));

        citizenIdentityTokenBuilder.setHolderOfKeyCertificate(getTokenHolderOfKeyCert());
        citizenIdentityTokenBuilder.setSigningVault(getTokenSigningVault());

        /* Optional */
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder();
        privilegesBuilder.addPrivilege(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + "1111111118", "urn:dk:healthcare:fmk:read");
        privilegesBuilder.addPrivilege(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + "1111111118", "urn:dk:healthcare:fmk:write");
        citizenIdentityTokenBuilder.setPrivileges(privilegesBuilder);

        BlurringInstructions10DomBuilder blurringBuilder =
                new BlurringInstructions10DomBuilder("5kZZLNQMNIkz1Y7tCDj3GQ==");
        blurringBuilder.addBlurringInstruction("29190925", Cvr, SpecificForPerson);
        blurringBuilder.addBlurringInstruction("29190941", Cvr, FromRelatedPerson);
        citizenIdentityTokenBuilder.setBlurringInstructions(blurringBuilder);

        return citizenIdentityTokenBuilder;
    }

    protected CitizenIdentityTokenBuilder createBuilder() {
        CitizenIdentityTokenBuilder citizenIdentityTokenBuilder = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()).createIdentityTokenBuilder();
        return createBuilder(citizenIdentityTokenBuilder);
    }

    protected void assertValidIdentityToken(CitizenIdentityToken citizenIdentityToken) {
        assertEquals("3",citizenIdentityToken.getAssuranceLevel());
        assertEquals("http://fmk-online.dk",citizenIdentityToken.getAudienceRestriction());
        assertEquals("2512484916",citizenIdentityToken.getCpr());
        assertNotNull(citizenIdentityToken.getID());
        assertEquals("http://pan.certifikat.dk/sts/services/SecurityTokenService",citizenIdentityToken.getIssuer());
        assertEquals(notBefore,citizenIdentityToken.getNotBefore());
        assertEquals(notOnOrAfter,citizenIdentityToken.getNotOnOrAfter());

        List<String> privileges = citizenIdentityToken.getPrivileges().getPrivileges(HealthcareSAMLAttributes.THROUGH_PROCURATION_BY_CPR + ":" + "1111111118");
        assertEquals("[urn:dk:healthcare:fmk:read, urn:dk:healthcare:fmk:write]", privileges.toString());

        BlurringInstructions10Impl blurringInstructions10 = (BlurringInstructions10Impl) citizenIdentityToken.getBlurringInstructions();
        List<BlurredOrganisation> blurredOrganisations = blurringInstructions10.getBlurredOrganisations();
        assertEquals("29190925", blurredOrganisations.get(0).getValue());
        assertEquals(Cvr, blurredOrganisations.get(0).getOrgType());
        assertEquals(SpecificForPerson, blurredOrganisations.get(0).getReason());

        assertEquals("29190941", blurredOrganisations.get(1).getValue());
        assertEquals(Cvr, blurredOrganisations.get(1).getOrgType());
        assertEquals(FromRelatedPerson, blurredOrganisations.get(1).getReason());

        assertEquals("DK-SAML-2.0",citizenIdentityToken.getSpecVersion());

        /*Fields from IdentityToken not used in CitizenIdentityToken*/
        assertEquals(null,citizenIdentityToken.getAuthorizationCode());
        assertEquals(null,citizenIdentityToken.getCommonName());
        assertEquals(null,citizenIdentityToken.getCvrNumberIdentifier());
        assertEquals(null,citizenIdentityToken.getEmail());
        assertEquals(null,citizenIdentityToken.getITSystemName());
        assertEquals(null,citizenIdentityToken.getOrganizationName());
        assertEquals(null,citizenIdentityToken.getSurName());
        assertEquals(null,citizenIdentityToken.getUserAuthenticationInstant());
        assertEquals(null,citizenIdentityToken.getUserEducationCode());	
    }

    protected X509Certificate getTokenHolderOfKeyCert() {
        return getMessageSigningVault().getSystemCredentialPair().getCertificate();
    }

    /**
     * Creates a <code>Federation</code> for test purposes.
     * 
     * @return A test <code>Federation</code> instance.
     */
    protected Federation getMockFederation() {
        return getMockFederation("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
    }
}
