/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.pki.*;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.GenericCredentialVault;
import org.hamcrest.BaseMatcher;
import org.hamcrest.Description;
import org.junit.Before;
import org.junit.BeforeClass;
import org.w3c.dom.Element;

import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Properties;

public abstract class AbstractModelTest {

	protected static GenericCredentialVault credentialVault;
	protected static final String EXPECTED_ASSURANCELEVEL = "3";
	protected static final String EXPECTED_AUTHORIZATIONCODE = "004PT";
	protected static final String EXPECTED_COMMONNAME = "Jens Sundbye Poulsen";
	protected static final String EXPECTED_CPR = "2702681273";
	protected static final String EXPECTED_CVR = "20688092";
	protected static final String EXPECTED_EMAIL = "jens@email.dk";
	protected static final String EXPECTED_GIVENNAME = "Jens Sundbye";
	protected static final String EXPECTED_ISSUER = "http://pan.certifikat.dk/sts/services/SecurityTokenService";
	protected static final String EXPECTED_ITSYSTEMNAME = "Harmoni/EMS";
	protected static final String EXPECTED_OCCUPATION = "overlæge";
	protected static final String EXPECTED_ORGANIZATION = "Lægehuset på bakken";
	protected static final String EXPECTED_SPECVERSION = "DK-SAML-2.0";
	protected static final String EXPECTED_SURNAME = "Poulsen";

	protected static final String EXPECTED_USEREDUCATIONCODE = "7170";
	protected static SOSIFactory sosiFactory;

	protected static final Date notBefore = d(-10);
	protected static final Date notOnOrAfter = d(10);

	@BeforeClass
	public static void setupAbstract() {
		credentialVault = CredentialVaultTestUtil.getVoces3CredentialVault();
	}

	@Before
	public void setUp() {
		sosiFactory = new SOSIFactory(getMockFederation(), credentialVault, System.getProperties());
	}

	protected static Date d(int minutesFromNow) {
		long l = minutesFromNow * 60L * 1000L;
		return d(l);
	}

	protected static Date d(long milliSecondsFromNow) {
		Calendar now = Calendar.getInstance();
		now.set(Calendar.MILLISECOND, 0);

		return new Date(now.getTimeInMillis() + milliSecondsFromNow);
	}

	/**
	 * Creates a new <code>Date</code> based on the input parameters.
	 * 
	 * @param year
	 *            The year value.
	 * @param month
	 *            The month (1 based) value.
	 * @param date
	 *            The date value.
	 * @param hour
	 *            The 24h clock value.
	 * @param minute
	 *            The minute value.
	 * @param second
	 *            The second value.
	 * @return The constructed <code>Date</code> object.
	 */
	protected static Date d(int year, int month, int date, int hour, int minute, int second) {
		Calendar cal = Calendar.getInstance();
		cal.set(Calendar.YEAR, year);
		cal.set(Calendar.MONTH, month - 1);
		cal.set(Calendar.DATE, date);
		cal.set(Calendar.HOUR_OF_DAY, hour);
		cal.set(Calendar.MINUTE, minute);
		cal.set(Calendar.SECOND, second);
		cal.set(Calendar.MILLISECOND, 0);
		return cal.getTime();
	}

	/**
	 * Creates a <code>Federation</code> for test purposes.
	 * 
	 * @return A test <code>Federation</code> instance.
	 */
	protected Federation getMockFederation() {
		return getMockFederation("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b");
	}

	/**
	 * Creates a <code>Federation</code> for test purposes.
	 * 
	 * @return A test <code>Federation</code> instance.
	 */
	protected Federation getMockFederation(String... stsSubjectSerialNumbers) {
		Properties properties = System.getProperties();
		final CertificationAuthority ca = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA,
				getCertificateStatusChecker(properties), getCertificateCache());
		return new MockFederation(properties, ca, stsSubjectSerialNumbers);
	}

	protected CertificateCache getCertificateCache() {
		return CredentialVaultTestUtil.getCertificateCacheForVoces3CredentialVaultCert();
	}

	protected CertificateStatusChecker getCertificateStatusChecker(Properties properties) {
		return new NaiveCertificateStatusChecker(properties);
	}

	public static class TestDOMInfoExtractor extends AbstractDOMInfoExtractor {
		public TestDOMInfoExtractor(Element dom) {
			super(dom);
		}
	}

	protected static RegexMatcher getRegexMatcher(String regex) {
		return new RegexMatcher(regex);
	}

	static class RegexMatcher extends BaseMatcher<String> {
		private final String regex;

		public RegexMatcher(String regex) {
			this.regex = regex;
		}

		public boolean matches(Object actual) {
			return (actual instanceof String) && ((String) actual).matches(regex);
		}

		public void describeTo(Description description) {
			description.appendText("matches(\"" + regex.replaceAll("\\\\", "\\\\\\\\") + "\")");
		}
	}

	protected CredentialVault getMessageSigningVault() {
		return CredentialVaultTestUtil.getCredentialVaultFromResource(System.getProperties(), "oces3/MOCES_gyldig.p12");
	}

	protected CredentialVault getTokenSigningVault() {
		return credentialVault;
	}

	protected CredentialVault getUntrustedTokenSigningVault() {
		return CredentialVaultTestUtil.getVoces2CredentialVault();
	}

	protected CredentialVault getExpiredTokenSigningVault() {
		throw new RuntimeException("TODO");
		// TODO SDS-6659 Change 'CredentialVaultTestUtil.VOCES_EXPIRED_PKCS12' to an expired OCES3 certififcate
		//return CredentialVaultTestUtil.getCredentialVaultFromPKCS12(CredentialVaultTestUtil.VOCES_EXPIRED_PKCS12, CredentialVaultTestUtil.VOCES_EXPIRED_PKCS12_PWD);
	}

	protected void remove(List<SignatureConfiguration.Reference> references, String identifier) {
		for (SignatureConfiguration.Reference reference : references) {
			if (reference.getURI().equals(identifier)) {
				references.remove(reference);
				break;
			}
		}
	}
}