/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;

public abstract class AbstractUserIDCardTest extends AbstractModelTest {
    /**
     * Creates a signed <code>UserIDCard</code>.<br />
     * This is usually done by the STS, but for use in unittests, we emulate the STS here.
     * 
     * @return A signed <code>UserIDCard</code>.
     */
    protected static UserIDCard createSignedIdCard() {
        UserIDCard userIdCard = createUserIDCard();
        final SecurityTokenRequest tokenRequest = sosiFactory.createNewSecurityTokenRequest();
        tokenRequest.setIDCard(userIdCard);
        userIdCard.sign(tokenRequest.serialize2DOMDocument(), sosiFactory.getSignatureProvider());
        return userIdCard;
    }

    /**
     * Creates a <code>UserIDCard</code> based on the following information.<br />
     * <table border="1">
     * <tr>
     * <td>CVR#</td>
     * <td>20688092</td>
     * </tr>
     * <tr>
     * <td>Organization</td>
     * <td>Lægehuset på bakken</td>
     * </tr>
     * <tr>
     * <td>CPR#</td>
     * <td>2702681273</td>
     * </tr>
     * <tr>
     * <td>Given name</td>
     * <td>Jens Sundbye</td>
     * </tr>
     * <tr>
     * <td>Surname</td>
     * <td>Poulsen</td>
     * </tr>
     * <tr>
     * <td>E-mail</td>
     * <td>jens@email.dk</td>
     * </tr>
     * <tr>
     * <td>Occupation</td>
     * <td>overlæge</td>
     * </tr>
     * <tr>
     * <td>Role</td>
     * <td>doctor</td>
     * </tr>
     * <tr>
     * <td>AuthorizationCode</td>
     * <td>004PT</td>
     * </tr>
     * </table>
     * 
     * @return New <code>UserIDCard</code> instance.
     */
    protected static UserIDCard createUserIDCard() {
        return createUserIDCard(SubjectIdentifierTypeValues.CVR_NUMBER);
    }

    /**
     * Creates a <code>UserIDCard</code> based on the following information.<br />
     * <table border="1">
     * <tr>
     * <td>CVR#</td>
     * <td>20688092</td>
     * </tr>
     * <tr>
     * <td>Organization</td>
     * <td>Lægehuset på bakken</td>
     * </tr>
     * <tr>
     * <td>CPR#</td>
     * <td>2702681273</td>
     * </tr>
     * <tr>
     * <td>Given name</td>
     * <td>Jens Sundbye</td>
     * </tr>
     * <tr>
     * <td>Surname</td>
     * <td>Poulsen</td>
     * </tr>
     * <tr>
     * <td>E-mail</td>
     * <td>jens@email.dk</td>
     * </tr>
     * <tr>
     * <td>Occupation</td>
     * <td>overlæge</td>
     * </tr>
     * <tr>
     * <td>Role</td>
     * <td>doctor</td>
     * </tr>
     * <tr>
     * <td>AuthorizationCode</td>
     * <td>004PT</td>
     * </tr>
     * </table>
     * 
     * @param careProviderType
     *            Type of identication of the care provider.
     * @return New <code>UserIDCard</code> instance.
     */
    protected static UserIDCard createUserIDCard(String careProviderType) {
        CareProvider careProvider = new CareProvider(careProviderType, EXPECTED_CVR, EXPECTED_ORGANIZATION);
        UserInfo userInfo = new UserInfo(EXPECTED_CPR, EXPECTED_GIVENNAME, EXPECTED_SURNAME, EXPECTED_EMAIL, EXPECTED_OCCUPATION, EXPECTED_USEREDUCATIONCODE, EXPECTED_AUTHORIZATIONCODE);
        UserIDCard idcard = sosiFactory.createNewUserIDCard(EXPECTED_ITSYSTEMNAME, userInfo, careProvider, AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, "hans@dampf.dk");
        return new UserIDCard(idcard, "newIssuer");
    }
}
