package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.BasicPrivileges12Tags;
import dk.sosi.seal.model.constants.BasicPrivilegesTags;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.util.Base64;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import static dk.sosi.seal.model.constants.NameSpaces.OIO_BASIC_PRIVILEGES_PROFILE_1_2;
import static org.junit.Assert.*;

public class BasicPriviligesTest {
	
	@Test(expected=NullPointerException.class)
	public void testNull() {
		Element elm = null;
		new BasicPrivileges(elm);
	}
	
	@Test
	public void testNoScopeNoPriviligeInDocument() {
		String soapXml = 
				"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" + 
				"<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" + 
				"  <soapenv:Header />\n" + 			 
				"  <soapenv:Body />\n" +
				"</soapenv:Envelope>";
		
		Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
		Element body = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body);
		
		BasicPrivileges basicPrivileges = new BasicPrivileges(body);
		
		Set<String> scopes = basicPrivileges.getScopes();
		assertTrue(scopes.isEmpty());
		
		List<String> privileges = basicPrivileges.getPrivileges("NonExistingScope");
		assertNull(privileges);
	}
	
	@Test
	public void testOneScopeNoPriviligeInDocument() {
		String soapXml = 
				"<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" + 
				"<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" + 
				"  <soapenv:Header />\n" + 			 
				"  <soapenv:Body>\n" +
				"    <bpp:PrivilegeList xmlns:bpp=\"http://itst.dk/oiosaml/basic_privilege_profile\">\n" + 
				"      <bpp:PrivilegeGroup Scope=\"MyScope\" />\n" + 
				"    </bpp:PrivilegeList>\n" +
				"  </soapenv:Body>\n" +			
				"</soapenv:Envelope>";
		
		Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
		Element privilegeList = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BasicPrivilegesTags.PRIVILEGE_LIST);
		
		BasicPrivileges basicPrivileges = new BasicPrivileges(privilegeList);

		Set<String> scopes = basicPrivileges.getScopes();
		assertEquals(1, scopes.size());
		assertTrue(scopes.contains("MyScope"));
		
		List<String> privileges = basicPrivileges.getPrivileges("MyScope");
		assertEquals(0, privileges.size());
	}
	
	@Test
	public void testOneScopeOnePrivilegeInDocument() {
		final String
				privilegeList =
				"    <bpp:PrivilegeList xmlns:bpp=\"http://itst.dk/oiosaml/basic_privilege_profile\">\n" +
						"      <bpp:PrivilegeGroup Scope=\"*scope*\">\n" +
						"        <bpp:Privilege>*privilige*</bpp:Privilege>\n" +
						"      </bpp:PrivilegeGroup>\n" +
						"    </bpp:PrivilegeList>\n",
				soapXml = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" +
						"<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
						"  <soapenv:Header />\n" +
						"  <soapenv:Body>\n" +
						privilegeList +
						"  </soapenv:Body>\n" +
						"</soapenv:Envelope>";
		
		Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
		Element elm = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BasicPrivilegesTags.PRIVILEGE_LIST);
		assertBasicPrivileges10(new BasicPrivileges(elm));

		String base64Encoded = Base64.getEncoder().encodeToString(privilegeList.getBytes());
		BasicPrivileges decodedBasicPrivileges = BasicPrivileges.decode(base64Encoded);
		assertBasicPrivileges10(decodedBasicPrivileges);
	}

	private static void assertBasicPrivileges10(BasicPrivileges basicPrivileges) {
		Set<String> scopes = basicPrivileges.getScopes();
		assertEquals(1, scopes.size());
		assertTrue(scopes.contains("*scope*"));

		List<String> privileges = basicPrivileges.getPrivileges("*scope*");
		assertEquals(1, privileges.size());
		assertTrue(privileges.contains("*privilige*"));
	}

	@Test
	public void testBasicPrivilegeVersion12() {
		final String
				basicPrivilegesScope = "urn:prefix:0501792275",
				privilege = "urn:dk:sundhedsdatastyrelsen:idsas:read:unblurred",
				constraintCvr = "12345678",
				privilegeList = "    <bpp:PrivilegeList xmlns:bpp=\"" + OIO_BASIC_PRIVILEGES_PROFILE_1_2 + "\">\n" +
							    "      <PrivilegeGroup Scope=\"" + basicPrivilegesScope + "\">\n" +
								"        <Privilege>" + privilege + "</Privilege>\n" +
								"        <Constraint Name=\"CVR\">" + constraintCvr + "</Constraint>\n" +
								"      </PrivilegeGroup>\n" +
								"    </bpp:PrivilegeList>\n",
				version12BasicPrivilegeXml = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" +
						"<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
						"  <soapenv:Header />\n" +
						"  <soapenv:Body>\n" +
						privilegeList +
						"  </soapenv:Body>\n" +
						"</soapenv:Envelope>";

		Document doc = XmlUtil.readXml(new Properties(), version12BasicPrivilegeXml, true);
		Element elm = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BasicPrivileges12Tags.PRIVILEGE_LIST);

		assertBasicPrivileges(new BasicPrivileges(elm), basicPrivilegesScope, privilege, constraintCvr);
		String base64Encoded = Base64.getEncoder().encodeToString(privilegeList.getBytes());
		BasicPrivileges decodedBasicPrivileges = BasicPrivileges.decode(base64Encoded);
		assertBasicPrivileges(decodedBasicPrivileges, basicPrivilegesScope, privilege, constraintCvr);
	}

	private static void assertBasicPrivileges(BasicPrivileges p,
								  String basicPrivilegesScope,
								  String privilege,
								  String constraintCvr) {
		List<String> ps = p.getPrivileges(basicPrivilegesScope);
		assertEquals(1, ps.size());
		assertEquals(privilege, ps.stream().findFirst().get());

		List<BasicPrivileges.Constraint> cs = p.getConstraints(basicPrivilegesScope);
		assertEquals(1, cs.size());
		assertEquals(new BasicPrivileges.Constraint("CVR", constraintCvr),
					 cs.stream().findFirst().get());
	}
}