package dk.sosi.seal.model;

import dk.sosi.seal.model.constants.BlurringInstructions10Tags;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.util.List;
import java.util.Properties;

import static dk.sosi.seal.model.BlurredOrganisation.OrgType.Cvr;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.FromRelatedPerson;
import static dk.sosi.seal.model.BlurredOrganisation.Reason.SpecificForPerson;

public class BlurringInstructions10Test {

    @Test(expected = NullPointerException.class)
    public void testNull() {
        Element elm = null;
        new BlurringInstructions10Impl(elm);
    }

    @Test(expected = ModelException.class)
    public void testNoSaltInDocument() {
        String soapXml =
                "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" +
                        "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
                        "  <soapenv:Header />\n" +
                        "  <soapenv:Body>\n" +
                        "     <bip:BlurringInstructions xmlns:bip=\"urn:dk:healthcare:saml:blurring_instruction_profile:1.0\">" +
                        "        <bip:BlurEmployeeNamesFromOrg orgType=\"CVR\" reason=\"specific_for_person\">29190925</bip:BlurEmployeeNamesFromOrg>" +
                        "        <bip:BlurEmployeeNamesFromOrg orgType=\"CVR\" reason=\"from_related_person\">29190941</bip:BlurEmployeeNamesFromOrg> " +
                        "     </bip:BlurringInstructions>" +
                        "  </soapenv:Body>\n" +
                        "</soapenv:Envelope>";

        Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
        Element blurringInstructions = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BlurringInstructions10Tags.BLURRING_INSTRUCTIONS);
        new BlurringInstructions10Impl(blurringInstructions);
    }

    @Test(expected = ModelException.class)
    public void testWrongOrgTypeInDocument() {
        String soapXml =
                "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" +
                        "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
                        "  <soapenv:Header />\n" +
                        "  <soapenv:Body>\n" +
                        "     <bip:BlurringInstructions xmlns:bip=\"urn:dk:healthcare:saml:blurring_instruction_profile:1.0\" currentSalt=\"5kZZLNQMNIkz1Y7tCDj3GQ==\">" +
                        "        <bip:BlurEmployeeNamesFromOrg orgType=\"KUK\" reason=\"specific_for_person\">29190925</bip:BlurEmployeeNamesFromOrg>" +
                        "     </bip:BlurringInstructions>" +
                        "  </soapenv:Body>\n" +
                        "</soapenv:Envelope>";

        Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
        Element blurringInstructions = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BlurringInstructions10Tags.BLURRING_INSTRUCTIONS);

        new BlurringInstructions10Impl(blurringInstructions);
    }

    @Test(expected = ModelException.class)
    public void testWrongReasonInDocument() {
        String soapXml =
                "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" +
                        "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
                        "  <soapenv:Header />\n" +
                        "  <soapenv:Body>\n" +
                        "     <bip:BlurringInstructions xmlns:bip=\"urn:dk:healthcare:saml:blurring_instruction_profile:1.0\" currentSalt=\"5kZZLNQMNIkz1Y7tCDj3GQ==\">" +
                        "        <bip:BlurEmployeeNamesFromOrg orgType=\"CVR\" reason=\"kuk_kuk\">29190925</bip:BlurEmployeeNamesFromOrg>" +
                        "     </bip:BlurringInstructions>" +
                        "  </soapenv:Body>\n" +
                        "</soapenv:Envelope>";

        Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
        Element blurringInstructions = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BlurringInstructions10Tags.BLURRING_INSTRUCTIONS);

        new BlurringInstructions10Impl(blurringInstructions);
    }



    @Test
    public void testZeroBlurredOrganisationInDocument() {
        String soapXml =
                "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" +
                        "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
                        "  <soapenv:Header />\n" +
                        "  <soapenv:Body>\n" +
                        "     <bip:BlurringInstructions xmlns:bip=\"urn:dk:healthcare:saml:blurring_instruction_profile:1.0\" currentSalt=\"5kZZLNQMNIkz1Y7tCDj3GQ==\"/>" +
                        "  </soapenv:Body>\n" +
                        "</soapenv:Envelope>";

        Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
        Element blurringInstructions = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BlurringInstructions10Tags.BLURRING_INSTRUCTIONS);

        BlurringInstructions10Impl blurringInstructions10 = new BlurringInstructions10Impl(blurringInstructions);

        List<BlurredOrganisation> orgs = blurringInstructions10.getBlurredOrganisations();
        Assert.assertEquals(0, orgs.size());
    }

    @Test
    public void testTwoBlurredOrganisationsInDocument() {
        String soapXml =
                "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" +
                        "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">\n" +
                        "  <soapenv:Header />\n" +
                        "  <soapenv:Body>\n" +
                        "     <bip:BlurringInstructions xmlns:bip=\"urn:dk:healthcare:saml:blurring_instruction_profile:1.0\"  currentSalt=\"5kZZLNQMNIkz1Y7tCDj3GQ==\">" +
                        "        <bip:BlurEmployeeNamesFromOrg orgType=\"CVR\" reason=\"specific_for_person\">29190925</bip:BlurEmployeeNamesFromOrg>" +
                        "        <bip:BlurEmployeeNamesFromOrg orgType=\"CVR\" reason=\"from_related_person\">29190941</bip:BlurEmployeeNamesFromOrg> " +
                        "     </bip:BlurringInstructions>" +
                        "  </soapenv:Body>\n" +
                        "</soapenv:Envelope>";

        Document doc = XmlUtil.readXml(new Properties(), soapXml, true);
        Element blurringInstructions = TagUtil.getFirstChildElementNS(doc.getDocumentElement(), SOAPTags.body, BlurringInstructions10Tags.BLURRING_INSTRUCTIONS);

        BlurringInstructions10Impl blurringInstructions10 = new BlurringInstructions10Impl(blurringInstructions);

        List<BlurredOrganisation> orgs = blurringInstructions10.getBlurredOrganisations();
        Assert.assertEquals(2, orgs.size());

        Assert.assertEquals("29190925", orgs.get(0).getValue());
        Assert.assertEquals(Cvr, orgs.get(0).getOrgType());
        Assert.assertEquals(SpecificForPerson, orgs.get(0).getReason());

        Assert.assertEquals("29190941", orgs.get(1).getValue());
        Assert.assertEquals(Cvr, orgs.get(1).getOrgType());
        Assert.assertEquals(FromRelatedPerson, orgs.get(1).getReason());
    }
}