/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.util.List;

import org.junit.Before;
import org.junit.Ignore;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.LibertyTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.WSATags;
import dk.sosi.seal.model.constants.WSTTags;
import dk.sosi.seal.model.constants.WSUTags;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.TestTagUtil;
import dk.sosi.seal.xml.XmlUtil;

public class CitizenLibertyReponseTest extends AbstractModelTest {
	@Rule
	public ExpectedException expectedException = ExpectedException.none();

	private Document document;
	
	@Before
	public void setUp() {
		super.setUp();

		document = XmlUtil.createEmptySoapDocument();
	}

	@Test
	// Parsning af valid response
	// - wsa:MessageID, wsa:Action, wsa:RelatesTo, wsu:Timestamp, sbf:Framework,
	// soap:Body, wsa:To (hvis det er til stede) er signerede
	public void t22_RequestOK() {
		final LibertyMessageDOMEnhancer enhancer = OIOIDWSFactory.createResponseDomEnhancer(getMessageSigningVault(), document, true);

		// wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		// TimeStamp
		// Framework
		// Body
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();
		
		final LibertyResponse response = OIOIDWSFactory.createResponse(document);

		assertNotNull(response.getCreatedTimestamp());
		
		System.out.println(XmlUtil.node2String(response.dom, true, true));
		
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Parsning af invalid response
	// (manglende signatur)
	public void t23_RequestOKParseInvalidResponseMissingSignature() {	
		final LibertyMessageDOMEnhancer enhancer = OIOIDWSFactory.createResponseDomEnhancer(getMessageSigningVault(), document, true);

		// wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		/* Remove Signature */
		NodeList signatures = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), DSTags.signature, false);
		assertEquals(1, signatures.getLength());
		TestTagUtil.removeFirst(document, DSTags.signature);

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		assertNotNull(response.getCreatedTimestamp());
	}

	@Test
	// Parsning af invalid response
	// (brudt signatur)
	public void t24_RequestOKParseInvalidResponseTamperedSignature() {
		final LibertyMessageDOMEnhancer enhancer = OIOIDWSFactory.createResponseDomEnhancer(getMessageSigningVault(), document, true);

		// wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		/* Tamper with content */
		final Node body = document.getElementsByTagNameNS(NameSpaces.SOAP_SCHEMA, SOAPTags.BODY_UNPREFIXED).item(0);
		body.setTextContent("foo");

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		assertNotNull(response.getCreatedTimestamp());
	}

	@Test
	// Validering af valid response
	// (wsa:MessageID, wsa:Action, wsa:RelatesTo, wsu:Timestamp, sbf:Framework,
	// soap:Body, wsa:To (hvis det er til stede) er signerede af en trusted part)
	public void t25_ResponseOKParsingAndValidationResponse() {
		final LibertyMessageDOMEnhancer enhancer = OIOIDWSFactory.createResponseDomEnhancer(getMessageSigningVault(), document, true);

		// wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		OIOIDWSFactory.createResponse(document).validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Parsning af request med
	// - brudt beskedssignatur
	public void t26_ResponseFailTamperedMessageSignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Liberty signature could not be validated");

		final LibertyMessageDOMEnhancer enhancer = OIOIDWSFactory.createResponseDomEnhancer(getMessageSigningVault(), document, true);

		// wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final Element body = TagUtil.getFirstChildElementNS(document.getDocumentElement(), SOAPTags.body);

		Element lifeTime = TagUtil.createElementNS(document, WSTTags.lifetime);
		lifeTime.setTextContent("theLifeTimeValue");
		body.appendChild(lifeTime);

		OIOIDWSFactory.createResponse(document).validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - signatur fra ikke trusted part
	public void t2x_ResponseFailUntrustedSignature() {
		expectedException.expect(ModelException.class);
		expectedException.expectMessage("The certificate that signed the security token is not trusted!");

		final LibertyMessageDOMEnhancer enhancer = OIOIDWSFactory.createResponseDomEnhancer(getUntrustedTokenSigningVault(), document, true);

		// wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final Element body = TagUtil.getFirstChildElementNS(document.getDocumentElement(), SOAPTags.body);

		Element lifeTime = TagUtil.createElementNS(document, WSTTags.lifetime);
		lifeTime.setTextContent("theLifeTimeValue");
		body.appendChild(lifeTime);

		OIOIDWSFactory.createResponse(document).validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	@Ignore("SDS-6659")
	// Validering af response med
	// - signatur med udløbet certifikat fra trusted part
	public void t27_ResponseFailSignatureIsExpired() {
		expectedException.expect(ModelException.class);
		expectedException.expectMessage("The certificate that signed the security token is not trusted!");

		final LibertyMessageDOMEnhancer enhancer = OIOIDWSFactory.createResponseDomEnhancer(getExpiredTokenSigningVault(), document, true);

		// wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final Element body = TagUtil.getFirstChildElementNS(document.getDocumentElement(), SOAPTags.body);

		Element lifeTime = TagUtil.createElementNS(document, WSTTags.lifetime);
		lifeTime.setTextContent("theLifeTimeValue");
		body.appendChild(lifeTime);

		OIOIDWSFactory.createResponse(document).validateSignatureAndTrust(getMockFederation("CVR:19343634-UID:1165813731180"));
	}

	@Test
	// Validering af response med
	// - manglende wsa:MessageID
	public void t28_RequestFailMissingMessagID() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find WS-Addressing 1.0 MessageID element");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "messageID");
				NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), WSATags.messageID);
				if (messageIDList != null && messageIDList.getLength() == 1) {
					Node messageID = messageIDList.item(0);
					messageID.getParentNode().removeChild(messageID);
				}

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - wsa:MessageID, som ikke er under signatur
	public void t29_RequestFailMissingMessageIDSignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#MessageID");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "messageID");

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - manglende wsa:MessageID
	public void t30_RequestFailMissingRelatesTo() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find WS-Addressing 1.0 RelatesTo element");
		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "relatesTo");
				NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), WSATags.relatesTo);
				if (messageIDList != null && messageIDList.getLength() == 1) {
					Node messageID = messageIDList.item(0);
					messageID.getParentNode().removeChild(messageID);
				}

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingMessageID("wsAddressingMessageID");
		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		// System.out.println(XmlUtil.node2String(document, true, false));

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - wsa:RelatesTo, som ikke er under signatur
	public void t31_RequestFailMissingRelatesToSignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#RelatesTo");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "relatesTo");

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - manglende wsu:Timestamp
	public void t32_RequestFailMissingTimestamp() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find Timestamp element");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "ts");
				NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), WSUTags.timestamp);
				if (messageIDList != null && messageIDList.getLength() == 1) {
					Node messageID = messageIDList.item(0);
					messageID.getParentNode().removeChild(messageID);
				}

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - wsu:Timestamp, som ikke er under signatur
	public void t33_RequestFailMissingTimestampSignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd#Timestamp");
		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "ts");

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - manglende sbf:Framework
	public void t34_RequestFailMissingFramework() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find Liberty Framework element");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "sbf");
				NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), LibertyTags.framework);
				if (messageIDList != null && messageIDList.getLength() == 1) {
					Node messageID = messageIDList.item(0);
					messageID.getParentNode().removeChild(messageID);
				}

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - sbf:Framework, som ikke er under signatur
	public void t35_RequestFailMissingFrameworkSignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element urn:liberty:sb#Framework");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "sbf");

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - manglende soap:Body
	public void t36_RequestFailMissingBody() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find Body element");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "body");
				NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), SOAPTags.body);
				if (messageIDList != null && messageIDList.getLength() == 1) {
					Node messageID = messageIDList.item(0);
					messageID.getParentNode().removeChild(messageID);
				}

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}

	@Test
	// Validering af response med
	// - soap:Body, som ikke er under signatur
	public void t37_RequestFailMissingBodySignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://schemas.xmlsoap.org/soap/envelope/#Body");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "body");

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}
	
	@Test
	// Validering af response med
	// - manglende soap:Body
	public void t38_RequestFailMissingBody() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find WS-Addressing 1.0 Action element");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "action");
				NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), WSATags.action);
				if (messageIDList != null && messageIDList.getLength() == 1) {
					Node messageID = messageIDList.item(0);
					messageID.getParentNode().removeChild(messageID);
				}

				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}
	
	@Test
	// Validering af response med
	// - soap:Body, som ikke er under signatur
	public void t39_RequestFailMissingBodySignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#Action");

		final LibertyMessageDOMEnhancer enhancer = new LibertyMessageDOMEnhancer(getMessageSigningVault(), document, true) {
			@Override
			protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
				/* Remove reference and tag */
				remove(references, "action");
				return super.getSignatureConfiguration(securityWsuId, references);
			}
		};

		enhancer.setWSAddressingAction("addressingAction");
		enhancer.setWSAddressingRelatesTo("wsAddressingRelatesTo");
		enhancer.setWSAddressingTo("wsAddressingTo");

		enhancer.enhanceAndSign();

		final LibertyResponse response = OIOIDWSFactory.createResponse(document);
		response.validateSignatureAndTrust(getMockFederation("UI:DK-E:G:35fb4742-9420-4c1b-8d74-5415a1eeb789"));
	}
}
