/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.security.cert.X509Certificate;
import java.util.List;

import org.junit.Before;
import org.junit.Ignore;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXParseException;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.SignatureConfiguration.Reference;
import dk.sosi.seal.model.constants.DSTags;
import dk.sosi.seal.model.constants.LibertyTags;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.WSATags;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.model.constants.WSTTags;
import dk.sosi.seal.model.constants.WSUTags;
import dk.sosi.seal.modelbuilders.LibertyRequestModelBuilder;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.pki.CertificateCache;
import dk.sosi.seal.pki.OCESTestCertificationAuthority;
import dk.sosi.seal.pki.PKIException;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.TestTagUtil;
import dk.sosi.seal.xml.XmlUtil;

/**
 */
@RunWith(Parameterized.class)
public class CitizenLibertyRequestTest extends AbstractCitizenIdentityTokenTest {
	@Parameter(0)
	public String suiteName;

	@Parameter(1)
	public GenericCredentialVault credentialVault;

	@Rule
	public ExpectedException expectedException = ExpectedException.none();

	private Document document;
	private CitizenIdentityToken identityToken;
	
	/*Setting up test parameters*/
	@Parameters(name = "{0}")
	public static Iterable<Object[]> data() {
		TestParameterBuilder builder = new TestParameterBuilder();
		builder.newTest("VOCES3").add(CredentialVaultTestUtil.getVoces3CredentialVault(System.getProperties()));
		return builder.build();
	}
	
	@Override
	protected CredentialVault getMessageSigningVault() {
		return credentialVault;
	}

	@Before
	public void setUp() {
		super.setUp();

		document = XmlUtil.createEmptySoapDocument();
		identityToken = createBuilder().build();
	}
	
	@Override
	protected CertificateCache getCertificateCache() {
		CertificateCache ref = super.getCertificateCache();
		return new CertificateCache() {
			@Override
			public X509Certificate getCertificate(Category category, String key) throws PKIException {
				if(category == Category.IntermediateCert && "http://ca1.cti-gov.dk/oces/issuing/1/cacert/issuing.cer".equals(key)) {
					return OCESTestCertificationAuthority.OCES_3_TEST_INTERMEDIATE_CERTIFICATE;
				}
				return ref.getCertificate(category, key);
			}
			
			@Override
			public void putCertificate(Category category, String key, X509Certificate certificate) throws PKIException {
				ref.putCertificate(category, key, certificate);
			}
		};
	}

	@Test
//    Parsning af valid request:
//    	- identitytoken gyldig i tid og signeret af trusted part
//    	- wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To (hvis det er til stede) og saml tokenet via en SecurityTokenReference er signerede 
//		- Holder-of-key relationen passer
	public void t1_RequestOK() {
		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		assertNotNull(request.getIdentityToken());

		assertValidIdentityToken(request.getIdentityToken());
	}

	@Test
	// Parsning af valid request:
	// - identitytoken gyldig i tid og signeret af trusted part
	// - wsa:MessageID, wsa:Action, wsu:Timestamp, sbf:Framework, soap:Body, wsa:To
	// (hvis det er til stede) og saml tokenet via en SecurityTokenReference er
	// signerede
	// - Holder-of-key relationen passer
	public void t1_RequestOKUsingSecurityTokenReference() {
		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.useSecurityTokenReference();
		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		final CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);
	}

	@Test
	// Parsning af request med
	// - Udløbet identitytoken
	public void t2_RequestFailIdentityTokenExpired() {
		expectedException.expect(ModelException.class);
		RegexMatcher regexMatcher = getRegexMatcher("OIOSAML token no longer valid - now: (.*). OIOSAML token validity end: (.*). Allowed clock drift: 0 seconds");
		expectedException.expectMessage(regexMatcher);

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		final CitizenIdentityTokenBuilder builder = new TestCitizenIdentityTokenBuilder(identityToken, getTokenHolderOfKeyCert(), getTokenSigningVault());
		/* Sæt identitytoken til expired! */
		builder.setNotOnOrAfter(d(-1L));

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(builder.build());

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		final CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);
	}

	@Test
	@Ignore("SDS-6659")
	// Parsning af request med
	// - identitytoken signeret med et trusted, men udløbet certifikat
	public void t3_RequestFailSignedWithTrustedButExpiredCeritificate() {
		expectedException.expect(ModelException.class);
		/* This message is also thrown when certificate is expired */
		expectedException.expectMessage("The certificate that signed the security token is not trusted!");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation("CVR:19343634-UID:1165813731180"), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		final CitizenIdentityTokenBuilder builder = new TestCitizenIdentityTokenBuilder(identityToken, getTokenHolderOfKeyCert(), getExpiredTokenSigningVault());

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(builder.build());

		enhancer.enhanceAndSign();

		/* Making sure certificate us trusted */
		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		final CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);
	}

	@Test
	// Parsning af request med
	// - Identitytoken signeret af en ikke-trusted part
	public void t4_RequestFailSignedByUntrustedPart() {
		expectedException.expect(ModelException.class);
		expectedException.expectMessage("The certificate that signed the security token is not trusted!");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		final CitizenIdentityTokenBuilder builder = new TestCitizenIdentityTokenBuilder(identityToken, getTokenHolderOfKeyCert(), getUntrustedTokenSigningVault());

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(builder.build());

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);
	}

	@Test
	// Parsning af request med
	// - Usigneret identitytoken
	public void t5_RequestFailUnsignedIdentityToken() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find ds:Signature element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestModelBuilder createRequestModelBuilder() {
				return new LibertyRequestModelBuilder(federation) {
					@Override
					public LibertyRequest build(Node node) {
						return new LibertyRequest(federation, node) {
							public CitizenIdentityToken getIdentityToken() {
								CitizenIdentityTokenBuilder citizenIdentityTokenBuilder = new CitizenIdentityTokenBuilder() {
									public CitizenIdentityToken build() throws ModelException {
										Document document = createDocument();
										// Do not valiate scheme
										return new CitizenIdentityToken((Element) document.getFirstChild(), false);
									};
								};
								// Needs to populare citizenIdentityTokenBuilder with values
								return createBuilder(citizenIdentityTokenBuilder).build();
							};
						};
					};
				};
			}
		};
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - Identityttoken med brudt signatur
	public void t6_RequestFailTamperedIdentityTokenSignature() {
		expectedException.expect(ModelException.class);
		expectedException.expectMessage("Signature on IdentityToken is invalid");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		final Node issuer = document.getElementsByTagNameNS(NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ISSUER).item(0);
		issuer.setTextContent("foo");

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - manglende wsa:MessageID
	public void t7_RequestFailMissingMessageID() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find WS-Addressing 1.0 MessageID element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference and tag */
						remove(references, "messageID");
						NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), WSATags.messageID);
						if (messageIDList != null && messageIDList.getLength() == 1) {
							Node messageID = messageIDList.item(0);
							messageID.getParentNode().removeChild(messageID);
						}

						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - wsa:MessageID, som ikke er under signatur
	public void t8_RequestFailMissingMessageIDSignatur() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#MessageID");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference */
						remove(references, "messageID");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - manglende wsu:Timestamp
	public void t9_RequestFailMissingTimestamp() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find Timestamp element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference and tag */
						remove(references, "ts");
						NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), WSUTags.timestamp);
						if (messageIDList != null && messageIDList.getLength() == 1) {
							Node messageID = messageIDList.item(0);
							messageID.getParentNode().removeChild(messageID);
						}

						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - wsu:Timestamp, som ikke er under signatur
	public void t10_RequestFailMissingTimestampSignatur() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd#Timestamp");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference */
						remove(references, "ts");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - manglende sbf:Framework
	public void t11_RequestFailMissingFramework() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find Liberty Framework element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference and tag */
						remove(references, "sbf");
						NodeList messageIDList = TagUtil.getElementsByTagNameNS(document.getDocumentElement(), LibertyTags.framework);
						if (messageIDList != null && messageIDList.getLength() == 1) {
							Node messageID = messageIDList.item(0);
							messageID.getParentNode().removeChild(messageID);
						}

						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - sbf:Framework, som ikke er under signatur
	public void t12_RequestFailMissingFrameworkSignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element urn:liberty:sb#Framework");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference */
						remove(references, "sbf");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - manglende saml:Assertion (og wsse:SecurityTokenReference)
	public void t13_RequestFailMissingAssertion() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find SAML assertion element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		TestTagUtil.removeFirst(document, SAMLTags.assertion);
		TestTagUtil.removeFirst(document, WSSETags.securityTokenReference);

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - manglende saml:Assertion (og wsse:SecurityTokenReference)
	public void t13_RequestFailMissingSecurityTokenReference() {
		expectedException.expect(ModelException.class);
		expectedException.expectMessage("Unable to validate the xmlSignature");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		Element signatureElm = TestTagUtil.getFirstChildElementNS(document.getDocumentElement(), SOAPTags.header, WSSETags.security, DSTags.signature);
		TestTagUtil.removeFirst(signatureElm, WSSETags.securityTokenReference);

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - manglende saml:Assertion (og wsse:SecurityTokenReference)
	public void t14_RequestFailMissingAssertionSignature() {
//		TODO: s.dk laver i øjeblikket requests uden reference til SecurityTokenReference
//		expectedException.expect(ModelBuildException.class);
//		expectedException.expectMessage("Missing Liberty signature on element http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd#SecurityTokenReference");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "str");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - manglende soap:Body
	public void t15_RequestFailMissingBody() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Liberty signature could not be validated");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		TestTagUtil.removeFirst(document, SOAPTags.body);

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - soap:Body, som ikke er under signatur
	public void t16_RequestFailMissingBodySignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://schemas.xmlsoap.org/soap/envelope/#Body");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "body");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	// Parsning af request med
	// - andre SOAP headere som også er under signatur
	public void t17_RequestOKWithOtherSOAPHeaderWithSignature() {
		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<Reference> references) {
						final Element header = TagUtil.getFirstChildElementNS(document.getDocumentElement(), SOAPTags.header);
						String lifeTimeId = ensureElement(header, WSTTags.lifetime, "theLifeTimeValue", true, "demo");
						references.add(new SignatureConfiguration.Reference(lifeTimeId, SignatureConfiguration.Type.DIRECT_REFERENCE_NOT_ENVELOPED));

						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		final CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);

		final Element lifetimeTag = request.getTag(SOAPTags.envelope, SOAPTags.header, WSTTags.lifetime);
		assertNotNull(lifetimeTag);
		assertEquals("demo", lifetimeTag.getAttribute("wsu:Id"));
		assertEquals("theLifeTimeValue", lifetimeTag.getTextContent());
	}

	@Test
	// Parsning af request med
	// - forkert holder-of-key relation
	public void t18_RequestFailWrongHolderOfKeyRelation() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Signing certificate and holder-of-key certificate relation do not match.");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		final CitizenIdentityTokenBuilder builder = new TestCitizenIdentityTokenBuilder(identityToken, getUntrustedTokenSigningVault().getSystemCredentialPair().getCertificate(), getTokenSigningVault());

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(builder.build());
		enhancer.useSecurityTokenReference();

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		final CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);
	}

	@Test
	@Ignore("SDS-6659")
	// Parsning af request med
	// - udløbet holder-of-key certifikat
	public void t19_RequestFailHolderOfKeyCertificateExpired() {
		expectedException.expect(ModelException.class);
		expectedException.expectMessage("The holder-of-key certificate is not trusted! Certificate is expired.");

		// TODO SDS-6659 To make the test work, change the stsSubjectSerialNumber given to getMockFederation
		//  to the matching serial number of the OCES3 certificate that will be returned by getExpiredTokenSigningVault
		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation("CVR:30808460-UID:25351738"), getExpiredTokenSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		final CitizenIdentityTokenBuilder builder = new TestCitizenIdentityTokenBuilder(identityToken, getExpiredTokenSigningVault().getSystemCredentialPair().getCertificate(), getTokenSigningVault());

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(builder.build());

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		final CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);
	}

	@Test
	// Parsning af request med
	// - udløbet holder-of-key certifikat
	public void t19_1_RequestFailHolderOfKeyCertificateMissingSubjectConfirmation() throws Throwable {
		expectedException.expect(SAXParseException.class);
		expectedException.expectMessage("cvc-complex-type.2.4.b: The content of element 'saml:Subject' is not complete. One of '{\"urn:oasis:names:tc:SAML:2.0:assertion\":SubjectConfirmation}' is expected");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference and tag */
						TestTagUtil.removeFirst(document, SAMLTags.subjectConfirmation);

						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		try {

			final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

			enhancer.setWSAddressingAction("http://foo.com#bar");
			enhancer.setIdentityToken(identityToken);

			enhancer.enhanceAndSign();

			factory.createRequest(document);
		} catch (ModelBuildException mbx) {
			throw mbx.getCause();
		}
	}

	@Test
	// Parsning af request med
	// - udløbet holder-of-key certifikat
	public void t19_2_RequestFailHolderOfKeyCertificateMissingSubjectConfirmationMethod() throws Throwable {
		expectedException.expect(SAXParseException.class);
		expectedException.expectMessage("cvc-complex-type.4: Attribute 'Method' must appear on element 'saml:SubjectConfirmation'");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						/* Remove reference and tag */
						Element subjectConfirmation = TagUtil.getFirstChildElementNS(document.getDocumentElement(), SOAPTags.header, WSSETags.security, SAMLTags.assertion, SAMLTags.subject, SAMLTags.subjectConfirmation);
						subjectConfirmation.removeAttribute("Method");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		try {
			final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

			enhancer.setWSAddressingAction("http://foo.com#bar");
			enhancer.setIdentityToken(identityToken);

			enhancer.enhanceAndSign();

			factory.createRequest(document);
		} catch (ModelBuildException mbx) {
			throw mbx.getCause();
		}
	}

	@Test
	// Parsning af request med
	// - brudt beskedssignatur
	public void t20_RequestFail() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Liberty signature could not be validated");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		final Element body = TagUtil.getFirstChildElementNS(document.getDocumentElement(), SOAPTags.body);
		// String lifeTimeId = ensureElement(body, WSTTags.lifetime, "theLifeTimeValue",
		// true, "demo");
		final Element lifeTime = TagUtil.createElementNS(document, WSTTags.lifetime);
		lifeTime.setTextContent("theLifeTimeValue");
		body.appendChild(lifeTime);

		final LibertyRequest request = factory.createRequest(document);

		assertNotNull(request);
		final CitizenIdentityToken citizenIdentityToken = request.getIdentityToken();
		assertNotNull(citizenIdentityToken);
	}

	@Test
	public void tx_MissingSecurityHeader() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find SAML assertion element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		TestTagUtil.removeFirst(document, WSSETags.security);

		factory.createRequest(document);
	}

	@Test
	public void tx_TamperedLibertySignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Liberty signature could not be validated");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault());
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		final Node body = document.getElementsByTagNameNS(NameSpaces.SOAP_SCHEMA, SOAPTags.BODY_UNPREFIXED).item(0);
		body.setTextContent("foo");

		factory.createRequest(document);
	}

	@Test
	public void tx_LibertySignatureMissingSignatureOnAction() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#Action");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "action");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	public void tx_LibertySignatureMissingSignatureOnWsaTo() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#To");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "to");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingTo("http://foo.dk");
		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	public void tx_CreatedTimestamp() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not parse wsu:Created element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), getMessageSigningVault()) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						Node createdTimestampNode = document.getElementsByTagNameNS(NameSpaces.WSU_SCHEMA, WSUTags.CREATED).item(0);
						createdTimestampNode.setTextContent("FOO");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};

		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);

		enhancer.enhanceAndSign();

		final LibertyRequest request = factory.createRequest(document);

		request.getCreatedTimestamp();
	}
}
