/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import org.xml.sax.InputSource;

class InputStreamCollection implements Closeable {

	public static abstract class Creator<E> {
		final ThreadLocal<Class<E>> CREATOR_TYPE = new ThreadLocal<Class<E>>();

		public abstract E create() throws RuntimeException;

		@SuppressWarnings("unchecked")
		public Class<E> type() {
			if (CREATOR_TYPE.get() == null) {
				try {
					Class<E> type = (Class<E>) this.getClass().getMethod("create").getReturnType();
					CREATOR_TYPE.set(type);
				} catch (NoSuchMethodException e) {
					throw new RuntimeException("Coould not find method create in implementation.", e);
				} catch (SecurityException e) {
					throw new RuntimeException("Not allowed to access method create on implementation", e);
				}
			}
			return CREATOR_TYPE.get();
		}
	}

	private static class InputStreamEntry implements Closeable {
		private final Path path;
		private final Map<Class<?>, Object> converts = new HashMap<>();

		private InputStreamEntry() {
			path = null;
		}
		
		private InputStreamEntry(URL url) {
			try {
				path = Paths.get(url.toURI());
			} catch (Exception e) {
				throw new IllegalArgumentException("Cannot find path for " + url, e);
			}
		}

		@SuppressWarnings("unchecked")
		private <E> E get(Creator<E> cr) {
			E result = null;
			if (!converts.containsKey(cr.type())) {
				result = cr.create();
				converts.put(cr.type(), result);
			} else {
				result = (E) converts.get(cr.type());
			}
			return result;
		}

		private ReadableByteChannel getReadableByteChannel() {
			return get(new Creator<ReadableByteChannel>() {
				@Override
				public ReadableByteChannel create() throws RuntimeException {
					if (path == null) {
						return null;
					}
					try {
						return Files.newByteChannel(path);
					} catch (IOException e) {
						throw new RuntimeException("Problem reading from path " + path, e);
					}
				}
			});
		}

		private InputStream getInputStream() {
			return get(new Creator<InputStream>() {
				@Override
				public InputStream create() throws RuntimeException {
					ReadableByteChannel readableByteChannel = getReadableByteChannel();
					return Channels.newInputStream(readableByteChannel);
				}
			});
		}

		@Override
		public void close() throws IOException {
			if (getInputStream() != null) {
				getInputStream().close();
			}
		}
	}

	private final SortedMap<String, InputStreamEntry> streams;

	private InputStreamCollection(Map<String, InputStreamEntry> streams) {
		this.streams = isSortedMap(streams) ? cast(streams) : create(streams);
	}

	private static <K, V> SortedMap<K, V> cast(Map<K, V> map) {
		return (SortedMap<K, V>) map;
	}

	private static boolean isSortedMap(Map<?, ?> map) {
		return map instanceof SortedMap<?, ?>;
	}

	private static <K, V> SortedMap<K, V> create(Map<K, V> map) {
		return new TreeMap<>(map);
	}

	static InputStreamCollection fromSystemResource(String... names) {

		final Map<String, InputStreamEntry> result = new HashMap<>();

		for (String name : names) {
			try {
				URL systemResource = ClassLoader.getSystemResource(name);
				result.put(name, new InputStreamEntry(systemResource));
			} catch (Exception e) {
				throw new RuntimeException("Problems loading resource " + name, e);
			}
		}

		return new InputStreamCollection(result);

	}

	@Override
	public void close() throws IOException {
		for (String name : streams.keySet()) {
			streams.get(name).close();
		}
	}

	private Path getPath(String name) {
		return streams.get(name).path;
	}

	public InputStream getInputStream(String name) {
		return streams.get(name).getInputStream();
	}

	public InputSource getInputSource(String name) {
		return streams.get(name).get(new Creator<InputSource>() {
			@Override
			public InputSource create() throws RuntimeException {
				InputStream inputStream = getInputStream(name);
				return new InputSource(inputStream);
			}
		});
	}

	public String getContent(String name) {
		Path path = getPath(name);

		try {
			byte[] readAllBytes = Files.readAllBytes(path);
			return new String(readAllBytes);
		} catch (IOException e) {
			throw new RuntimeException("Could not read content from " + path, e);
		}
	}
}