/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.model;

import java.util.List;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.WSSETags;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.xml.XmlUtil;

/**
 */
public class LibertyRequestTest extends AbstractIdentityTokenTest {

	@Rule
	public ExpectedException expectedException = ExpectedException.none();

	private Document document;
	private IdentityToken identityToken;

	@Before
	public void setUp() {
		super.setUp();
		
		document = XmlUtil.createEmptySoapDocument();
		identityToken = createBuilder().build();
	}

	@Test
	public void testIdentityTokenWillFailWhenValidatedByLibertyRequest() {
		//Fejler ikke - rettet så Binary Security Token og Holder-of-key ikke forbindes.
		//expectedException.expect(ModelBuildException.class);
		//expectedException.expectMessage("Could not find holder-of-key certificate element");
		
		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault);
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);
		
		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();

		factory.createRequest(document);
	}

	@Test
	public void testTamperedIdentityTokenSignature() {
		expectedException.expect(ModelException.class);
		expectedException.expectMessage("Signature on IdentityToken is invalid");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault);
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();

		final Node issuer = document.getElementsByTagNameNS(NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ISSUER).item(0);
		issuer.setTextContent("foo");

		factory.createRequest(document);
	}

	@Test
	public void testMissingIdentityToken() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find SAML assertion element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault);
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();

		final Node assertion = document.getElementsByTagNameNS(NameSpaces.SAML2ASSERTION_SCHEMA, SAMLTags.ASSERTION).item(0);
		assertion.getParentNode().removeChild(assertion);

		factory.createRequest(document);
	}

	@Test
	public void testMissingSecurityHeader() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Could not find SAML assertion element");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault);
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();

		final Node security = document.getElementsByTagNameNS(NameSpaces.WSSE_SCHEMA, WSSETags.SECURITY).item(0);
		security.getParentNode().removeChild(security);

		factory.createRequest(document);
	}

	@Test
	public void testTamperedLibertySignature() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Liberty signature could not be validated");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault);
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		
		enhancer.setIdentityToken(identityToken);
		enhancer.enhanceAndSign();

		final Node body = document.getElementsByTagNameNS(NameSpaces.SOAP_SCHEMA, SOAPTags.BODY_UNPREFIXED).item(0);
		body.setTextContent("foo");

		factory.createRequest(document);
	}

	@Test
	public void testLibertySignatureMissingSignatureOnMessageID() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#MessageID");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "messageID");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};
		
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();
		
		factory.createRequest(document);
	}

	@Test
	public void testLibertySignatureMissingSignatureOnAction() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#Action");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "action");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};
		
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		enhancer.enhanceAndSign();
		
		factory.createRequest(document);
	}

	@Test
	public void testLibertySignatureMissingSignatureOnWsaTo() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://www.w3.org/2005/08/addressing#To");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "to");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};
		
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingTo("http://foo.dk");
		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();
		
		factory.createRequest(document);
	}

	@Test
	public void testLibertySignatureMissingSignatureOnFrameworkHeader() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element urn:liberty:sb#Framework");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "sbf");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};
		
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingTo("http://foo.dk");
		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();
		
		factory.createRequest(document);
	}

	@Test
	public void testLibertySignatureMissingSignatureOnTimestamp() {
		expectedException.expect(ModelBuildException.class);
		expectedException.expectMessage("Missing Liberty signature on element http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd#Timestamp");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "ts");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};
		
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);
		
		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();
		
		factory.createRequest(document);
	}

	@Test
	public void testLibertySignatureMissingSignatureOnReferenceToIdentityToken() {
		//expectedException.expect(ModelBuildException.class);
		//expectedException.expectMessage("Missing Liberty signature on element http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd#SecurityTokenReference");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "str");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};
		
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();
		
		System.out.println(XmlUtil.node2String(document, true, true));
		
		factory.createRequest(document);
	}

	@Test
	public void testLibertySignatureMissingSignatureOnBody() {
		expectedException.expect(ModelBuildException.class);
		expectedException
				.expectMessage("Missing Liberty signature on element http://schemas.xmlsoap.org/soap/envelope/#Body");

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault) {
			@Override
			public LibertyRequestDOMEnhancer createRequestDOMEnhancer(Document document) {
				return new LibertyRequestDOMEnhancer(credentialVault, document) {
					@Override
					protected SignatureConfiguration getSignatureConfiguration(String securityWsuId, List<SignatureConfiguration.Reference> references) {
						remove(references, "body");
						return super.getSignatureConfiguration(securityWsuId, references);
					}
				};
			}
		};
		
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();
		
		factory.createRequest(document);
	}

	@Test // This test is no longer valid as Token Expiry is not validated against
			// LibertyTimestamp anymore
	public void testIdentityTokenExpired() {
		expectedException.expect(ModelException.class);
		RegexMatcher regexMatcher = getRegexMatcher(
				"OIOSAML token no longer valid - now: (.*). OIOSAML token validity end: (.*). Allowed clock drift: 0 seconds");
		expectedException.expectMessage(regexMatcher);

		identityToken = createExpiredIdentityToken();

		final OIOIDWSFactory factory = new OIOIDWSFactory(getMockFederation(), credentialVault);
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);

		enhancer.setWSAddressingAction("http://foo.com#bar");
		enhancer.setIdentityToken(identityToken);
		
		enhancer.enhanceAndSign();
		
		factory.createRequest(document);
	}

	protected void remove(List<SignatureConfiguration.Reference> references, String identifier) {
		for (SignatureConfiguration.Reference reference : references) {
			if (reference.getURI().equals(identifier)) {
				references.remove(reference);
				break;
			}
		}
	}

	private IdentityToken createExpiredIdentityToken() {
		UserIDCard uidc = createUserIDCard();

		IdentityTokenBuilder itb = new IdentityTokenBuilder(sosiFactory.getCredentialVault());
		itb.setAudienceRestriction("http://fmk-online.dk");
		itb.setNotBefore(d(-10));
		itb.setNotOnOrAfter(d(-5));
		itb.setUserIdCard(uidc);
		itb.setIssuer("http://pan.certifikat.dk/sts/services/SecurityTokenService");
		
		return itb.build();
	}

}
