/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.model.mitid;

import static org.junit.Assert.*;

import java.math.BigInteger;
import java.sql.Date;
import java.time.Instant;

import org.junit.Test;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import dk.sosi.seal.model.AssuranceLevel.NIST;
import dk.sosi.seal.model.AssuranceLevel.NSIS;
import dk.sosi.seal.model.OIO3BSTSAMLAssertion;
import dk.sosi.seal.model.OIOH2BSTSAMLAssertion;
import dk.sosi.seal.model.OIOH3BSTSAMLAssertion;
import dk.sosi.seal.util.SOSITestUtils;
import dk.sosi.seal.xml.XmlUtil;

import javax.xml.crypto.dsig.keyinfo.X509IssuerSerial;

/**
 */
public class OIOSAML3AssertionTest {

    private static final String BASE_DIR =
            System.getProperty("OIOSAML3AssertionTest.baseDir", "/oiosaml-examples/OIOSAML3Assertion");

    public static final Document OIOH2BST = getDocument("OIOH2BST-example.xml");
    public static final Document OIOH3BST = getDocument("OIOH3BST-example.xml");
    public static final Document OIO3BST = getDocument("OIO3BST-unencrypted-example.xml");

    public static final Document OIO3BSTAssertionWithIssuerSerial = getDocument("OIO3BST-issuer-serial-example.xml");
    public static final Document OIO3BSTAssertionWithEncryptedKeyAndIssuerSerial = getDocument("OIO3BST-encrypted-key-issuer-serial-example.xml");

    private static Document getDocument(String file) {
        assertNotNull(BASE_DIR);

        String path;
        if (BASE_DIR.endsWith("/") || file.startsWith("/")) {
            path = BASE_DIR.concat(file);
        } else {
            path = BASE_DIR.concat("/").concat(file);
        }

        InputSource inputSource = new InputSource(OIOSAML3AssertionTest.class.getResourceAsStream(path));
        Document document = XmlUtil.readXml(System.getProperties(), inputSource, false);
        return document;
    }

    @Test
    public void testOIOH2BST() throws Exception {
        OIOH2BSTSAMLAssertion assertion = new OIOH2BSTSAMLAssertion(OIOH2BST.getDocumentElement());
        assertEquals(NIST.NIVEAU_3.toString(), assertion.getAssuranceLevel());
        assertEquals("https://sts.sosi.dk/", assertion.getAudienceRestriction());
        //assertNull("Forventede ikke Certificate Issuer på OIOSAML2Assertion", assertion.getCertificateIssuer());
        //assertNull("Forventede ikke Certificate Serial på OIOSAML2Assertion", assertion.getCertificateSerial());
        //assertNull(assertion.getCommonName()); //dk.sosi.seal.model.ModelException: Mandatory
        assertEquals("1802602810", assertion.getCpr());
        assertEquals("20301823", assertion.getCvrNumberIdentifier());
        //assertEquals("", assertion.getDeliveryNotOnOrAfter()); //java.lang.IllegalArgumentException: DateTimeString
        // cannot be null or empty
        //assertEquals("", assertion.getEmail()); //dk.sosi.seal.model.ModelException: Mandatory 'email' SAML
        // attribute (urn:oid:0.9.2342.19200300.100.1.3) is missing
        assertEquals("_f3070cce-b0ce-4025-b374-ada158cb137c", assertion.getID());
        assertEquals("https://seblogin.nsi.dk/runtime/", assertion.getIssuer());
        //assertEquals("", assertion.getNotBefore());	//java.lang.IllegalArgumentException: DateTimeString cannot be
        // null or empty
        assertEquals(Date.from(Instant.parse("2020-11-13T12:22:50Z")), assertion.getNotOnOrAfter());
        assertEquals("Korsbæk Kommune", assertion.getOrganizationName());
        //assertEquals("", assertion.getRecipient());
        assertEquals("1231593107593", assertion.getRidNumberIdentifier());
        assertEquals(SOSITestUtils.NEW_NEMLOGIN_IDP_CERTIFICATE, assertion.getSigningCertificate());
        assertEquals("DK-SAML-2.0", assertion.getSpecVersion());
        assertEquals("KorsbaekKommune\\MSK", assertion.getSubjectNameID());
        assertEquals("urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified", assertion.getSubjectNameIDFormat());
        //assertEquals("", assertion.getSurName()); //Fejler da den forventes at være mandatory ift. eksisterende
        // OIOSAMLAssertion
        //assertNull("Forventede ikke UID på OIOSAML2Assertion", assertion.getUID());
        //assertEquals("", assertion.getUserAuthenticationInstant()); //dk.sosi.seal.model.ModelException: Date
        // element cannot be null
        //assertNull("Forventede ikke User Certificate på OIOSAML2Assertion", assertion.getUserCertificate());
        //assertNull("Forventede ikke User ID Card på OIOSAML2Assertion", assertion.getUserIDCard());
    }

    @Test
    public void testOIOH3BST() throws Exception {
        OIOH3BSTSAMLAssertion assertion = new OIOH3BSTSAMLAssertion(OIOH3BST.getDocumentElement());
        /* Required */
        validateOIOH3BSTAssertion(assertion);
    }

    public static void validateOIOH3BSTAssertion(OIOH3BSTSAMLAssertion assertion) {

        assertEquals(NSIS.Substantial.toString(), assertion.getAssuranceLevel());
        assertEquals("https://sts.sosi.dk/", assertion.getAudienceRestriction());
        //assertNull("Forventede ikke Certificate Issuer på OIOSAML3Assertion", assertion.getCertificateIssuer());
        //assertNull("Forventede ikke Certificate Serial på OIOSAML3Assertion", assertion.getCertificateSerial());
        //assertNull(assertion.getCommonName()); //dk.sosi.seal.model.ModelException: Mandatory
        assertNull(assertion.getCpr());
        assertEquals("20301823", assertion.getCvrNumberIdentifier());
        //assertEquals("", assertion.getDeliveryNotOnOrAfter()); //java.lang.IllegalArgumentException: DateTimeString
        // cannot be null or empty
        //assertEquals("", assertion.getEmail());	//dk.sosi.seal.model.ModelException: Mandatory 'email' SAML
        // attribute (urn:oid:0.9.2342.19200300.100.1.3) is missing
        assertEquals("_f3070cce-b0ce-4025-b374-ada158cb137c", assertion.getID());
        assertEquals("https://idp.korsbaek-kommune.dk", assertion.getIssuer());
        //assertEquals("", assertion.getNotBefore());	//java.lang.IllegalArgumentException: DateTimeString cannot be
        // null or empty
        assertEquals(Date.from(Instant.parse("2020-11-13T12:22:50Z")), assertion.getNotOnOrAfter());
        assertEquals("Korsbæk Kommune", assertion.getOrganizationName());
        //assertEquals("", assertion.getRecipient());
        assertNull(assertion.getRidNumberIdentifier());
        assertEquals(SOSITestUtils.NEW_NEMLOGIN_IDP_CERTIFICATE, assertion.getSigningCertificate());
        assertEquals("OIO-SAML-3.0", assertion.getSpecVersion());
        assertEquals("OIO-SAML-H-3.0", assertion.getSpecVersionAdditional());
        assertEquals("KorsbaekKommune\\MSK", assertion.getSubjectNameID());
        assertEquals("urn:oasis:names:tc:SAML:2.0:nameid-format:persistent", assertion.getSubjectNameIDFormat());
        //assertEquals("", assertion.getSurName()); //Fejler da den forventes at være mandatory ift. eksisterende
        // OIOSAMLAssertion
        //assertNull("Forventede ikke UID på OIOSAML3Assertion", assertion.getUID());
        //assertEquals("", assertion.getUserAuthenticationInstant()); //dk.sosi.seal.model.ModelException: Date
        // element cannot be null
        //assertNull("Forventede ikke User Certificate på OIOSAML3Assertion", assertion.getUserCertificate());
        //assertNull("Forventede ikke User ID Card på OIOSAML3Assertion", assertion.getUserIDCard());
    }

    @Test
    public void testOIO3BST() throws Exception {
        OIO3BSTSAMLAssertion assertion = new OIO3BSTSAMLAssertion(OIO3BST.getDocumentElement());
        /* Required */
        assertEquals(NSIS.Substantial.toString(), assertion.getAssuranceLevel());
        assertEquals("https://sts.sosi.dk/", assertion.getAudienceRestriction());
        //assertNull("Forventede ikke Certificate Issuer på OIOSAML3Assertion", assertion.getCertificateIssuer());
        //assertNull("Forventede ikke Certificate Serial på OIOSAML3Assertion", assertion.getCertificateSerial());
        //assertNull(assertion.getCommonName()); //dk.sosi.seal.model.ModelException: Mandatory
        assertNull(assertion.getCpr());
        assertEquals("20301823", assertion.getCvrNumberIdentifier());
        //assertEquals("", assertion.getDeliveryNotOnOrAfter()); //java.lang.IllegalArgumentException: DateTimeString
        // cannot be null or empty
        //assertEquals("", assertion.getEmail());	//dk.sosi.seal.model.ModelException: Mandatory 'email' SAML
        // attribute (urn:oid:0.9.2342.19200300.100.1.3) is missing
        assertEquals("_f3070cce-b0ce-4025-b374-ada158cb137c", assertion.getID());
        //assertEquals("https://idp.korsbaek-kommune.dk", assertion.getIssuer());
        //assertEquals("", assertion.getNotBefore());	//java.lang.IllegalArgumentException: DateTimeString cannot be
        // null or empty
        assertEquals(Date.from(Instant.parse("2020-11-13T12:22:50Z")), assertion.getNotOnOrAfter());
        assertEquals("Korsbæk Kommune", assertion.getOrganizationName());
        //assertEquals("", assertion.getRecipient());
        assertNull(assertion.getRidNumberIdentifier());
        assertEquals(SOSITestUtils.NEW_NEMLOGIN_IDP_CERTIFICATE, assertion.getSigningCertificate());
        assertEquals("OIO-SAML-3.0", assertion.getSpecVersion());
        assertEquals("https://data.gov.dk/model/core/eid/professional/uuid/323e4567-e89b-12d3-a456-426655440000",
                assertion.getSubjectNameID());
        assertEquals("urn:oasis:names:tc:SAML:2.0:nameid-format:persistent", assertion.getSubjectNameIDFormat());
        //assertEquals("", assertion.getSurName()); //Fejler da den forventes at være mandatory ift. eksisterende
        // OIOSAMLAssertion
        //assertNull("Forventede ikke UID på OIOSAML3Assertion", assertion.getUID());
        //assertEquals("", assertion.getUserAuthenticationInstant()); //dk.sosi.seal.model.ModelException: Date
        // element cannot be null
        //assertNull("Forventede ikke User Certificate på OIOSAML3Assertion", assertion.getUserCertificate());
        //assertNull("Forventede ikke User ID Card på OIOSAML3Assertion", assertion.getUserIDCard());
    }

    @Test
    public void givenAssertionWithX509IssuerSerialDataThenGetHolderOfKeyCertificateShouldReturnHok() {
        OIO3BSTSAMLAssertion assertion =
                new OIO3BSTSAMLAssertion(OIO3BSTAssertionWithIssuerSerial.getDocumentElement());
        X509IssuerSerial hok = assertion
                .getHolderOfKeyCertificate()
                .get()
                .certificate.match(
                        u -> { throw new RuntimeException("Either should contain issuer serial and not certificate"); },
                        issuerSerial -> issuerSerial);
        assertEquals("CN=TRUST2408 Systemtest XXXIV CA, O=TRUST2408, C=DK", hok.getIssuerName());
        assertEquals(BigInteger.valueOf(1604479732), hok.getSerialNumber());
    }

    @Test
    public void givenAssertionWithX509IssuerSerialDataInEncryptedKeyElementThenGetHolderOfKeyCertificateShouldReturnHok() {
        OIO3BSTSAMLAssertion assertion =
                new OIO3BSTSAMLAssertion(OIO3BSTAssertionWithEncryptedKeyAndIssuerSerial.getDocumentElement());
        X509IssuerSerial hok = assertion
                .getHolderOfKeyCertificate()
                .get()
                .certificate.match(
                        u -> { throw new RuntimeException("Either should contain issuer serial and not certificate"); },
                        issuerSerial -> issuerSerial);
        assertEquals("CN=TRUST2408 Systemtest XXXIV CA, O=TRUST2408, C=DK", hok.getIssuerName());
        assertEquals(BigInteger.valueOf(1604479732), hok.getSerialNumber());
    }
}
