package dk.sosi.seal.model.mitid.builder;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.AssuranceLevel.NIST;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIO2BSTCitizenSAMLAssertion;
import dk.sosi.seal.model.OIO2BSTCitizenSAMLAssertionBuilder;
import dk.sosi.seal.model.OIOBSTSAMLAssertion;
import dk.sosi.seal.model.OIOBootstrapToIdentityTokenRequest;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.modelbuilders.OIOBootstrapToIdentityTokenRequestModelBuilder;
import dk.sosi.seal.security.HolderOfKey;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import java.io.InputStream;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Properties;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class OIO2BSTCitizenTest {
    private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();

    public Date validTo;
    public CredentialVault holderOfKeyVault;
    public CredentialVault issuerSigningVault;

    @Before
    public void setup() {
        holderOfKeyVault = CredentialVaultTestUtil.getFoces3CredentialVault();
        issuerSigningVault = CredentialVaultTestUtil.getVoces3CredentialVault();
        validTo = Date.from(Instant.now().plus(2, ChronoUnit.HOURS));
    }

    private OIO2BSTCitizenSAMLAssertionBuilder getOIO2BSTCitizenSAMLAssertionBuilder() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setIssuer("https://seblogin.nsi.dk/runtime/");
        builder.setNameId("dk:gov:saml:attribute:CprNumberIdentifier:1802602810");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience("https://sts.sosi.dk/");
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(issuerSigningVault);
        builder.setAssuranceLevel(NIST.NIVEAU_3);
        builder.setCpr("1802602810");
        return builder;
    }

    private void validate(OIOBSTSAMLAssertion assertion) {
        validate(assertion, true);
    }

    private void validate(OIOBSTSAMLAssertion assertion, boolean validateCertificates) {
        assertEquals("Uventet type", "OIO2BST_CITIZEN", assertion.getType().toString());

        assertEquals("Uventet Issuer", "https://seblogin.nsi.dk/runtime/", assertion.getIssuer());
        assertEquals("Uventet Name ID", "dk:gov:saml:attribute:CprNumberIdentifier:1802602810",
                assertion.getSubjectNameID());
        assertEquals("Uventet Audience", "https://sts.sosi.dk/", assertion.getAudienceRestriction());
        //Tjekker tid uden MS
        assertEquals("Uventet NotOnOrAfter date", validTo.toString(), assertion.getNotOnOrAfter().toString());
        assertEquals("Uventet Assurance Level", "3", assertion.getAssuranceLevel());
        assertEquals("Uventet CPR", "1802602810", assertion.getCpr());


        //Test getMetoder der ikke er sat i GIVEN
        assertEquals("Uventet Name ID format", "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent",
                assertion.getSubjectNameIDFormat());
        assertNotNull("Forventede DOM objekt fra Assertion", assertion.getDOM());
        assertNotNull("Forventede ID fra Assertion", assertion.getID());
        assertEquals("Uventet Spec version", "DK-SAML-2.0", assertion.getSpecVersion());
        assertNull("Forventede ikke UUID fra Assertion", assertion.getUuid());

        //Test getAttribute
        assertNull("Uventet resultat af NonExistingAttribute", assertion.getAttribute("NonExistingAttribute"));
        assertNotNull("Uventet resultat af OIO2BSTCitizenSAMLAssertion.Attribute.CPR",
                assertion.getAttribute(OIO2BSTCitizenSAMLAssertion.Attribute.CPR.getName()));

        if (validateCertificates) {
            HolderOfKey hok = assertion.getHolderOfKeyCertificate()
                    .orElseThrow(() -> new RuntimeException("Should be some"));
            assertTrue("Uventet Holder Of Key Certificate",
                    hok.matches(holderOfKeyVault.getSystemCredentialPair().getCertificate()));
            assertEquals("Uventet Signing Certificate",
                    issuerSigningVault.getSystemCredentialPair().getCertificate(),
                    assertion.getSigningCertificate());
            assertion.validateSignatureAndTrust(issuerSigningVault);
        }
    }

    @Test
    public void testSAMLAssertionBuilder() {
        //GIVEN
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();

        //WHEN
        OIO2BSTCitizenSAMLAssertion assertion = builder.build();

        //THEN
        validate(assertion);
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithInvalidSigningCertificate() {
        //GIVEN
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();

        //WHEN
        OIO2BSTCitizenSAMLAssertion assertion = builder.build();

        //THEN
        assertion.validateSignatureAndTrust(CredentialVaultTestUtil.getMoces3CredentialVault());
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingAssuranceLevel() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setAssuranceLevel(null);
        builder.build();
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingCpr() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setCpr(null);
        builder.build();
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingIssuer() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setIssuer(null);
        builder.build();
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingNameId() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setNameId(null);
        builder.build();
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingAudience() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setAudience(null);
        builder.build();
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingNotOnOrAfter() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setNotOnOrAfter(null);
        builder.build();
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingHolderOfCertificate() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setHolderOfKeyCertificate(null);
        builder.build();
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingSigningVault() {
        OIO2BSTCitizenSAMLAssertionBuilder builder = getOIO2BSTCitizenSAMLAssertionBuilder();
        builder.setSigningVault(null);
        builder.build();
    }

    @Test
    public void testSAMLAssertionFromElement() throws Exception {
        //GIVEN
        InputStream exampleXml = getClass().getResourceAsStream(
                "/oiosaml-examples/OIOSAML3Assertion/OIO2BST-unencrypted-borger-example.xml");
        Document exampleDocument = XmlUtil.readXml(new Properties(), new InputSource(exampleXml), false);
        validTo = new GregorianCalendar(2020, Calendar.NOVEMBER, 13, 1, 0).getTime();

        //WHEN
        OIO2BSTCitizenSAMLAssertion assertion = new OIO2BSTCitizenSAMLAssertion(exampleDocument.getDocumentElement());

        //THEN
        validate(assertion, false);
    }

    @Test
    public void testBST2IDWSDOMBuilder() {
        //GIVEN
        OIO2BSTCitizenSAMLAssertionBuilder assertionBuilder = getOIO2BSTCitizenSAMLAssertionBuilder();

        OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder =
                OIOSAML_FACTORY.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
        requestBuilder.setOIOBootstrapToken(assertionBuilder.build());
        requestBuilder.setAudience("https://sts.sosi.dk/");
        requestBuilder.setCPRNumberClaim("1802602810");
        requestBuilder.setEncryptionKey(issuerSigningVault.getSystemCredentialPair().getCertificate().getPublicKey());
        requestBuilder.setSigningVault(holderOfKeyVault);

        //WHEN
        Document build = requestBuilder.build();

        //THEN
        assertNotNull(build);
    }

    @Test
    public void testSAMLAssertionToIDCardRequestDOMAndModelBuilderWithInvalidHolderOfKeyRelation() {
        //GIVEN
        OIO2BSTCitizenSAMLAssertionBuilder assertionBuilder = getOIO2BSTCitizenSAMLAssertionBuilder();

        OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder =
                OIOSAML_FACTORY.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
        requestBuilder.setOIOBootstrapToken(assertionBuilder.build());
        requestBuilder.setAudience("https://sts.sosi.dk/");
        requestBuilder.setCPRNumberClaim("1802602810");
        requestBuilder.setSigningVault(issuerSigningVault);

        //WHEN
        Document build = requestBuilder.build();
        OIOBootstrapToIdentityTokenRequestModelBuilder modelBuilder =
                OIOSAML_FACTORY.createOIOBootstrapToIdentityTokenRequestModelBuilder();
        OIOBootstrapToIdentityTokenRequest request = modelBuilder.build(build);

        //THEN
        try {
            request.validateHolderOfKeyRelation();
            throw new AssertionError("validateHolderOfKeyRelation should fail");
        } catch (ModelBuildException e) {
            assertTrue(
                    e.getMessage().contains("Signing certificate and holder-of-key certificate relation do not match"));
        }
    }

    @Test
    public void testSAMLAssertionToIDCardRequestDOMAndModelBuilder() {
        //GIVEN
        OIO2BSTCitizenSAMLAssertionBuilder assertionBuilder = getOIO2BSTCitizenSAMLAssertionBuilder();

        OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder =
                OIOSAML_FACTORY.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
        requestBuilder.setOIOBootstrapToken(assertionBuilder.build());

        requestBuilder.setAudience("https://sts.sosi.dk/");
        requestBuilder.setCPRNumberClaim("1802602810");
        requestBuilder.setEncryptionKey(issuerSigningVault.getSystemCredentialPair().getCertificate().getPublicKey());
        requestBuilder.setSigningVault(holderOfKeyVault);


        //WHEN
        Document build = requestBuilder.build();
        OIOBootstrapToIdentityTokenRequestModelBuilder modelBuilder =
                OIOSAML_FACTORY.createOIOBootstrapToIdentityTokenRequestModelBuilder();
        OIOBootstrapToIdentityTokenRequest request = modelBuilder.build(build);

        //THEN
        assertTrue(request.isAssertionEncrypted());
        request.decryptAssertion(issuerSigningVault.getSystemCredentialPair().getPrivateKey());
        request.validateSignature();
        request.validateSignatureAndTrust(holderOfKeyVault);
        request.validateHolderOfKeyRelation();

        assertEquals("Uventet Audience", "https://sts.sosi.dk/", request.getAppliesTo());
        assertEquals("Uventet Action", "http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue",
                request.getAction());
        assertNotNull("Forventede Context uuid i Request", request.getContext());
        assertNotNull("Forventede Message uuid i Request", request.getMessageID());
        assertEquals("Uventet Signing Certificate", holderOfKeyVault.getSystemCredentialPair().getCertificate(),
                request.getSigningCertificate());
        assertEquals("Uventet cpr number claim", "1802602810", request.getCPRNumberClaim());

        assertTrue(request.isAssertionEncrypted());
        OIOBSTSAMLAssertion assertion = request.getOIOBSTSAMLAssertion();
        validate(assertion);

        assertEquals("Uventet indhold af Claims", "{dk:gov:saml:attribute:CprNumberIdentifier=1802602810}",
                request.getClaimMap().toString());
    }
}
