package dk.sosi.seal.model.mitid.builder;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.AssuranceLevel.NIST;
import dk.sosi.seal.model.*;
import dk.sosi.seal.security.HolderOfKey;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;

import static org.junit.Assert.*;

public class OIO3BSTTest {
    private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();

    public Date validTo;
    public CredentialVault holderOfKeyVault;
    public CredentialVault issuerSigningVault;

    @Before
    public void setup() {
        holderOfKeyVault = CredentialVaultTestUtil.getFoces3CredentialVault();
        issuerSigningVault = CredentialVaultTestUtil.getVoces3CredentialVault();

        validTo = Date.from(Instant.now().plus(2, ChronoUnit.HOURS));
    }

    private OIO3BSTSAMLAssertionBuilder getOIO3BSTSAMLAssertionBuilder() {
        //GIVEN
        OIO3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO3BSTSAMLAssertionBuilder();
        //Assertion
        builder.setIssuer("https://sts.nemlog-in.dk");
        //Subject
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        //Conditions
        builder.setAudience("https://sts.sosi.dk/");

        builder.setNotOnOrAfter(validTo);
        //Signature
        builder.setSigningVault(issuerSigningVault);
        //AttributeStatemnent
        builder.setAssuranceLevel(NIST.NIVEAU_3);
        builder.setCpr("1802602810");
        builder.setUuid("urn:uuid:323e4567-e89b-12d3-a456-426655440000");
        builder.setCvr("20301823");
        builder.setOrganizationName("Korsbæk Kommune");

        return builder;
    }

    private void validate(OIOBSTSAMLAssertion assertion) {
        System.out.println(XmlUtil.node2String(assertion.getDOM(), true, false));
        assertEquals("Uventet type", "OIO3BST", assertion.getType().toString());


        assertEquals("Uventet Issuer", "https://sts.nemlog-in.dk", assertion.getIssuer());
        assertEquals("Uventet Name ID", "KorsbaekKommune\\MSK", assertion.getSubjectNameID());
        HolderOfKey hok = assertion.getHolderOfKeyCertificate()
                .orElseThrow(() -> new RuntimeException("Should be some"));
        assertTrue("Uventet Holder Of Key Certificate",
                hok.matches(holderOfKeyVault.getSystemCredentialPair().getCertificate()));
        assertEquals("Uventet Audience", "https://sts.sosi.dk/", assertion.getAudienceRestriction());
        //Tjekker tid uden MS
        assertEquals("Uventet NotOnOrAfter date", validTo.toString(), assertion.getNotOnOrAfter().toString());
        assertEquals("Uventet Signing Certificate", issuerSigningVault.getSystemCredentialPair().getCertificate(),
                assertion.getSigningCertificate());
        //assertEquals("Uventet Assurance Level", "Substantial", assertion.getAssuranceLevel().toString());
        assertEquals("Uventet Assurance Level", "3", assertion.getAssuranceLevel().toString());
        assertEquals("Uventet CPR", "1802602810", assertion.getCpr());
        assertNull("Forventede ikke RID i Assertion", assertion.getRidNumberIdentifier());
        assertEquals("Uventet CVR", "20301823", assertion.getCvrNumberIdentifier());
        assertEquals("Uventet Organization Name", "Korsbæk Kommune", assertion.getOrganizationName());
        assertNull("Forventede ikke Privileges fra Assertion", assertion.getBasicPrivileges());

        //Test getMetoder der ikke er sat i GIVEN
        assertEquals("Uventet Name ID format", "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent",
                assertion.getSubjectNameIDFormat());
        assertNotNull("Forventede DOM objekt fra Assertion", assertion.getDOM());
        assertNotNull("Forventede ID fra Assertion", assertion.getID());
        assertEquals("Uventet Spec version", "OIO-SAML-3.0", assertion.getSpecVersion());
        assertEquals("Uventet UUID fra Assertion", "urn:uuid:323e4567-e89b-12d3-a456-426655440000",
                assertion.getUuid());

        //Test getAttribute
        assertNull("Uventet resultat af NonExistingAttribute", assertion.getAttribute("NonExistingAttribute"));
        assertNull("Uventet resultat af OIO3BSTSAMLAssertion.Attribute.CVR",
                assertion.getAttribute(OIOH2BSTSAMLAssertion.Attribute.CVR.getName()));
        assertNotNull("Uventet resultat af OIOH3BSTSAMLAssertion.Attribute.CVR",
                assertion.getAttribute(OIOH3BSTSAMLAssertion.Attribute.CVR.getName()));
        assertNotNull("Uventet resultat af OIO3BSTSAMLAssertion.Attribute.CVR",
                assertion.getAttribute(OIO3BSTSAMLAssertion.Attribute.CVR.getName()));
    }

    @Test
    public void testSAMLAssertionBuilder() {
        //GIVEN
        OIO3BSTSAMLAssertionBuilder builder = getOIO3BSTSAMLAssertionBuilder();

        //WHEN
        OIO3BSTSAMLAssertion assertion = builder.build();

        //THEN
        validate(assertion);
    }

    @Test
    public void testBuildValidatesAssertion() {
        //GIVEN
        OIO3BSTSAMLAssertionBuilder builder = Mockito.spy(getOIO3BSTSAMLAssertionBuilder());

        //WHEN
        builder.build();

        //THEN
        Mockito.verify(builder, Mockito.atLeastOnce()).validateBeforeBuild();
    }

    @Test
    public void testMissingHolderOfKey() throws Exception {
        OIO3BSTSAMLAssertionBuilder builder = getOIO3BSTSAMLAssertionBuilder();
        builder.setHolderOfKeyCertificate(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("holderOfCertificate is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingSigningVault() throws Exception {
        OIO3BSTSAMLAssertionBuilder builder = getOIO3BSTSAMLAssertionBuilder();
        builder.setSigningVault(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("signingVault is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingNameId() throws Exception {
        OIO3BSTSAMLAssertionBuilder builder = getOIO3BSTSAMLAssertionBuilder();
        builder.setNameId(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("nameId is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingAudience() throws Exception {
        OIO3BSTSAMLAssertionBuilder builder = getOIO3BSTSAMLAssertionBuilder();
        builder.setAudience(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("audience is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingNotOnOrAfter() throws Exception {
        OIO3BSTSAMLAssertionBuilder builder = getOIO3BSTSAMLAssertionBuilder();
        builder.setNotOnOrAfter(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("notOnOrAfter is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingAttributes() throws Exception {
        for (OIO3BSTSAMLAssertion.Attribute attribute : OIO3BSTSAMLAssertion.Attribute.values()) {
            if (attribute.isMandatory()) {
                OIO3BSTSAMLAssertionBuilder builder = getOIO3BSTSAMLAssertionBuilder();
                OIOBSTSAMLAssertionHelper.removeAttribute(builder, attribute);

                try {
                    builder.validateBeforeBuild();
                    fail("Validation should fail for missing attribute " + attribute.getId());
                } catch (ModelException e) {
                    assertTrue(e.getMessage(), e.getMessage()
                            .contains("Assertion attribute " + attribute.getName() + " is mandatory - but was null"));
                }
            }
        }
    }
}
