package dk.sosi.seal.model.mitid.builder;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.dombuilders.DOMBuilderHelper;
import dk.sosi.seal.model.dombuilders.OIOBSTSAMLAssertionToIDCardRequestDOMBuilder;
import dk.sosi.seal.modelbuilders.OIOBSTSAMLAssertionToIDCardRequestModelBuilder;
import dk.sosi.seal.security.HolderOfKey;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Assume;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.mockito.Mockito;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;

import static org.junit.Assert.*;

@RunWith(Parameterized.class)
public class OIOBSTSAMLAssertionToIDCardRequestTest {
    private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();
    private final OIOBSTSAMLAssertionToIDCardRequestModelBuilder requestModelBuilder =
            OIOSAML_FACTORY.createOIOBSTSAMLAssertionToIDCardRequestModelBuilder();
    private static final String uuid = "urn:uuid:323e4567-e89b-12d3-a456-426655440000";
    private static final String audience = "https://sts.sosi.dk/";
    private static final Date validTo = Date.from(Instant.now().plus(2, ChronoUnit.HOURS));
    public static final String cvr = "20301823";
    public static final String organizationName = "Korsbæk Kommune";
    public static final String cpr = "0707764321";
    public static final String userRole = "7170";
    public static final String authorizationCode = "ZXCVB";
    private static String issuer;
    private static String assuranceLevel;

    @Parameterized.Parameter(0)
    public String suiteName;

    @Parameterized.Parameter(1)
    public CredentialVault holderOfKeyVault;

    @Parameterized.Parameter(2)
    public CredentialVault issuerSigningVault;
    private AbstractOIOBSTSAMLAssertionBuilder assertionBuilder;
    private OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder;
    private OIOBSTSAMLAssertion.Type tokenType;

    /* Setting up test parameters */
    @Parameterized.Parameters(name = "{0}")
    public static Iterable<Object[]> data() {
        // CredentialVault oces2HolderOfKeyVault = CredentialVaultTestUtil.getFocesCredentialVault();
        CredentialVault oces3HolderOfKeyVault = CredentialVaultTestUtil.getFoces3CredentialVault();
        // CredentialVault oces2SigningVault = CredentialVaultTestUtil.getVocesCredentialVault();
        CredentialVault oces3SigningVault = CredentialVaultTestUtil.getVoces3CredentialVault();

        TestParameterBuilder builder = new TestParameterBuilder();
        builder.newTest(OIOBSTSAMLAssertion.Type.OIO3BST.name()).add(oces3HolderOfKeyVault, oces3SigningVault);
        builder.newTest(OIOBSTSAMLAssertion.Type.OIOH3BST.name()).add(oces3HolderOfKeyVault, oces3SigningVault);
        // builder.newTest(OIOBSTSAMLAssertion.Type.OIOH2BST.name()).add(oces2HolderOfKeyVault, oces2SigningVault); TODO SDS-6659

        return builder.build();
    }

    @Before
    public void setup() {
        switch (OIOBSTSAMLAssertion.Type.valueOf(suiteName)) {
            case OIO3BST:
                assertionBuilder = createOIO3BSTSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                requestBuilder = requestBuilderFromAssertionBuilder(holderOfKeyVault, assertionBuilder);
                requestBuilder.setEncryptionKey(
                        issuerSigningVault.getSystemCredentialPair().getCertificate().getPublicKey());
                break;
            case OIOH3BST:
                assertionBuilder = createOIOH3BSTSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                requestBuilder = requestBuilderFromAssertionBuilder(holderOfKeyVault, assertionBuilder);
                break;
            case OIOH2BST:
                assertionBuilder = createOIOH2BSTSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                requestBuilder = requestBuilderFromAssertionBuilder(holderOfKeyVault, assertionBuilder);
                requestBuilder.setEncryptionKey(
                        issuerSigningVault.getSystemCredentialPair().getCertificate().getPublicKey());
                break;
            default:
                throw new RuntimeException("Unrecognized token type");
        }
        tokenType = OIOBSTSAMLAssertion.Type.valueOf(suiteName);
    }

    @Test
    public void testValidRequest() {
        //GIVEN
        Document requestDocument = requestBuilder.build();

        //WHEN
        OIOBSTSAMLAssertionToIDCardRequest request = requestModelBuilder.build(requestDocument);

        //THEN
        if (tokenType != OIOBSTSAMLAssertion.Type.OIOH3BST) {
            request.decryptAssertion(issuerSigningVault.getSystemCredentialPair().getPrivateKey());
        }
        request.validateSignature();
        request.validateSignatureAndTrust(holderOfKeyVault);
        request.validateHolderOfKeyRelation();

        assertEquals("Uventet Audience", "https://fmk", request.getAppliesTo());
        assertEquals("Uventet Action", "http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue",
                request.getAction());
        assertNotNull("Forventede Context uuid i Request", request.getContext());
        assertEquals("Uventet IT System Name", "Korsbæk Kommunes IT systemer", request.getITSystemName());
        assertNotNull("Forventede Message uuid i Request", request.getMessageID());
        assertEquals("Uventet Signing Certificate", holderOfKeyVault.getSystemCredentialPair().getCertificate(),
                request.getSigningCertificate());
        assertEquals("Uventet Subject Name ID", "Mads_Skjern", request.getSubjectNameId());
        assertEquals("Uventet User Role", userRole, request.getUserRole());
        assertEquals("Uventet Authorization Code", authorizationCode, request.getUserAuthorizationCode());
        assertEquals("Uventet indhold af Claims",
                "{medcom:UserRole=7170, medcom:ITSystemName=Korsbæk Kommunes IT systemer, " +
                        "sosi:SubjectNameID=Mads_Skjern, medcom:UserAuthorizationCode=ZXCVB}",
                request.getClaimMap().toString());

        validate(request.getOIOBSTSAMLAssertion());
    }

    @Test
    public void testDecryptionWithWrongKey() throws Exception {
        Assume.assumeTrue(tokenType != OIOBSTSAMLAssertion.Type.OIOH3BST); // OIOH3BST request not encrypted
        //GIVEN
        Document requestDocument = requestBuilder.build();

        //WHEN
        OIOBSTSAMLAssertionToIDCardRequest request = requestModelBuilder.build(requestDocument);

        //THEN
        try {
            request.decryptAssertion(holderOfKeyVault.getSystemCredentialPair()
                    .getPrivateKey()); // assertionSigningVault was used to encrypt
            fail("decrption should fail");
        } catch (Exception e) {
            assertEquals("Unable to decrypt element", e.getMessage());
        }
    }

    @Test
    public void testUnknownEncryptionMethod() throws Exception {
        Assume.assumeTrue(tokenType != OIOBSTSAMLAssertion.Type.OIOH3BST); // OIOH3BST request not encrypted
        //GIVEN
        Document requestDocument = requestBuilder.build();
        Node encryptedData = requestDocument.getElementsByTagName("xenc:EncryptedData").item(0);
        XmlUtil.getFirstChildElement(encryptedData, "EncryptionMethod")
                .setAttribute("Algorithm", "http://www.w3.org/2009/xmlenc11#aes128-gcm");


        //WHEN
        OIOBSTSAMLAssertionToIDCardRequest request = requestModelBuilder.build(requestDocument);

        //THEN
        try {
            request.decryptAssertion(issuerSigningVault.getSystemCredentialPair().getPrivateKey());
            fail("decrption should fail");
        } catch (Exception e) {
            assertEquals("Unable to decrypt element. The encryption algorithm might not be supported.", e.getMessage());
        }
    }

    @Test
    public void testInvalidHolderOfKeyRelation() throws Exception {
        //GIVEN
        assertionBuilder.setHolderOfKeyCertificate(issuerSigningVault.getSystemCredentialPair().getCertificate());
        requestBuilder.setOIOBSTSAMLAssertion(assertionBuilder.build());
        Document requestDocument = requestBuilder.build();

        //WHEN
        OIOBSTSAMLAssertionToIDCardRequest request = requestModelBuilder.build(requestDocument);
        if (request.isAssertionEncrypted()) {
            request.decryptAssertion(issuerSigningVault.getSystemCredentialPair().getPrivateKey());
        }

        //THEN
        try {
            request.validateHolderOfKeyRelation();
            fail("holder-of-key validation should fail");
        } catch (Exception e) {
            assertTrue(e.getMessage(),
                    e.getMessage().contains("Signing certificate and holder-of-key certificate relation do not match"));
        }
    }

    @Test
    public void testUntrustedSignature() throws Exception {
        //GIVEN
        assertionBuilder.setHolderOfKeyCertificate(issuerSigningVault.getSystemCredentialPair().getCertificate());
        requestBuilder.setSigningVault(issuerSigningVault);
        Document requestDocument = requestBuilder.build();

        //WHEN
        OIOBSTSAMLAssertionToIDCardRequest request = requestModelBuilder.build(requestDocument);

        //THEN
        try {
            request.validateSignatureAndTrust(holderOfKeyVault);
            fail("validation should fail");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains(
                    "The certificate that signed the security token is not trusted! Signing certificate was:"));
        }
    }

    @Test
    public void testBrokenMessageSignature() throws Exception {
        //GIVEN
        Document requestDocument = requestBuilder.build();
        Element claims =
                getFirstChild(requestDocument.getDocumentElement(), "http://docs.oasis-open.org/ws-sx/ws-trust/200512",
                        "Claims");
        Element claimValue = getFirstChild(claims, "http://docs.oasis-open.org/wsfed/authorization/200706", "Value");
        claimValue.getFirstChild().setNodeValue("modified");

        //WHEN
        OIOBSTSAMLAssertionToIDCardRequest request = requestModelBuilder.build(requestDocument);

        //THEN
        try {
            request.validateSignature();
            fail("signature validation should fail");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("signature could not be validated"));
        }
    }

    @Test
    public void testMissingMessageSignature() throws Exception {
        testMissingMessageSignature("body");
        testMissingMessageSignature("ts");
        testMissingMessageSignature("messageID");
        testMissingMessageSignature("action");
    }

    private void testMissingMessageSignature(String messageUri) throws Exception {
        //GIVEN
        Document requestDocument = DOMBuilderHelper.buildWithoutSigning(requestBuilder);
        new MissingMessageSignatureSigner(holderOfKeyVault, requestDocument).skipUri(messageUri).enhanceAndSign();
        MissingMessageSignatureValidator messageSkippingValidator =
                new MissingMessageSignatureValidator(new EmptyCredentialVault(),
                        requestDocument.getDocumentElement()).skipUri(messageUri);

        //WHEN
        OIOBSTSAMLAssertionToIDCardRequest request = requestModelBuilder.build(requestDocument);

        //THEN
        messageSkippingValidator.validateSignature(); // Should not fail
        try {
            request.validateSignature();
            fail("signature validation should fail");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("Missing Liberty signature on element "));
        }
    }

    @Test
    public void testMissingAssertion() throws Exception {
        requestBuilder.setOIOBSTSAMLAssertion(null);
        try {
            requestBuilder.build();
            fail("Build should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("OIOBSTSAMLAssertion is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingAudience() throws Exception {
        requestBuilder.setAudience(null);
        try {
            requestBuilder.build();
            fail("Build should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("audience is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingSigningVault() throws Exception {
        requestBuilder.setSigningVault(null);
        try {
            requestBuilder.build();
            fail("Build should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("signingVault is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingITSystemName() throws Exception {
        requestBuilder.setITSystemName(null);
        try {
            requestBuilder.build();
            fail("Build should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("ITSystemName is mandatory - but was null"));
        }
    }

    @Test
    public void testRequestBuilderValidatesAssertion() throws Exception {
        AbstractOIOBSTSAMLAssertionBuilder assertionBuilderSpy = Mockito.spy(assertionBuilder);
        requestBuilder.setOIOSAMLAssertion(assertionBuilderSpy);
        requestBuilder.build();

        Mockito.verify(assertionBuilderSpy, Mockito.atLeastOnce()).validateBeforeBuild();
    }

    private void validate(OIOBSTSAMLAssertion assertion) {
        assertEquals("Uventet type", tokenType.name(), assertion.getType().toString());
        assertEquals("Uventet Issuer", issuer, assertion.getIssuer());
        assertEquals("Uventet Name ID", "KorsbaekKommune\\MSK", assertion.getSubjectNameID());
        HolderOfKey hok = assertion.getHolderOfKeyCertificate()
                .orElseThrow(() -> new RuntimeException("Should be some"));
        assertTrue("Uventet Holder Of Key Certificate",
                hok.matches(holderOfKeyVault.getSystemCredentialPair().getCertificate()));
        assertEquals("Uventet Audience", audience, assertion.getAudienceRestriction());
        //Tjekker tid uden MS
        assertEquals("Uventet NotOnOrAfter date", validTo.toString(), assertion.getNotOnOrAfter().toString());
        assertEquals("Uventet Signing Certificate", issuerSigningVault.getSystemCredentialPair().getCertificate(),
                assertion.getSigningCertificate());
        assertEquals("Uventet Assurance Level", assuranceLevel, assertion.getAssuranceLevel());
        if (tokenType != OIOBSTSAMLAssertion.Type.OIOH3BST) {
            assertEquals("Uventet CPR", cpr, assertion.getCpr());
        }
        assertNull("Forventede ikke RID i Assertion", assertion.getRidNumberIdentifier());
        assertEquals("Uventet CVR", cvr, assertion.getCvrNumberIdentifier());
        assertEquals("Uventet Organization Name", organizationName, assertion.getOrganizationName());

        //Test getMetoder der ikke er sat i GIVEN
        String expectedNameIdFormat =
                tokenType == OIOBSTSAMLAssertion.Type.OIOH2BST ? "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified"
                        : "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent";
        assertEquals("Uventet Name ID format", expectedNameIdFormat, assertion.getSubjectNameIDFormat());
        assertNotNull("Forventede DOM objekt fra Assertion", assertion.getDOM());
        assertNotNull("Forventede ID fra Assertion", assertion.getID());
        String expectedVersion = tokenType == OIOBSTSAMLAssertion.Type.OIOH2BST ? "DK-SAML-2.0" : "OIO-SAML-3.0";
        assertEquals("Uventet Spec version", expectedVersion, assertion.getSpecVersion());
        assertEquals("Uventet UUID fra Assertion", uuid, assertion.getUuid());

        if (tokenType == OIOBSTSAMLAssertion.Type.OIO3BST) {
            assertNull("Forventede ikke Privileges fra Assertion", assertion.getBasicPrivileges());
        } else {
            assertNotNull("Forventede Privileges fra Assertion", assertion.getBasicPrivileges());
            assertEquals("Uventet Privilege", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3",
                    assertion.getBasicPrivileges().getPrivileges("urn:dk:gov:saml:cvrNumberIdentifier:20301823")
                            .get(0));
        }

        //Test getAttribute
        assertNull("Uventet resultat af NonExistingAttribute", assertion.getAttribute("NonExistingAttribute"));
        switch (assertion.getClass().getName()) {
            case "OIOH2BSTSAMLAssertion":
                assertNotNull("Uventet resultat af OIOH2BSTSAMLAssertion.Attribute.CVR",
                        assertion.getAttribute(OIOH2BSTSAMLAssertion.Attribute.CVR.getName()));
                break;
            case "OIOH3BSTSAMLAssertion":
                assertNotNull("Uventet resultat af OIOH3BSTSAMLAssertion.Attribute.CVR",
                        assertion.getAttribute(OIOH3BSTSAMLAssertion.Attribute.CVR.getName()));
                break;
            case "OIO3BSTSAMLAssertion":
                assertNotNull("Uventet resultat af OIO3BSTSAMLAssertion.Attribute.CVR",
                        assertion.getAttribute(OIO3BSTSAMLAssertion.Attribute.CVR.getName()));
                break;
        }
    }

    private static OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilderFromAssertionBuilder(
            CredentialVault signingVault, AbstractOIOBSTSAMLAssertionBuilder assertionBuilder) {
        OIOBSTSAMLAssertionToIDCardRequestDOMBuilder requestBuilder =
                OIOSAML_FACTORY.createOIOBSTSAMLAssertionToIDCardRequestDOMBuilder();
        requestBuilder.setOIOBSTSAMLAssertion(assertionBuilder.build());
        requestBuilder.setAudience("https://fmk");
        requestBuilder.setITSystemName("Korsbæk Kommunes IT systemer");
        requestBuilder.setSigningVault(signingVault);
        requestBuilder.setUserRole(userRole);
        requestBuilder.setUserAuthorizationCode(authorizationCode);
        requestBuilder.setSubjectNameID("Mads_Skjern");
        return requestBuilder;
    }

    private static OIO3BSTSAMLAssertionBuilder createOIO3BSTSAMLAssertionBuilder(CredentialVault holderOfKeyVault,
                                                                                 CredentialVault signingVault) {
        OIO3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO3BSTSAMLAssertionBuilder();
        issuer = "https://oio3bst-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        AssuranceLevel.NIST level = AssuranceLevel.NIST.NIVEAU_4;
        assuranceLevel = level.toString();
        builder.setAssuranceLevel(level);
        builder.setUuid(uuid);
        builder.setCvr(cvr);
        builder.setOrganizationName(organizationName);
        builder.setCpr(cpr);
        return builder;
    }

    private static OIOH3BSTSAMLAssertionBuilder createOIOH3BSTSAMLAssertionBuilder(CredentialVault holderOfKeyVault,
                                                                                   CredentialVault signingVault) {
        OIOH3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH3BSTSAMLAssertionBuilder();
        issuer = "https://oioh3bst-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        AssuranceLevel.NSIS level = AssuranceLevel.NSIS.High;
        assuranceLevel = level.toString();
        builder.setAssuranceLevel(level);
        builder.setUuid(uuid);
        builder.setCvr(cvr);
        builder.setOrganizationName(organizationName);
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(BasicPrivileges.Profile.V12);
        privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:20301823",
                "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3");
        builder.setPrivileges(privilegesBuilder);
        return builder;
    }

    private static OIOH2BSTSAMLAssertionBuilder createOIOH2BSTSAMLAssertionBuilder(CredentialVault holderOfKeyVault,
                                                                                   CredentialVault signingVault) {
        OIOH2BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH2BSTSAMLAssertionBuilder();
        issuer = "https://oioh2bst-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        AssuranceLevel.NIST level = AssuranceLevel.NIST.NIVEAU_4;
        assuranceLevel = level.toString();
        builder.setAssuranceLevel(level);
        builder.setUuid(uuid);
        builder.setCvr(cvr);
        builder.setOrganizationName(organizationName);
        builder.setCpr(cpr);
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(BasicPrivileges.Profile.V12);
        privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:20301823",
                "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3");
        builder.setPrivileges(privilegesBuilder);
        return builder;
    }

    private static Element getFirstChild(Element ancestor, String tagNS, String tagName) {
        return (Element) getChildren(ancestor, tagNS, tagName).item(0);
    }

    private static NodeList getChildren(Element ancestor, String tagNS, String tagName) {
        return tagNS != null ? ancestor.getElementsByTagNameNS(tagNS, tagName) : ancestor.getElementsByTagName(tagName);
    }
}
