package dk.sosi.seal.model.mitid.builder;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.OIOBSTSAMLAssertion;
import dk.sosi.seal.model.OIOBootstrapToken;
import dk.sosi.seal.model.OIOBootstrapTokenBuilder;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.security.HolderOfKey;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import org.junit.Before;
import org.junit.Test;
import org.w3c.dom.Document;

import java.util.Calendar;

import static org.junit.Assert.*;

public class OIOBootstrapTokenTest {
    private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();

    public CredentialVault holderOfKeyVault;
    public CredentialVault issuerSigningVault;

    @Before
    public void setup() {
        holderOfKeyVault = CredentialVaultTestUtil.getFoces3CredentialVault();
        issuerSigningVault = CredentialVaultTestUtil.getVoces3CredentialVault();
    }

    private OIOBootstrapTokenBuilder getOIOBootstrapTokenBuilder(CredentialVault signingVault) {
        OIOBootstrapTokenBuilder builder =
                new OIOBootstrapTokenBuilder(signingVault, "https://saml.test-nemlog-in.dk/");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setSubjectName(
                "C=DK,O=LAKESIDE A/S // CVR:25450442,CN=Sårjournal TEST læge,Serial=CVR:25450442-RID:73570260");
        builder.setAudienceRestriction("https://bootstrap.sts.nemlog-in.dk/");
        builder.setCpr("12345678");
        return builder;
    }

    private void validate(OIOBSTSAMLAssertion assertion) {
        assertEquals("Uventet type", "OIO2BST_LEGACY", assertion.getType().toString());

        assertEquals("Uventet Issuer", "https://saml.test-nemlog-in.dk/", assertion.getIssuer());
        assertEquals("Uventet Name ID",
                "C=DK,O=LAKESIDE A/S // CVR:25450442,CN=Sårjournal TEST læge,Serial=CVR:25450442-RID:73570260",
                assertion.getSubjectNameID());
        assertEquals("Uventet Audience", "https://bootstrap.sts.nemlog-in.dk/", assertion.getAudienceRestriction());
        Calendar expireTime = Calendar.getInstance();
        expireTime.add(Calendar.HOUR_OF_DAY, 1);
        assertTrue("Uventet NotOnOrAfter date", assertion.getNotOnOrAfter().before(expireTime.getTime()));

        assertEquals("Uventet Name ID format", "urn:oasis:names:tc:SAML:1.1:nameid-format:X509SubjectName",
                assertion.getSubjectNameIDFormat());
        assertNotNull("Forventede DOM objekt fra Assertion", assertion.getDOM());
        assertNotNull("Forventede ID fra Assertion", assertion.getID());
        assertEquals("DK-SAML-2.0", assertion.getSpecVersion());
        assertEquals("12345678", assertion.getCpr());
        assertEquals("3", assertion.getAssuranceLevel());
        assertNull("Forventede ikke UUID fra Assertion", assertion.getUuid());

        //Test getAttribute
        assertNull("Uventet resultat af NonExistingAttribute", assertion.getAttribute("NonExistingAttribute"));

        HolderOfKey hok = assertion.getHolderOfKeyCertificate()
                .orElseThrow(() -> new RuntimeException("Should be some"));
        assertTrue("Uventet Holder Of Key Certificate",
                hok.matches(holderOfKeyVault.getSystemCredentialPair().getCertificate()));
        assertion.validateSignatureAndTrust(issuerSigningVault);
    }

    @Test
    public void testSAMLAssertionBuilder() {
        //GIVEN
        OIOBootstrapTokenBuilder builder = getOIOBootstrapTokenBuilder(issuerSigningVault);

        //WHEN
        OIOBootstrapToken assertion = builder.build();

        //THEN
        validate(assertion);
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithInvalidSigningCertificate() {
        //GIVEN
        OIOBootstrapTokenBuilder builder = getOIOBootstrapTokenBuilder(issuerSigningVault);

        //WHEN
        OIOBootstrapToken assertion = builder.build();

        //THEN
        assertion.validateSignatureAndTrust(CredentialVaultTestUtil.getVoces2CredentialVault());
    }

    @Test(expected = ModelException.class)
    public void testSAMLAssertionBuilderWithMissingIssuer() {
        OIOBootstrapTokenBuilder builder = getOIOBootstrapTokenBuilder(issuerSigningVault);
        builder.setIssuer(null);
        builder.build();
    }

    @Test
    public void testBST2IDWSDOMBuilder() {
        //GIVEN
        OIOBootstrapTokenBuilder assertionBuilder = getOIOBootstrapTokenBuilder(issuerSigningVault);

        OIOBootstrapToIdentityTokenRequestDOMBuilder requestBuilder =
                OIOSAML_FACTORY.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
        requestBuilder.setOIOBootstrapToken(assertionBuilder.build());
        requestBuilder.setAudience("https://sts.sosi.dk/");
        requestBuilder.setCPRNumberClaim("1802602810");
        requestBuilder.setEncryptionKey(issuerSigningVault.getSystemCredentialPair().getCertificate().getPublicKey());
        requestBuilder.setSigningVault(holderOfKeyVault);

        //WHEN
        Document build = requestBuilder.build();

        //THEN
        assertNotNull(build);
    }
}
