package dk.sosi.seal.model.mitid.builder;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.AssuranceLevel.NIST;
import dk.sosi.seal.model.BasicPrivileges.Profile;
import dk.sosi.seal.model.*;
import dk.sosi.seal.security.HolderOfKey;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;

import static org.junit.Assert.*;

public class OIOH2BSTTest {
    private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();

    public Date validTo;
    public CredentialVault holderOfKeyVault;
    public CredentialVault issuerSigningVault;

    @Before
    public void setup() {
        holderOfKeyVault = CredentialVaultTestUtil.getFoces3CredentialVault();
        issuerSigningVault = CredentialVaultTestUtil.getVoces3CredentialVault();

        validTo = Date.from(Instant.now().plus(2, ChronoUnit.HOURS));
    }

    private OIOH2BSTSAMLAssertionBuilder getOIOH2BSTSAMLAssertionBuilder(boolean withPrivileges) {
        //GIVEN
        OIOH2BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH2BSTSAMLAssertionBuilder();
        //Assertion
        builder.setIssuer("https://seblogin.nsi.dk/runtime/");
        //Subject
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        //Conditions
        builder.setAudience("https://sts.sosi.dk/");

        builder.setNotOnOrAfter(validTo);
        //Signature
        builder.setSigningVault(issuerSigningVault);
        //AttributeStatemnent
        builder.setAssuranceLevel(NIST.NIVEAU_3);
        builder.setCpr("1802602810");
        builder.setRid("1231593107593");
        builder.setCvr("20301823");
        builder.setOrganizationName("Korsbæk Kommune");

        if (withPrivileges) {
            //privilege
            BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
            privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:20301823",
                    "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3");
            builder.setPrivileges(privilegesBuilder);

        }

        return builder;
    }

    private void validate(OIOBSTSAMLAssertion assertion, boolean withPrivileges) {
        assertEquals("Uventet type", "OIOH2BST", assertion.getType().toString());

        assertEquals("Uventet Issuer", "https://seblogin.nsi.dk/runtime/", assertion.getIssuer());
        assertEquals("Uventet Name ID", "KorsbaekKommune\\MSK", assertion.getSubjectNameID());
        HolderOfKey hok = assertion.getHolderOfKeyCertificate()
                .orElseThrow(() -> new RuntimeException("Should be some"));
        assertTrue("Uventet Holder Of Key Certificate",
                hok.matches(holderOfKeyVault.getSystemCredentialPair().getCertificate()));
        assertEquals("Uventet Audience", "https://sts.sosi.dk/", assertion.getAudienceRestriction());
        //Tjekker tid uden MS
        assertEquals("Uventet NotOnOrAfter date", validTo.toString(), assertion.getNotOnOrAfter().toString());
        assertEquals("Uventet Signing Certificate", issuerSigningVault.getSystemCredentialPair().getCertificate(),
                assertion.getSigningCertificate());
        assertEquals("Uventet Assurance Level", "3", assertion.getAssuranceLevel().toString());
        assertEquals("Uventet CPR", "1802602810", assertion.getCpr());
        assertEquals("Uventet Organization Name", "Korsbæk Kommune", assertion.getOrganizationName());
        if (withPrivileges) {
            assertNotNull("Forventede Privileges fra Assertion", assertion.getBasicPrivileges());
            assertEquals("Uventet Privilege", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3",
                    assertion.getBasicPrivileges().getPrivileges("urn:dk:gov:saml:cvrNumberIdentifier:20301823")
                            .get(0));
        } else {
            assertNull("Forventede ikke Privileges fra Assertion", assertion.getBasicPrivileges());
        }


        //Test getMetoder der ikke er sat i GIVEN
        assertEquals("Uventet Name ID format", "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                assertion.getSubjectNameIDFormat());
        assertNotNull("Forventede DOM objekt fra Assertion", assertion.getDOM());
        assertNotNull("Forventede ID fra Assertion", assertion.getID());
        assertEquals("Uventet Spec version", "DK-SAML-2.0", assertion.getSpecVersion());
        assertNull("Forventede ikke UUID fra Assertion", assertion.getUuid());

        //Test getAttribute
        assertNull("Uventet resultat af NonExistingAttribute", assertion.getAttribute("NonExistingAttribute"));
        assertNotNull("Uventet resultat af OIOH2BSTSAMLAssertion.Attribute.CPR",
                assertion.getAttribute(OIOH2BSTSAMLAssertion.Attribute.CVR.getName()));
        assertNull("Uventet resultat af OIOH3BSTSAMLAssertion.Attribute.CPR",
                assertion.getAttribute(OIOH3BSTSAMLAssertion.Attribute.CVR.getName()));
        assertNull("Uventet resultat af OIO3BSTSAMLAssertion.Attribute.CPR",
                assertion.getAttribute(OIO3BSTSAMLAssertion.Attribute.CVR.getName()));
    }

    @Test
    public void testMinimalSAMLAssertionBuilder() {
        //GIVEN
        boolean withPrivileges = false;
        OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(withPrivileges);

        //WHEN
        OIOH2BSTSAMLAssertion assertion = builder.build();

        //THEN
        validate(assertion, withPrivileges);
    }

    @Test
    public void testSAMLAssertionBuilder() {
        //GIVEN
        boolean withPrivileges = true;
        OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(withPrivileges);

        //WHEN
        OIOH2BSTSAMLAssertion assertion = builder.build();

        //THEN
        validate(assertion, withPrivileges);
    }

    @Test
    public void testBuildValidatesAssertion() {
        //GIVEN
        OIOH2BSTSAMLAssertionBuilder builder = Mockito.spy(getOIOH2BSTSAMLAssertionBuilder(false));

        //WHEN
        builder.build();

        //THEN
        Mockito.verify(builder, Mockito.atLeastOnce()).validateBeforeBuild();
    }

    @Test
    public void testMissingHolderOfKey() throws Exception {
        OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(false);
        builder.setHolderOfKeyCertificate(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("holderOfCertificate is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingSigningVault() throws Exception {
        OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(false);
        builder.setSigningVault(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("signingVault is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingNameId() throws Exception {
        OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(false);
        builder.setNameId(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("nameId is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingAudience() throws Exception {
        OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(false);
        builder.setAudience(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("audience is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingNotOnOrAfter() throws Exception {
        OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(false);
        builder.setNotOnOrAfter(null);
        try {
            builder.validateBeforeBuild();
            fail("Validation should fail");
        } catch (ModelException e) {
            assertTrue(e.getMessage(), e.getMessage().contains("notOnOrAfter is mandatory - but was null"));
        }
    }

    @Test
    public void testMissingAttributes() throws Exception {
        for (OIOH2BSTSAMLAssertion.Attribute attribute : OIOH2BSTSAMLAssertion.Attribute.values()) {
            if (attribute.isMandatory()) {
                OIOH2BSTSAMLAssertionBuilder builder = getOIOH2BSTSAMLAssertionBuilder(false);
                OIOBSTSAMLAssertionHelper.removeAttribute(builder, attribute);

                try {
                    builder.validateBeforeBuild();
                    fail("Validation should fail for missing attribute " + attribute.getId());
                } catch (ModelException e) {
                    assertTrue(e.getMessage(), e.getMessage()
                            .contains("Assertion attribute " + attribute.getName() + " is mandatory - but was null"));
                }
            }
        }
    }
}
