package dk.sosi.seal.model.mitid.builder;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.AssuranceLevel.NSIS;
import dk.sosi.seal.model.BasicPrivileges.Profile;
import dk.sosi.seal.model.*;
import dk.sosi.seal.security.HolderOfKey;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;

import static org.junit.Assert.*;

public class OIOH3BSTTest {
	private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();
	
	public Date validTo;
	public CredentialVault holderOfKeyVault;
	public CredentialVault issuerSigningVault;

	private Date issuerInstant = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'").parse("2022-12-24T20:17:46.384Z");

	public OIOH3BSTTest() throws ParseException {
	}

	@Before
	public void setup() {
		holderOfKeyVault = CredentialVaultTestUtil.getFoces3CredentialVault();
		issuerSigningVault = CredentialVaultTestUtil.getVoces3CredentialVault();
		
		validTo = Date.from(Instant.now().plus(2, ChronoUnit.HOURS));
	}
	
	private OIOH3BSTSAMLAssertionBuilder getOIOH3BSTSAMLAssertionBuilder(boolean withPrivileges) throws Exception{
		//GIVEN
		OIOH3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH3BSTSAMLAssertionBuilder();
		//Assertion
		builder.setIssuer("https://seblogin.nsi.dk/runtime/");
		//Subject
		builder.setNameId("KorsbaekKommune\\MSK");
		builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
		//Conditions
		builder.setAudience("https://sts.sosi.dk/");
		builder.setNotOnOrAfter(validTo);

		// AuthnStatement
		builder.setIssuerInstant(issuerInstant);

		//Signature
		builder.setSigningVault(issuerSigningVault);
		//AttributeStatemnent
		builder.setAssuranceLevel(NSIS.Substantial);
		builder.setUuid("urn:uuid:323e4567-e89b-12d3-a456-426655440000");
		builder.setCvr("20301823");

		builder.setOrganizationName("Korsbæk Kommune");
		
		if(withPrivileges) {
			//privilege
			BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(Profile.V12);
			privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:20301823", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3");
			builder.setPrivileges(privilegesBuilder);
		}
		
		return builder;
	}
	
	private void validate(OIOBSTSAMLAssertion assertion, boolean withPrivileges) {
		assertEquals(OIOH3BSTSAMLAssertion.class, assertion.getClass());
		OIOH3BSTSAMLAssertion oioh3bstAssertion = (OIOH3BSTSAMLAssertion) assertion;
		assertEquals("Uventet type", "OIOH3BST", oioh3bstAssertion.getType().toString());
		assertEquals("Uventet spec version", "OIO-SAML-3.0", oioh3bstAssertion.getSpecVersion());
		assertEquals("Uventet additional spec version", "OIO-SAML-H-3.0", oioh3bstAssertion.getSpecVersionAdditional());

		assertEquals("Uventet Issuer", "https://seblogin.nsi.dk/runtime/", oioh3bstAssertion.getIssuer());
		assertEquals("Uventet Name ID", "KorsbaekKommune\\MSK", oioh3bstAssertion.getSubjectNameID());
		HolderOfKey hok = oioh3bstAssertion.getHolderOfKeyCertificate()
				.orElseThrow(() -> new RuntimeException("Should be some"));
		assertTrue("Uventet Holder Of Key Certificate", hok.matches(holderOfKeyVault.getSystemCredentialPair().getCertificate()));
		assertEquals("Uventet Audience", "https://sts.sosi.dk/", oioh3bstAssertion.getAudienceRestriction());
		//Tjekker tid uden MS
		assertEquals("Uventet NotOnOrAfter date", validTo.toString(), oioh3bstAssertion.getNotOnOrAfter().toString());
		assertEquals("Uventet Signing Certificate", issuerSigningVault.getSystemCredentialPair().getCertificate(), oioh3bstAssertion.getSigningCertificate());
		assertEquals("Uventet Assurance Level", "Substantial", oioh3bstAssertion.getAssuranceLevel());
		assertNull("Forventer ikke CPR i OIOH3BST", oioh3bstAssertion.getCpr());
		assertNull("Forventer ikke RID i OIOH3BST", oioh3bstAssertion.getRidNumberIdentifier());
		assertEquals("Uventet CVR", "20301823", oioh3bstAssertion.getCvrNumberIdentifier());
		assertEquals("Uventet Organization Name", "Korsbæk Kommune",oioh3bstAssertion.getOrganizationName());
		if(withPrivileges) {
			assertNotNull("Forventede Privileges fra Assertion", oioh3bstAssertion.getBasicPrivileges());
			assertEquals("Uventet Privilege", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3", oioh3bstAssertion.getBasicPrivileges().getPrivileges("urn:dk:gov:saml:cvrNumberIdentifier:20301823").get(0));
		} else {
			assertNull("Forventede ikke Privileges fra Assertion", oioh3bstAssertion.getBasicPrivileges());
		}
		
		
		//Test getMetoder der ikke er sat i GIVEN
		assertEquals("Uventet Name ID format", "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent",oioh3bstAssertion.getSubjectNameIDFormat());
		assertNotNull("Forventede DOM objekt fra Assertion", oioh3bstAssertion.getDOM());
		assertNotNull("Forventede ID fra Assertion", oioh3bstAssertion.getID());
		assertEquals("Uventet Spec version", "OIO-SAML-3.0",oioh3bstAssertion.getSpecVersion());
		assertEquals("Uventet uuid", "urn:uuid:323e4567-e89b-12d3-a456-426655440000", oioh3bstAssertion.getUuid());
		
		//Test getAttribute
		assertNull("Uventet resultat af NonExistingAttribute", oioh3bstAssertion.getAttribute("NonExistingAttribute"));
		assertNull("Uventet resultat af OIOH2BSTSAMLAssertion.Attribute.CPR", oioh3bstAssertion.getAttribute(OIOH2BSTSAMLAssertion.Attribute.SPEC_VER.getName()));
		assertNotNull("Uventet resultat af OIOH3BSTSAMLAssertion.Attribute.CPR",oioh3bstAssertion.getAttribute(OIOH3BSTSAMLAssertion.Attribute.SPEC_VER.getName()));
		assertNotNull("Uventet resultat af OIO3BSTSAMLAssertion.Attribute.CPR",oioh3bstAssertion.getAttribute(OIO3BSTSAMLAssertion.Attribute.SPEC_VER.getName()));

		// Test AuthnStatement
		assertEquals("Uventet dato for IssuerInstant", issuerInstant.toString(), oioh3bstAssertion.getIssuerInstant().toString());

	}
	@Test
	public void testMinimalSAMLAssertionBuilder() throws Exception {
		//GIVEN
		boolean withPrivileges = false;
		OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(withPrivileges);
		
		//WHEN
		OIOH3BSTSAMLAssertion assertion = builder.build();
		
		//THEN
		validate(assertion, withPrivileges);
	}
	
	@Test
	public void testSAMLAssertionBuilder() throws Exception {
		//GIVEN
		boolean withPrivileges = true;		
		OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(withPrivileges);
				
		//WHEN
		OIOH3BSTSAMLAssertion assertion = builder.build();
		System.out.println(XmlUtil.node2String(assertion.getDOM(), true, false));
		

		//THEN
		validate(assertion, withPrivileges);
	}

	@Test
	public void testBuildValidatesAssertion() throws  Exception{
		//GIVEN
		OIOH3BSTSAMLAssertionBuilder builder = Mockito.spy(getOIOH3BSTSAMLAssertionBuilder(false));

		//WHEN
		builder.build();

		//THEN
		Mockito.verify(builder, Mockito.atLeastOnce()).validateBeforeBuild();
	}

	@Test
	public void testMissingHolderOfKey() throws Exception {
		OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(false);
		builder.setHolderOfKeyCertificate(null);
		try {
			builder.validateBeforeBuild();
			fail("Validation should fail");
		} catch (ModelException e) {
			assertTrue(e.getMessage(), e.getMessage().contains("holderOfCertificate is mandatory - but was null"));
		}
	}

	@Test
	public void testMissingSigningVault() throws Exception {
		OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(false);
		builder.setSigningVault(null);
		try {
			builder.validateBeforeBuild();
			fail("Validation should fail");
		} catch (ModelException e) {
			assertTrue(e.getMessage(), e.getMessage().contains("signingVault is mandatory - but was null"));
		}
	}

	@Test
	public void testMissingNameId() throws Exception {
		OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(false);
		builder.setNameId(null);
		try {
			builder.validateBeforeBuild();
			fail("Validation should fail");
		} catch (ModelException e) {
			assertTrue(e.getMessage(), e.getMessage().contains("nameId is mandatory - but was null"));
		}
	}

	@Test
	public void testMissingAudience() throws Exception {
		OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(false);
		builder.setAudience(null);
		try {
			builder.validateBeforeBuild();
			fail("Validation should fail");
		} catch (ModelException e) {
			assertTrue(e.getMessage(), e.getMessage().contains("audience is mandatory - but was null"));
		}
	}

	@Test
	public void testMissingNotOnOrAfter() throws Exception {
		OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(false);
		builder.setNotOnOrAfter(null);
		try {
			builder.validateBeforeBuild();
			fail("Validation should fail");
		} catch (ModelException e) {
			assertTrue(e.getMessage(), e.getMessage().contains("notOnOrAfter is mandatory - but was null"));
		}
	}

	@Test
	public void testMissingAttributes() throws Exception {
		for (OIOH3BSTSAMLAssertion.Attribute attribute : OIOH3BSTSAMLAssertion.Attribute.values()) {
			if (attribute.isMandatory()) {
				OIOH3BSTSAMLAssertionBuilder builder = getOIOH3BSTSAMLAssertionBuilder(false);
				OIOBSTSAMLAssertionHelper.removeAttribute(builder, attribute);

				try {
					builder.validateBeforeBuild();
					fail("Validation should fail for missing attribute " + attribute.getId());
				} catch (ModelException e) {
					assertTrue(e.getMessage(), e.getMessage().contains("Assertion attribute " + attribute.getName() + " is mandatory - but was null"));
				}
			}
		}
	}
}
