package dk.sosi.seal.model.mitid.builder;

import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.AbstractOIOBSTSAMLAssertionBuilder;
import dk.sosi.seal.model.AssuranceLevel;
import dk.sosi.seal.model.BasicPrivileges;
import dk.sosi.seal.model.BasicPrivilegesDOMBuilder;
import dk.sosi.seal.model.OIO2BSTCitizenSAMLAssertion;
import dk.sosi.seal.model.OIO2BSTCitizenSAMLAssertionBuilder;
import dk.sosi.seal.model.OIO3BSTCitizenSAMLAssertion;
import dk.sosi.seal.model.OIO3BSTCitizenSAMLAssertionBuilder;
import dk.sosi.seal.model.OIO3BSTSAMLAssertion;
import dk.sosi.seal.model.OIO3BSTSAMLAssertionBuilder;
import dk.sosi.seal.model.OIOBSTSAMLAssertion;
import dk.sosi.seal.model.OIOH2BSTSAMLAssertion;
import dk.sosi.seal.model.OIOH2BSTSAMLAssertionBuilder;
import dk.sosi.seal.model.OIOH3BSTSAMLAssertion;
import dk.sosi.seal.model.OIOH3BSTSAMLAssertionBuilder;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.lang.reflect.InvocationTargetException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;

import static dk.sosi.seal.model.OIOBSTSAMLAssertion.Type.OIO2BST_CITIZEN;
import static dk.sosi.seal.model.OIOBSTSAMLAssertion.Type.OIO3BST;
import static dk.sosi.seal.model.OIOBSTSAMLAssertion.Type.OIO3BST_CITIZEN;
import static dk.sosi.seal.model.OIOBSTSAMLAssertion.Type.OIOH2BST;
import static dk.sosi.seal.model.OIOBSTSAMLAssertion.Type.OIOH3BST;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

@RunWith(Parameterized.class)
public class OIOXBSTTest {
    private static final OIOSAMLFactory OIOSAML_FACTORY = new OIOSAMLFactory();
    private AbstractOIOBSTSAMLAssertionBuilder assertionBuilder;
    private OIOBSTSAMLAssertion.Type tokenType;

    private static final String uuid = "urn:uuid:323e4567-e89b-12d3-a456-426655440000";
    private static final String audience = "https://sts.sosi.dk/";
    private static final Date validTo = Date.from(Instant.now().plus(2, ChronoUnit.HOURS));
    public static final String cvr = "20301823";
    public static final String organizationName = "Korsbæk Kommune";
    public static final String cpr = "0707764321";
    public static final String userRole = "7170";
    private static String issuer;

    @Parameterized.Parameter(0)
    public String suiteName;

    @Parameterized.Parameter(1)
    public CredentialVault holderOfKeyVault;

    @Parameterized.Parameter(2)
    public CredentialVault issuerSigningVault;

    @Parameterized.Parameter(3)
    public Class<OIOBSTSAMLAssertion> assertionClass;

    /* Setting up test parameters */
    @Parameterized.Parameters(name = "{0}")
    public static Iterable<Object[]> data() {
        // CredentialVault oces2HolderOfKeyVault = CredentialVaultTestUtil.getFocesCredentialVault();
        CredentialVault oces3HolderOfKeyVault = CredentialVaultTestUtil.getFoces3CredentialVault();
        // CredentialVault oces2SigningVault = CredentialVaultTestUtil.getVocesCredentialVault();
        CredentialVault oces3SigningVault = CredentialVaultTestUtil.getVoces3CredentialVault();

        TestParameterBuilder builder = new TestParameterBuilder();
        builder.newTest(OIO3BST.name()).add(oces3HolderOfKeyVault, oces3SigningVault, OIO3BSTSAMLAssertion.class);
        builder.newTest(OIOH3BST.name()).add(oces3HolderOfKeyVault, oces3SigningVault, OIOH3BSTSAMLAssertion.class);
//        builder.newTest(OIOH2BST.name()).add(oces2HolderOfKeyVault, oces3SigningVault, OIOH2BSTSAMLAssertion.class); // TODO SDS-6659
//        builder.newTest(OIO2BST_CITIZEN.name()).add(oces2HolderOfKeyVault, oces2SigningVault, OIO2BSTCitizenSAMLAssertion.class); // TODO SDS-6659
//        builder.newTest(OIO3BST_CITIZEN.name()).add(oces2HolderOfKeyVault, oces2SigningVault, OIO3BSTCitizenSAMLAssertion.class); // TODO SDS-6659
        return builder.build();
    }

    @Before
    public void setup() {
        switch (OIOBSTSAMLAssertion.Type.valueOf(suiteName)) {
            case OIO3BST :
                assertionBuilder = createOIO3BSTSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                break;
            case OIOH3BST :
                assertionBuilder = createOIOH3BSTSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                break;
            case OIOH2BST :
                assertionBuilder = createOIOH2BSTSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                break;
            case OIO2BST_CITIZEN :
                assertionBuilder = createOIO2BSTCitizenSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                break;
            case OIO3BST_CITIZEN :
                assertionBuilder = createOIO3BSTCitizenSAMLAssertionBuilder(holderOfKeyVault, issuerSigningVault);
                break;
            default :
                throw new RuntimeException("Unrecognized token type");

        }
        tokenType = OIOBSTSAMLAssertion.Type.valueOf(suiteName);
    }

    @Test
    public void testExpiredToken() throws Exception {
        assertionBuilder.setNotOnOrAfter(Date.from(Instant.now().minus(2, ChronoUnit.HOURS)));
        OIOBSTSAMLAssertion token = assertionBuilder.build();

        try {
            token.validateTimestamp();
            fail("Validation should fail");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().matches("OIOSAML token no longer valid - now: .* OIOSAML token validity end: .* Allowed clock drift: 0 seconds"));
        }
    }

    @Test
    public void testInvalidSignature() throws Exception {
        Element dom = assertionBuilder.build().getDOM().getDocumentElement();
        dom.getElementsByTagName("saml:NameID").item(0).setTextContent("Break signature");
        OIOBSTSAMLAssertion token = assertionClass.getConstructor(Element.class).newInstance(dom);

        try {
            token.validateSignatureAndTrust(issuerSigningVault);
            fail("Validation should fail");
        } catch (Exception e) {
            assertEquals("Signature on OIOSAMLAssertion is invalid", e.getMessage());
        }
    }

    @Test
    public void testUnknownSignatureMethod() throws Exception {
        Element dom = assertionBuilder.build().getDOM().getDocumentElement();
        ((Element) dom.getElementsByTagName("ds:SignatureMethod").item(0)).setAttribute("Algorithm", "http://www.w3.org/2001/04/xmldsig-more#ecdsa-sha256");
        OIOBSTSAMLAssertion token = assertionClass.getConstructor(Element.class).newInstance(dom);

        try {
            token.validateSignatureAndTrust(issuerSigningVault);
            fail("Validation should fail");
        } catch (Exception e) {
            assertTrue(e.getMessage(), e.getMessage().contains("Unable to validate the xmlSignature. The requested algorithm http://www.w3.org/2001/04/xmldsig-more#ecdsa-sha256 does not exist"));
        }
    }

    @Test
    public void testInvalidSamlToken() throws Exception {
        Document dom = assertionBuilder.build().getDOM();
        Element invalidElement = dom.createElement("saml:invalidTag");
        dom.getElementsByTagName("saml:Subject").item(0).appendChild(invalidElement);

        try {
            assertionClass.getConstructor(Element.class).newInstance(dom.getDocumentElement());
            fail("Schema validation should fail");
        } catch (InvocationTargetException e) {
            Throwable cause = e.getCause();
            assertTrue(cause.getMessage(), cause.getMessage().contains("Invalid content was found starting with element 'saml:invalidTag'"));
        }
    }

    @Test
    public void testInvalidXml() throws Exception {
        Document dom = assertionBuilder.build().getDOM();
        Element invalidElement = dom.createElement("saml:invalidTag");
        dom.getElementsByTagName("saml:Subject").item(0).appendChild(invalidElement);

        try {
            assertionClass.getConstructor(Element.class).newInstance(dom.getDocumentElement());
            fail("Schema validation should fail");
        } catch (InvocationTargetException e) {
            Throwable cause = e.getCause();
            assertTrue(cause.getMessage(), cause.getMessage().contains("Invalid content was found starting with element 'saml:invalidTag'"));
        }
    }

    @Test
    public void testMissingMandatoryAttribute() throws Exception {
        Document assertionDoc = assertionBuilder.build().getDOM();
        String mandatoryAttribute =
                tokenType == OIO3BST ? OIO3BSTSAMLAssertion.Attribute.CVR.getName() :
                tokenType == OIOH3BST ? OIOH3BSTSAMLAssertion.Attribute.CVR.getName() :
                tokenType == OIOH2BST ? OIOH2BSTSAMLAssertion.Attribute.CVR.getName() :
                tokenType == OIO2BST_CITIZEN ? OIO2BSTCitizenSAMLAssertion.Attribute.CPR.getName() :
                tokenType == OIO3BST_CITIZEN ? OIO3BSTCitizenSAMLAssertion.Attribute.CPR.getName() : null;
        String attributeValue = mandatoryAttribute.toLowerCase().contains("cvr") ? "cvr" : "cpr";

        removeAttribute(assertionDoc, mandatoryAttribute);
        OIOBSTSAMLAssertion token = assertionClass.getConstructor(Element.class).newInstance(assertionDoc.getDocumentElement());

        try {
            String ignore = attributeValue.equals("cvr") ? token.getCvrNumberIdentifier() : token.getCpr();
            fail("Should throw exception");
        } catch (Exception e) {
            assertEquals("Mandatory '" + attributeValue + "' SAML attribute (" + mandatoryAttribute + ") is missing", e.getMessage());
        }
    }

    private void removeAttribute(Document assertionDoc, String attributeName) {
        NodeList attributes = assertionDoc.getDocumentElement().getElementsByTagName("saml:Attribute");
        for (int i = 0; i < attributes.getLength(); i++) {
            Node attribute = attributes.item(i);
            String currentAttributeName = attribute.getAttributes().getNamedItem("Name").getTextContent();

            if (currentAttributeName.equals(attributeName)) {
                attribute.getParentNode().removeChild(attribute);
            }
        }
    }

    @Test
    public void testUntrustedSignature() throws Exception {
        assertionBuilder.setSigningVault(holderOfKeyVault);
        OIOBSTSAMLAssertion token = assertionBuilder.build();

        try {
            token.validateSignatureAndTrust(issuerSigningVault);
            fail("Validation should fail");
        } catch (Exception e) {
            assertTrue(e.getMessage().contains("The certificate that signed the security token is not trusted! Signing certificate was:"));
        }
    }

    private static OIO3BSTSAMLAssertionBuilder createOIO3BSTSAMLAssertionBuilder(CredentialVault holderOfKeyVault, CredentialVault signingVault) {
        OIO3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO3BSTSAMLAssertionBuilder();
        issuer = "https://oio3bst-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        AssuranceLevel.NIST level = AssuranceLevel.NIST.NIVEAU_4;
        builder.setAssuranceLevel(level);
        builder.setUuid(uuid);
        builder.setCvr(cvr);
        builder.setOrganizationName(organizationName);
        builder.setCpr(cpr);
        return builder;
    }

    private static OIOH3BSTSAMLAssertionBuilder createOIOH3BSTSAMLAssertionBuilder(CredentialVault holderOfKeyVault, CredentialVault signingVault) {
        OIOH3BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH3BSTSAMLAssertionBuilder();
        issuer = "https://oioh3bst-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        AssuranceLevel.NSIS level = AssuranceLevel.NSIS.High;
        builder.setAssuranceLevel(level);
        builder.setUuid(uuid);
        builder.setCvr(cvr);
        builder.setOrganizationName(organizationName);
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(BasicPrivileges.Profile.V12);
        privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:20301823", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3");
        builder.setPrivileges(privilegesBuilder);
        return builder;
    }

    private static OIOH2BSTSAMLAssertionBuilder createOIOH2BSTSAMLAssertionBuilder(CredentialVault holderOfKeyVault, CredentialVault signingVault) {
        OIOH2BSTSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIOH2BSTSAMLAssertionBuilder();
        issuer = "https://oioh2bst-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        AssuranceLevel.NIST level = AssuranceLevel.NIST.NIVEAU_4;
        builder.setAssuranceLevel(level);
        builder.setUuid(uuid);
        builder.setCvr(cvr);
        builder.setOrganizationName(organizationName);
        builder.setCpr(cpr);
        BasicPrivilegesDOMBuilder privilegesBuilder = new BasicPrivilegesDOMBuilder(BasicPrivileges.Profile.V12);
        privilegesBuilder.addPrivilege("urn:dk:gov:saml:cvrNumberIdentifier:20301823", "urn:dk:healthcare:national-federation-role:code:41003:value:PlejeAssR3");
        builder.setPrivileges(privilegesBuilder);
        return builder;
    }

    private static OIO2BSTCitizenSAMLAssertionBuilder createOIO2BSTCitizenSAMLAssertionBuilder(CredentialVault holderOfKeyVault, CredentialVault signingVault) {
        OIO2BSTCitizenSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO2BSTCitizenSAMLAssertionBuilder();
        issuer = "https://oio2bst-citizen-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        builder.setAssuranceLevel(AssuranceLevel.NIST.NIVEAU_4);
        builder.setCpr(cpr);
        return builder;
    }

    private static OIO3BSTCitizenSAMLAssertionBuilder createOIO3BSTCitizenSAMLAssertionBuilder(CredentialVault holderOfKeyVault, CredentialVault signingVault) {
        OIO3BSTCitizenSAMLAssertionBuilder builder = OIOSAML_FACTORY.createOIO3BSTCitizenSAMLAssertionBuilder();
        issuer = "https://oio3bst-citizen-issuer.dk";
        builder.setIssuer(issuer);
        builder.setNameId("KorsbaekKommune\\MSK");
        builder.setHolderOfKeyCertificate(holderOfKeyVault.getSystemCredentialPair().getCertificate());
        builder.setAudience(audience);
        builder.setNotOnOrAfter(validTo);
        builder.setSigningVault(signingVault);
        builder.setAssuranceLevel(AssuranceLevel.NSIS.Substantial);
        builder.setCpr(cpr);
        return builder;
    }
}
