/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.pki;

import java.io.IOException;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;
import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.SecurityProviderUtil;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.modelbuilders.SignatureInvalidModelBuildException;
import dk.sosi.seal.pki.impl.HashMapCertificateCache;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

import static org.junit.Assert.*;

public class TestAbstractOCESCertificationAuthority {

	private Properties properties;
	private boolean bcAdded;

	@Before
	public void setUp() throws Exception {
		BasicConfigurator.configure(new NullAppender());
		bcAdded = SecurityProviderUtil.addBCAsProvider();
		properties = SignatureUtil.setupCryptoProviderForJVM();
	}

	@After
	public void tearDown() throws Exception {
		if (bcAdded) {
			SecurityProviderUtil.removeBCAsProvider();
		}
		BasicConfigurator.resetConfiguration();
	}

	/* OCES 1*/
	@Test
	public void testExpiredOCES1ProductionCertificate() {
		CertificationAuthority testCA = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());
		CertificationAuthority prodCA = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());
		X509Certificate cert = CertificateParser.asCertificate(XmlUtil.fromBase64(InlinedTestCertificates.OCES_1_PRODUCTION_CERT));
		assertFalse(testCA.isValid(cert));
		assertFalse(prodCA.isValid(cert));
	}

	@Test
	@Ignore("SDS-6659: This is moved from the 'testOCES1ProductionRootCertificate' test - should it be rewritten to a OCES3 test?")
	public void testOCES3ProductionRootCertificate() {
		CertificationAuthority testCA = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());
		X509Certificate cert = OCESCertificationAuthority.OCES_3_ROOT_CERTIFICATES.get(0); // SDS-6659 This has been changed to the OCES3 root certificate instead of OCES1, so the test is failing now

		try {
			testCA.isValid(cert);
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Exception cause", "The supplied certificate  with DN 'CN=TDC OCES CA, O=TDC, C=DK' is not a OCES Test certificate", ex.getMessage());
		}
	}

	@Test
	public void testOCES1ProductionRootCertificate() {
		NaiveCertificateStatusChecker certificateStatusChecker = new NaiveCertificateStatusChecker(properties);
		CertificationAuthority prodCA = CertificationAuthorityFactory.create(properties,
																			 CertificationAuthorityFactory.OCES_CA,
																			 certificateStatusChecker,
																			 new HashMapCertificateCache());
		PKIException ex = assertThrows(PKIException.class,
									   () -> prodCA.isValid(OCESCertificationAuthority.OCES_1_ROOT_CERTIFICATE));
		assertEquals("The supplied certificate with DN 'CN=TDC OCES CA, O=TDC, C=DK' is an OCES1 certificate. OCES1 " +
							 "certificates are no longer supported.", ex.getMessage());
	}

	@Test
	public void testOCES1TestRootCertificate() {
		CertificationAuthority testCA = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());
		CertificationAuthority prodCA = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());
		X509Certificate cert = OCESTestCertificationAuthority.OCES_1_TEST_ROOT_CERTIFICATE;

		try {
			testCA.isValid(cert);
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Exception cause", "The supplied certificate with DN 'CN=TDC OCES Systemtest CA II, O=TDC, C=DK' is an OCES1 certificate. OCES1 certificates are no longer supported.", ex.getMessage());
		}

		try {
			prodCA.isValid(cert);
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Exception cause", "The supplied certificate  with DN 'CN=TDC OCES Systemtest CA II, O=TDC, C=DK' is not a OCES Production certificate", ex.getMessage());
		}

	}

	/* OCES 2*/
	@Test
	@Ignore("SDS-6659")
	public void testOCES2CertificateValidationIG() throws Exception {
		AbstractOCESCertificationAuthority ca = (AbstractOCESCertificationAuthority) CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new NaiveCertificateStatusChecker(properties), getCertificateCacheForIG());
		X509Certificate cert = loadMoces2CertificateIG();
		// Trust to IG removed
		try {
			ca.getAndValidateIntermediateOCES2Certificate(cert);
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Exception cause", "Intermediate certificate not issued by OCES Test root certificate", ex.getMessage());
		}
	}

	@Test
	public void testNegativeCheckerUserCertificate() throws IOException {
		CertificationAuthority ca = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new MockCertificateStatusChecker(1), CredentialVaultTestUtil.getCertificateCacheForVoces3CredentialVault());
		X509Certificate cert = CredentialVaultTestUtil.getVoces3CredentialVault().getSystemCredentialPair().getCertificate();
		assertFalse(ca.isValid(cert));
	}

	@Test
	public void testNegativeCheckerIntermediateCertificate() throws IOException {
		CertificationAuthority ca = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new MockCertificateStatusChecker(0), CredentialVaultTestUtil.getCertificateCacheForVoces3CredentialVault());
		X509Certificate cert = CredentialVaultTestUtil.getVoces3CredentialVault().getSystemCredentialPair().getCertificate();
		try {

			ca.isValid(cert);
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Exception cause", "Intermediate certificate is revoked", ex.getMessage());
		}
	}

	@Test
	public void testRetractedIntermediateCertificate() throws PKIException, IOException {
		PublicKey pub = CredentialVaultTestUtil.getMoces3CredentialVault().getSystemCredentialPair().getCertificate()
				.getPublicKey();
		Date notAfter = new Date(System.currentTimeMillis() - 1 * 60 * 1000);
		Date notBefore = new Date(System.currentTimeMillis() - 10 * 60 * 1000);
		final X509Certificate intermediateCertificate = OCESTestHelper.issueCertificate("cn=Thomas,o=Test,c=DK",
				"thomas@signaturgruppen.dk", pub, loadIntermediateCertificate().getPublicKey(), notBefore, notAfter);

		final CertificateCache mockCache = new CertificateCache() {
			public void putCertificate(Category category, String key, X509Certificate certificate) throws PKIException {
			}

			public X509Certificate getCertificate(Category category, String key) throws PKIException {
				return intermediateCertificate;
			}
		};
		CertificationAuthority ca = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new MockCertificateStatusChecker(0), mockCache);

		try {
			ca.isValid(loadMoces3Certificate());
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Dates invalid", "Intermediate certificate not valid in time", ex.getMessage());
		}
	}

	@Test
	@Ignore("SDS-6659")
	public void testInvalidUserCertificates() throws PKIException, IOException {
		AbstractOCESCertificationAuthority ca = (AbstractOCESCertificationAuthority) CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());

		try {
			// Intermediate certificates don't have an AIA extension at all
			ca.isValid(loadIntermediateCertificate());
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Cause", "Invalid certificate - Authority Information Access (1.3.6.1.5.5.7.1.1) not found.",
					ex.getMessage());
		}

		try {
			// Certificate has an AIA extension but is missing the 'CA Issuers' part
			ca.getAndValidateIntermediateOCES2Certificate(loadInvalidMoces2CertificateIG());
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Cause", "Invalid certificate - CA Issuers (1.3.6.1.5.5.7.48.2) not found under Authority Information Access.",
					ex.getMessage());
		}
	}

	@Test
	public void testInvalidIntermediateCertificate() throws PKIException, IOException {
		PublicKey pub = CredentialVaultTestUtil.getMoces3CredentialVault().getSystemCredentialPair().getCertificate()
				.getPublicKey();
		final X509Certificate intermediateCertificate = OCESTestHelper.issueCertificate("cn=Thomas,o=Test,c=DK",
																						"thomas@signaturgruppen.dk", pub, loadIntermediateCertificate().getPublicKey());

		final CertificateCache mockCache = new CertificateCache() {

			public void putCertificate(Category category, String key, X509Certificate certificate) throws PKIException {
			}

			public X509Certificate getCertificate(Category category, String key) throws PKIException {
				return intermediateCertificate;
			}
		};
		CertificationAuthority ca = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new MockCertificateStatusChecker(0), mockCache);

		try {
			ca.isValid(loadMoces3Certificate());
			fail("Exception expected");
		} catch (PKIException ex) {
			assertEquals("Dates invalid", "Intermediate certificate not issued by OCES Test root certificate", ex.getMessage());
		}
	}

	@Test
	public void testSOSIFactoryWithOCES3Certificate() throws Exception {
		// Client side!!!
		SOSIFactory clientFactory = CredentialVaultTestUtil.createOCES3SOSIFactory();

		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "12345", "someOrg");
		IDCard clientIdCard = clientFactory.createNewSystemIDCard("SOSITEST", careProvider,
				AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, clientFactory.getCredentialVault()
				.getSystemCredentialPair().getCertificate(), null);

		// Create request
		Request clientRequest = clientFactory.createNewRequest(true, "oces3testflow");
		clientRequest.setIDCard(clientIdCard);

		// Make request document.
		Document clientDocument = clientRequest.serialize2DOMDocument(XmlUtil.createEmptyDocument());

		// Assert signature of client document
		Node signature = clientDocument.getElementsByTagNameNS(NameSpaces.DSIG_SCHEMA, "Signature").item(0);
		assertNotNull("Signature expected not null", signature);
		assertTrue(SignatureUtil.validate(signature, clientFactory.getFederation(), clientFactory.getCredentialVault(),
				true));

		// Serialize document to XML string
		String clientDocumentXMLString = XmlUtil.node2String(clientDocument);

		// Server side!!!
		SOSIFactory serverFactory = CredentialVaultTestUtil.createOCES3SOSIFactory();
		Request serverRequest = serverFactory.deserializeRequest(clientDocumentXMLString);

		IDCard serverIdCard = serverRequest.getIDCard();
		assertEquals(clientIdCard, serverIdCard);

		clientDocumentXMLString = clientDocumentXMLString.replaceAll("1\\.0\\.1", "1.0");
		try {
			serverFactory.deserializeRequest(clientDocumentXMLString);
			fail("Expected SignatureInvalidModelBuildException");
		} catch (SignatureInvalidModelBuildException e) {
			assertEquals("Signature could not be validated", e.getMessage());
		}
	}

	@Test
	public void testRootCertificates() {
		/* ROD CERTIFIKATER PROD*/
		AbstractOCESCertificationAuthority ocesCa =
				(AbstractOCESCertificationAuthority) CertificationAuthorityFactory.create
				(properties, CertificationAuthorityFactory.OCES_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());
		/* OCES 1 - Prod*/
		assertEquals(new DistinguishedName("C=DK,O=TDC,CN=TDC OCES CA"), new DistinguishedName(ocesCa.getOCES1RootCertificate().getSubjectX500Principal()));

		/* OCES 2 - Prod*/
		assertEquals(new DistinguishedName("C=DK,O=TRUST2408,CN=TRUST2408 OCES Primary CA"), new DistinguishedName(ocesCa.getOCES2RootCertificate().getSubjectX500Principal()));

		/* OCES 3 - Prod*/
		List<X509Certificate> oces3CaList = ocesCa.getOCES3RootCertificates();
		assertEquals(1, oces3CaList.size());
		assertEquals(new DistinguishedName("C=DK, O=Den Danske Stat, CN=Den Danske Stat OCES rod-CA"), new DistinguishedName(ocesCa.getOCES3RootCertificates().get(0).getSubjectX500Principal()));

		/* ROD CERTIFIKATER TEST*/
		AbstractOCESCertificationAuthority ocesTestCa =
				(AbstractOCESCertificationAuthority) CertificationAuthorityFactory.create
				(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new NaiveCertificateStatusChecker(properties), new HashMapCertificateCache());
		/* OCES 1 - Test*/
		assertEquals(new DistinguishedName("C=DK,O=TDC,CN=TDC OCES Systemtest CA II"), new DistinguishedName(ocesTestCa.getOCES1RootCertificate().getSubjectX500Principal()));

		/* OCES 2 - Test*/
		assertEquals(new DistinguishedName("C=DK,O=TRUST2408,CN=TRUST2408 Systemtest VII Primary CA"), new DistinguishedName(ocesTestCa.getOCES2RootCertificate().getSubjectX500Principal()));

		/* OCES 3 - Test*/
		List<X509Certificate> oces3TestCertificates = ocesTestCa.getOCES3RootCertificates();

		List<DistinguishedName> expectedTestDNs = new ArrayList<DistinguishedName>();
		expectedTestDNs.add(new DistinguishedName("C=DK, O=Den Danske Stat, OU=Test - certifikat.local, CN=Den Danske Stat OCES rod-CA"));
		expectedTestDNs.add(new DistinguishedName("C=DK, O=Den Danske Stat, OU=Test - cti, CN=Den Danske Stat OCES rod-CA"));

		for (X509Certificate oces3RootCertificate : oces3TestCertificates) {
			DistinguishedName actualDN = new DistinguishedName(oces3RootCertificate.getSubjectX500Principal());
			assertTrue("Uventet OCES3 test rod-certifikat fundet:"+actualDN, expectedTestDNs.remove(actualDN));
		}
		assertTrue("Forventede at finde OCES3 test rod-certifikat(er):"+Arrays.toString(expectedTestDNs.toArray()), expectedTestDNs.isEmpty());
	}

	private static X509Certificate loadMoces2CertificateIG() throws IOException {
		return CredentialVaultTestUtil.loadCertificate("oces2/IG/tc01-valid-moces-cpr1.pkcs12");
	}

	private static CertificateCache getCertificateCacheForIG() throws IOException {
		CertificateCache certificateCache = new HashMapCertificateCache();
		certificateCache.putCertificate(CertificateCache.Category.IntermediateCert, "http://m.aia.systemtest10.trust2408.com/systemtest10-ca.cer", loadIntermediateCertificateIG());
		return certificateCache;
	}

	/* pp */ static X509Certificate loadInvalidMoces2CertificateIG() throws IOException {
		return CredentialVaultTestUtil.loadCertificate("oces2/IG/tc06-valid-moces-no-intermediatecert.pkcs12");
	}

	private static X509Certificate loadIntermediateCertificateIG() throws IOException {
		return CertificateParser.asCertificate(CredentialVaultTestUtil.readResource("oces2/IG/intermediateCerts/systemtest10-ca.cer"));
	}

	private static X509Certificate loadIntermediateCertificate() throws IOException {
		return CertificateParser.asCertificate(CredentialVaultTestUtil.readResource("oces3/intermediateCerts/issuing.cer"));
	}

	static X509Certificate loadMoces3Certificate() throws IOException {
		return CredentialVaultTestUtil.loadCertificate("oces3/MOCES_gyldig.p12");
	}

	/* OCES 3*/
	@Test
	public void testOCES3CertificateValidation() throws Exception {
		CertificationAuthority ca = CertificationAuthorityFactory.create(properties, CertificationAuthorityFactory.OCES_SYSTEMTEST_CA, new NaiveCertificateStatusChecker(properties), CredentialVaultTestUtil.getCertificateCacheForVoces3CredentialVault());
		X509Certificate cert = CredentialVaultTestUtil.getVoces3CredentialVault().getSystemCredentialPair().getCertificate();

		assertTrue(ca.isValid(cert));
	}
}