/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.pki;

import dk.sosi.seal.vault.CredentialVaultTestUtil;
import org.junit.Test;

import java.math.BigInteger;

import static org.junit.Assert.*;

/**
 */
public class TestCertificateInfo {

    @Test
    public void testIsProbableCertificateInfoString() {
        assertTrue(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo},IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN=foo},IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN=foo},IssuerDN={bar,CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("DN={foo},IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SUBJECTDN={foo},IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo},Issuer={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={},IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo},IssuerDN={},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo},IssuerDN={bar},CertSerial={}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={{}},IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo {} bar},IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo},IssuerDN={bar}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("IssuerDN={bar},CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString("SubjectDN={foo} IssuerDN={bar} CertSerial={1234}"));
        assertFalse(CertificateInfo.isProbableCertificateInfoString(""));
        assertFalse(CertificateInfo.isProbableCertificateInfoString(null));
    }

    @Test
    public void testConstructionFromString() {
        String certInfoString = "SubjectDN={C=DK},IssuerDN={CN=TRUST2408 Systemtest X CA,C=SE},CertSerial={133445229}";
        CertificateInfo certificateInfo = CertificateInfo.fromString(certInfoString);
        assertEquals("DK", certificateInfo.getSubjectDN().getCountry());
        assertEquals("SE", certificateInfo.getIssuerDN().getCountry());
        assertEquals(new BigInteger("133445229"), certificateInfo.getCertificateSerialNumber());
        assertEquals(certInfoString, certificateInfo.toString());
    }

    @Test(expected = IllegalArgumentException.class)
    public void testConstructionFromInvalidString() {
        CertificateInfo.fromString("something");
    }

    @Test(expected = IllegalArgumentException.class)
    public void testConstructingFromNullString() {
        CertificateInfo.fromString(null);
    }

    @Test(expected = IllegalArgumentException.class)
    public void testConstructingFromEmptyString() {
        CertificateInfo.fromString("");
    }
    
    @Test
    public void testOces3() {
        CertificateInfo info = new CertificateInfo(CredentialVaultTestUtil.getVoces3CredentialVault().getSystemCredentialPair().getCertificate());

        assertEquals("UI:DK-O:G:c70b0207-162e-4d3d-a7f1-a19a8e07d99b", info.getSubjectDN().getSubjectSerialNumber());
        assertEquals("VOCES_gyldig", info.getSubjectDN().getCommonName());
        assertEquals("Testorganisation nr. 94354969", info.getSubjectDN().getOrganization());
        assertEquals("DK", info.getSubjectDN().getCountry());
        assertNull(info.getSubjectDN().getOrganizationalUnit());

        assertNull(info.getIssuerDN().getSubjectSerialNumber());
        assertEquals("Den Danske Stat OCES udstedende-CA 1", info.getIssuerDN().getCommonName());
        assertEquals("Den Danske Stat", info.getIssuerDN().getOrganization());
        assertEquals("DK", info.getIssuerDN().getCountry());
        assertEquals("Test - cti", info.getIssuerDN().getOrganizationalUnit());

        assertEquals(new BigInteger("621110847107449959243323626634601512349711670190"), info.getCertificateSerialNumber());

        CertificateInfo deserializedInfo = CertificateInfo.fromString(info.toString());

        assertEquals(info.getSubjectDN(), deserializedInfo.getSubjectDN());
        assertEquals(info.getIssuerDN(), deserializedInfo.getIssuerDN());
        assertEquals(info.getCertificateSerialNumber(), deserializedInfo.getCertificateSerialNumber());
    }

}
