package dk.sosi.seal.pki;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.util.Collections;

import javax.security.auth.x500.X500Principal;

import org.junit.Test;

public class TestDistinguishedNameMoces3 {
	/* Tests */
	@Test(expected = NullPointerException.class)
	public void testNoDistinguisedNameString() {
		try {
			String dnStr = null;
			new DistinguishedName(dnStr);
			fail("Forventede fejl");
		} catch (Exception e) {
			throw e;
		}
	}

	@Test(expected = NullPointerException.class)
	public void testNoPrincipal() {
		try {
			X500Principal principal = null;
			new DistinguishedName(principal);
			fail("Forventede fejl");
		} catch (Exception e) {
			throw e;
		}
	}

	@Test
	public void testEmptyPrincipal() {
		try {
			X500Principal principal = new X500Principal("");
			DistinguishedName actual = new DistinguishedName(principal);
			assertEquals("Forkert CommonName", null, actual.getCommonName());
			assertEquals("Forkert Country", null, actual.getCountry());
			assertEquals("Forkert Organization", null, actual.getOrganization());
			assertEquals("Forkert OrganizationalUnit", null, actual.getOrganizationalUnit());
			assertEquals("Forkert OrganizationalUnits", null, actual.getOrganizationalUnits());
			assertEquals("Forkert OrganizationIdentifier", null, actual.getOrganizationIdentifier());
			assertEquals("Forkert SubjectSerialNumber", null, actual.getSubjectSerialNumber());
			assertEquals("Forkert GivenName", null, actual.getGivenName());
			assertEquals("Forkert SurName", null, actual.getSurName());
			assertEquals("Forkert Pseudonym", null, actual.getPseudonym());
		} catch (Exception e) {
			throw e;
		}
	}

	@Test(expected = PKIException.class)
	public void testMoces3PrincipalWithGivenNameOnly() {
		try {
			X500Principal principal = new X500Principal("C=2.5.4.6, OID.2.5.4.97=2.5.4.97, O=2.5.4.10, SERIALNUMBER=2.5.4.5, GIVENNAME=2.5.4.42, CN=2.5.4.3", Collections.singletonMap("PSEUDONYM", "2.5.4.65"));
			new DistinguishedName(principal);
			fail("Forventede fejl");
		} catch (Exception e) {
			assertEquals("Failed to parse DN: SurName SHALL be present if givenName is present.", e.getMessage());
			throw e;
		}
	}

	@Test
	public void testMoces3PrincipalWithGivenAndSurName() {
		try {
			X500Principal principal = new X500Principal("C=2.5.4.6, OID.2.5.4.97=2.5.4.97, O=2.5.4.10, SERIALNUMBER=2.5.4.5, SURNAME=2.5.4.4, GIVENNAME=2.5.4.42, CN=2.5.4.3", Collections.singletonMap("PSEUDONYM", "2.5.4.65"));
			DistinguishedName actual = new DistinguishedName(principal);
			assertEquals("Forkert CommonName", "2.5.4.3", actual.getCommonName());
			assertEquals("Forkert Country", "2.5.4.6", actual.getCountry());
			assertEquals("Forkert Organization", "2.5.4.10", actual.getOrganization());
			assertEquals("Forkert OrganizationalUnit", null, actual.getOrganizationalUnit());
			assertEquals("Forkert OrganizationalUnits", null, actual.getOrganizationalUnits());
			assertEquals("Forkert OrganizationIdentifier", "2.5.4.97", actual.getOrganizationIdentifier());
			assertEquals("Forkert SubjectSerialNumber", "2.5.4.5", actual.getSubjectSerialNumber());
			assertEquals("Forkert GivenName", "2.5.4.42", actual.getGivenName());
			assertEquals("Forkert SurName", "2.5.4.4", actual.getSurName());
			assertEquals("Forkert Pseudonym", null, actual.getPseudonym());
		} catch (Exception e) {
			throw e;
		}
	}

	@Test(expected = PKIException.class)
	public void testMoces3PrincipalWithGivenNameAndPseudonym() {
		try {
			X500Principal principal = new X500Principal("C=2.5.4.6, OID.2.5.4.97=2.5.4.97, O=2.5.4.10, SERIALNUMBER=2.5.4.5, GIVENNAME=2.5.4.42, CN=2.5.4.3, PSEUDONYM=2.5.4.65", Collections.singletonMap("PSEUDONYM", "2.5.4.65"));
			new DistinguishedName(principal);
			fail("Forventede fejl");
		} catch (Exception e) {
			assertEquals("Failed to parse DN: SurName SHALL be present if givenName is present.", e.getMessage());
			throw e;
		}
	}

	@Test(expected = PKIException.class)
	public void testMoces3PrincipalWithGivenAndSurNameAndPseudonym() {
		try {
			X500Principal principal = new X500Principal("C=2.5.4.6, OID.2.5.4.97=2.5.4.97, O=2.5.4.10, SERIALNUMBER=2.5.4.5, SURNAME=2.5.4.4, GIVENNAME=2.5.4.42, CN=2.5.4.3, PSEUDONYM=2.5.4.65", Collections.singletonMap("PSEUDONYM", "2.5.4.65"));
			new DistinguishedName(principal);
			fail("Forventede fejl");
		} catch (Exception e) {
			assertEquals("Failed to parse DN: Pseudonym MUST not be present if givenName and surname is present.", e.getMessage());
			throw e;
		}
	}

	@Test(expected = PKIException.class)
	public void testMoces3PrincipalWithSurNameOnly() {
		try {
			X500Principal principal = new X500Principal("C=2.5.4.6, OID.2.5.4.97=2.5.4.97, O=2.5.4.10, SERIALNUMBER=2.5.4.5, SURNAME=2.5.4.4, CN=2.5.4.3", Collections.singletonMap("PSEUDONYM", "2.5.4.65"));
			new DistinguishedName(principal);
			fail("Forventede fejl");
		} catch (Exception e) {
			assertEquals("Failed to parse DN: GivenName SHALL be present if surName is present.", e.getMessage());
			throw e;
		}
	}

	@Test(expected = PKIException.class)
	public void testMoces3PrincipalWithSurNameAndPseudonym() {
		try {
			X500Principal principal = new X500Principal("C=2.5.4.6, OID.2.5.4.97=2.5.4.97, O=2.5.4.10, SERIALNUMBER=2.5.4.5, SURNAME=2.5.4.4, CN=2.5.4.3, PSEUDONYM=2.5.4.65", Collections.singletonMap("PSEUDONYM", "2.5.4.65"));
			new DistinguishedName(principal);
			fail("Forventede fejl");
		} catch (Exception e) {
			assertEquals("Failed to parse DN: GivenName SHALL be present if surName is present.", e.getMessage());
			throw e;
		}
	}

	@Test
	public void testMoces3PrincipalWithPseudonym() {
		try {
			X500Principal principal = new X500Principal("C=2.5.4.6, OID.2.5.4.97=2.5.4.97, O=2.5.4.10, SERIALNUMBER=2.5.4.5, CN=2.5.4.3, PSEUDONYM=2.5.4.65", Collections.singletonMap("PSEUDONYM", "2.5.4.65"));
			DistinguishedName actual = new DistinguishedName(principal);
			assertEquals("Forkert CommonName", "2.5.4.3", actual.getCommonName());
			assertEquals("Forkert Country", "2.5.4.6", actual.getCountry());
			assertEquals("Forkert Organization", "2.5.4.10", actual.getOrganization());
			assertEquals("Forkert OrganizationalUnit", null, actual.getOrganizationalUnit());
			assertEquals("Forkert OrganizationalUnits", null, actual.getOrganizationalUnits());
			assertEquals("Forkert OrganizationIdentifier", "2.5.4.97", actual.getOrganizationIdentifier());
			assertEquals("Forkert SubjectSerialNumber", "2.5.4.5", actual.getSubjectSerialNumber());
			assertEquals("Forkert GivenName", null, actual.getGivenName());
			assertEquals("Forkert SurName", null, actual.getSurName());
			assertEquals("Forkert Pseudonym", "2.5.4.65", actual.getPseudonym());
		} catch (Exception e) {
			throw e;
		}
	}
}
