/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.seal.pki;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Properties;

import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

import dk.sosi.seal.pki.impl.intermediate.HashMapIntermediateCertificateCache;
import dk.sosi.seal.vault.renewal.KeyGenerator;

/**
 */
@SuppressWarnings("deprecation")
public class TestOCESCertificateResolver {

	private OCESCertificateResolver resolver;

	@Before
	public void setUp() throws Exception {
		final IntermediateCertificateCache cache = new HashMapIntermediateCertificateCache();
		
		resolver = new OCESCertificateResolver(cache);
	}

	@After
	public void tearDown() throws Exception {
		resolver = null;
	}

	@Test
	public void testNull() {
		try {
			resolver.getIssuingCertificate(null);
			fail("IllegalArgumentException expected");
		} catch (IllegalArgumentException e) {
			//OK
		}
	}
	
	@Test
	public void testOces3RootCertificates() throws IOException {
		List<X509Certificate> oces3TestRootCertificates = OCESTestCertificationAuthority.OCES_3_TEST_ROOT_CERTIFICATES;
		for (X509Certificate x509Certificate : oces3TestRootCertificates) {
			assertEquals(x509Certificate, resolver.getIssuingCertificate(x509Certificate));			
		}
		
		List<X509Certificate> oces3RootCertificates = OCESCertificationAuthority.OCES_3_ROOT_CERTIFICATES;
		for (X509Certificate x509Certificate : oces3RootCertificates) {
			assertEquals(x509Certificate, resolver.getIssuingCertificate(x509Certificate));	
		}
	}
	
	@Test
	public void testOces3IntermediateCertificate() throws IOException {
		X509Certificate oces3TestRootCertificates = OCESTestCertificationAuthority.OCES_3_TEST_INTERMEDIATE_CERTIFICATE;
		//OCESTestCertificationAuthority.OCES_3_TEST_ROOT_CERTIFICATES.get(0)=OCES3_TEST_ROOT_CERTIFICATE_BASE_64
		assertEquals(OCESTestCertificationAuthority.OCES_3_TEST_ROOT_CERTIFICATES.get(0), resolver.getIssuingCertificate(oces3TestRootCertificates));	
	}

	@Test
	public void testOces2RootCertificates() throws IOException {
		assertEquals(OCESTestCertificationAuthority.OCES_2_TEST_ROOT_CERTIFICATE, resolver.getIssuingCertificate(OCESTestCertificationAuthority.OCES_2_TEST_ROOT_CERTIFICATE));
		assertEquals(OCESCertificationAuthority.OCES_2_ROOT_CERTIFICATE, resolver.getIssuingCertificate(OCESCertificationAuthority.OCES_2_ROOT_CERTIFICATE));
	}

	@Test
	public void failingtestOces3Certificate() throws IOException {
		final X509Certificate mocesCertififcate = TestAbstractOCESCertificationAuthority.loadMoces3Certificate();
		final X509Certificate intermediateCertificate = resolver.getIssuingCertificate(mocesCertififcate);
		assertEquals(OCESTestCertificationAuthority.OCES_3_TEST_ROOT_CERTIFICATES.get(0), resolver.getIssuingCertificate(intermediateCertificate));
	}

	@Test
	@Ignore("SDS-6659")
	public void testOces2InvalidCertificate() throws IOException {
		//OCES2 cert without reference to intermediate certificate
		try {
			resolver.getIssuingCertificate(TestAbstractOCESCertificationAuthority.loadInvalidMoces2CertificateIG());
			fail("PKIException expected");
		} catch (PKIException e) {
			assertEquals("Invalid certificate - CA Issuers (1.3.6.1.5.5.7.48.2) not found under Authority Information Access.", e.getMessage());
		}
	}

	@Test
	public void testNonOcesCertificate() throws Exception {
		Properties properties = System.getProperties();
		PKITestCA ca = new PKITestCA();

		KeyGenerator kg = new KeyGenerator("1", properties);
		kg.setKeySize(512);
		kg.generateKeyPair();

		X509Certificate certificate = OCESTestHelper.issueCertificate("cn=TestCert,o=Test,c=DK", null, kg.getPublicKey(), ca.getRootCertificate().getPublicKey());

		try {
			resolver.getIssuingCertificate(certificate);
			fail("PKIException expected");
		} catch (PKIException e) {
			String errMsg = e.getMessage();
			assertTrue("Expected error message", errMsg.startsWith("Unable to resolve issuing certificate with DN:"));
		}
	}

	@Test
	public void testOces1TestCertificate() throws Exception { // TODO SDS-6659 Should this be removed?
		try {
			resolver.getIssuingCertificate(OCESTestCertificationAuthority.OCES_1_TEST_ROOT_CERTIFICATE);
			fail("PKIException expected");
		} catch (PKIException e) {
			assertEquals("The supplied certificate issued by: 'CN=TDC OCES Systemtest CA II,O=TDC,C=DK' is an OCES1 certificate which is no longer supported", e.getMessage());
		}
	}

	@Test
	public void testOces1ProdCertificate() throws Exception { // TODO SDS-6659 Should this be removed?
		try {
			resolver.getIssuingCertificate(OCESCertificationAuthority.OCES_1_ROOT_CERTIFICATE);
			fail("PKIException expected");
		} catch (PKIException e) {
			assertEquals("The supplied certificate issued by: 'CN=TDC OCES CA,O=TDC,C=DK' is an OCES1 certificate which is no longer supported", e.getMessage());
		}
	}

}
