package dk.sosi.seal.pki;

import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Assert;
import org.junit.Test;

import java.security.cert.X509Certificate;
import java.util.Properties;

import static dk.sosi.seal.vault.CredentialVaultTestUtil.*;

public class TestSosiFederationMatcher {
    private static final String OCES3_PROD_FEDERATION_CERT = "MIIGZzCCBJugAwIBAgIUHiP4yPL9wlSf/nVnyxC66bNmBv4wQQYJKoZIhvcNAQEK\n" +
            "MDSgDzANBglghkgBZQMEAgEFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgEF\n" +
            "AKIDAgEgMFYxLTArBgNVBAMMJERlbiBEYW5za2UgU3RhdCBPQ0VTIHVkc3RlZGVu\n" +
            "ZGUtQ0EgMTEYMBYGA1UECgwPRGVuIERhbnNrZSBTdGF0MQswCQYDVQQGEwJESzAe\n" +
            "Fw0yMzA0MTgxMDMyMDFaFw0yNjA0MTcxMDMyMDBaMIGeMR0wGwYDVQQDDBRTT1NJ\n" +
            "IEZlZGVyYXRpb24gY05TUDE3MDUGA1UEBRMuVUk6REstTzpHOmM5MzdkYTEyLWI3\n" +
            "Y2MtNGQxNS04Njg2LWQyZjQ4ZGM4MTY5YzEeMBwGA1UECgwVU3VuZGhlZHNkYXRh\n" +
            "c3R5cmVsc2VuMRcwFQYDVQRhDA5OVFJESy0zMzI1Nzg3MjELMAkGA1UEBhMCREsw\n" +
            "ggGiMA0GCSqGSIb3DQEBAQUAA4IBjwAwggGKAoIBgQCtPNYOxUqR8YPimnmIt83p\n" +
            "5ljLnqypL7y5ihRx+mwWxoQtJsFcrlB3tj8Ald0sCqFa2RYVtLkqZ1qROfRWZ4Yw\n" +
            "W5MW6TEVaesxAZBTZ+ha/KRLNOb+6Axg/zKCjAWxIa75C3aa/zESALczmu73QO3H\n" +
            "o51/Jj2z/pybb0765GCTj3FL7u21W/Efen8zJbYLBJiLJsW7eKG34243VZBkBuNd\n" +
            "h4uWcDXrju9RQJ5DDbXbdRq635anCJnHopX5uLTM+43plGC+k92T0pRg4ekybe57\n" +
            "kbVVGYrGdMlqao2byaYi2kXJ633xg/3NrHEfKAIqJmDjY+iHVresgPbIidBNHNMA\n" +
            "6CpsdryLPRbuT08l8P7C5mo6zLCES6qEPGSm929HRIt7Le3BaZ+TvZPGLrPdamPZ\n" +
            "12///jU0ewD9BVG0pRkIKRMw677oq6GNNIH+n4TRj4HUurKCQjpAo/+aZH3NrxSZ\n" +
            "01tzEnIb3K3fn8NK1GMNbEQlm5I+9Dxvy4UjR0GGa4cCAwEAAaOCAXowggF2MAwG\n" +
            "A1UdEwEB/wQCMAAwHwYDVR0jBBgwFoAUTAHiynO8w744Cjg9NrBcdJx7l7kwcwYI\n" +
            "KwYBBQUHAQEEZzBlMD8GCCsGAQUFBzAChjNodHRwOi8vY2ExLmdvdi5kay9vY2Vz\n" +
            "L2lzc3VpbmcvMS9jYWNlcnQvaXNzdWluZy5jZXIwIgYIKwYBBQUHMAGGFmh0dHA6\n" +
            "Ly9jYTEuZ292LmRrL29jc3AwIQYDVR0gBBowGDAIBgYEAI96AQEwDAYKKoFQgSkB\n" +
            "AQEDBzA7BggrBgEFBQcBAwQvMC0wKwYIKwYBBQUHCwIwHwYHBACL7EkBAjAUhhJo\n" +
            "dHRwczovL3VpZC5nb3YuZGswQQYDVR0fBDowODA2oDSgMoYwaHR0cDovL2NhMS5n\n" +
            "b3YuZGsvb2Nlcy9pc3N1aW5nLzEvY3JsL2lzc3VpbmcuY3JsMB0GA1UdDgQWBBTT\n" +
            "pbTHGeSa3eYeRIhuQG+6BMYnDDAOBgNVHQ8BAf8EBAMCBaAwQQYJKoZIhvcNAQEK\n" +
            "MDSgDzANBglghkgBZQMEAgEFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgEF\n" +
            "AKIDAgEgA4IBgQB5PWaO6Jt9cUkUahE33elW1as8tbG3iUfFQVtV6bu0wsJRLf1y\n" +
            "oqmf77iegixK+vt7+3KPsIOQZ/Tn361mvsS0V9Bn2utmnEU7bFen5Cdr1lbAIVy+\n" +
            "/aSmaHdxp3DGEFfGwf1nWB1kFX+G6XDjpG0cPgeZRx1/bSA5dOPH9hR5XMDSj/Bf\n" +
            "18HUZLEFbf4qis7jYs3gpdyGN+iKWiecWDdeMoDgF51xNXpxEGS4Cz1GhobM0cFn\n" +
            "3hhFcMsADcgCb2JmT8AgzE6AYoOfz/yHfvBXVurfA3QOgI8fKWuqnS4YNqGiih/3\n" +
            "W0BLFSWXgByWa/PkR7vY4CHWk60y83Az7+s19HzZAiE613jzi2guatwqu7WrG50P\n" +
            "lDguj5YNzCF4bJVqeHNnQFYvFVfevviHw3NyYSgCL6GxISrSc7IX3HK90JL81n/e\n" +
            "cfp4vld1KG+JLxeNeDQO8itwQh2HNttLklR8yQT0mAjXCO3rm4CHV2tIUS8kwh3n\n" +
            "VnRcp8WB4T4+6tU=";

    private static final String OCES3_TEST_WRONG_COMMON_NAME_CERT = "MIIGljCCBMqgAwIBAgIUFjMz0kZutbxV4vQFIjBRWd4GA2EwQQYJKoZIhvcNAQEK\n" +
            "MDSgDzANBglghkgBZQMEAgEFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgEF\n" +
            "AKIDAgEgMGsxLTArBgNVBAMMJERlbiBEYW5za2UgU3RhdCBPQ0VTIHVkc3RlZGVu\n" +
            "ZGUtQ0EgMTETMBEGA1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNr\n" +
            "ZSBTdGF0MQswCQYDVQQGEwJESzAeFw0yMzAyMjEwODUyMjJaFw0yNjAyMjAwODUy\n" +
            "MjFaMIGsMSswKQYDVQQDDCJTZWFsLkphdmEgVGVzdCBmdW5rdGlvbnNjZXJ0aWZp\n" +
            "a2F0MTcwNQYDVQQFEy5VSTpESy1POkc6OWZlMTg3NGItODM5NS00NWE4LTk3OGUt\n" +
            "YWYzNjM3MTViMzdlMR4wHAYDVQQKDBVTdW5kaGVkc2RhdGFzdHlyZWxzZW4xFzAV\n" +
            "BgNVBGEMDk5UUkRLLTMzMjU3ODcyMQswCQYDVQQGEwJESzCCAaIwDQYJKoZIhvcN\n" +
            "AQEBBQADggGPADCCAYoCggGBAL6t9gPPr2fGDegaMpUNA1uitq7eQ6jZ67jj9cdq\n" +
            "fmRescBSpxDMrDHShxHZ/BmxEsoSVN47DPkl2BmQdiq7O0iijJHbOpurz2qoBskJ\n" +
            "pvL7UCkWmykZJtBvnqkOZEgNKivuEOS7fkFa9lDPWRh+wt8s8es7V0nOo2C2uMoP\n" +
            "VYMJOuEZc8HVeFmHw2zYuAlwLfBVzv6qnk3VVfg9vLezpCOHl/pi9MWfDoXIePeR\n" +
            "ohOVIOQ7vbQTSMsHTMSyKN9DAWjaLe03vhowT3R3m/v1lQ18qtOXZk1yD3XQs9YV\n" +
            "GyQY2OHyUJWfljvY/s/zpscxJXMAyqiRwIWmid95aFhyZY+U4YTpKrhMGd6GAn5s\n" +
            "4QFvadvCjiQ5Crkl8A39OLsl+358l47T9xJKmF1V+QADkkxfqVERu8ffMJCWFjOq\n" +
            "nbgpA7EiIc1cmzwYJlR5vTbQO/A6ePCJa7mQmpF1ryFfiORS/pVm+Dfv0/U7pmPu\n" +
            "cJvVVCdVldV7kBjX4slCXjEmkQIDAQABo4IBhjCCAYIwDAYDVR0TAQH/BAIwADAf\n" +
            "BgNVHSMEGDAWgBR/KJ/ZcZlC4nXn1zV2Lk0IJW12XjB7BggrBgEFBQcBAQRvMG0w\n" +
            "QwYIKwYBBQUHMAKGN2h0dHA6Ly9jYTEuY3RpLWdvdi5kay9vY2VzL2lzc3Vpbmcv\n" +
            "MS9jYWNlcnQvaXNzdWluZy5jZXIwJgYIKwYBBQUHMAGGGmh0dHA6Ly9jYTEuY3Rp\n" +
            "LWdvdi5kay9vY3NwMCEGA1UdIAQaMBgwCAYGBACPegEBMAwGCiqBUIEpAQEBAwcw\n" +
            "OwYIKwYBBQUHAQMELzAtMCsGCCsGAQUFBwsCMB8GBwQAi+xJAQIwFIYSaHR0cHM6\n" +
            "Ly91aWQuZ292LmRrMEUGA1UdHwQ+MDwwOqA4oDaGNGh0dHA6Ly9jYTEuY3RpLWdv\n" +
            "di5kay9vY2VzL2lzc3VpbmcvMS9jcmwvaXNzdWluZy5jcmwwHQYDVR0OBBYEFEDv\n" +
            "1OtrQ92fs2gX0aijSJ8b3NgPMA4GA1UdDwEB/wQEAwIFoDBBBgkqhkiG9w0BAQow\n" +
            "NKAPMA0GCWCGSAFlAwQCAQUAoRwwGgYJKoZIhvcNAQEIMA0GCWCGSAFlAwQCAQUA\n" +
            "ogMCASADggGBAG/q0zoJsii+Nx6jIpRksTTX5zaanx55hg6jSwpv1RoTNr52lQca\n" +
            "pzE1vuh2/HDh2mvc0/jywucm5MNk+HvA96CqzoYlfjUvq/jxyxT5/L7i//pV9r1q\n" +
            "qPFG/VX0u9WWcCqP9qUp2or3kMLAKzZCj1XGljQwqJ65WYTRE0pziIS7y//Xk4su\n" +
            "qyfU0EA/RRVTrl4fGB/AkRv7MAgCeNA06NMdkx8ilM0QT4PvxDuMDaprcncB3o40\n" +
            "gQMoDVPUHQVS1sTTn46/8qpAuN3qYJIpxQoDbZfbo4XLx1P4QHsVKOW6oDB4++V7\n" +
            "/fuPSXDUVl94aoLTtkca1TiYCTtyCxWX7FnDd5KfbpE8uDyVFvvsgM4LVg/dIU0I\n" +
            "/oJvMu1O4QJjMag0CUKuxMSrrWKCrkdi7NHzOXafQHMkD2/8VNVN6fuc/wlE+3HM\n" +
            "pfkN2pRNpD5SHV9LOok9QP/54oRAg61fYkCN4+a3glelriqQ1ieTvp5e6ysZL3L9\n" +
            "Azps2B2jZK4f+Q==";

    private static final String OCES3_TEST_WRONG_CVR_CERT = "MIIGvDCCBPCgAwIBAgIUEQ9OVGJs9BBGSahwCjva2LbBThYwQQYJKoZIhvcNAQEK\n" +
            "MDSgDzANBglghkgBZQMEAgEFAKEcMBoGCSqGSIb3DQEBCDANBglghkgBZQMEAgEF\n" +
            "AKIDAgEgMGsxLTArBgNVBAMMJERlbiBEYW5za2UgU3RhdCBPQ0VTIHVkc3RlZGVu\n" +
            "ZGUtQ0EgMTETMBEGA1UECwwKVGVzdCAtIGN0aTEYMBYGA1UECgwPRGVuIERhbnNr\n" +
            "ZSBTdGF0MQswCQYDVQQGEwJESzAeFw0yMzAyMjAxMjMwNDVaFw0yNjAyMTkxMjMw\n" +
            "NDRaMIGwMScwJQYDVQQDDB5TT1NJIFRlc3QgRmVkZXJhdGlvbiB3cm9uZyBjdnIx\n" +
            "NzA1BgNVBAUTLlVJOkRLLU86Rzo4ZTRmNzllNC1kYjkzLTRhM2ItODI4Ny02NmIw\n" +
            "ZGVhMWFiZmExJjAkBgNVBAoMHVRlc3RvcmdhbmlzYXRpb24gbnIuIDk4MDIxODM4\n" +
            "MRcwFQYDVQRhDA5OVFJESy05ODAyMTgzODELMAkGA1UEBhMCREswggGiMA0GCSqG\n" +
            "SIb3DQEBAQUAA4IBjwAwggGKAoIBgQDniNmsBaWCaHOmvJ1ZHnFp2qb1ah7YBWPy\n" +
            "0nMBoGbHSAsYGP4V40z9IBOCHCFye9DvXwscnIaERPASnr2FNQBtOnA6jJk+lJd6\n" +
            "tsJgt+qk9miMRXBKfjWKY1Sdnf/tyJMDRBIkJJDNbH3k3eLvvnDvfhoh4LSczOMh\n" +
            "JLnNyviyHn6Wd+BHCvxemqlnkT0Z4FNQvcxFeAEiIaczsscWlAVKEydlrFnB/XcP\n" +
            "cR8TRRDdVvLF8qHFJURvSGxlE2Aa/5/SeNI33ATEqqOrYP2BqPTL/FkdO6+MgRgt\n" +
            "QSecXMt9dk3jJTqK/t0ffezmUzr5DIBKM0iX+0JeNFWNIGhginVDfDHCGhgQ2EPN\n" +
            "4eSbP8c7wPHRuZD9w0gPWpYEJWrPAviFDlEpJ0zWf/aQw48VdIZVgT3gB0uYzcda\n" +
            "6chGQE2PYsRZDKsTegcdvj3twmgABcbfnYrLLv9uJAyJizSiiv9dh8DLvJSIr2XD\n" +
            "CM9SGrNX1WeGnf9uMnMsKEMCCLNAKD0CAwEAAaOCAagwggGkMAwGA1UdEwEB/wQC\n" +
            "MAAwHwYDVR0jBBgwFoAUfyif2XGZQuJ159c1di5NCCVtdl4wewYIKwYBBQUHAQEE\n" +
            "bzBtMEMGCCsGAQUFBzAChjdodHRwOi8vY2ExLmN0aS1nb3YuZGsvb2Nlcy9pc3N1\n" +
            "aW5nLzEvY2FjZXJ0L2lzc3VpbmcuY2VyMCYGCCsGAQUFBzABhhpodHRwOi8vY2Ex\n" +
            "LmN0aS1nb3YuZGsvb2NzcDAgBgNVHREEGTAXgRVuc3Auc3VwcG9ydEBhcm9zaWku\n" +
            "ZGswIQYDVR0gBBowGDAIBgYEAI96AQEwDAYKKoFQgSkBAQEDBzA7BggrBgEFBQcB\n" +
            "AwQvMC0wKwYIKwYBBQUHCwIwHwYHBACL7EkBAjAUhhJodHRwczovL3VpZC5nb3Yu\n" +
            "ZGswRQYDVR0fBD4wPDA6oDigNoY0aHR0cDovL2NhMS5jdGktZ292LmRrL29jZXMv\n" +
            "aXNzdWluZy8xL2NybC9pc3N1aW5nLmNybDAdBgNVHQ4EFgQU9cc/TAX316X6k6q+\n" +
            "1RpI2NQGz14wDgYDVR0PAQH/BAQDAgXgMEEGCSqGSIb3DQEBCjA0oA8wDQYJYIZI\n" +
            "AWUDBAIBBQChHDAaBgkqhkiG9w0BAQgwDQYJYIZIAWUDBAIBBQCiAwIBIAOCAYEA\n" +
            "eQfwTmPBBvctdE65dRt2tdUcZoTPDKo6/TXs45FKyf4UmPLXSSlh6L5dll6KcvQo\n" +
            "FH0blFSOon/AtqihI5XZn9WUHocvydfVhmxJVr10ZAW9X5XHEktDLsHnHb8ahUW+\n" +
            "58z0Lysg1W00IkEwlVLQ67Wnz4p0wlHfAvXv8/n+zRZTLNZWjCdjf0H7wbRwBkK6\n" +
            "D3Mhtc6686KghbkNSQWKXF/z8HJ9BUtzvIqEGBlU5UkdaTxM8BHybuzf46ZPXI0M\n" +
            "beQWEUPp6cBDJ7z1GK7z/5uAxblMnCZU4MUPVzLzMb78br8AWWQ24WRQuVXjW0TL\n" +
            "gq5N3pz/ci0kkmbmRFAwG5RI2bY1YfvXQzGbeX/zG8IGXN/BvD+vRoGL2lpQHH04\n" +
            "nYEmaGvxKz3BlEbkFX1N5kZ6dbD1n22Jnts0r70H9G28cK5ESBsebhb3a8giygw5\n" +
            "WyQwltr21Q2EJ8TFNPtv1p7vvXLpfVcSl4PN4wpDtoGkV7oQYOxbwH0pRG1D2cUF";


    @Test
    public void testOces3TestSOSITestFederationValid() {
        // U5 OCES3 SOSI Test føderationscertifikat
        X509Certificate certificate = CertificateParser.asCertificate(XmlUtil.fromBase64(OCES3_TEST_FEDERATION_CERT));

        SOSITestFederation sosiTestFederation = new SOSITestFederation(new Properties());
        Assert.assertTrue(sosiTestFederation.isValidSTSCertificate(certificate));
    }

    @Test
    public void testOces3TestSOSIFederationInvalid() {
        // U6 OCES3 SOSI Test føderationscertifikat
        X509Certificate certificate = CertificateParser.asCertificate(XmlUtil.fromBase64(OCES3_TEST_FEDERATION_CERT));

        SOSIFederation sosiFederation = new SOSIFederation(new Properties());
        Assert.assertFalse(sosiFederation.isValidSTSCertificate(certificate));
    }

    @Test
    public void testOces3ProdSOSITestFederationInvalid() {
        // U7 OCES3 SOSI Prodføderationscertifikat
        X509Certificate certificate = CertificateParser.asCertificate(XmlUtil.fromBase64(OCES3_PROD_FEDERATION_CERT));

        SOSITestFederation sosiTestFederation = new SOSITestFederation(new Properties());
        Assert.assertFalse(sosiTestFederation.isValidSTSCertificate(certificate));
    }

    @Test
    public void testOces3ProdSOSIFederationValid() {
        // U8 OCES3 SOSI Prodføderationscertifikat
        X509Certificate certificate = CertificateParser.asCertificate(XmlUtil.fromBase64(OCES3_PROD_FEDERATION_CERT));

        SOSIFederation sosiFederation = new SOSIFederation(new Properties());
        Assert.assertTrue(sosiFederation.isValidSTSCertificate(certificate));
    }

    @Test
    public void testOces3TestWrongCommonNameInvalid() {
        // U9 OCES3 Test certifikatmed et andet præfiks end "SOSI Test Federation"
        X509Certificate certificate = CertificateParser.asCertificate(XmlUtil.fromBase64(OCES3_TEST_WRONG_COMMON_NAME_CERT));

        SOSITestFederation sosiTestFederation = new SOSITestFederation(new Properties());
        Assert.assertFalse(sosiTestFederation.isValidSTSCertificate(certificate));
    }

    @Test
    public void testOces3TestWrongCVRInvalid() {
        // U10 OCES3 Test certifikat med Common Name som starter med"SOSI Test Federation", men som er udstedt af et andet CVR-nummer
        X509Certificate certificate = CertificateParser.asCertificate(XmlUtil.fromBase64(OCES3_TEST_WRONG_CVR_CERT));

        SOSITestFederation sosiTestFederation = new SOSITestFederation(new Properties());
        Assert.assertFalse(sosiTestFederation.isValidSTSCertificate(certificate));
    }
}