package dk.sosi.seal.sign;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.axis.utils.ByteArrayOutputStream;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicHeader;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.util.Properties;
import java.util.Scanner;
import java.util.UUID;

import static dk.sosi.seal.vault.GenericCredentialVault.KeyStoreType;
import static org.junit.Assert.assertNull;

public class SealSignTest {

    private static String                   audience = "https://minlog";
    private static String                   cpr      = "3108732076";
    private static String                   pid      = "PID:9208-2002-2-098076399641";
    
    
    private ClasspathCredentialVault credentialVault;
    private OIOIDWSFactory factory;
    private IdentityToken identityToken;
    private DocumentBuilderFactory documentBuilderFactory;

    @BeforeClass
    public static void setupLog4J() {
        BasicConfigurator.configure();
    }

    @Before
    public void setup() throws IOException {
        credentialVault = new ClasspathCredentialVault(new Properties(),
                                                       "oces3/FOCES_gyldig.p12",
                                                       "Test1234",
                                                       KeyStoreType.Pkcs12);
        SOSITestFederation federation = new SOSITestFederation(SignatureUtil.setupCryptoProviderForJVM());
        factory = new OIOIDWSFactory(federation, credentialVault);
        BootstrapTokenHelper bootstrapTokenHelper = new BootstrapTokenHelper();

        // Build a selfsigned bootstrap token simulating a call to an actual IdP (like nemlogin or others)
        OIOSAMLAssertion bootstrapOioSamlAssertion = bootstrapTokenHelper.buildBootstrapAssertion(pid);
        identityToken = getIdentityToken(bootstrapOioSamlAssertion, audience, pid, cpr, null);
        
        documentBuilderFactory = DocumentBuilderFactory.newInstance();
        documentBuilderFactory.setNamespaceAware(true);
    }

    
    @Test
    @Ignore("SDS-6659")
    public void signAndVerifyDocumentWithRebuilding() throws SAXException, IOException, ParserConfigurationException, TransformerException {
        signAndVerify(true);
    }
    
    @Test
    @Ignore("SDS-6659")
    public void signAndVerifyDocumentWithoutRebuilding() throws SAXException, IOException, ParserConfigurationException, TransformerException {
        signAndVerify(false);
    }


    private void signAndVerify(boolean streamAndRebuildBeforeSigning) throws SAXException, IOException, ParserConfigurationException, TransformerException {
        Document document = getSOAP11Document();        

        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        Transformer transformer = transformerFactory.newTransformer();

        if (streamAndRebuildBeforeSigning) {
            // Validation fails if this transformation is not performed - https://jira.nspop.dk/browse/SDS-3970
            ByteArrayOutputStream intermediateDocumentBytes = new ByteArrayOutputStream();
            transformer.transform(new DOMSource(document), new StreamResult(intermediateDocumentBytes));
            document = documentBuilderFactory.newDocumentBuilder().parse(new ByteArrayInputStream(intermediateDocumentBytes.toByteArray()));
        }

        // Add the OioIdws IdentityToken header and WS Adressing headers to the SOAP
        // header and sign the entire message
        LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(document);
        enhancer.setIdentityToken(identityToken);
        enhancer.setWSAddressingMessageID(UUID.randomUUID().toString());
        enhancer.setWSAddressingAction("http://nspop.dk/2020/03/19/sealsign/test#RequestServiceCall");
        enhancer.setWSAddressingRelatesTo("not related");
        enhancer.enhanceAndSign();

        // Serialize to a byte array and wrap in an InputStream
        transformer = transformerFactory.newTransformer();
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        transformer.transform(new DOMSource(document), new StreamResult(os));
        InputStream overtheWireStream = new ByteArrayInputStream(os.toByteArray());
        Logger.getLogger("WireBytes").info(getAsString(new ByteArrayInputStream(os.toByteArray())));

        DocumentBuilderFactory receivingDocumentBuilderFactory = DocumentBuilderFactory.newInstance();
        receivingDocumentBuilderFactory.setNamespaceAware(true);
        Document responseDocument = receivingDocumentBuilderFactory.newDocumentBuilder().parse(overtheWireStream);

        Exception parseException = null;
        try {
            LibertyResponse libertyResponse = OIOIDWSFactory.createResponse(responseDocument);
            libertyResponse.validateSignatureAndTrust(credentialVault);
        } catch (ModelException e) {
            parseException = e;
        }
        assertNull(parseException);
    }
    
    private Document getSOAP11Document() throws ParserConfigurationException {
        
        Document document = documentBuilderFactory.newDocumentBuilder().newDocument();
        Element soapEnvelope = document.createElementNS("http://schemas.xmlsoap.org/soap/envelope/", "soapenv:Envelope");
        Element soapBody = document.createElementNS("http://schemas.xmlsoap.org/soap/envelope/", "soapenv:Body");
        Element soapContent = document.createElementNS("http://nspop.dk/2020/03/19/sealsign/test", "ah:RequestServiceCall");
        document.appendChild(soapEnvelope);
        soapEnvelope.appendChild(soapBody);
        soapBody.appendChild(soapContent);
        return document;
    }

    private IdentityToken getIdentityToken(OIOSAMLAssertion bootstrapTokenAssertion, String audience, String pid, String cpr, String procurationCpr) throws IOException {
        OIOBootstrapToken bootstrapToken = new OIOBootstrapToken(bootstrapTokenAssertion.getDOM().getDocumentElement());
        OIOBootstrapToIdentityTokenRequestDOMBuilder oioBootstrapToIdentityTokenRequestDOMBuilder = new OIOBootstrapToIdentityTokenRequestDOMBuilder();
        oioBootstrapToIdentityTokenRequestDOMBuilder.setSigningVault(credentialVault);
        oioBootstrapToIdentityTokenRequestDOMBuilder.setAudience(audience);
        oioBootstrapToIdentityTokenRequestDOMBuilder.setCPRNumberClaim(cpr);
        oioBootstrapToIdentityTokenRequestDOMBuilder.setOIOBootstrapToken(bootstrapToken);
        Document bootstrapRequest = oioBootstrapToIdentityTokenRequestDOMBuilder.build();
        Document stsBootstrapResponse = XmlUtil.readXml(new Properties(), sendRequest("http://test2.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws", "Bst2Idws", XmlUtil.node2String(bootstrapRequest, false, true)), false);
        OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
        return oiosamlFactory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(stsBootstrapResponse).getIdentityToken();
    }
   
    public static String sendRequest(String url, String soapAction, String postBody) throws IOException {
        HttpClientBuilder b = HttpClientBuilder.create();
        try (CloseableHttpClient httpClient = b.build()) {
            HttpPost httpPost = new HttpPost(url);
            httpPost.addHeader(new BasicHeader("Content-Type", "text/xml; charset=utf-8"));
            httpPost.addHeader(new BasicHeader("SOAPAction", "\"" + soapAction + "\""));
            httpPost.addHeader(new BasicHeader("X-NSI-GW-PARTITION", "test"));
            httpPost.addHeader(new BasicHeader("SOSIGW_PARTITION", "test"));
            httpPost.setEntity(new StringEntity(postBody, "utf-8"));
            HttpResponse response = httpClient.execute(httpPost);
            String result = "";
            int statusCode = response.getStatusLine().getStatusCode();
            if (statusCode == 200 || statusCode == 500) {
                InputStream is = null;
                try {
                    is = response.getEntity().getContent();
                    result = getAsString(is);
                } finally {
                    if (is != null) {
                        is.close();
                    }
                }
            } else if (statusCode == 202) {
                return ""; // Accepted without any response
            } else {
                result = response.getStatusLine().getReasonPhrase();
                throw new IOException("HTTP POST failed (" + statusCode + "): " + result);
            }
            return result;
        }
    }

    public static String getAsString(InputStream inputStream) throws UnsupportedEncodingException, IOException {
        try (Scanner scanner = new Scanner(inputStream, StandardCharsets.UTF_8.name())) {
            return scanner.useDelimiter("JustTheWholeStringAsASingleTokenPlease").next();
        }
    }
}