/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.seal.util;

import dk.sosi.seal.pki.testobjects.CredentialVaultAdapter;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.CertificateParser;
import dk.sosi.seal.xml.XmlUtil;

import java.io.InputStream;
import java.security.cert.X509Certificate;
import java.util.Scanner;

/**
 */
public class SOSITestUtils {

    // SSR: CVR:34051178-UID:1336055110451
    /**
     * OCES1
     * 
     * Common Name: Digitaliseringsstyrelsen - Digst - NemLog-in Tes
     * Subject Alternative Names: email:nemlogin@digst.dk
     * Organization: Digitaliseringsstyrelsen // CVR:34051178
     * Country: DK
     * Valid From: May 7, 2012
     * Valid To: May 7, 2014
     * Issuer: TDC OCES Systemtest CA II, TDC
     * Serial Number: 1077419680 (0x40381ea0)
     */
    private final static String OLD_NEMLOGIN_IDP_CERTIFICATE_STRING =
            "MIIFrTCCBRagAwIBAgIEQDgeoDANBgkqhkiG9w0BAQUFADA/MQswCQYDVQQGEwJESzEMMAoGA1UE" +
                    "ChMDVERDMSIwIAYDVQQDExlUREMgT0NFUyBTeXN0ZW10ZXN0IENBIElJMB4XDTEyMDUwNzEzMDIz" +
                    "N1oXDTE0MDUwNzEzMzIzN1owgaIxCzAJBgNVBAYTAkRLMTEwLwYDVQQKEyhEaWdpdGFsaXNlcmlu" +
                    "Z3NzdHlyZWxzZW4gLy8gQ1ZSOjM0MDUxMTc4MWAwJQYDVQQFEx5DVlI6MzQwNTExNzgtVUlEOjEz" +
                    "MzYwNTUxMTA0NTEwNwYDVQQDEzBEaWdpdGFsaXNlcmluZ3NzdHlyZWxzZW4gLSBEaWdzdCAtIE5l" +
                    "bUxvZy1pbiBUZXMwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDnV90kgwkzKmuCtfAb" +
                    "2OORg6DufFoLekR1jDfRyf8qYIkAGEELYiNuatR8o29/gEWibNy/W9i0cMNH44e9t6cPf6GXqqQl" +
                    "FxbV6SWuX4LPMsq7kJiWw+kkV3Ac66jsKVEJduSxHRKfNiFq3Fg4AAFn7TsvQSAapHZXUwMpTNXf" +
                    "qq6BkO9RvJjkMlj3TpDGIgiRMWyQmH84RsL/SjgiLDbkesCToWr8GylQVCascIfihwxxDNM3QJmz" +
                    "t7oZuYX1+uZkQayq57vZxqGldvnhTuAjCNt7nN76Doy584D7kzPU2vrad0sNUrcvPzzHQBj2aKgw" +
                    "uhColBOtLOjcELtKn0C7AgMBAAGjggLMMIICyDAOBgNVHQ8BAf8EBAMCA7gwKwYDVR0QBCQwIoAP" +
                    "MjAxMjA1MDcxMzAyMzdagQ8yMDE0MDUwNzEzMzIzN1owRgYIKwYBBQUHAQEEOjA4MDYGCCsGAQUF" +
                    "BzABhipodHRwOi8vdGVzdC5vY3NwLmNlcnRpZmlrYXQuZGsvb2NzcC9zdGF0dXMwggEDBgNVHSAE" +
                    "gfswgfgwgfUGCSkBAQEBAQEBAzCB5zAvBggrBgEFBQcCARYjaHR0cDovL3d3dy5jZXJ0aWZpa2F0" +
                    "LmRrL3JlcG9zaXRvcnkwgbMGCCsGAQUFBwICMIGmMAoWA1REQzADAgEBGoGXVERDIFRlc3QgQ2Vy" +
                    "dGlmaWthdGVyIGZyYSBkZW5uZSBDQSB1ZHN0ZWRlcyB1bmRlciBPSUQgMS4xLjEuMS4xLjEuMS4x" +
                    "LjEuMy4gVERDIFRlc3QgQ2VydGlmaWNhdGVzIGZyb20gdGhpcyBDQSBhcmUgaXNzdWVkIHVuZGVy" +
                    "IE9JRCAxLjEuMS4xLjEuMS4xLjEuMS4zLjAcBglghkgBhvhCAQ0EDxYNb3JnYW5XZWJOb0RpcjAc" +
                    "BgNVHREEFTATgRFuZW1sb2dpbkBkaWdzdC5kazCBlwYDVR0fBIGPMIGMMFegVaBTpFEwTzELMAkG" +
                    "A1UEBhMCREsxDDAKBgNVBAoTA1REQzEiMCAGA1UEAxMZVERDIE9DRVMgU3lzdGVtdGVzdCBDQSBJ" +
                    "STEOMAwGA1UEAxMFQ1JMMzIwMaAvoC2GK2h0dHA6Ly90ZXN0LmNybC5vY2VzLmNlcnRpZmlrYXQu" +
                    "ZGsvb2Nlcy5jcmwwHwYDVR0jBBgwFoAUHJgJRxpMOLkQxQQpW/H0ToBqzH4wHQYDVR0OBBYEFKhW" +
                    "lVALv2E2JYC8SQPu2jvRXVjqMAkGA1UdEwQCMAAwGQYJKoZIhvZ9B0EABAwwChsEVjcuMQMCA6gw" +
                    "DQYJKoZIhvcNAQEFBQADgYEAQlw2hwwrFX9rehfxsMDkiOJGLJcVmSpK9bA/qDUCr8EU4aBfBE+o" +
                    "IHaAQxmrG9Rsvyhf1rUE+Cir2GQsS5gcG9zIhwEwtCitQutM7dOJ1eIbApmHF8bHike8eouerTcr" +
                    "oqM6EY8TbfEkDEtDuQxk+GqbCh5gpKg243ZLZiG3Aes=";

    //SERIALNUMBER=CVR:34051178-UID:83384970 + CN=Digitaliseringsstyrelsen - NemLog-in Test, O=Digitaliseringsstyrelsen // CVR:34051178, C=DK
    /**
     * OCES2
     * 
     * Common Name: Digitaliseringsstyrelsen - NemLog-in Test
     * Subject Alternative Names: email:nemlogin@digst.dk
     * Organization: Digitaliseringsstyrelsen // CVR:34051178
     * Country: DK
     * Valid From: May 5, 2014
     * Valid To: May 5, 2017
     * Issuer: TRUST2408 Systemtest XIX CA, TRUST2408
     * Serial Number: 1393492915 (0x530f03b3)
     */
    private final static String NEW_NEMLOGIN_IDP_CERTIFICATE_STRING =
            "MIIGRTCCBS2gAwIBAgIEUw8DszANBgkqhkiG9w0BAQsFADBHMQswCQYDVQQGEwJESzESMBAGA1UE" +
                    "CgwJVFJVU1QyNDA4MSQwIgYDVQQDDBtUUlVTVDI0MDggU3lzdGVtdGVzdCBYSVggQ0EwHhcNMTQw" +
                    "NTA1MTMzNTU4WhcNMTcwNTA1MTMzNTEyWjCBljELMAkGA1UEBhMCREsxMTAvBgNVBAoMKERpZ2l0" +
                    "YWxpc2VyaW5nc3N0eXJlbHNlbiAvLyBDVlI6MzQwNTExNzgxVDAgBgNVBAUTGUNWUjozNDA1MTE3" +
                    "OC1VSUQ6ODMzODQ5NzAwMAYDVQQDDClEaWdpdGFsaXNlcmluZ3NzdHlyZWxzZW4gLSBOZW1Mb2ct" +
                    "aW4gVGVzdDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALnCmDRMztjDckSupQBLcEzr" +
                    "RRJnAFxzEFdB7Cj6ApMQ/YxqKzfL/TSIr3v2mdgQNnsJGz91YbAteDPRHR/K1W3kqoIX/qH2uXDz" +
                    "HK+qi4YD9D8s4MnHAt02x6t0TgKQGjn1XO6lgLQ563DjtgD2fdPm9USV2Lkxe5ofNRG7yvWowBWj" +
                    "XKia8D64k6zSzoHKdPz6GCy9S0NmwIyJE0sJavcfwxT3/ia0g63/xD77SteT4H/OR/DLis7FLnfk" +
                    "Lp8yrd5xAk4nEGizmjrg2OVJmIMMPK6PQdw+/lqSdgaPDxMD6yoIwWshux5Rup1+piMLg852odHR" +
                    "6EhUzjEsi9DnWWcCAwEAAaOCAucwggLjMA4GA1UdDwEB/wQEAwIEsDCBlwYIKwYBBQUHAQEEgYow" +
                    "gYcwPAYIKwYBBQUHMAGGMGh0dHA6Ly9vY3NwLnN5c3RlbXRlc3QxOS50cnVzdDI0MDguY29tL3Jl" +
                    "c3BvbmRlcjBHBggrBgEFBQcwAoY7aHR0cDovL3YuYWlhLnN5c3RlbXRlc3QxOS50cnVzdDI0MDgu" +
                    "Y29tL3N5c3RlbXRlc3QxOS1jYS5jZXIwggEgBgNVHSAEggEXMIIBEzCCAQ8GDSsGAQQBgfRRAgQG" +
                    "AwQwgf0wLwYIKwYBBQUHAgEWI2h0dHA6Ly93d3cudHJ1c3QyNDA4LmNvbS9yZXBvc2l0b3J5MIHJ" +
                    "BggrBgEFBQcCAjCBvDAMFgVEYW5JRDADAgEBGoGrRGFuSUQgdGVzdCBjZXJ0aWZpa2F0ZXIgZnJh" +
                    "IGRlbm5lIENBIHVkc3RlZGVzIHVuZGVyIE9JRCAxLjMuNi4xLjQuMS4zMTMxMy4yLjQuNi4zLjQu" +
                    "IERhbklEIHRlc3QgY2VydGlmaWNhdGVzIGZyb20gdGhpcyBDQSBhcmUgaXNzdWVkIHVuZGVyIE9J" +
                    "RCAxLjMuNi4xLjQuMS4zMTMxMy4yLjQuNi4zLjQuMBwGA1UdEQQVMBOBEW5lbWxvZ2luQGRpZ3N0" +
                    "LmRrMIGpBgNVHR8EgaEwgZ4wPKA6oDiGNmh0dHA6Ly9jcmwuc3lzdGVtdGVzdDE5LnRydXN0MjQw" +
                    "OC5jb20vc3lzdGVtdGVzdDE5LmNybDBeoFygWqRYMFYxCzAJBgNVBAYTAkRLMRIwEAYDVQQKDAlU" +
                    "UlVTVDI0MDgxJDAiBgNVBAMMG1RSVVNUMjQwOCBTeXN0ZW10ZXN0IFhJWCBDQTENMAsGA1UEAwwE" +
                    "Q1JMMjAfBgNVHSMEGDAWgBTMAlUM5IF0ryBU1REUV5yRUjh/oDAdBgNVHQ4EFgQUwm9c3oUHE/zZ" +
                    "/43g4RUhswnMVAowCQYDVR0TBAIwADANBgkqhkiG9w0BAQsFAAOCAQEACLh3Ovvljv4b/Ywf/8Wx" +
                    "oB2y50Oqt8rpwXZp+no4d5tLqIMTSAlQxL0lAf4Qm4e6tF5m/55+dLwxw5/Dqwa0bQXHt98vJjSB" +
                    "YLQH6rwfDzNmVGimO1n84k4MMYY449ykjqRNDfCS3+5+zV/An4CH9eUhvB0AHHWbD6eALw39sPGx" +
                    "c5kHADTOdJ5SboSm9DHYdLLt9k8HyxrHIkcJApLWPgyFmkE0+8jtuQQluN62F5+j5d53oTKinHEd" +
                    "7adM0ea537vNf5uBGu6h9OTXlhZwM9tlnrsTYQTTAIzdxGPlpD9Zvo5nmJHwILdRonm8rZf3vAm5" +
                    "9/U6+Cht4+X2l5zxyg==";

    /**
     * OCES2
     * 
     * Common Name: NemLog-in ADFS Test (funktionscertifikat)
     * Organization: Digitaliseringsstyrelsen // CVR:34051178
     * Country: DK
     * Valid From: April 20, 2017
     * Valid To: April 20, 2020
     * Issuer: TRUST2408 Systemtest XIX CA, TRUST2408
     * Serial Number: 1478034417 (0x581903f1)
     */
    private final static String NEWEST_NEMLOGIN_IDP_CERTIFICATE_STRING = 
            "MIIGKTCCBRGgAwIBAgIEWBkD8TANBgkqhkiG9w0BAQsFADBHMQswCQYDVQQGEwJE" +
                    "SzESMBAGA1UECgwJVFJVU1QyNDA4MSQwIgYDVQQDDBtUUlVTVDI0MDggU3lzdGVt" +
                    "dGVzdCBYSVggQ0EwHhcNMTcwNDIwMDgxODU2WhcNMjAwNDIwMDgxODI2WjCBljEL" +
                    "MAkGA1UEBhMCREsxMTAvBgNVBAoMKERpZ2l0YWxpc2VyaW5nc3N0eXJlbHNlbiAv" +
                    "LyBDVlI6MzQwNTExNzgxVDAgBgNVBAUTGUNWUjozNDA1MTE3OC1GSUQ6NTY5NDA0" +
                    "MTMwMAYDVQQDDClOZW1Mb2ctaW4gQURGUyBUZXN0IChmdW5rdGlvbnNjZXJ0aWZp" +
                    "a2F0KTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAJu6nF+J1+5LfnOc" +
                    "F1IetAM6RD9WA+5SPbfmMZFctyg4yifZL2ppkW7XUB5AUdZsFHpqFbZkTKbCeZFJ" +
                    "IpiraeBDQZtHt7mdHtL8/UF262FCGZdIqP77Kct323rhYcQqd9HfU5iHJZJ1D+H2" +
                    "zNdhqxnEGsTHivvg5Q2qtTUNh+nZqoAeBxPKEg3pHkhrs4QXqPPK+yBEcO8skhqd" +
                    "pMuHBWzo0ciUKz/S8oqO7y8wySqR2p6tZCVIMJg6gqwNeUzbxOPRy1MDk2KXNXdI" +
                    "6xGXtYQ/nUNnxSUl99/Dx0NxEXxMynZ6dJIL0RtqTfawtmGeR25LC3iE0WoLaeNB" +
                    "gTnmyj8CAwEAAaOCAsswggLHMA4GA1UdDwEB/wQEAwIDuDCBlwYIKwYBBQUHAQEE" +
                    "gYowgYcwPAYIKwYBBQUHMAGGMGh0dHA6Ly9vY3NwLnN5c3RlbXRlc3QxOS50cnVz" +
                    "dDI0MDguY29tL3Jlc3BvbmRlcjBHBggrBgEFBQcwAoY7aHR0cDovL2YuYWlhLnN5" +
                    "c3RlbXRlc3QxOS50cnVzdDI0MDguY29tL3N5c3RlbXRlc3QxOS1jYS5jZXIwggEg" +
                    "BgNVHSAEggEXMIIBEzCCAQ8GDSsGAQQBgfRRAgQGBAIwgf0wLwYIKwYBBQUHAgEW" +
                    "I2h0dHA6Ly93d3cudHJ1c3QyNDA4LmNvbS9yZXBvc2l0b3J5MIHJBggrBgEFBQcC" +
                    "AjCBvDAMFgVEYW5JRDADAgEBGoGrRGFuSUQgdGVzdCBjZXJ0aWZpa2F0ZXIgZnJh" +
                    "IGRlbm5lIENBIHVkc3RlZGVzIHVuZGVyIE9JRCAxLjMuNi4xLjQuMS4zMTMxMy4y" +
                    "LjQuNi40LjIuIERhbklEIHRlc3QgY2VydGlmaWNhdGVzIGZyb20gdGhpcyBDQSBh" +
                    "cmUgaXNzdWVkIHVuZGVyIE9JRCAxLjMuNi4xLjQuMS4zMTMxMy4yLjQuNi40LjIu" +
                    "MIGrBgNVHR8EgaMwgaAwPKA6oDiGNmh0dHA6Ly9jcmwuc3lzdGVtdGVzdDE5LnRy" +
                    "dXN0MjQwOC5jb20vc3lzdGVtdGVzdDE5LmNybDBgoF6gXKRaMFgxCzAJBgNVBAYT" +
                    "AkRLMRIwEAYDVQQKDAlUUlVTVDI0MDgxJDAiBgNVBAMMG1RSVVNUMjQwOCBTeXN0" +
                    "ZW10ZXN0IFhJWCBDQTEPMA0GA1UEAwwGQ1JMMTA2MB8GA1UdIwQYMBaAFMwCVQzk" +
                    "gXSvIFTVERRXnJFSOH+gMB0GA1UdDgQWBBRn40IAgUyGNBLIe9WVpg4dp2w6rzAJ" +
                    "BgNVHRMEAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQANLBO8hF+ZyxkHZryhp6hLIPi9" +
                    "tH2TWZNYvv/Y9NW/tWWmVYkgNNAecwpYsm5EzMAmlJiYF0KxZZ0jMIcyUM1q6aG8" +
                    "TOnym8nDKw7kXNegLjH7ZO/uhPabRtCaJmyrYUnwRjgHVmB2h8rsfqV3ACdCAROG" +
                    "8OSSgFdMs9hTHLQLOBxnbhXD+Ohv8taArKnXvgSEb4jbYa78SJ+WD/8Ov3CIlm/i" +
                    "ABbTwPOu1oqa+2c3mURzgEZZlUAb8sor+0Ig1QW3zicDcIVVVk+EmV9SUMxOR437" +
                    "wTJ2KYvU6XziTupCgMWLmrY52lIJ+uS4XMNmxoc60ajUyUZXFYEhVQkH85Qv";

    /**
     * OCES2
     * 
     * Common Name: NemLog-in ADFS Test (funktionscertifikat)
     * Organization: Digitaliseringsstyrelsen // CVR:34051178
     * Country: DK
     * Valid From: February 25, 2020
     * Valid To: February 25, 2023
     * Issuer: TRUST2408 Systemtest XXXIV CA, TRUST2408
     * Serial Number: 1558728194 (0x5ce84e02)
     */
    private static final String NEMLOGIN_IDP_CERTIFICATE_2020_2023_STRING = 
            "MIIGLDCCBRSgAwIBAgIEXOhOAjANBgkqhkiG9w0BAQsFADBJMQswCQYDVQQGEwJE" +
                    "SzESMBAGA1UECgwJVFJVU1QyNDA4MSYwJAYDVQQDDB1UUlVTVDI0MDggU3lzdGVt" +
                    "dGVzdCBYWFhJViBDQTAeFw0yMDAyMjUyMTIwMDVaFw0yMzAyMjUyMTE3MjFaMIGW" +
                    "MQswCQYDVQQGEwJESzExMC8GA1UECgwoRGlnaXRhbGlzZXJpbmdzc3R5cmVsc2Vu" +
                    "IC8vIENWUjozNDA1MTE3ODFUMCAGA1UEBRMZQ1ZSOjM0MDUxMTc4LUZJRDo1Njk0" +
                    "MDQxMzAwBgNVBAMMKU5lbUxvZy1pbiBBREZTIFRlc3QgKGZ1bmt0aW9uc2NlcnRp" +
                    "ZmlrYXQpMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwhqXoxvnrg1R" +
                    "8GSAOmWdvR0lrdRjTss9zzJc/j0zHeamjsOyMdTc787Qdy2yre74ImzbBIRGi7HI" +
                    "3PIP59PebFZnVOUa+/Y1enuQ8HZgJnaZqMCxPrQ+SuU3iD99WTvWtz6x+W1BLGiI" +
                    "JK5Xkz8y3Ya82OZ25LwU+VXEovYEmPFNECV7z9yWkCcLZrfF4mwKbIdr96o9uFav" +
                    "tiThMSD/XkD40ahdvDap13tV+HBBIVwa0gAjYe1Cz2jYPRvCVCy5adDgoDrP7WTS" +
                    "ipvga7A9Y8WIZ52nDLgUlq67rK416trp6zThXAyzpDS98e99MVPy32Ovpqvw8phd" +
                    "LqT6Tq8y3wIDAQABo4ICzDCCAsgwDgYDVR0PAQH/BAQDAgO4MIGXBggrBgEFBQcB" +
                    "AQSBijCBhzA8BggrBgEFBQcwAYYwaHR0cDovL29jc3Auc3lzdGVtdGVzdDM0LnRy" +
                    "dXN0MjQwOC5jb20vcmVzcG9uZGVyMEcGCCsGAQUFBzAChjtodHRwOi8vZi5haWEu" +
                    "c3lzdGVtdGVzdDM0LnRydXN0MjQwOC5jb20vc3lzdGVtdGVzdDM0LWNhLmNlcjCC" +
                    "ASAGA1UdIASCARcwggETMIIBDwYNKwYBBAGB9FECBAYEAzCB/TAvBggrBgEFBQcC" +
                    "ARYjaHR0cDovL3d3dy50cnVzdDI0MDguY29tL3JlcG9zaXRvcnkwgckGCCsGAQUF" +
                    "BwICMIG8MAwWBURhbklEMAMCAQEagatEYW5JRCB0ZXN0IGNlcnRpZmlrYXRlciBm" +
                    "cmEgZGVubmUgQ0EgdWRzdGVkZXMgdW5kZXIgT0lEIDEuMy42LjEuNC4xLjMxMzEz" +
                    "LjIuNC42LjQuMy4gRGFuSUQgdGVzdCBjZXJ0aWZpY2F0ZXMgZnJvbSB0aGlzIENB" +
                    "IGFyZSBpc3N1ZWQgdW5kZXIgT0lEIDEuMy42LjEuNC4xLjMxMzEzLjIuNC42LjQu" +
                    "My4wgawGA1UdHwSBpDCBoTA8oDqgOIY2aHR0cDovL2NybC5zeXN0ZW10ZXN0MzQu" +
                    "dHJ1c3QyNDA4LmNvbS9zeXN0ZW10ZXN0MzQuY3JsMGGgX6BdpFswWTELMAkGA1UE" +
                    "BhMCREsxEjAQBgNVBAoMCVRSVVNUMjQwODEmMCQGA1UEAwwdVFJVU1QyNDA4IFN5" +
                    "c3RlbXRlc3QgWFhYSVYgQ0ExDjAMBgNVBAMMBUNSTDE0MB8GA1UdIwQYMBaAFM1s" +
                    "aJc5chmkNatk6vQRo4GH+Gk7MB0GA1UdDgQWBBRBCneiJ1Rbm3TO7S7J+lDgRq3I" +
                    "FDAJBgNVHRMEAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQCh4sfwYmyapqcQ/hfXye+m" +
                    "qJQa9pTktjU6OVZ5XwqvnMSRAwDaix6jfkW+s1CDBRdHY/vpUEczOKW6piPd6x0O" +
                    "cRrKghyK138yN8472EuNb3v+A4g5AvKgtrRIojRtzUHOOQyl7e5um49WQKEsl8FV" +
                    "DxFPbWjzkLHoDsgnvyLZdOR00QjKJsKYVK6Xct1Rsh5zJZE9/LHAkKx0lBCMT2Uj" +
                    "rYqIj3qR+yKQLutLJxhfaAiLj9h00dVNhUQKTb6cfgPkr/+41lg7CvaVas+wDhQD" +
                    "SrsaTkF54V9u+h3q8+BC3IW6ObMOz84Ws+Rbol5d7pVkVJemvYD5uqCCA+YLG3Bj";

    public final static X509Certificate OLD_NEMLOGIN_IDP_CERTIFICATE = CertificateParser.asCertificate(XmlUtil.fromBase64(OLD_NEMLOGIN_IDP_CERTIFICATE_STRING));

    public final static X509Certificate NEW_NEMLOGIN_IDP_CERTIFICATE = CertificateParser.asCertificate(XmlUtil.fromBase64(NEW_NEMLOGIN_IDP_CERTIFICATE_STRING));

    public final static X509Certificate NEWEST_NEMLOGIN_IDP_CERTIFICATE = CertificateParser.asCertificate(XmlUtil.fromBase64(NEWEST_NEMLOGIN_IDP_CERTIFICATE_STRING));

    public static String readXMLStreamAndRemoveFormatting(InputStream is) {
        String xml = new Scanner(is, "UTF-8").useDelimiter("\\A").next();
        return XmlUtil.removeFormatting(xml);
    }

    public static CredentialVault getOldIdPTrustVault() {
        return new CredentialVaultAdapter() {
            public boolean isTrustedCertificate(X509Certificate certificate) {
                return OLD_NEMLOGIN_IDP_CERTIFICATE.equals(certificate);
            }
        };
    }

    public static CredentialVault getNewIdPTrustVault() {
        return new CredentialVaultAdapter() {
            public boolean isTrustedCertificate(X509Certificate certificate) {
                return NEW_NEMLOGIN_IDP_CERTIFICATE.equals(certificate);
            }
        };
    }

    public static CredentialVault getNewestIdPTrustVault() {
        return new CredentialVaultAdapter() {
            public boolean isTrustedCertificate(X509Certificate certificate) {
                return NEWEST_NEMLOGIN_IDP_CERTIFICATE.equals(certificate);
            }
        };
    }

}
