/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.testtools.testrequest;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.DGWSConstants;
import dk.sosi.seal.model.constants.MedComTags;
import dk.sosi.seal.model.constants.SOSIAttributes;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import javax.xml.transform.TransformerException;
import java.util.Date;
import java.util.Properties;

public abstract class MessageGenerator {

    public String OUTPUTFOLDER;
    public String JKSFOLDER;
    public boolean SAVEOUTPUT;

    protected Properties properties;

    public MessageGenerator(Properties properties) {
        this.properties = properties;

        SAVEOUTPUT = Boolean.valueOf(properties.getProperty(
                "sositest:saveoutput", "true")).booleanValue();

        OUTPUTFOLDER = properties.getProperty(
                "sositest:outputfolder", "target");

        JKSFOLDER = properties.getProperty("sositest:jksfolder",
                "jks");
    }

    public WSTError getError(Document document) {
        try {
            Node node;
            node = XPathAPI.selectSingleNode(document, "//soapenv:Fault");
            if (node == null) {
                return null;
            }
            String faultCode = XPathAPI.selectSingleNode(node, ".//faultcode")
                    .getFirstChild().getNodeValue();
            String faultString = XPathAPI.selectSingleNode(node,
                    ".//faultstring").getFirstChild().getNodeValue();

            String faultActor = XPathAPI.selectSingleNode(node, ".//faultactor").getFirstChild().getNodeValue();

            return new WSTError(faultCode, faultString, faultActor);
        } catch (TransformerException e) {
            throw new MessageGeneratorException("get error failed!", e);
        }
    }

    public Document setLevelType(Document dom, int level, String type) {
        Document clone = domClone(dom);
        setSamlAttribute(clone, SOSIAttributes.AUTHENTICATION_LEVEL, level + "");
        setSamlAttribute(clone, SOSIAttributes.IDCARD_TYPE, type);


        try {
            // TODO move system, user into constants
            if ("system".equals(type)) {
                Node userLog = XPathAPI.selectSingleNode(clone,
                        "//saml:AttributeStatement[@id='UserLog']");
                userLog.getParentNode().removeChild(userLog);
            }
        } catch (TransformerException e) {
            throw new MessageGeneratorException("set saml attribute failed!", e);
        }

        return clone;
    }

    public void setSecurityLevel(Document dom, int level) {
        try {
            // medcom:SecurityLevel
            Node node = XPathAPI.selectSingleNode(dom, "//"+MedComTags.SECURITY_LEVEL_PREFIXED);
            if(node != null) node.getFirstChild().setNodeValue(""+level);
        } catch (TransformerException e) {
            throw new MessageGeneratorException("set saml attribute failed!", e);
        }
    }

    public Document createSTSRequestTemplateDOM(STSMessageGenerator.VERSION version) {
        return createSTSRequestTemplateDOM("99900499"); // Cvr from MOCES_cpr_gyldig.p12
    }

    public Document createSTSRequestTemplateDOM(String cvr) {
        String stsRequestTemplateFilename = properties.getProperty("sositest:stsrequesttemplate");
        if(stsRequestTemplateFilename != null) {
            return MessageGeneratorHelper.getAdjustedDocument(stsRequestTemplateFilename);
        } else {
            Properties props = System.getProperties();
            props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_DGWS_VERSION, DGWSConstants.VERSION_1_0_1);
            SOSIFactory factory = new SOSIFactory(new EmptyCredentialVault(), props);
            SecurityTokenRequest req = factory.createNewSecurityTokenRequest();
            long starttime = System.currentTimeMillis() - 5 * 60 * 1000;
            Date createdDate = new Date(starttime);
            Date expiryDate = new Date(starttime + 24 * 60 * 60 * 1000);
            IDCard idcard = createTemplateIdCard(factory, cvr, AuthenticationLevel.MOCES_TRUSTED_USER, createdDate, expiryDate);
            req.setIDCard(idcard);
            return req.serialize2DOMDocument();
        }
    }

    public Document createExpiredSTSResponseTemplateDOM(STSMessageGenerator.VERSION version) {
        return createExpiredSTSResponseTemplateDOM("99900499"); // Cvr from MOCES_cpr_gyldig.p12
    }

    public Document createExpiredSTSResponseTemplateDOM(String cvr) {
        ClasspathCredentialVault signingVault = new ClasspathCredentialVault(System.getProperties(), "NSP_Test_Service_Consumer_sds.p12", "Test1234");
        SOSIFactory factory = new SOSIFactory(signingVault, properties);
        SecurityTokenResponse response = factory.createNewSecurityTokenResponse(DGWSConstants.VERSION_1_0_1, null);
        long starttime = System.currentTimeMillis() - 5 * 60 * 1000;
        Date createdDate = new Date(starttime);
        Date expiryDate = new Date(starttime + 60 * 1000);
        IDCard idcard = createTemplateIdCard(factory, cvr, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, createdDate, expiryDate);
        response.setIDCard(idcard);
        return response.serialize2DOMDocument();
    }

    private IDCard createTemplateIdCard(SOSIFactory factory, String cvr, AuthenticationLevel authenticationLevel, Date createdDate, Date expiryDate) {
        UserInfo userInfo = new UserInfo("1802602810", "Anders", "And", "aand@andeby.dk", "coolDuck", "doctor", "ZXCVB");
        SystemInfo systemInfo = new SystemInfo(new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, cvr, "orgName"), "SOSITEST");
        return new UserIDCard(factory.getDGWSVersion(), null, XmlUtil.createGUID(), authenticationLevel, null, factory.getIssuer(), systemInfo, userInfo, createdDate, expiryDate, null, null, null);
    }

    public Document domClone(Document dom) {
        return (Document) dom.cloneNode(true);
    }

    protected void setSamlAttribute(Document doc, String id, String value) {
        try {
            Node node = XPathAPI.selectSingleNode(doc,
                    "//saml:Attribute[@Name='" + id + "']/saml:AttributeValue");
            node.getFirstChild().setNodeValue(value);
        } catch (TransformerException e) {
            throw new MessageGeneratorException("set saml attribute failed!", e);
        }
    }

    protected Document sign(Document doc, CredentialVault vault) {
        return MessageGeneratorHelper.sign(doc, vault);
    }
}
