/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.testtools.testrequest;

import dk.sosi.seal.model.constants.SOSIAttributes;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;

import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import javax.xml.transform.TransformerException;
import java.util.Properties;

public class ProviderMessageGenerator extends MessageGenerator {

	public final CredentialVault vaultInvalidIssuerSTS;
	public final CredentialVault vaultExpiredSTS;

	private STSMessageGenerator smg;
	private Document expiredSTSResponse;
	private Document stsRequestTemplate;

	public ProviderMessageGenerator(Properties properties) {
		this(properties, STSMessageGenerator.VERSION.OCES3);
	}

	public ProviderMessageGenerator(Properties properties, STSMessageGenerator.VERSION version) {
		super(properties);

		vaultInvalidIssuerSTS = new ClasspathCredentialVault(properties,"jks/invalidIssuerSTSVault.jks", "Test1234");
		vaultExpiredSTS = new ClasspathCredentialVault(properties,"jks/expiredSTSVault.jks", "Test1234");
		smg = new STSMessageGenerator(properties, version);

		stsRequestTemplate = createSTSRequestTemplateDOM(version);
		expiredSTSResponse = createExpiredSTSResponseTemplateDOM(version);
	}

	public void generateLvl1SystemValidIdCard(Document providerRequest) {
		int level = 1;
		Document dom1system = smg.setLevelType(stsRequestTemplate, level, "system");
		MessageGeneratorHelper.copyIdCard(dom1system, providerRequest);
		smg.setSecurityLevel(providerRequest, level);
	}

	/**
	 * Replaces IDCard on providerRequest
	 *
	 * @param providerRequest
	 */
	public void generateLvl3SystemValidIdCard(Document providerRequest) {
		int level = 3;
		Document dom3system = smg.setLevelType(stsRequestTemplate, level, "system");
		Document validStsRequest = smg.generateVocesValidLevel3System(dom3system);
		Document stsResponse = smg.execute(validStsRequest);
		MessageGeneratorHelper.copyIdCard(stsResponse, providerRequest);
		smg.setSecurityLevel(providerRequest, level);
	}

	public void generateLvl3UserValidIdCard(Document providerRequest) {
		int level = 3;
		Document dom3user = smg.setLevelType(stsRequestTemplate, level, "user");
		Document validStsRequest = smg.generateVocesValidLevel3User(dom3user);
		Document stsResponse = smg.execute(validStsRequest);
		MessageGeneratorHelper.copyIdCard(stsResponse, providerRequest);
		smg.setSecurityLevel(providerRequest, level);
	}

	public void generateLvl4UserValidIdCard(Document providerRequest) {
		int level = 4;
		Document dom4user = smg.setLevelType(stsRequestTemplate, level, "user");
		Document validStsRequest = smg.generateMocesValidLevel4User(dom4user);
		Document stsResponse = smg.execute(validStsRequest);
		MessageGeneratorHelper.copyIdCard(stsResponse, providerRequest);
		smg.setSecurityLevel(providerRequest, level);
	}

	public void generateIDCardExpired(Document providerRequest) {
		Document stsResponse = smg.domClone(expiredSTSResponse);
		MessageGeneratorHelper.copyIdCard(stsResponse, providerRequest);

		try {
			Node node = XPathAPI.selectSingleNode(expiredSTSResponse,
					"//saml:Attribute[@Name='" + SOSIAttributes.AUTHENTICATION_LEVEL + "']/saml:AttributeValue");
			int level = Integer.parseInt(node.getFirstChild().getNodeValue());
			smg.setSecurityLevel(providerRequest, level);
		} catch (TransformerException e) {
			throw new MessageGeneratorException("set saml attribute failed!", e);
		}
	}

	public void generateIDCardUnsigned(Document providerRequest) {
		int level = 4;
		Document dom4user = smg.setLevelType(stsRequestTemplate, level, "user");
		MessageGeneratorHelper.copyIdCard(dom4user, providerRequest);
		smg.setSecurityLevel(providerRequest, level);
	}

	public void generateCorruptedUserInfoIdCard(Document providerRequest) {
		generateLvl4UserValidIdCard(providerRequest);
		// remove userinfo
		try {
			Node userLog = XPathAPI.selectSingleNode(providerRequest,
					"//saml:AttributeStatement[@id='UserLog']");
			userLog.getParentNode().removeChild(userLog);
		} catch (TransformerException e) {
			throw new MessageGeneratorException("element deletion failed",e);
		}
	}

	public void generateCorruptedRoleIdCard(Document providerRequest) {
		generateLvl4UserValidIdCard(providerRequest);
		// remove rolevalue
		try {
			Node role = XPathAPI.selectSingleNode(providerRequest,
					"//saml:Attribute[@Name='medcom:UserRole']");
			role.getParentNode().removeChild(role);
		} catch (TransformerException e) {
			throw new MessageGeneratorException("element deletion failed",e);
		}
	}

	public void generateIDCardSTSCertWrongSSN(Document providerRequest) {
		int level = 4;
		// this is signed with users cert - thus wrong ssn
		Document dom4user = smg.setLevelType(stsRequestTemplate, level, "user");
		Document validStsRequest = smg.generateMocesValidLevel4User(dom4user);
		MessageGeneratorHelper.copyIdCard(validStsRequest, providerRequest);
		smg.setSecurityLevel(providerRequest, level);
	}

	public void generateIDCardSTSCertInvalidIssuer(Document providerRequest) {
		generateLvl4UserValidIdCard(providerRequest);
		// remove signature and add the invalid one
		MessageGeneratorHelper.removeSignature(providerRequest);
		sign(providerRequest, vaultInvalidIssuerSTS);
	}

	public void generateIDCardSTSCertExpired(Document providerRequest) {
		generateLvl4UserValidIdCard(providerRequest);
		// remove signature and add the expired one
		MessageGeneratorHelper.removeSignature(providerRequest);
		sign(providerRequest, vaultExpiredSTS);
	}

	protected void writeXmlFile(Document doc, String id) {
		if(SAVEOUTPUT) {
			MessageGeneratorHelper.writeXmlFile(doc, OUTPUTFOLDER + "/providerrequests/"
					+ id + ".xml");
		}
	}


	public STSMessageGenerator getSTSMessageGenerator() {
		return smg;
	}
}