/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.testtools.testrequest;

import dk.sosi.seal.model.constants.MedcomAttributes;
import dk.sosi.seal.model.constants.SOSIAttributes;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Calendar;
import java.util.Properties;

public class STSMessageGenerator extends MessageGenerator {
	
	public enum VERSION { OCES3 }

    private final String defaultUrl;

	public final CredentialVault vaultMoces;
	public final CredentialVault vaultVoces;
	public final CredentialVault vaultExpiredVoces;
	public final CredentialVault vaultInvalidIssuerVoces;
	public final CredentialVault vaultRevokedVoces;
	public final CredentialVault vaultExpiredMoces;
	public final CredentialVault vaultInvalidIssuerMoces;
	public final CredentialVault vaultRevokedMoces;
	private final VERSION version;

	public STSMessageGenerator(Properties properties, VERSION version) {
		super(properties);
		this.version = version;

		vaultMoces = CredentialVaultTestUtil.getCredentialVaultFromResource(properties, "oces3/MOCES_cpr_gyldig.p12");
		vaultExpiredMoces = null; // TODO SDS-6659: Findes ikke endnu
		vaultRevokedMoces =
				CredentialVaultTestUtil.getCredentialVaultFromResource(properties, "oces3/MOCES_spaerret.p12");
		vaultInvalidIssuerMoces = null; // TODO SDS-6659: Er ikke lavet

		vaultVoces = CredentialVaultTestUtil.getCredentialVaultFromResource(properties, "oces3/VOCES_gyldig.p12");
		vaultExpiredVoces = null; // TODO SDS-6659: Findes ikke endnu;
		vaultRevokedVoces =
				CredentialVaultTestUtil.getCredentialVaultFromResource(properties, "oces3/VOCES_spaerret.p12");
		vaultInvalidIssuerVoces =
				null; //CredentialVaultTestUtil.getCredentialVaultFromResource(properties, "oces3/VOCES_invalid_issuer
		// .p12"); // TODO SDS-6659

		defaultUrl = "http://test1-cnsp.ekstern-test.nspop.dk:8080/sts/services/NewSecurityTokenService";
	}

	public VERSION getVersion() { return version; }

	public Document generateInvalidIdCardTypeLevelCombination(Document document) {
		return sign(document, vaultMoces);
	}

	public Document generateMocesValidLevel4User(Document document) {
		return sign(document, vaultMoces);
	}

	public Document generateVocesValidLevel3System(Document document) {
		return sign(document, vaultVoces);
	}

	public Document generateVocesValidLevel3User(Document document) {
		return sign(document, vaultVoces);
	}

	public Document generateInvalidCareProviderID(Document document) {
		setSamlAttribute(document, MedcomAttributes.CARE_PROVIDER_ID, "00000000");
		return sign(document, vaultVoces);
	}

	public Document generateInvalidUserCivilRegistrationNumber(Document document) {
		setSamlAttribute(document, MedcomAttributes.USER_CIVIL_REGISTRATION_NUMBER,
				"0000000000");
		return sign(document, vaultMoces);
	}

	public Document generateInvalidUserAuthorizationCode(Document document) {
		setSamlAttribute(document, MedcomAttributes.USER_AUTHORIZATION_CODE, "00000");
		return sign(document, vaultMoces);
	}

	public Document generateInvalidRole(Document document) {
		setSamlAttribute(document, MedcomAttributes.USER_ROLE, "healer");
		return sign(document, vaultMoces);
	}

	public Document generateMocesLevel3(Document document) {
		return sign(document, vaultMoces);
	}

	public Document generateMocesRevoked(Document document) {
		return sign(document, vaultRevokedMoces);
	}

	public Document generateMocesInvalidIssuer(Document document) {
		return sign(document, vaultInvalidIssuerMoces);
	}

	public Document generateMocesExpired(Document document) {
		return sign(document, vaultExpiredMoces);
	}

	public Document generateMocesInvalidSignature(Document document) {
		Document doc = sign(document, vaultMoces);
		setSamlAttribute(document, MedcomAttributes.USER_EMAIL_ADDRESS,
				"alice@malice.dk");
		return doc;
	}

	public Document generateVocesLevel4(Document document) {
		return sign(document, vaultVoces);
	}

	public Document generateVocesRevoked(Document document) {
		return sign(document, vaultRevokedVoces);
	}

	public Document generateVocesInvalidIssuer(Document document) {
		return sign(document, vaultInvalidIssuerVoces);
	}

	public Document generateVocesExpired(Document document) {
		return sign(document, vaultExpiredVoces);
	}

	public Document generateVocesInvalidSignature(Document document) {
		Document doc = sign(document, vaultVoces);
		setSamlAttribute(document, MedcomAttributes.IT_SYSTEM_NAME, "malice");
		return doc;
	}

	public Document generateInvalidGyldighed(Document document) {
		try {
			Calendar c = Calendar.getInstance();
			c.add(Calendar.DAY_OF_MONTH, 2);
			String dateStringMuchLater = MessageGeneratorHelper.sdf.format(c.getTime());

			Element elm;
			elm = (Element) XPathAPI.selectSingleNode(document,	"//saml:Conditions");
			elm.setAttributeNS(null, "NotOnOrAfter", dateStringMuchLater);
			return sign(document, vaultVoces);
		} catch (TransformerException e) {
			throw new MessageGeneratorException(e.getMessage(), e);
		}
	}

	public Document generateInvalidAutentifikationLevel(Document document) {
		return sign(document, vaultVoces);
	}

	public Document generateInvalidIdCardType(Document document) {
		return sign(document, vaultVoces);
	}

	public Document generateInvalidXmlRequest(Document document) {
		try {
			Node node = XPathAPI.selectSingleNode(document, "//saml:Attribute[@Name='" + SOSIAttributes.IDCARD_VERSION + "']");
			node.getParentNode().removeChild(node);
			return sign(document, vaultVoces);
		} catch (TransformerException e) {
			throw new MessageGeneratorException(e.getMessage(), e);
		}
	}

	public Document generateInvalidVersion(Document document) {
		setSamlAttribute(document, SOSIAttributes.IDCARD_VERSION, "0.8");
		return sign(document, vaultVoces);
	}

	public Document process(Document req, String id) {
		if(SAVEOUTPUT) MessageGeneratorHelper.writeXmlFile(req, OUTPUTFOLDER + "/requests/" + id + "Request.xml");
		Document resp = execute(req);
		if(SAVEOUTPUT) MessageGeneratorHelper.writeXmlFile(resp, OUTPUTFOLDER + "/responses/" + id + "Response.xml");
		return resp;
	}

	public Document execute(Document req) {
		try {
			Source source = new DOMSource(req);
			StringWriter sw = new StringWriter();
			Result result = new StreamResult(sw);
			Transformer xformer;
			xformer = TransformerFactory.newInstance().newTransformer();
			xformer.transform(source, result);
			return execute(sw.getBuffer().toString());
		} catch (TransformerConfigurationException e) {
			throw new MessageGeneratorException("execute failed", e);
		} catch (TransformerException e) {
			throw new MessageGeneratorException("execute failed", e);
		}
	}

	public Document execute(String data) {
		try {
			String urlString = System.getProperty("sts.url", defaultUrl);
			SoapClient sc = new SoapClient(new URL(urlString));
			Document res = sc.execute(data);
			return res;
		} catch (MalformedURLException e) {
			throw new MessageGeneratorException("url failed", e);
		}
	}
}