/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package dk.sosi.testtools.testrequest;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.SignatureConfiguration.KeyInfoType;
import dk.sosi.seal.model.constants.SOAPTags;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.modelbuilders.OIOBootstrapToIdentityTokenResponseModelBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.vault.GenericCredentialVault;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;
import org.junit.Ignore;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import static org.junit.Assert.*;

/**
 * Created by chg on 02/03/2017.
 */
public class TestCitizenFlow {
	
	private static boolean doPrint = false;

	private static final String testCpr = "0501792275";

    private static final String IDWS_REQUEST =
            "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:ns=\"http://sundhedsdatastyrelsen.dk/skr/2021/06/02\">" +
    			"<soapenv:Header/>" +
    			"<soapenv:Body>" +
        			"<ns:GetPersonalDataCardRequest>" +
            			"<id assigningAuthorityName=\"CPR\" extension=\"" + testCpr + "\" root=\"1.2.208.176.1.2\"/>" +
        			"</ns:GetPersonalDataCardRequest>" +
    			"</soapenv:Body>" +
			"</soapenv:Envelope>";

	private static final String IDWS_SOAP_ACTION = "http://sundhedsdatastyrelsen.dk/skr/2021/06/02#GetPersonalDataCard";

	private static final String STS_ENDPOINT = "http://test1-cnsp.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws";

	private static final String IDWS_ENDPOINT = "http://test1-cnsp.ekstern-test.nspop.dk:8080/skr/skr";

	@Test
	@Ignore("SDS-6659")
	public void testCitizenSkrInvocation() throws IOException, SOAPException {

		// Build signed bootstrap token (instead of NemLog-in issued bootstrap token)
		OIOSAMLAssertionBuilder assertionBuilder = new OIOSAMLAssertionBuilder() {

			// Tweaks to generate the desired token
			@Override
			protected Element createAttributeStatement(Document doc) {
				return null;
			}

			@Override
			protected Node createAuthnStatement() {
				return null;
			}

			@Override
			public void validateBeforeBuild() {
				this.certificateInfo = CertificateInfo.fromString("SubjectDN={C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503},IssuerDN={C=ignored},CertSerial={1234}");
			}
		};

		CredentialVault idpVault = new ClasspathCredentialVault(System.getProperties(), "TEST trusted IdP SOSI alias (til bootstrap token).jks", "Test1234");
		assertionBuilder.setUserIdCard(createUserIDCard(idpVault));
		assertionBuilder.setSigningVault(idpVault);
		assertionBuilder.setIssuer("TEST trusted IdP");
		assertionBuilder.setRecipientURL("https://sosi");
		assertionBuilder.setAudienceRestriction("https://bootstrap.sts.nspop.dk/");
		Date now = new Date();
		assertionBuilder.setDeliveryNotOnOrAfter(new Date(now.getTime() + 5 * 60 * 1000));
		assertionBuilder.setNotBefore(new Date(now.getTime() - 5 * 1000));
		assertionBuilder.setNotOnOrAfter(new Date(now.getTime() + 5 * 60 * 1000));

		OIOSAMLAssertion bootstrapToken = assertionBuilder.build();

		// WS-Trust invocation
		OIOSAMLFactory factory = new OIOSAMLFactory();
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setCPRNumberClaim(testCpr);
		domBuilder.setAudience("https://fsk");
		final CredentialVault spVault = new ClasspathCredentialVault(System.getProperties(), "TEST whitelisted SP SOSI alias.jks", "Test1234");
		domBuilder.setSigningVault(spVault);
		domBuilder.setOIOBootstrapToken(new OIOBootstrapToken(bootstrapToken.getDOM().getDocumentElement()));

		Document wstRequest = domBuilder.build();

		printNode("WS-Trust request", wstRequest);

		Document wstResponse = new SoapClient(new URL(STS_ENDPOINT)).execute(XmlUtil.node2String(wstRequest));

		final IdentityToken identityToken = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder()
				.build(wstResponse)
				.getIdentityToken();

		printNode("WS-Trust response", wstResponse);

        Document skrResponse = callSkr(identityToken);

		printNode("SKR Response", skrResponse);
		
		final String
				expected = "GetPersonalDataCardResponse",
				response = XmlUtil.node2String(skrResponse, false, false);

		SOAPBody sb = MessageFactory
				.newInstance()
				.createMessage(null, new ByteArrayInputStream(response.getBytes()))
				.getSOAPBody();
		assertFalse(sb.hasFault() ? sb.getFault().getFaultString() : "", sb.hasFault());
		assertTrue("Forventede "+ expected +" i svar fra SKR", response.contains(expected));
	}


	@Test
	@Ignore("SDS-6659")
	public void testCitizenSkrInvocationWithCitizenIdentityToken()
			throws IOException, URISyntaxException, SOAPException {
		final CredentialVault idpVault = new ClasspathCredentialVault(System.getProperties(), "TEST trusted IdP SOSI alias (til bootstrap token).jks", "Test1234");
		final CredentialVault spVault = new ClasspathCredentialVault(System.getProperties(), "TEST whitelisted SP SOSI alias.jks", "Test1234");
		final GenericCredentialVault idwsTestVault = CredentialVaultTestUtil.getCredentialVaultFromResource(System.getProperties(), "fmk_idws_test.p12");
		final File fmk_idws_test_2023 = new File(Thread.currentThread().getContextClassLoader().getResource("fmk_idws_test_2023.pem").toURI());
		idwsTestVault.addTrustedCertificate(fmk_idws_test_2023, "fmk_idws_test_2023");
		
		final CitizenIdentityToken identityToken = callSTS(idpVault, spVault, "https://fsk");

		final String
				response = XmlUtil.node2String(callSkr(identityToken)),
				expected = "GetPersonalDataCardResponse";
		SOAPBody sb = MessageFactory
				.newInstance()
				.createMessage(null, new ByteArrayInputStream(response.getBytes()))
				.getSOAPBody();
		assertFalse(sb.hasFault() ? sb.getFault().getFaultString() : "", sb.hasFault());
		assertTrue("Forventede " + expected + " i svar fra SKR", response.contains(expected));
	}
	
	@Test
	@Ignore("SDS-6659")
	public void testCitizenDemoServiceInvocationWithCitizenIdentityToken() throws IOException {
		String endpointAddress = "http://localhost:8080/fmk12/idws/MedicineCardIdwsEndpoint";
		String soapAction = "http://testrequest.testtools.sosi.dk/DemoService/GetMedicineCardRequest";

		/**/
		final CredentialVault idpVault = new ClasspathCredentialVault(System.getProperties(), "TEST trusted IdP SOSI alias (til bootstrap token).jks", "Test1234");
		final CredentialVault spVault = new ClasspathCredentialVault(System.getProperties(), "TEST whitelisted SP SOSI alias.jks", "Test1234");
		final CredentialVault fmkIdwsTestVault = CredentialVaultTestUtil.getCredentialVaultFromResource(System.getProperties(), "fmk_idws_test.p12");

		/*STS*/
		final CitizenIdentityToken identityToken = callSTS(idpVault, spVault);
		
		Document demoDocument =  XmlUtil.createEmptySoapDocument();
		Element data = TagUtil.addTag(demoDocument.getDocumentElement(), SOAPTags.envelope, SOAPTags.body, TagUtil.create("data"));
		TagUtil.setText(data, "text");
		TagUtil.setAttribute(data, "attr", "test");
		
		/**/
		LibertyResponse response = callDemoService(new URL(endpointAddress), soapAction, spVault, KeyInfoType.WSS_REFERENCE, demoDocument, identityToken);

		assertTrue("Forventede <data> i svar fra FMK", response.toString().contains("<data>"));

		assertNotNull(response.getMessageID());
		assertNotNull(response.getCreatedTimestamp());	
		assertNotNull(response.getSigningCertificate());

		response.validateSignatureAndTrust(fmkIdwsTestVault);
		
		assertSame(KeyInfoType.WSS_REFERENCE, response.getKeyInfoType());
	}
	
	@Test
	@Ignore("SDS-6659")
	public void testCitizenDemoServiceInvocationWithCitizenIdentityToken2() throws IOException {
		String endpointAddress = "http://localhost:8080/fmk12/idws/MedicineCardIdwsEndpoint";
		String soapAction = "http://testrequest.testtools.sosi.dk/DemoService/GetMedicineCardRequest";
	
		/**/
		final CredentialVault idpVault = new ClasspathCredentialVault(System.getProperties(), "TEST trusted IdP SOSI alias (til bootstrap token).jks", "Test1234");
		final CredentialVault spVault = new ClasspathCredentialVault(System.getProperties(), "TEST whitelisted SP SOSI alias.jks", "Test1234");
		final CredentialVault fmkIdwsTestVault = CredentialVaultTestUtil.getCredentialVaultFromResource(System.getProperties(), "fmk_idws_test.p12");

		/*STS*/
		final CitizenIdentityToken identityToken = callSTS(idpVault, spVault);
		
		Document demoDocument =  XmlUtil.createEmptySoapDocument();
		Element data = TagUtil.addTag(demoDocument.getDocumentElement(), SOAPTags.envelope, SOAPTags.body, TagUtil.create("data"));
		TagUtil.setText(data, "text");
		TagUtil.setAttribute(data, "attr", "test");
		
		/**/
		LibertyResponse response = callDemoService(new URL(endpointAddress), "\""+soapAction+"\"", spVault, KeyInfoType.WSS_KEYIDENTIFIER, demoDocument, identityToken);

		assertTrue("Forventede <data> i svar fra FMK", response.toString().contains("<data>"));

		assertNotNull(response.getMessageID());
		assertNotNull(response.getCreatedTimestamp());	
		assertNotNull(response.getSigningCertificate());

		response.validateSignatureAndTrust(fmkIdwsTestVault);
		
		assertSame(KeyInfoType.WSS_REFERENCE, response.getKeyInfoType());
	}

	private CitizenIdentityToken callSTS(final CredentialVault idpVault, final CredentialVault spVault)
			throws MalformedURLException {
		return callSTS(idpVault, spVault, "https://fmk");
	}

	private CitizenIdentityToken callSTS(final CredentialVault idpVault, final CredentialVault spVault, String audience)
			throws MalformedURLException {
		// Build signed bootstrap token (instead of NemLog-in issued bootstrap token)
		OIOSAMLAssertionBuilder assertionBuilder = new OIOSAMLAssertionBuilder() {

			// Tweaks to generate the desired token
			@Override
			protected Element createAttributeStatement(Document doc) {
				return null;
			}

			@Override
			protected Node createAuthnStatement() {
				return null;
			}

			@Override
			public void validateBeforeBuild() {
				this.certificateInfo = CertificateInfo.fromString("SubjectDN={C=DK,O=Ingen organisatorisk tilknytning,CN=Lars Larsen,Serial=PID:9208-2002-2-514358910503},IssuerDN={C=ignored},CertSerial={1234}");
			}

		};

		assertionBuilder.setUserIdCard(createUserIDCard(idpVault));
		assertionBuilder.setSigningVault(idpVault);
		assertionBuilder.setIssuer("TEST trusted IdP");
		assertionBuilder.setRecipientURL("https://sosi");
		assertionBuilder.setAudienceRestriction("https://bootstrap.sts.nspop.dk/");
		Date now = new Date();
		assertionBuilder.setDeliveryNotOnOrAfter(new Date(now.getTime() + 5 * 60 * 1000));
		assertionBuilder.setNotBefore(new Date(now.getTime() - 5 * 1000));
		assertionBuilder.setNotOnOrAfter(new Date(now.getTime() + 5 * 60 * 1000));

		OIOSAMLAssertion bootstrapToken = assertionBuilder.build();

		// WS-Trust invocation
		OIOSAMLFactory factory = new OIOSAMLFactory();
		OIOBootstrapToIdentityTokenRequestDOMBuilder domBuilder = factory.createOIOBootstrapToIdentityTokenRequestDOMBuilder();
		domBuilder.setCPRNumberClaim("0501792275");
		domBuilder.setAudience(audience);
		domBuilder.setSigningVault(spVault);
		domBuilder.setOIOBootstrapToken(new OIOBootstrapToken(bootstrapToken.getDOM().getDocumentElement()));

		Document wstRequest = domBuilder.build();

		printNode("WS-Trust request", wstRequest);
		
		SoapClient soapClient = new SoapClient(new URL(STS_ENDPOINT));

		Document wstResponse = soapClient.execute(XmlUtil.node2String(wstRequest));

		OIOBootstrapToIdentityTokenResponseModelBuilder modelBuilder = factory.createOIOBootstrapToIdentityTokenResponseModelBuilder();
		OIOBootstrapToIdentityTokenResponse tokenResponse = modelBuilder.build(wstResponse);
		printNode("WS-Trust response", wstResponse);
		
		final Federation federation = new SOSITestFederation(SignatureUtil.setupCryptoProviderForJVM());
		tokenResponse.validateSignatureAndTrust(federation);

		final CitizenIdentityToken identityToken = (CitizenIdentityToken) tokenResponse.getIdentityToken();
		
		return identityToken;
	}

	public Document callSkr(IdentityToken identityToken) throws MalformedURLException {
		final CredentialVault spVault = new ClasspathCredentialVault(System.getProperties(), "TEST whitelisted SP SOSI alias.jks", "Test1234");

		Document skrRequest = XmlUtil.readXml(System.getProperties(), IDWS_REQUEST, false);

		LibertyRequestDOMEnhancer domEnhancer = new LibertyRequestDOMEnhancer(spVault, skrRequest) {

			//FMK cannot handle STR-Transforms, refer directly to identity token to let it be part of the message signature
			@Override
			protected void addTokenToSecurityHeader(Element securityHeader, List<SignatureConfiguration.Reference> references) {
				securityHeader.appendChild(document.importNode(identityToken.getDOM().getDocumentElement(), true));
			}

		};

		domEnhancer.setIdentityToken(identityToken);
		domEnhancer.setWSAddressingAction(IDWS_SOAP_ACTION);
		domEnhancer.useSecurityTokenReference();

		domEnhancer.enhanceAndSign();

		printNode("SKR request", skrRequest);

		SoapClient skrClient = new SoapClient(new URL(IDWS_ENDPOINT));
		skrClient.setSOAPAction(IDWS_SOAP_ACTION);
		return skrClient.execute(XmlUtil.node2String(skrRequest));
	}


	private static Schema getSchemas(SchemaFactory factory, String... urls) throws SAXException {
		List<String> urlList = Arrays.asList(urls);
		StreamSource[] streamSources = urlList.stream().map(url -> new StreamSource(url)).toArray(StreamSource[]::new);
		return factory.newSchema(streamSources);
	}
    
	public LibertyResponse callDemoService(URL endpoint, String soapAction, CredentialVault vault, KeyInfoType keyInfoType, Document request, CitizenIdentityToken token) throws MalformedURLException {
//		System.setProperty("com.sun.xml.ws.transport.http.client.HttpTransportPipe.dump", "true");
//		System.setProperty("com.sun.xml.internal.ws.transport.http.client.HttpTransportPipe.dump", "true");
//		System.setProperty("com.sun.xml.ws.transport.http.HttpAdapter.dump", "true");
//		System.setProperty("com.sun.xml.internal.ws.transport.http.HttpAdapter.dump", "true");
//		System.setProperty("com.sun.xml.internal.ws.transport.http.HttpAdapter.dumpTreshold", "999999");
		
		/*Client side enhancing of request document*/
		final OIOIDWSFactory factory = new OIOIDWSFactory(vault);
		final LibertyRequestDOMEnhancer domEnhancer = factory.createRequestDOMEnhancer(request);

		domEnhancer.setIdentityToken(token);
		domEnhancer.setWSAddressingAction(soapAction);
		domEnhancer.setWSAddressingTo("demo service");
		/*Sign document*/
		domEnhancer.enhanceAndSign();

		
		try (EndpointServer<DemoService> server = EndpointServer.start(endpoint, new DemoServiceImpl())){
			assertTrue(server.isPublished());
			assertEquals("http://schemas.xmlsoap.org/wsdl/soap/http", server.getBinding().getBindingID());
			
			EndpointClientBuilder<DemoService> clientBuilder = server.getClientBuilder();
			
			clientBuilder.setNamespaceURI("http://www.dkma.dk/medicinecard/xml.schema/2015/01/01/E1");
			clientBuilder.setServicePart("GetMedicineCardRequest");
			clientBuilder.setPortName("GetMedicineCardPort");
			
			EndpointClient<DemoService> client = clientBuilder.build();
			
			printNode("DEMO request", request);
			Document response = client.execute(XmlUtil.node2String(request), new EndpointClient.Executor<Document, DemoService>() {
				@Override
				public Class<DemoService> getServiceInterface() {
					return DemoService.class;
				}
				
				@Override
				public Document execute(DemoService t, String input) {
					String output = t.GetMedicineCard(input);
					StringReader sr = new StringReader(output);

					org.xml.sax.InputSource inputSource = new InputSource(sr);

					DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
					factory.setValidating(false);
					factory.setNamespaceAware(true);
					Document document;
					try {
						DocumentBuilder builder = factory.newDocumentBuilder();
						document = builder.parse(inputSource);
					} catch (SAXException sxe) {
						Exception x = sxe;
						if (sxe.getException() != null) {
							x = sxe.getException();
						}
						throw new MessageGeneratorException(x.getMessage(),x);
					} catch (ParserConfigurationException pce) {
						throw new MessageGeneratorException(pce.getMessage(),pce);
					} catch (IOException e) {
						throw new MessageGeneratorException("Error during soap invocation: " + e.getMessage(), e);
					}
					return document;

				}
			});	
			printNode("DEMO response", response);
			
			return new LibertyResponse(response) {
				@Override
				public String toString() {
					return XmlUtil.node2String(response, true, true);
				}
			};
		}
	}

	private void printNode(String key, Document document) {
		if(doPrint) {
			System.out.println("------------------- " + key + " -------------------");
//			Node cloneNode = document.cloneNode(true);
//			Node envelope = cloneNode.getFirstChild();
//			NodeList childNodes = envelope.getChildNodes();
//			
//			for (int i = 0; i < childNodes.getLength(); i++) {
//				Node item = childNodes.item(i);
//				if(item.getNodeName().contains("Body")) {
//					envelope.removeChild(item);
//				}
//				
//			}
			System.out.println(XmlUtil.node2String(document, true, true));
			System.out.println("----------------------------------------------------");
		}
	}

	private UserIDCard createUserIDCard(CredentialVault credentialVault) {
		SOSIFactory sosiFactory = new SOSIFactory(credentialVault, System.getProperties());
		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "30808460", "Lægehuset på bakken");
		UserInfo userInfo = new UserInfo(testCpr, "Hans", "Dampf", null, null, "7170", "341KY");
		String alternativeIdentifier = new CertificateInfo(credentialVault.getSystemCredentialPair().getCertificate()).toString();
		UserIDCard userIDCard = sosiFactory.createNewUserIDCard("IT-System", userInfo, careProvider, AuthenticationLevel.MOCES_TRUSTED_USER, null, null, null, alternativeIdentifier);
		Request tmpRequest = sosiFactory.createNewRequest(false, "flow");
		tmpRequest.setIDCard(userIDCard);
		userIDCard.sign(tmpRequest.serialize2DOMDocument(), sosiFactory.getSignatureProvider());
		return userIDCard;
	}
}