package dk.sosi.testtools.testrequest;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPConstants;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;

import dk.sosi.seal.model.BlurringInstructions10Impl;
import dk.sosi.seal.pki.SOSITestFederation;
import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.AbstractCitizenIdentityTokenTest;
import dk.sosi.seal.model.BasicPrivileges;
import dk.sosi.seal.model.CitizenIdentityToken;
import dk.sosi.seal.model.LibertyMessageDOMEnhancer;
import dk.sosi.seal.model.LibertyRequest;
import dk.sosi.seal.model.LibertyRequestDOMEnhancer;
import dk.sosi.seal.model.LibertyResponse;
import dk.sosi.seal.model.constants.Tag;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.TagUtil;
import dk.sosi.seal.xml.XmlUtil;

public class TestCitizenServer extends AbstractCitizenIdentityTokenTest {
	private static final String MY_SOAP_APP_REQUEST = 	"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
														"<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">" + 
														"<soapenv:Header>" + 
														"<myApp:Header xmlns:myApp=\"http://demo.dk/custom\">my_app_header</myApp:Header>" + 
														"</soapenv:Header>" + 
														"<soapenv:Body>" + 
														"<myApp:App xmlns:myApp=\"http://demo.dk/custom\">" + 
														"<myApp:Field_1>value1</myApp:Field_1>" + 
														"<myApp:Field_2 custom_attribute_2=\"value2\" />" + 
														"</myApp:App>" + 
														"</soapenv:Body>" + 
														"</soapenv:Envelope>";
	
	private static final String MY_SOAP_APP_REPONSE = 	"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
														"<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\">" + 
														"<soapenv:Header>" + 
														"<myApp:Header xmlns:myApp=\"http://demo.dk/custom\">my_app_header</myApp:Header>" + 
														"</soapenv:Header>" + 
														"<soapenv:Body>" + 
														"<myApp:Response xmlns:myApp=\"http://demo.dk/custom\">" + 
														"<myApp:Status>It is working!</myApp:Status>" + 
														"</myApp:Response>" + 
														"</soapenv:Body>" + 
														"</soapenv:Envelope>";

	private SOAPMessage soapRequestMessage = null;
	private SOAPMessage soapResponseMessage = null;
	
	@Before
	public void setup() throws IOException, SOAPException {
		/* Client will build up custom soap request */
		Document custom_app_request = XmlUtil.readXml(System.getProperties(), MY_SOAP_APP_REQUEST, false);

		/* IdentityToken will be fetched from STS */
		

		CitizenIdentityToken citizenIdentityToken = super.createBuilder()
				.setIssuer("TEST1-NSP-STS")
				.setAudienceRestriction("https://myApp")
				.build();

		/* Setting up enhancer */
		final OIOIDWSFactory factory = new OIOIDWSFactory(spVault);
		final LibertyRequestDOMEnhancer enhancer = factory.createRequestDOMEnhancer(custom_app_request);
		enhancer.setIdentityToken(citizenIdentityToken); // Required
		enhancer.setWSAddressingAction("myApp.request");// Required
		//enhancer.setWSAddressingMessageID("?");//Required - one will be created if not set be client
		//enhancer.setWSAddressingRelatesTo(""); //Optional
		enhancer.setWSAddressingTo("https://myApp"); //Optional
		enhancer.enhanceAndSign();

		MessageFactory messageFactory = MessageFactory.newInstance(SOAPConstants.SOAP_1_1_PROTOCOL);
		soapRequestMessage = messageFactory.createMessage(null, new ByteArrayInputStream(XmlUtil.node2String(custom_app_request).getBytes()));		
	}
	
	private static final CredentialVault stsTestVault = new ClasspathCredentialVault(System.getProperties(), "NSP_Test_Service_Consumer_sds.p12", "Test1234");
	@Override
	protected CredentialVault getTokenSigningVault() {
		return stsTestVault;
	}
	
	private static final CredentialVault spVault = new ClasspathCredentialVault(System.getProperties(), "TEST whitelisted SP SOSI alias.jks", "Test1234");	
	@Override
	protected CredentialVault getMessageSigningVault() {
		return spVault;
	}

	@Test
	@Ignore("SDS-6659")
	public void testServerSideParsingAndValidation() throws SOAPException, IOException {
		/* Server will setup OIOIDWSFactory */
		final Federation federation = new SOSITestFederation(new Properties(), "UI:DK-O:G:8d3fa047-c77e-47e4-bdd2-e91488610ce6");
		final CredentialVault serverTestVault = CredentialVaultTestUtil.getCredentialVaultFromResource(System.getProperties(), "fmk_idws_test.p12");

		final OIOIDWSFactory oioidwsFactory = new OIOIDWSFactory(federation, serverTestVault);
		
		
		// ****************** HANDLING INPUT ******************
		// Server side input/request document
		SOAPPart soapPart = soapRequestMessage.getSOAPPart();
		SOAPEnvelope soapEnvelope = soapPart.getEnvelope();
		SOAPBody soapBody = soapRequestMessage.getSOAPBody();
		SOAPHeader soapHeader = soapRequestMessage.getSOAPHeader();
		
		/*Parse request document into IDWS request*/
		/*Validation is done in parsing. So if any error - no request object is created*/
		LibertyRequest request = oioidwsFactory.createRequest(soapEnvelope);
		
		/*Work on request*/
		String messageID = request.getMessageID();
		assertNotNull(messageID); // Client did not set MessageID - but in cannot be null
		String to = request.getTo();
		assertEquals("https://myApp", to);
		
		/*Work with CitizenIdentityToken*/
		CitizenIdentityToken identityToken = request.getIdentityToken();
		assertNotNull(identityToken); // Must be there!
		String cpr = identityToken.getCpr();
		assertEquals("2512484916", cpr);
		BasicPrivileges basicPriviliges = identityToken.getPrivileges();
		assertNotNull(basicPriviliges);

		BlurringInstructions10Impl blurringInstructions10 = (BlurringInstructions10Impl) identityToken.getBlurringInstructions();
		assertNotNull(blurringInstructions10);

		/*Work with priviliges*/
		Set<String> scopes = basicPriviliges.getScopes();
		for (String scope : scopes) {
			List<String> privileges = basicPriviliges.getPrivileges(scope);
			
			assertEquals("urn:dk:healthcare:saml:actThroughProcurationBy:cprNumberIdentifier:1111111118", scope);
			
			assertEquals("urn:dk:healthcare:fmk:read", privileges.get(0));
			assertEquals("urn:dk:healthcare:fmk:write", privileges.get(1));			
		}
		
		List<String> privilegeList = basicPriviliges.getPrivileges("urn:dk:healthcare:saml:actThroughProcurationBy:cprNumberIdentifier:1111111118");
		assertNotNull(privilegeList);
		assertEquals(2, privilegeList.size());
		
		/*Work in my app*/
		Tag myAppTag = TagUtil.create("http://demo.dk/custom", "myApp","App");
		Element myApp = TagUtil.getFirstChildElementNS(soapBody, myAppTag);
		String myAppRequestAsString = XmlUtil.node2String(myApp, false, false);
		assertEquals("<myApp:App xmlns:myApp=\"http://demo.dk/custom\"><myApp:Field_1>value1</myApp:Field_1><myApp:Field_2 custom_attribute_2=\"value2\"/></myApp:App>", myAppRequestAsString);
		
		Tag myAppHeaderTag = TagUtil.create("http://demo.dk/custom", "myApp","Header");
		Element myHeader = TagUtil.getFirstChildElementNS(soapHeader, myAppHeaderTag);
		String myAppRequestHeaderAsString = XmlUtil.node2String(myHeader, false, false);
		assertEquals("<myApp:Header xmlns:myApp=\"http://demo.dk/custom\">my_app_header</myApp:Header>", myAppRequestHeaderAsString);
		
		// ****************** CREATING OUTPUT ******************
		Document outputDocument = XmlUtil.readXml(System.getProperties(), MY_SOAP_APP_REPONSE, false);
		
		/*Sign repsonse with test vault*/
		LibertyMessageDOMEnhancer enhancer = oioidwsFactory.createResponseDomEnhancer(outputDocument, true);
		enhancer.setWSAddressingAction("myApp.response"); //Required
		//enhancer.setWSAddressingMessageID(); //Required - one will be made if not set be server
		enhancer.setWSAddressingRelatesTo(request.getTo()); //Required - Validation will fail if not set
		//enhancer.setWSAddressingTo("client"); //Optional
		
		enhancer.enhanceAndSign(); // Enhance and sign output document into IDWS valid reponse
		
		/*create response message*/
		MessageFactory messageFactory = MessageFactory.newInstance(SOAPConstants.SOAP_1_1_PROTOCOL);
		soapResponseMessage = messageFactory.createMessage(null, new ByteArrayInputStream(XmlUtil.node2String(outputDocument).getBytes()));
		//return responseSoapMessage
		//System.out.println(XmlUtil.node2String(responseSoapMessage.getSOAPPart().getEnvelope(), true, false));
	}
	
	@After
	public void validate() throws SOAPException {
		final CredentialVault serverTestVault = CredentialVaultTestUtil.getCredentialVaultFromResource(System.getProperties(), "fmk_idws_test.p12");
		
		// ****************** HANDLING INPUT ******************
		// Server side output/response document
		SOAPPart soapPart = soapResponseMessage.getSOAPPart();
		SOAPEnvelope soapEnvelope = soapPart.getEnvelope();
		SOAPBody soapBody = soapResponseMessage.getSOAPBody();
		SOAPHeader soapHeader = soapResponseMessage.getSOAPHeader();
			
		// Not required to validate that response is valid. The client could make the validation
		LibertyResponse response = OIOIDWSFactory.createResponse(soapEnvelope);
		response.validateSignatureAndTrust(serverTestVault);
		
		/*Work with my app*/
		Tag myAppHeaderTag = TagUtil.create("http://demo.dk/custom", "myApp","Header");
		Element myHeader = TagUtil.getFirstChildElementNS(soapHeader, myAppHeaderTag);
		String myAppRequestHeaderAsString = XmlUtil.node2String(myHeader, false, false);
		assertEquals("<myApp:Header xmlns:myApp=\"http://demo.dk/custom\">my_app_header</myApp:Header>", myAppRequestHeaderAsString);
		
		/*Work in my app*/
		Tag myAppResponseTag = TagUtil.create("http://demo.dk/custom", "myApp","Response");
		Tag myAppStatusTag = TagUtil.create("http://demo.dk/custom", "myApp","Status");
		Element myAppStatus = TagUtil.getFirstChildElementNS(soapBody, myAppResponseTag, myAppStatusTag);
		assertNotNull(myAppStatus);
		assertEquals("It is working!", myAppStatus.getTextContent());		
	}
}
