/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.testtools.testrequest;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.SignatureException;
import java.util.Properties;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;
import org.w3c.dom.Document;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Message;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SecurityTokenResponse;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.UserInfo;
import dk.sosi.seal.model.constants.DGWSConstants;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.test.TestParameterBuilder;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.testtools.testrequest.STSMessageGenerator.VERSION;

/**
 * @author chg
 *
 */
@RunWith(Parameterized.class)
public class TestDGWSVersions {

    @Parameter(0)
    public String suiteName;

    @Parameter(1)
    public CredentialVault vocesVault;

    @Parameter(2)
    public CredentialVault mocesVault;

    @Parameter(3)
    public STSMessageGenerator.VERSION stsMessageVersion;

    /*Setting up test parameters*/
    @Parameters(name = "{0}")
    public static Iterable<Object[]> data() {
        TestParameterBuilder builder = new TestParameterBuilder();
        builder.newTest("OCES3").add(CredentialVaultTestUtil.getVoces3CredentialVault(), CredentialVaultTestUtil.getMoces3CredentialVault(), STSMessageGenerator.VERSION.OCES3);
        return builder.build();
    }


    @Test
    public void testDGWS_1_0() throws Exception {
        dgwsVersionTest(DGWSConstants.VERSION_1_0);
    }

    @Test
    public void testDGWS_1_0_1() throws Exception {
        dgwsVersionTest(DGWSConstants.VERSION_1_0_1);
    }

    private void dgwsVersionTest(String version) throws Exception {
        Properties properties = SignatureUtil.setupCryptoProviderForJVM();
        properties.put(SOSIFactory.PROPERTYNAME_SOSI_DGWS_VERSION, version);
        Federation federation = new SOSITestFederation(properties);
        SOSIFactory factory = new SOSIFactory(federation, vocesVault, properties);
        STSMessageGenerator messageGenerator = new STSMessageGenerator(properties, stsMessageVersion);

        SecurityTokenRequest request = factory.createNewSecurityTokenRequest();

        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "30808460", "someOrg");
        IDCard idCard = factory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
        request.setIDCard(idCard);

        Document requestDoc = request.serialize2DOMDocument();
        Document responseDoc = messageGenerator.execute(XmlUtil.node2String(requestDoc));
        SecurityTokenResponse response = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(responseDoc));

        assertTrue(response.isFault());
        assertEquals(version, response.getDGWSVersion());

        // Uses bogus CVR number that doesn't match the certificate's CVR in order to provoke a fault response from the STS
        careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, "XXXXXXX", "someOrg");
        idCard = factory.createNewSystemIDCard("SOSITEST", careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, null, null);
        request.setIDCard(idCard);

        requestDoc = request.serialize2DOMDocument();
        responseDoc = messageGenerator.execute(XmlUtil.node2String(requestDoc));
        response = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(responseDoc));

        assertTrue(response.isFault());
        assertEquals(version, response.getDGWSVersion());

        String cvrNumber = "30808460";

        factory = new SOSIFactory(federation, mocesVault, properties);
        idCard = factory.createNewUserIDCard(
                "SOSITEST",
                new UserInfo(null, "Jan", "Riis", "jan<at>lakeside.dk", "hacker", "doctor", "ZXCVB"),
                new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, cvrNumber, "Hansens Praksis"),
                AuthenticationLevel.MOCES_TRUSTED_USER,
                null,
                null,
                factory.getCredentialVault().getSystemCredentialPair().getCertificate(),
                null);
        request.setIDCard(idCard);


        requestDoc = request.serialize2DOMDocument();
        String signature = createSignature(factory, request, requestDoc);
        request.getIDCard().injectSignature(signature, factory.getCredentialVault().getSystemCredentialPair().getCertificate());

        responseDoc = messageGenerator.execute(XmlUtil.node2String(requestDoc));
        response = factory.deserializeSecurityTokenResponse(XmlUtil.node2String(responseDoc));

        assertTrue(response.isFault());
        assertEquals(version, response.getDGWSVersion());

    }

    private String createSignature(SOSIFactory factory, Message message, Document doc) throws NoSuchAlgorithmException, InvalidKeyException,
    SignatureException {

        byte[] siBytes = message.getIDCard().getBytesForSigning(doc);
        //		Sign the bytes with a private key (in this case the key in credential
        //		vault, but that makes no difference)
        Signature jceSign;
        try {
            jceSign = Signature.getInstance("SHA1withRSA",SignatureUtil.getCryptoProvider(factory.getProperties(),
                    SOSIFactory.PROPERTYNAME_SOSI_CRYPTOPROVIDER_SHA1WITHRSA));
        } catch (NoSuchProviderException e) {
            throw new ModelException("No Such Provider", e);
        }
        PrivateKey key = factory.getCredentialVault().getSystemCredentialPair().getPrivateKey();
        jceSign.initSign(key);
        jceSign.update(siBytes);
        return XmlUtil.toBase64(jceSign.sign());
    }



}
