/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
package dk.sosi.testtools.testrequest;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.modelbuilders.ModelBuildException;
import dk.sosi.seal.modelbuilders.SignatureInvalidModelBuildException;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVaultTestUtil;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.testtools.testrequest.STSMessageGenerator.VERSION;
import junit.framework.TestCase;
import org.w3c.dom.Document;

import java.io.File;
import java.util.Properties;

import static org.junit.Assert.fail;


public class TestProviderMessageGenerator // TODO SDS-6659: extends TestCase
{

	Document requestTemplate;
	ProviderMessageGenerator pmg;
	Properties properties;
	SOSIFactory factory;

	protected void setUp() {
		if (pmg == null) {
            properties = SignatureUtil.setupCryptoProviderForJVM();
			properties.putAll(System.getProperties());
			pmg = new ProviderMessageGenerator(properties, VERSION.OCES3);
			STSMessageGenerator smg = pmg.getSTSMessageGenerator();
			new File(smg.OUTPUTFOLDER + "/requests/").mkdirs();
			new File(smg.OUTPUTFOLDER + "/responses/").mkdirs();
			new File(smg.OUTPUTFOLDER + "/providerrequests/").mkdirs();

			String filename = properties.getProperty("sositest:providerrequesttemplate", "requests/lvl4providerrequest.xml");
			requestTemplate = MessageGeneratorHelper.parseXmlFile(filename);
			factory = CredentialVaultUtil.createSOSIFactoryWithTestSTSFederation(properties);
		}
	}

// T0
	// TODO SDS-6659
	public void _testLvl1SystemValidIdCard() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateLvl1SystemValidIdCard(pr);
		pmg.writeXmlFile(pr, "generateLvl1SystemValidIdCard");
		factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
	}

	// TODO SDS-6659
	public void _testLvl3SystemValidIdCard() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateLvl3SystemValidIdCard(pr);
		pmg.writeXmlFile(pr, "generateLvl3SystemValidIdCard");
		factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
	}

	// TODO SDS-6659
	public void _testLvl3UserValidIdCard() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateLvl3UserValidIdCard(pr);
		pmg.writeXmlFile(pr, "generateLvl3UserValidIdCard");
		factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
	}

	// TODO SDS-6659
	public void _testLvl4UserValidIdCard() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateLvl4UserValidIdCard(pr);
		pmg.writeXmlFile(pr, "generateLvl4UserValidIdCard");
		factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
	}

//	 T1
	// TODO SDS-6659
	public void _testIDCardSTSCertWrongSSN() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateIDCardSTSCertWrongSSN(pr);
		pmg.writeXmlFile(pr, "generateIDCardSTSCertWrongSSN");
		try {
			factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
			fail();
		} catch (ModelException e) {
			// success
		}
	}

	// TODO SDS-6659
	public void _testIDCardSTSCertExpired() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateIDCardSTSCertExpired(pr);
		pmg.writeXmlFile(pr, "generateIDCardSTSCertExpired");
		try {
			factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
			fail();
		} catch (ModelException e) {
			// success
		}
	}

	// TODO SDS-6659
	public void _testIDCardSTSCertInvalidIssuer() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateIDCardSTSCertInvalidIssuer(pr);
		pmg.writeXmlFile(pr, "generateIDCardSTSCertInvalidIssuer");
		try {
			factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
			fail();
		} catch (ModelException e) {
			// success
		}
	}

//	 T2
	// TODO SDS-6659
	public void _testIDCardUnsigned() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateIDCardUnsigned(pr);
		pmg.writeXmlFile(pr, "generateUnsignedIdCard");
		try {
			factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
			fail();
		} catch (SignatureInvalidModelBuildException e) {
			// success
		}
	}

//	 T3
	// TODO SDS-6659
	public void _testIDCardExpired() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateIDCardExpired(pr);
		pmg.writeXmlFile(pr, "generateExpiredIdCard");
		// Id card is not signed with real sts test certificate, so the MockFederation must be used for validation
		SOSIFactory fac = new SOSIFactory(new SOSITestFederation(properties, "CVR:46837428-UID:27910135"), CredentialVaultTestUtil.getMoces3CredentialVault(), properties);
		fac.deserializeRequest(XmlUtil.node2String(pr, false, true));
	}

//	 T4
	// TODO SDS-6659
	public void _testIDCardCorruptedUserInfo() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateCorruptedUserInfoIdCard(pr);
		pmg.writeXmlFile(pr, "generateCorruptedUserInfoIdCard");
		try {
			factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
			fail();
		} catch (ModelBuildException e) {
			// succcess
		}
	}

	// TODO SDS-6659
	public void _testIDCardCorruptedRole() {
		Document pr = pmg.domClone(requestTemplate);
		pmg.generateCorruptedRoleIdCard(pr);
		pmg.writeXmlFile(pr, "generateCorruptedRoleIdCard");
		try {
			factory.deserializeRequest(XmlUtil.node2String(pr, false, true));
			fail();
		} catch (ModelException e) {
			// succcess
		}
	}
}